/*
 *   Driver for the XDX 1212 IO PCI card
 *
 *	Copyright (c) 2001 Haroldo Gamal <gamal@alternex.com.br>
 */

#include <linux/delay.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/slab.h>
#include <linux/wait.h>
#include <linux/module.h>
#include <linux/mutex.h>
#include <linux/firmware.h>
#include <linux/io.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/uaccess.h>
#include <linux/kernel.h>

#include <sound/info.h>
#include <sound/control.h>
#include <sound/pcm_params.h>

#include "xdxaudio_hal.h"
#include "xdx_ddcci.h"
#include "../display/output/hdmi/xdx_hdmi_accel.h"
#include "../xdx_irq.h"

//#define XDX_HDMI_AUDIO_DEBUG

#if defined(XDX_HDMI_AUDIO_DEBUG)
#define XDX_HDMI_AUDIO_PRINT(dev, fmt, ...)	dev_info(dev, fmt, ##__VA_ARGS__)
#else
#define XDX_HDMI_AUDIO_PRINT(dev, fmt, ...)	do { } while (0)
#endif

#define XDX_HDMI_AUDIO0_SYSTEM_NAME				"xdx-audio0"
#define XDX_HDMI_AUDIO1_SYSTEM_NAME				"xdx-audio1"
#define XDX_HDMI_VDMA1CH3_SYSTEM_NAME				"xdx-audio-vdma1"
#define XDX_HDMI_AUDIO_SYSTEM_NAME				"xdx-hdmi-audio"

#define XDX_GDDR0_BASE_ADDR					(0x100000000)
#define XDX_B0_AUDIOIF_FRAME_BUF_BYTES_16K			(0X04000)

#define XDX_B0_AUDIOIF_FRAME_BUF_BYTES				(XDX_B0_AUDIOIF_FRAME_BUF_BYTES_16K)

#define B0_AUDIOIF_SUB_FRAME_IN_BYTES        		(1536)
#define B0_AUDIOIF_SUB_FRAME_OU_BYTES        		(2048)
#define XDX_AUDIO_INPUT_FRAME_BYTES             	(XDX_B0_AUDIOIF_FRAME_BUF_BYTES*B0_AUDIOIF_SUB_FRAME_IN_BYTES/B0_AUDIOIF_SUB_FRAME_OU_BYTES)
#define XDX_AUDIO_OUTPUT_FRAME_BYTES             	(XDX_B0_AUDIOIF_FRAME_BUF_BYTES)
#define XDX_AUDIO_BUFFER_BYTES             			(XDX_AUDIO_INPUT_FRAME_BYTES*16)

#define XDX_B0_AUDIOIF_BUF_PTR_1					(0X7)
#define XDX_B0_AUDIOIF_BUF_PTR_2					(0Xf)

#define XDX_AUDIOIF0_INTR_STS                   (0x48)
#define XDX_AUDIOIF0_INTR_STS_CLR               (0x4C)



#define XDX_PF1_IO_AUDIO_CONFIG_BASE                    (0xe00000)
#define XDX_PF1_IO_AUDIO_CONFIG_SIZE                    (0x100000)
#define XDX_PF1_IO_DOORBELL_CONFIG_BASE                 (0x100520)
#define XDX_PF1_IO_DOORBELL_CONFIG_SIZE                 (0x100000)
#define XDX_PF1_IO_ACPU_HOST_BASE                       (0x100540)
#define XDX_PF1_IO_ACPU_HOST_INT_BASE                   (0x1002ec)
#define XDX_PF1_IO_ACPU_HOST_CLEAR_INT_BASE             (0x1001cc)

#define XDX_CMU_CONFIG                                  (0xa8)
#define XDX_GLOBAL_SWDISABLE                            (0x44)
#define XDX_PKTSCHED_PKT_EN                             (0xaa8)
#define XDX_AUDIO_PKT                                   (0xbc4)
#define XDX_PKT_AUDI_CONTENTS1                          (0xc24)
#define XDX_PKT_AUDI_CONTENTS2                          (0xc28)
#define XDX_AUDPKT_ACR_CONTROL0                         (0xe40)
#define XDX_AUDPKT_ACR_CONTROL1                         (0xe44)
#define XDX_AUDIO_INTERFACE                             (0x820)


#define XDX_LINK_QP_CLOCK                               (0x2002)
#define XDX_GLOBAL_SWDISABLE_VALUE                      (0x0)
#define XDX_PKTSCHED_PKT_EN_VALUE                       (0xf00e)
#define XDX_AUDIO_PKT_VALUE                             (0xc035d)
#define XDX_AUDPKB_2CHNS_S1                             (0x16f)
#define XDX_AUDPKB_2CHNS_S2                             (0x100)
#define XDX_AUDPKB_4CHNS_S1                             (0x36a)
#define XDX_AUDPKB_4CHNS_S2                             (0x101)
#define XDX_AUDPKB_8CHNS_S1                             (0xf172b)
#define XDX_AUDPKB_8CHNS_S2                             (0x11f)
#define XDX_AUDIO_INTERFACE_2CHS                        (0x1011)


#define OSReadHWReg8(addr, off)  ((u8)readb((u8 __iomem *)(addr) + (off)))
#define OSReadHWReg16(addr, off) ((u16)readw((u8 __iomem *)(addr) + (off)))
#define OSReadHWReg32(addr, off) ((u32)readl((u8 __iomem *)(addr) + (off)))

/* Little endian support only */
#define OSReadHWReg64(addr, off) \
    ({ \
     __typeof__(addr) _addr = addr; \
     __typeof__(off) _off = off; \
     (u64) \
     ( \
       ( (u64)(readl((u8 __iomem *)(_addr) + (_off) + 4)) << 32) \
       | readl((u8 __iomem *)(_addr) + (_off)) \
       ); \
     })

#define OSWriteHWReg8(addr, off, val)  writeb((u8)(val), (u8 __iomem *)(addr) + (off))
#define OSWriteHWReg16(addr, off, val) writew((u16)(val), (u8 __iomem *)(addr) + (off))
#define OSWriteHWReg32(addr, off, val) writel((u32)(val), (u8 __iomem *)(addr) + (off))
/* Little endian support only */
#define OSWriteHWReg64(addr, off, val) do \
{ \
    __typeof__(addr) _addr = addr; \
    __typeof__(off) _off = off; \
    __typeof__(val) _val = val; \
    writel((u32)((_val) & 0xffffffff), (u8 __iomem *)(_addr) + (_off)); \
    writel((u32)(((u64)(_val) >> 32) & 0xffffffff), (u8 __iomem *)(_addr) + (_off) + 4); \
} while (0)

/// define for DP
#define FRAME_CNT_PER_BLOCK  192
#define MIN_BLOCK_SIZE       1536
#define GET_BIT(X, BP)      (((X)>>(BP))&1)
#define AUDIO_FIFO_SIZE      32768

#define MAX_BUFFER_BYTES_A0    30720
#define MAX_BUFFER_BYTES_B0    36864

#define __V volatile

#define AUDIOIF_REG_INIT_CNT	(12)
#define AUDIO_RESERVE_VRAM	0xfc000000//4G-64M
#define AUDIO_RESERVE_VRAM_LIMIT	0x1000000//16M


uint32_t audioif_reg_val[AUDIOIF_REG_INIT_CNT][2] = {
	{0x00, 0x0010}, 
	{0x14, (AUDIO_RESERVE_VRAM + 0x100000000)>>21}, 
	{0x08, 0x0}, 
	{0x0c, 0x0000}, 
	{0x04, 0x4020318}, 
	{0x40, 0x1f},
	{0x20, 0x108000},
	{0x24, 0xb},
	{0x28, 0x0},
	{0x2c, 0x0},
	{0x30, 0x0},
	{0x34, 0x0}};

uint32_t xdx_b0_audioif_buf_dst_addr_1 = 0;
uint32_t xdx_b0_audioif_buf_dst_addr_2 = 0;

uint32_t xdx_b0_bpcuv_uStatus[2][6] = {{0x00108000, 0xb, 0, 0, 0, 0}, {0x00208000, 0xb, 0, 0, 0, 0}};


typedef enum
{
	cH02AAudioSubCmdForAllZeroTest = 0,
	cH02AAudioSubCmdPllConfig,
	cH02AAudioSubCmdMax
} H02AAudioSubCmdType_T;
#if 0
typedef union
{
    __V u32 all;
    struct
    {
        __V u32 hdmi_bypass       : 1;
        __V u32 dp_iis_sel        : 1;
        __V u32 hdmi_cdiv         : 1;
        __V u32 precision         : 1;
        __V u32 chnl_num          : 2;
        __V u32 buffer_num        : 3;
        __V u32 threshold         : 2;
        __V u32 info_mode         : 2;
        __V u32 axi_qos           : 4;
        __V u32 to_thrd_sel       : 3;
    };
}CTRL_T;


typedef union
{
    __V u32 all;
    struct
    {
        __V u32 wptr        : 8;
        __V u32 rptr        : 8;
    };
}PTR_T;

typedef union
{
    __V u32 all;
    struct
    {
        __V u32 wptr_up      : 1;
    };
}PTR_UP_T;

typedef union
{
    __V u32 all;
    struct
    {
        __V u32 run      : 1;
    };
}TXEN_T;

typedef union
{
    __V u32 all;
    struct
    {
        __V u32 base    : 19;
    };
}BUF_BASE_T;

typedef union
{
    __V u32 all;
    struct
    {
        __V u32 iis_idle       : 1;
        __V u32 empty          : 1;
        __V u32 aempty         : 1;
        __V u32 bus_timeout    : 1;
        __V u32 bus_err        : 1;
    };
}INTR_EN_T;

typedef union
{
    __V u32 all;
    struct
    {
        __V u32 iis_idle       : 1;
        __V u32 empty          : 1;
        __V u32 aempty         : 1;
        __V u32 bus_timeout    : 1;
        __V u32 bus_err        : 1;
        __V u32 axi_rchnl_idle        : 1;
    };
}INTR_RAW_STS_T;

typedef union
{
    __V u32 all;
    struct
    {
        __V u32 iis_idle       : 1;
        __V u32 empty          : 1;
        __V u32 aempty         : 1;
        __V u32 bus_timeout    : 1;
        __V u32 bus_err        : 1;
    };
}INTR_STS_T;

typedef union
{
    __V u32 all;
    struct
    {
        __V u32 iis_idle       : 1;
        __V u32 empty          : 1;
        __V u32 aempty         : 1;
        __V u32 bus_timeout    : 1;
        __V u32 bus_err        : 1;
    };
}INTR_STS_CLR_T;

typedef struct
{
   __V CTRL_T           CTRL;              //Offset 0x04
   __V PTR_T            PTR;              //Offset 0x08
   __V PTR_UP_T         PTR_UP;              //Offset 0x0C
   __V TXEN_T           TXEN;                  //Offset 0x10
   __V BUF_BASE_T       BUF_BASE;                //Offset 0x14
   __V INTR_EN_T        INTR_EN;                //Offset 0x40
   __V INTR_RAW_STS_T   INTR_RAW_STS;                //Offset 0x44
   __V INTR_STS_T       INTR_STS;                 //offset 0x48      
   __V INTR_STS_CLR_T   INTR_STS_CLR;                 //offset 0x4c      
} Audio0_Regs_T;
#endif
typedef union
{
    __V u32 all;
    struct
    {
        __V u32 hdmi_bypass : 1;   //bit 0
        __V u32 resv0       : 2;   //bit 1-2
        __V u32 hdmi_cdiv   : 1;   //bit 3
        __V u32 resv1       : 1;   //bit 4
        __V u32 chnl_num    : 2;   //bit 5
        __V u32 resv2       : 1;   //bit 7
        __V u32 aempty_thrd : 2;   //bit 8
        __V u32 dp_iis_sel  : 1;   //bit 10
        __V u32 pl_shift    : 5;   //bit 11-15
        __V u32 resv3       : 16;  //bit 16-31
    };
}CNTL_T;

typedef union
{
    __V u32 all;
    struct
    {
        __V u32 iis_idle    : 1;
        __V u32 empty       : 1;
        __V u32 aempty      : 1;
        __V u32 resv        : 29;
    };
}IE_T;

typedef union
{
    __V u32 all;
    struct
    {
        __V u32 iis_idle    : 1;
        __V u32 empty       : 1;
        __V u32 aempty      : 1;
        __V u32 bclk_stable : 1;
        __V u32 resv        : 28;
    };
}RIS_T;

typedef union
{
    __V u32 all;
    struct
    {
        __V u32 iis_idle    : 1;
        __V u32 empty       : 1;
        __V u32 aempty      : 1;
        __V u32 resv        : 29;
    };
}MIS_T;

typedef union
{
    __V u32 all;
    struct
    {
        __V u32 hdmi_bclk_en: 1;
        __V u32 resv        : 31;
    };
}CKEN_T;

typedef union
{
    __V u32 all;
    struct
    {
        __V u32 iis_en      : 1;
        __V u32 resv        : 31;
    };
}TXEN_T;

typedef struct
{
   __V CNTL_T CNTL;              //Offset 0x00
   __V CKEN_T CKEN;              //Offset 0x04
   __V TXEN_T TXEN;              //Offset 0x08
   __V IE_T IE;                  //Offset 0x0C
   __V RIS_T RIS;                //Offset 0x10
   __V MIS_T MIS;                //Offset 0x14
} Audio_Regs_T;

/* DMA Command */
typedef union
{
    uint64_t all;
    struct
    {
        uint64_t srcAddr_47_3:45;   // Source Address[47:3]
        uint64_t srcAddrMode:2;     // Source Address Mode:0-HOST_SVA, 1-HOST_SPA, 2-GPU_SVA, 3-GPU_GPA
        uint64_t intrEn:1;          // Interrupt Enable: 1-set intsts when cmd cmpl
        uint64_t lengthUnit:2;      // Tranfer Length Unit:0-8B, 1-64B, 2-1KB, 3-1MB
        uint64_t commandTag:6;      // Command Tag
        uint64_t opCode:8;          // Operation Code
    };
}DMACmdDw0_T;

typedef union
{
    uint64_t all;
    struct
    {
        uint64_t dstAddr_47_3:45;       // Destination Address[47:3]
        uint64_t dstAddrMode:2;         // Destination Address Mode:0-HOST_SVA, 1-HOST_SPA, 2-GPU_SVA, 3-GPU_GPA
        uint64_t rsvd:1;
        uint64_t transferLen:16;        // Transfer Length
    };
}DMACmdDw1_T;

typedef union
{
    uint64_t all;
    struct
    {
        uint64_t patternLen:8;          // Pattern Length (Unit Byte) (RealLength - 1)
        uint64_t patternLoc:8;          // Pattern Location (4Byte Align)
        uint64_t pcieFuncNum:1;         // PCIE Physical Function Number: PF0 or PF1
        uint64_t rsvd1:1;
        uint64_t pcieDontSnoop:1;       // PCIE TLP Snoop Attribute:1-no snoop
        uint64_t rsvd2:45;
    };
}DMACmdDw2_T;

typedef union
{
    uint32_t cmd32[6];
    uint64_t cmd64[3];
    struct
    {
        DMACmdDw0_T dw0;
        DMACmdDw1_T dw1;
        DMACmdDw2_T dw2;
    };
}DMACmd_T;

typedef union
{
	__V u64 all;
	struct
	{
		__V u32 low32;
		__V u32 high32;
	};
	struct
	{
		__V u64 done   : 1;//----
		__V u64 cmd    : 3;//    |---- Header
		__V u64 subCmd : 4;//----
		__V u64 resv   : 24;//----
		__V u64 pixclock   : 32;//-------- Payload
	};
} DBCmd_T;

typedef union
{
	__V u64 all;
	struct
	{
		__V u32 low32;
		__V u32 high32;
	};
	struct
	{
		__V u64 done   : 1;//----
		__V u64 cmd    : 3;//    |---- Header
		__V u64 subCmd : 4;//----
		__V u64 data   : 56;//-------- Payload
	};
} DBCmd_a0f1_T;

typedef union
{
	uint64_t all;
	struct
	{
		uint64_t srcAddr_47_3:45;   // 44:0 Source Address[47:3]
		uint64_t LogicOp:3;         // 47:45 Logic Operation: 0:None 1:NOT 2:XOR 3:OR 4:AND
		uint64_t rsvd:2;            // 49:48 Rsvd
		uint64_t Tag:6;             // 55:50 Command Tag
		uint64_t opCode:8;          // 63:56 Operation Code
	};
} DMAB0CmdDw0_T;

typedef union
{
	uint64_t all;
	struct
	{
		uint64_t Xstep_SrcDescNum:16;  // 15:0 XStep(OP80/81/84) or SrcDescriptorNum(OP82/83)
		uint64_t DstAddr_47_3:45;   // 60:16 Destination Address
		uint64_t LenUnit:3;         // 63:61 Length Unit: 8B/16B/32B/64B/1KB
	};
} DMAB0CmdDw1_T;

typedef union
{
	uint64_t all;
	struct
	{
		uint64_t PcieFuncNum:1;                 // 0
		uint64_t PcieDontSnoop:1;               // 1
		uint64_t IntrEn:1;                      // 2
		uint64_t rsvd1:1;                       // 3
		uint64_t DstrideLow_PatternLen:8;       // 11:4 DstStride[7:0](OP84) or PatternLocation
		uint64_t DstrideHigh_PatternLoc:8;      // 19:12 DstStride[15:8](OP84) or PatternLocation
		uint64_t Sstride_DWriteLenLow:16;       // 35:20 SrcStride(OP84) or DirectWriteLen[15:0]
		uint64_t Ystep_DstDescNum_DstXferLen_DWriteLenHigh:16;  // 51:36 Ystep(OP84) or DstDescNum(OP81/83)
                                                                       //       or DstXferLen(OP82) or DirectWriteLen[24:16]
		uint64_t DstAddrMode:4;     // 55:52 Destination Address Mode
		uint64_t SrcAddrMode:4;     // 59:56 Source Address Mode
		uint64_t rsvd2:4;           // 63:60 Rsvd
	};
} DMAB0CmdDw2_T;

typedef union
{
	uint32_t cmd32[6];
	uint64_t cmd64[3];
	struct
	{
		DMAB0CmdDw0_T dw0;
		DMAB0CmdDw1_T dw1;
		DMAB0CmdDw2_T dw2;
	};
} VDMA_B0_CmdData_T;

typedef enum
{
	cPlayBackStatusNone = 0,
	cPlayBackStart,
	cPlayBackPause,
	cPlayBackStop,
	cPlayBackMax,
} PlayBackStatus_T;

typedef struct {
	uint32_t rsv:2;
	uint32_t data2:8;
	uint32_t data1:8;
	uint32_t data0:8;
	uint32_t v:1;
	uint32_t u:1;
	uint32_t c:1;
	uint32_t p:1;
	uint32_t pr:2;
} DPSubFrame_T;

typedef struct {
	uint32_t rsv:3;
	uint32_t data2:8;
	uint32_t data1:8;
	uint32_t data0:8;
	uint32_t v:1;
	uint32_t u:1;
	uint32_t c:1;
	uint32_t p:1;
	uint32_t b:1;
} HDMISubFrame_T;

typedef union {
	DPSubFrame_T DpSubFrame;
	HDMISubFrame_T HdmiSubFrame;
	uint32_t UnPacketFrame;
} ADIFSubFrame_T;

typedef struct
{
	uint32_t hdmiClk;
	uint32_t nValue;
	uint32_t ctsValue;
} HDMIArcTable_T;

typedef struct
{
	uint32_t dataRate;
	uint32_t tableNum;
	HDMIArcTable_T HdmiArcTbl[14];
	uint32_t otherClkNvalue;
} AdHdmiAcrCfg_T;

/*****************************************************************************************/
/*                                Static Function Statement                             */
/*****************************************************************************************/
static int xdx_hdmi_audio_volumeInfo(struct snd_kcontrol *kcontrol, struct snd_ctl_elem_info *uinfo);
static int  xdx_hdmi_audio_volumeGet(struct snd_kcontrol *kcontrol, struct snd_ctl_elem_value *u);
static int xdx_hdmi_audio_volumePut(struct snd_kcontrol *kcontrol, struct snd_ctl_elem_value *u);
int xdx_audio_drv_hotPlug(struct xdx_device *xdev, PlayDrv module, PlayCon status);

#if defined(XDX_HDMI_AUDIO_DEBUG)
static void dump_runtime(struct snd_pcm_substream *substream)
{
	struct snd_xdx_hdmi_audio *xdx_hdmi_audio_ptr = snd_pcm_substream_chip(substream);
	struct snd_pcm_runtime *runtime = substream->runtime;
	XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d] avail_max=0x%lx\n", __func__, __LINE__, runtime->avail_max);
	XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d] hw_ptr_base=0x%lx\n", __func__, __LINE__, runtime->hw_ptr_base);
	XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d] hw_ptr_interrupt=0x%lx\n", __func__, __LINE__, runtime->hw_ptr_interrupt);

	XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d] access=%d\n", __func__, __LINE__, runtime->access);
	XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d] format=%d\n", __func__, __LINE__, runtime->format);
	XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d] subformat=%d\n", __func__, __LINE__, runtime->subformat);
	XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d] rate=%d\n", __func__, __LINE__, runtime->rate);
	XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d] channels=%d\n", __func__, __LINE__, runtime->channels);
	XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d] period_size=%ld\n", __func__, __LINE__, runtime->period_size);
	XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d] periods=%d\n", __func__, __LINE__, runtime->periods);
	XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d] buffer_size=%ld\n", __func__, __LINE__, runtime->buffer_size);
	XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d] period_step=%d\n", __func__, __LINE__, runtime->period_step);
	XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d] frame_bits=%d\n", __func__, __LINE__, runtime->frame_bits);
	XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d] sample_bits=%d\n", __func__, __LINE__, runtime->sample_bits);
	//XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d] sleep_min=%d\n", __func__, __LINE__, runtime->sleep_min);
	XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d] start_threshold=%ld\n", __func__, __LINE__, runtime->start_threshold);
	XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d] stop_threshold=%ld\n", __func__, __LINE__, runtime->stop_threshold);
	XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d] dma_bytes=0x%lx\n", __func__, __LINE__, runtime->dma_bytes);
	//XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d] mmap_count=%d\n", __func__, __LINE__, runtime->mmap_count);
}
#endif

static const struct snd_pcm_hardware snd_xdx_hdmi_audio_playback_hw =
{
	.info =			(SNDRV_PCM_INFO_MMAP |
				SNDRV_PCM_INFO_MMAP_VALID |
				SNDRV_PCM_INFO_INTERLEAVED |
				SNDRV_PCM_INFO_BLOCK_TRANSFER),
	.formats =		(SNDRV_PCM_FMTBIT_S16_LE | SNDRV_PCM_FMTBIT_S24_LE),
	.rates =		(SNDRV_PCM_RATE_44100 |
				SNDRV_PCM_RATE_48000),
	.rate_min =		44100,
	.rate_max =		48000,
	.channels_min =		2, //X0101_MIN_CHANNELS,
	.channels_max =		2,//X0101_MAX_CHANNELS,
	.buffer_bytes_max =	36*1024, //X0101_MAX_BUF_SIZE,
	.period_bytes_min =	1536,//X0101_MIN_CHANNELS * 2 * kPlayBufferFrames,
	.period_bytes_max =	12*1024,//X0101_MAX_CHANNELS * 2 * kPlayBufferFrames,
	.periods_min =		1,//X0101_PERIODS,
	.periods_max =		1024,//X0101_PERIODS,
};

static const struct snd_kcontrol_new snd_xdx_hdmi_control[] =
{
	{
		.access =	SNDRV_CTL_ELEM_ACCESS_READ | SNDRV_CTL_ELEM_ACCESS_WRITE,
		.iface =	SNDRV_CTL_ELEM_IFACE_MIXER,
		.name =		"Monitor Left And Right Channel Volume",
		.info =		xdx_hdmi_audio_volumeInfo,
		.get =		xdx_hdmi_audio_volumeGet,
		.put =		xdx_hdmi_audio_volumePut,
	}
};

typedef enum
{
	VIDEO_TYPE_HDMI,
	VIDEO_TYPE_DP,
	VIDEO_TYPE_MAX,
} VIDEO_TYPE_MODE;

typedef enum
{
	PCM_BWIDTH_16,
	PCM_BWIDTH_24,
	PCM_BWIDTH_32,
	PCM_BWIDTH_MAX,
} PCM_BWIDTH_MODE;

typedef enum
{
	PCM_CHNUM_20,
	PCM_CHNUM_21,
	PCM_CHNUM_51,
	PCM_CHNUM_71,
	PCM_CHNUM_MAX,
} PCM_CHNUM_MODE;

static void snd_xdx_hdmi_audio_free_pcm(struct snd_pcm *pcm)
{
	struct snd_xdx_hdmi_audio *xdx_hdmi_audio_ptr = pcm->private_data;

	XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d] snd_xdx_hdmi_audio_free_pcm.\n", __func__, __LINE__);

	xdx_hdmi_audio_ptr->pcm = NULL;
}

static uint32_t period_bytes[] = {1536, 3072, 4608, 6144};
static struct snd_pcm_hw_constraint_list constraints_period_bytes = {
	.count = ARRAY_SIZE(period_bytes),
	.list = period_bytes,
	.mask = 0,
};

static int snd_xdx_hdmi_audio_playback_open(struct snd_pcm_substream *substream)
{
	struct snd_xdx_hdmi_audio *xdx_hdmi_audio_ptr = snd_pcm_substream_chip(substream);
	struct snd_pcm_runtime *runtime = substream->runtime;
	struct xdx_device *xdev = xdx_hdmi_audio_ptr->xdev;

	XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d]\n", __func__, __LINE__);


	runtime->hw = snd_xdx_hdmi_audio_playback_hw;
	xdx_hdmi_audio_ptr->playback_substream = substream;
	xdx_hdmi_audio_ptr->pos = 0;
	xdx_hdmi_audio_ptr->fifo_bytes = 0;
	xdx_hdmi_audio_ptr->fifo_empty = 1;
	xdx_hdmi_audio_ptr->dma_pending_cnt = 0;
	xdx_hdmi_audio_ptr->transfer_flag = 0;
	xdx_hdmi_audio_ptr->dmaTag = 0;
	xdx_hdmi_audio_ptr->vdma1ch3_finish = 0;
	if (xdev->asic_type == CHIP_PANGU_A0) {
		snd_pcm_hw_constraint_single(runtime, SNDRV_PCM_HW_PARAM_BUFFER_BYTES, MAX_BUFFER_BYTES_A0);
		snd_pcm_hw_constraint_list(runtime, 0, SNDRV_PCM_HW_PARAM_PERIOD_BYTES, &constraints_period_bytes);
	} else if (xdev->asic_type == CHIP_PANGU_B0)
		snd_pcm_hw_constraint_single(runtime, SNDRV_PCM_HW_PARAM_BUFFER_BYTES, MAX_BUFFER_BYTES_B0);

	snd_pcm_hw_constraint_single(runtime, SNDRV_PCM_HW_PARAM_SAMPLE_BITS, 32);

	return 0;
}

static int snd_xdx_hdmi_audio_playback_close(struct snd_pcm_substream *substream)
{
#if defined(XDX_HDMI_AUDIO_DEBUG)
	struct snd_xdx_hdmi_audio *xdx_hdmi_audio_ptr = snd_pcm_substream_chip(substream);
	XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d]\n", __func__, __LINE__);
#endif
	return 0;
}

static void ADIFOutputSelConfig(struct snd_xdx_hdmi_audio *xdx_hdmi_audio_ptr)
{
	uint32_t value;
	volatile uint32_t* pDpAudioCfgReg;
	volatile uint32_t* pDpAudioSdpCfg0Reg;
	volatile uint32_t* pDpAudioSdpCfg1Reg;

	if (2 == xdx_hdmi_audio_ptr->channels || 0 == xdx_hdmi_audio_ptr->channels) {
		if (16 == xdx_hdmi_audio_ptr->bit_width)
			value = 0x12001202;
		else
			value = 0x12001302;
	} else if (4 == xdx_hdmi_audio_ptr->channels) {
		if (16 == xdx_hdmi_audio_ptr->bit_width)
			 value = 0x12002206;
		else
			value = 0x12002306;
	} else if (8 == xdx_hdmi_audio_ptr->channels) {
		if (16 == xdx_hdmi_audio_ptr->bit_width)
			value = 0x1200221e;
		else
			value = 0x1200231e;
	} else {
		XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "Channel number error!\n");
		return;
	}

	/// Setup the DP register
	pDpAudioCfgReg = (volatile uint32_t*)(xdx_hdmi_audio_ptr->iobase_dp0 + 0x400);
	*pDpAudioCfgReg = value;

	pDpAudioSdpCfg0Reg =  (volatile uint32_t*)(xdx_hdmi_audio_ptr->iobase_dp0 + 0x504);
	*pDpAudioSdpCfg0Reg = 0x80000003;
    
	pDpAudioSdpCfg1Reg =  (volatile uint32_t*)(xdx_hdmi_audio_ptr->iobase_dp0 +  0x500);
	*pDpAudioSdpCfg1Reg = 0xc0000003;
}

//{audio_rate, the number of table, {hdmi_clock, n, cts}, {the value of other hdmi_clock to n}}
const AdHdmiAcrCfg_T audioRateCfg[]=
{
	{32000, 14,
	{{25175,4576,28125},{25200,4096,25200},{27000,4096,27000},{27027,4096,27027},{54000,4096,54000},{54054,4096,54054},
	{74176,11648,210937},{742500,4096,74250},{148352,11648,421875},{148500,4096,148500},{296703,5824,421875},
	{297000,3072,222750},{593407,5824,843750},{594000,3072,445500}},
	4096},

	{64000, 14,
	{{25175,9152,28125},{25200,8192,25200},{27000,8192,27000},{27027,8192,27027},{54000,8192,54000},{54054,8192,54054},
	{74176,23296,210937},{742500,8192,74250},{148352,23296,421875},{148500,8192,148500},{296703,11648,421875},
	{297000,8192,297000},{593407,11648,84375},{594000,8192,594000}},
	8192},

	{128000, 14,
	{{25175,18304,28125},{25200,16384,25200},{27000,16384,27000},{27027,16384,27027},{54000,16384,54000},{54054,16384,54054},
	{74176,46592,210937},{742500,16384,74250},{148352,46592,421875},{148500,16384,148500},{296703,23296,421875},
	{297000,16384,297000},{593407,23296,843750},{594000,16384,594000}},
	16384},

	{256000, 14,
	{{25175,36608,28125},{25200,327668,25200},{27000,32768,27000},{27027,32768,27027},{54000,32768,54000},{54054,32768,54054},
	{74176,93184,210937},{742500,32768,74250},{148352,93184,421875},{148500,32768,148500},{296703,46592,421875},
	{297000,32768,297000},{593407,46592,843750},{594000,32768,594000}},
	32768},

	{44100, 14,
	{{25175,7007,31250},{25200,6272,28000},{27000,6272,30000},{27027,6272,30030},{54000,6272,60000},{54054,6272,60060},
	{74176,17836,234375},{742500,6272,82500},{148352,8918,234375},{148500,6272,165000},{296703,4459,234375},
	{297000,4704,247500},{593407,8918,937500},{594000,9408,990000}},
	6272},

	{88200, 14,
	{{25175,14014,31250},{25200,12544,28000},{27000,12544,30000},{27027,12544,30030},{54000,12544,60000},{54054,12544,60060},
	{74176,35672,234375},{742500,12544,82500},{148352,17836,234375},{148500,12544,165000},{296703,8918,234375},
	{297000,9408,247500},{593407,17836,937500},{594000,18816,990000}},
	12544},

	{176400, 14,
	{{25175,28028,31250},{25200,25088,28000},{27000,25088,30000},{27027,25088,30030},{54000,25088,60000},{54054,25088,60060},
	{74176,71344,234375},{742500,25088,82500},{148352,35672,234375},{148500,25088,165000},{296703,17836,234375},
	{297000,18816,247500},{593407,35672,937500},{594000,37632,990000}},
	25088},

	{352800, 14,
	{{25175,56056,31250},{25200,50176,28000},{27000,50176,30000},{27027,50176,30030},{54000,50176,60000},{54054,50176,60060},
	{74176,142688,234375},{742500,50176,82500},{148352,71344,234375},{148500,50176,165000},{296703,35672,234375},
	{297000,37632,247500},{593407,71344,937500},{594000,75264,990000}},
	50176},

	{48000, 14,
	{{25175,6864,28125},{25200,6144,25200},{27000,6144,27000},{27027,6144,27027},{54000,6144,54000},{54054,6144,54054},
	{74176,11648,140625},{742500,6144,74250},{148352,5824,140625},{148500,6144,148500},{296703,5824,281250},
	{297000,5120,247500},{593407,5824,562500},{594000,6144,594000}},
	6144},

	{96000, 14,
	{{25175,13728,28125},{25200,12288,25200},{27000,12288,27000},{27027,12288,27027},{54000,12288,54000},{54054,12288,54054},
	{74176,23296,140625},{742500,12288,74250},{148352,11648,140625},{148500,12288,148500},{296703,11648,281250},
	{297000,10240,247500},{593407,11648,562500},{594000,12288,594000}},
	12288},

	{192000, 14,
	{{25175,27456,28125},{25200,24576,25200},{27000,24576,27000},{27027,24576,27027},{54000,24576,54000},{54054,24576,54054},
	{74176,46592,140625},{742500,24576,74250},{148352,23296,140625},{148500,24576,148500},{296703,23296,281250},
	{297000,20480,247500},{593407,23296,562500},{594000,24576,594000}},
	24576},

	{384000, 14,
	{{25175,54912,28125},{25200,49152,25200},{27000,49152,27000},{27027,49152,27027},{54000,49152,54000},{54054,49152,54054},
	{74176,93184,140625},{742500,49152,74250},{148352,46592,140625},{148500,49152,148500},{296703,46592,281250},
	{297000,40960,247500},{593407,46592,562500},{594000,49152,594000}},
	49152},
};
int xdx_hdmi_audio_setRate(struct snd_xdx_hdmi_audio *xdx_hdmi_audio_ptr)
{
	uint32_t audio_rate = xdx_hdmi_audio_ptr->rate;
	uint32_t hdmi_clock;
	uint32_t s1, s2, i, j, n, cts, cfg;

	if (audio_rate == 0) {
		XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "Audio Rate is 0!\n");
		return 0;
	}

	if (xdx_hdmi_audio_ptr->hdmi->hdmi_mode == NULL)
		hdmi_clock = 148500;
	else
		hdmi_clock = xdx_hdmi_audio_ptr->hdmi->hdmi_mode->clock;

	if (2 == xdx_hdmi_audio_ptr->channels || 0 == xdx_hdmi_audio_ptr->channels) {
		s1 = XDX_AUDPKB_2CHNS_S1;
		s2 = XDX_AUDPKB_2CHNS_S2;
		cfg = XDX_AUDIO_INTERFACE_2CHS;
	} else if (4 == xdx_hdmi_audio_ptr->channels) {
		s1 = XDX_AUDPKB_4CHNS_S1;
		s2 = XDX_AUDPKB_4CHNS_S2;
	} else if (8 == xdx_hdmi_audio_ptr->channels) {
		s1 = XDX_AUDPKB_8CHNS_S1;
		s2 = XDX_AUDPKB_8CHNS_S2;
	} else {
		XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "Channel number error!\n");
		return -1;
	}

	for (i = 0; i < sizeof(audioRateCfg)/sizeof(AdHdmiAcrCfg_T); i++) {
		if (audio_rate == audioRateCfg[i].dataRate) {
			for (j = 0; j < audioRateCfg[i].tableNum; j++) {
				if (hdmi_clock == audioRateCfg[i].HdmiArcTbl[j].hdmiClk) {
					n = audioRateCfg[i].HdmiArcTbl[j].nValue;
					cts = audioRateCfg[i].HdmiArcTbl[j].ctsValue;
					XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d] n=%d, cts=%d, hdmi_clock=%d, audio_rate=%d\n", __func__, __LINE__, n, cts, hdmi_clock, audio_rate);
					break;
				}
			}
			if (j == audioRateCfg[i].tableNum) {
				n = audioRateCfg[i].otherClkNvalue;
				cts = n*hdmi_clock/128/audio_rate*1000;
				XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d] n=%d, cts=%d, hdmi_clock=%d, audio_rate=%d\n", __func__, __LINE__, n, cts, hdmi_clock, audio_rate);
			}
			if (j > audioRateCfg[i].tableNum) {
				XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d] Input pixel error!\n", __func__, __LINE__);
				return -1;
			}
			break;
		}
	}
	if (i >= sizeof(audioRateCfg)/sizeof(AdHdmiAcrCfg_T)) {
		XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d] Input rate error!\n", __func__, __LINE__);
		return -1;
	}

	spin_lock(&xdx_hdmi_audio_ptr->hdmi->hdmi_reg_lock);

	xdx_hdmicore_write(xdx_hdmi_audio_ptr->hdmi, XDX_CMU_CONFIG, XDX_LINK_QP_CLOCK);
	XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d] Read XDX_CMU_CONFIG: 0x%x\n", __func__, __LINE__, xdx_hdmicore_read(xdx_hdmi_audio_ptr->hdmi, XDX_CMU_CONFIG));
	xdx_hdmicore_write(xdx_hdmi_audio_ptr->hdmi, XDX_GLOBAL_SWDISABLE, XDX_GLOBAL_SWDISABLE_VALUE);
	XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d] Read XDX_GLOBAL_SWDISABLE: 0x%x\n", __func__, __LINE__, xdx_hdmicore_read(xdx_hdmi_audio_ptr->hdmi, XDX_GLOBAL_SWDISABLE));
	xdx_hdmicore_write(xdx_hdmi_audio_ptr->hdmi, XDX_PKTSCHED_PKT_EN, XDX_PKTSCHED_PKT_EN_VALUE);
	XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d] Read XDX_PKTSCHED_PKT_EN: 0x%x\n", __func__, __LINE__, xdx_hdmicore_read(xdx_hdmi_audio_ptr->hdmi, XDX_PKTSCHED_PKT_EN));
	xdx_hdmicore_write(xdx_hdmi_audio_ptr->hdmi, XDX_AUDIO_PKT, XDX_AUDIO_PKT_VALUE);
	XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d] Read XDX_AUDIO_PKT: 0x%x\n", __func__, __LINE__, xdx_hdmicore_read(xdx_hdmi_audio_ptr->hdmi, XDX_AUDIO_PKT));

	xdx_hdmicore_write(xdx_hdmi_audio_ptr->hdmi, XDX_PKT_AUDI_CONTENTS1, s1);
	XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d] Read XDX_PKT_AUDI_CONTENTS1: 0x%x\n", __func__, __LINE__, xdx_hdmicore_read(xdx_hdmi_audio_ptr->hdmi, XDX_PKT_AUDI_CONTENTS1));
	xdx_hdmicore_write(xdx_hdmi_audio_ptr->hdmi, XDX_PKT_AUDI_CONTENTS2, s2);
	XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d] Read XDX_PKT_AUDI_CONTENTS2: 0x%x\n", __func__, __LINE__, xdx_hdmicore_read(xdx_hdmi_audio_ptr->hdmi, XDX_PKT_AUDI_CONTENTS2));
	xdx_hdmicore_write(xdx_hdmi_audio_ptr->hdmi, XDX_AUDPKT_ACR_CONTROL0, n);
	XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d] Read XDX_AUDPK_ACR_CONTROL0: 0x%x\n", __func__, __LINE__, xdx_hdmicore_read(xdx_hdmi_audio_ptr->hdmi, XDX_AUDPKT_ACR_CONTROL0));
	xdx_hdmicore_write(xdx_hdmi_audio_ptr->hdmi, XDX_AUDPKT_ACR_CONTROL1, 0x2 | (cts<<4));
	XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d] Read XDX_ADPKT_ACR_CONTROL1: 0x%x\n", __func__, __LINE__, xdx_hdmicore_read(xdx_hdmi_audio_ptr->hdmi, XDX_AUDPKT_ACR_CONTROL1));
	xdx_hdmicore_write(xdx_hdmi_audio_ptr->hdmi, XDX_AUDIO_INTERFACE, cfg);
	XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d] Read XDX_AUDIO_INTERFACE: 0x%x\n", __func__, __LINE__, xdx_hdmicore_read(xdx_hdmi_audio_ptr->hdmi, XDX_AUDIO_INTERFACE));

	spin_unlock(&xdx_hdmi_audio_ptr->hdmi->hdmi_reg_lock);
	return 0;
}

static int xdx_hdmi_audio_blkSize(int channels)
{
	uint32_t blkSize;
	if (channels == 2)
		blkSize = 1536;
	else if (channels == 4)
		blkSize = 3072;
	else if	(channels == 6)
		blkSize = 4608;
	else if	(channels == 8)
		blkSize = 6144;
	else
		blkSize = 1536;
	return blkSize;
}

static int snd_xdx_hdmi_audio_hw_params(struct snd_pcm_substream *substream,
                                 struct snd_pcm_hw_params *params)
{
	struct snd_xdx_hdmi_audio *xdx_hdmi_audio_ptr;
	DBCmd_T dbCmd;
	DBCmd_a0f1_T dbCmd_a0f1;
	uint32_t TimeOut;
	uint32_t dbvalue = 0;
	struct xdx_device *xdev;

	xdx_hdmi_audio_ptr = snd_pcm_substream_chip(substream);
	xdev = xdx_hdmi_audio_ptr->xdev;
	XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d]\n", __func__, __LINE__);

	XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d] access=%d\n", __func__, __LINE__, params_access(params));
	XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d] format=%d\n", __func__, __LINE__, params_format(params));
	XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d] subformat=%d\n", __func__, __LINE__, params_subformat(params));
	XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d] rate=%d\n", __func__, __LINE__, params_rate(params));
	XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d] channels=%d\n", __func__, __LINE__, params_channels(params));
	XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d] width=%d\n", __func__, __LINE__, params_width(params));
	XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d] phsical_width=%d\n", __func__, __LINE__, params_physical_width(params));
	XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d] period_bytes=%d\n", __func__, __LINE__, params_period_bytes(params));
	XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d] buffer_bytes=%d\n", __func__, __LINE__, params_buffer_bytes(params));

	xdx_hdmi_audio_ptr->bit_width = params_width(params);
	xdx_hdmi_audio_ptr->channels = params_channels(params);
	xdx_hdmi_audio_ptr->rate = params_rate(params);
	xdx_hdmi_audio_ptr->blkSize = xdx_hdmi_audio_blkSize(xdx_hdmi_audio_ptr->channels);
	xdx_hdmi_audio_ptr->last_bytes = 0;

	if (xdev->shareram_size == XDX_AUDIO_BAR4_16M) {
		switch (params_rate(params)) {
			case 44100:
				dbCmd_a0f1.data = 44100;
				break;
			case 48000:
				dbCmd_a0f1.data = 48000;
				break;
			default:
				break;
		}
		XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d] dbCmd_a0f1.all=0x%llx\n", __func__, __LINE__, dbCmd_a0f1.all);
		OSWriteHWReg64(xdx_hdmi_audio_ptr->iobase, 0x100588, dbCmd_a0f1.all);
	} else if (xdev->shareram_size == XDX_AUDIO_BAR4_1M) {
		dbCmd.pixclock = params_rate(params);
		dbCmd.subCmd = cH02AAudioSubCmdPllConfig;
		XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d] dbCmd.all=0x%llx\n", __func__, __LINE__, dbCmd.all);
		OSWriteHWReg64(xdx_hdmi_audio_ptr->iobase_acpu, 0x100588, dbCmd.all);
	}

	TimeOut = 10000;
	while(1) {
		if (xdev->shareram_size == XDX_AUDIO_BAR4_16M)
			dbvalue = OSReadHWReg32(xdx_hdmi_audio_ptr->iobase, 0x100588);
		else if (xdev->shareram_size == XDX_AUDIO_BAR4_1M)
			dbvalue = OSReadHWReg32(xdx_hdmi_audio_ptr->iobase_acpu, 0x100588);
		if ((dbvalue & 1) == 1)
			break;

		TimeOut--;
		if (TimeOut == 0) {
			XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d] dbCmd fail\n", __func__, __LINE__);
			return -1;
		}
		udelay(1000);
	}

	if (xdx_hdmi_audio_ptr->audioif_video_connect == XDX_AUDIO_MODULE_DP)
		ADIFOutputSelConfig(xdx_hdmi_audio_ptr);
	else {
		///set audio configure
		if (xdx_hdmi_audio_ptr->hdmi)
			xdx_hdmi_audio_setRate(xdx_hdmi_audio_ptr);
	}

	return snd_pcm_lib_malloc_pages(substream, params_buffer_bytes(params));
}

static void xdx_hdmi_audio_a0_Enable(struct snd_xdx_hdmi_audio *xdx_hdmi_audio_ptr, PlayDrv audioif_video_connect, bool enable)
{
	Audio_Regs_T *pAudioReg = (Audio_Regs_T *)(xdx_hdmi_audio_ptr->iobase_audio0);
	XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d]\n", __func__, __LINE__);

	mutex_lock(xdx_hdmi_audio_ptr->hw_mutex);
	if (enable == 0) {
		pAudioReg->CNTL.all = 0;
		pAudioReg->CKEN.all = 0;
		pAudioReg->TXEN.all = 0;
		pAudioReg->IE.all = 0;
	} else {
		if (xdx_hdmi_audio_ptr->audioif_video_connect == XDX_AUDIO_MODULE_HDMI) {
			pAudioReg->CNTL.hdmi_cdiv = 1;
			pAudioReg->CNTL.dp_iis_sel = 0;
			pAudioReg->CKEN.hdmi_bclk_en = 1;
			pAudioReg->IE.all = 0x6;
			pAudioReg->TXEN.iis_en = 1;
		} else if (xdx_hdmi_audio_ptr->audioif_video_connect == XDX_AUDIO_MODULE_DP) {
			pAudioReg->CNTL.hdmi_cdiv = 0;
			pAudioReg->CNTL.dp_iis_sel = 1;
			pAudioReg->CKEN.hdmi_bclk_en = 1;
			pAudioReg->IE.all = 0x6;
			pAudioReg->TXEN.iis_en = 1;
		} else {
			pAudioReg->CNTL.hdmi_cdiv = 1;
			pAudioReg->CNTL.dp_iis_sel = 0;
			pAudioReg->CKEN.hdmi_bclk_en = 1;
			pAudioReg->IE.all = 0x6;
			pAudioReg->TXEN.iis_en = 1;
		}
	}
	mutex_unlock(xdx_hdmi_audio_ptr->hw_mutex);

	XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d]\n", __func__, __LINE__);
}


static void xdx_hdmi_audio_b0_Enable(struct snd_xdx_hdmi_audio *xdx_hdmi_audio_ptr, PlayDrv audioif_video_connect, bool enable)
{
	int i = 0;
	uint32_t *pReg = NULL;
	XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d]\n", __func__, __LINE__);

	mutex_lock(xdx_hdmi_audio_ptr->hw_mutex);

	if (enable == 0) {
		pReg = (uint32_t *)(xdx_hdmi_audio_ptr->iobase_audio0 + 0x10);
		*pReg = 0;
		OSWriteHWReg32(xdx_hdmi_audio_ptr->iobase_audio0, 0x40, 0x0);

	} else {
		if (xdx_hdmi_audio_ptr->audioif_video_connect == XDX_AUDIO_MODULE_HDMI){
			for (i = 1; i < AUDIOIF_REG_INIT_CNT; i++) {
				pReg = (uint32_t *)(xdx_hdmi_audio_ptr->iobase_audio0 + audioif_reg_val[i][0]);
				*pReg = audioif_reg_val[i][1];
			}
			OSWriteHWReg32(xdx_hdmi_audio_ptr->iobase_audio0, 0x4, 0x4020318);
			XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d]\n", __func__, __LINE__);

		} else if (xdx_hdmi_audio_ptr->audioif_video_connect == XDX_AUDIO_MODULE_DP) {
			for (i = 1; i < AUDIOIF_REG_INIT_CNT; i ++) {
				pReg = (uint32_t *)(xdx_hdmi_audio_ptr->iobase_audio0 + audioif_reg_val[i][0]);
				*pReg = audioif_reg_val[i][1];
			}
			OSWriteHWReg32(xdx_hdmi_audio_ptr->iobase_audio0, 0x4,  0x402031a);
			XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d]\n", __func__, __LINE__);

		} else {
			for (i = 1; i < AUDIOIF_REG_INIT_CNT; i++) {
				pReg = (uint32_t *)(xdx_hdmi_audio_ptr->iobase_audio0 + audioif_reg_val[i][0]);
				*pReg = audioif_reg_val[i][1];
			}
			OSWriteHWReg32(xdx_hdmi_audio_ptr->iobase_audio0, 0x4, 0x4020318);
			XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d]\n", __func__, __LINE__);
		}

	}
	mutex_unlock(xdx_hdmi_audio_ptr->hw_mutex);

	XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d]\n", __func__, __LINE__);
}


static int snd_xdx_hdmi_audio_hw_free(struct snd_pcm_substream *substream)
{
	struct snd_xdx_hdmi_audio *xdx_hdmi_audio_ptr = snd_pcm_substream_chip(substream);
	struct xdx_device *xdev = xdx_hdmi_audio_ptr->xdev;

	uint32_t time_close, timeout_close, time_now;
	timeout_close = msecs_to_jiffies(300);
	XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d]\n", __func__, __LINE__);

	if (xdx_hdmi_audio_ptr->status != cPlayBackStop)
		xdx_hdmi_audio_ptr->status = cPlayBackStop;

	if (xdev->asic_type == CHIP_PANGU_B0)
		xdx_hdmi_audio_b0_Enable(xdx_hdmi_audio_ptr, xdx_hdmi_audio_ptr->audioif_video_connect, 0);
	time_close = get_jiffies_64();
	while (xdx_hdmi_audio_ptr->transfer_flag == 1) {
		udelay(10);
		time_now = get_jiffies_64();
		if (time_now > time_close + timeout_close) {
	    		XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d]Wait transfer done timeout!\n", __func__, __LINE__);
	    		break;
		}
	}

	time_close = get_jiffies_64();
	while (!xdx_hdmi_audio_ptr->fifo_empty || xdx_hdmi_audio_ptr->dma_pending_cnt) {
		udelay(10);
		time_now = get_jiffies_64();
		if (time_now > time_close + timeout_close) {
			XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d]Wait fifo_empty timeout!\n", __func__, __LINE__);
			break;
		}
	}
	if (xdev->asic_type == CHIP_PANGU_A0)
		xdx_hdmi_audio_a0_Enable(xdx_hdmi_audio_ptr, xdx_hdmi_audio_ptr->audioif_video_connect, 0);

	XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d]\n", __func__, __LINE__);

	return snd_pcm_lib_free_pages(substream);
}

static int snd_xdx_hdmi_audio_prepare(struct snd_pcm_substream *substream)
{
	struct snd_xdx_hdmi_audio *xdx_hdmi_audio_ptr = snd_pcm_substream_chip(substream);
	struct xdx_device *xdev = xdx_hdmi_audio_ptr->xdev;
#if defined(XDX_HDMI_AUDIO_DEBUG)
	XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d]\n", __func__, __LINE__);
#endif


#if defined(XDX_HDMI_AUDIO_DEBUG)
	dump_runtime(substream);
#endif

	if (xdev->asic_type == CHIP_PANGU_A0)
		xdx_hdmi_audio_a0_Enable(xdx_hdmi_audio_ptr, xdx_hdmi_audio_ptr->audioif_video_connect, 1);
	else if (xdev->asic_type == CHIP_PANGU_B0)
		xdx_hdmi_audio_b0_Enable(xdx_hdmi_audio_ptr, xdx_hdmi_audio_ptr->audioif_video_connect, 1);
	return 0;
}

static void ADIFGetStatusData(uint8_t ucChnlIndex, uint32_t ucSmplRate, uint32_t *uStsBuffer)
{
	uint32_t value = 0x8000;
	uint8_t code = 0;
	uint32_t *pDst = uStsBuffer;


	if (uStsBuffer == NULL)
		return;

	memset(pDst, 0x0, 24);//192bit

	if (0 == (ucChnlIndex%2))//left channel
		value |= 0x100000;
	else
		value |= 0x200000;

	switch (ucSmplRate) {
		case PCM_SAMPLEBRATE_32K:
			code = 0x3;
			break;
		case PCM_SAMPLEBRATE_64K://TODO
			code = 0x3;
			break;
		case PCM_SAMPLEBRATE_128K://TODO
			code = 0x3;
			break;
		case PCM_SAMPLEBRATE_441K:
			code = 0x0;
			break;
		case PCM_SAMPLEBRATE_882K:
			code = 0x8;
			break;
		case PCM_SAMPLEBRATE_1764K:
			code = 0xC;
			break;
		case PCM_SAMPLEBRATE_48K:
			code = 0x2;
			break;
		case PCM_SAMPLEBRATE_96K:
			code = 0xA;
			break;
		case PCM_SAMPLEBRATE_192K:
			code = 0xE;
			break;
		default:
			break;
	}

	value |= (code << 24);

	*pDst++ = value;

	*pDst = 0xB;//24bit

	return;
}

static uint8_t AdifEvenCheck(uint32_t data, uint32_t len)
{
	uint32_t i;
	uint32_t check = 0;

	for (i = 0; i < len; i++) {
		check += GET_BIT(data, i);
	}

	if (0 == (check % 2))
		return 0;
	else
		return 1;
}

static int xdx_hdmi_audio_prepareDataForDp(struct snd_xdx_hdmi_audio *xdx_hdmi_audio_ptr, uint64_t src, uint64_t dst, uint32_t *size)
{
	uint32_t i, frameCnt, cl, count, blkSize;
	uint32_t uAudioStatus[8][8];
	uint8_t *srcPtr = (uint8_t *)src;
	ADIFSubFrame_T *dstData = (ADIFSubFrame_T *)dst;

	XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d]\n", __func__, __LINE__);
	XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "Src = 0x%llx, Dst = 0x%llx, Count = %d\n", src, dst, *size);

	if (*size % MIN_BLOCK_SIZE != 0 || *size <= 0) {
		XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "Audio data size is not 1536 bytes aligned or greater than 0.\n");
		return -1;
	}

	//init channel status data
	for (cl = 0; cl < xdx_hdmi_audio_ptr->channels; cl++) {
		//channel number (2,4,8)
		ADIFGetStatusData(cl, xdx_hdmi_audio_ptr->rate, &uAudioStatus[cl][0]);
		XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "Audio Status %d=[%x, %x, %x, %x,%x, %x]\n", cl, uAudioStatus[cl][0], uAudioStatus[cl][1],\
			uAudioStatus[cl][2], uAudioStatus[cl][3], uAudioStatus[cl][4], uAudioStatus[cl][5]);
	}

	blkSize = xdx_hdmi_audio_ptr->blkSize;
	count = *size / blkSize;

	for (i = 0; i < count; i++) {
		//packet data for XDMA transmit
		XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "Audio DP Data Block (%d)\n", i);
		for (frameCnt = 0; frameCnt < FRAME_CNT_PER_BLOCK; frameCnt++) {
			//IEC60958 audio block data
			for (cl = 0; cl < xdx_hdmi_audio_ptr->channels; cl++) {
				//channel number (2,4,8)
				dstData->UnPacketFrame = 0;
				dstData->DpSubFrame.c = GET_BIT(uAudioStatus[cl][frameCnt / 32], frameCnt % 32);
				if (xdx_hdmi_audio_ptr->bit_width == 24)
					dstData->DpSubFrame.data2 = *srcPtr++;
				dstData->DpSubFrame.data1 = *srcPtr++;
				dstData->DpSubFrame.data0 = *srcPtr++;
				if (xdx_hdmi_audio_ptr->bit_width == 24)
					srcPtr++;   /// Pass 8 bits of dummy data
				else if (xdx_hdmi_audio_ptr->bit_width == 16) {
					/// Pass 16 bits of dummy data
					srcPtr++;
					srcPtr++;
				}

				dstData->DpSubFrame.p = AdifEvenCheck(dstData->UnPacketFrame, 29);
				if ((cl & 0x1) == 0) {
					if (frameCnt == 0)
						dstData->DpSubFrame.pr = 0;
					else
						dstData->DpSubFrame.pr = 1;
				} else
					dstData->DpSubFrame.pr = 2;

				dstData++;
			}
		}
	}

	return 0;
}


static int xdx_hdmi_audio_prepareDataForHdmi(struct snd_xdx_hdmi_audio *xdx_hdmi_audio_ptr, uint64_t src, uint64_t dst, uint32_t *size)
{
	uint32_t i, frameCnt, cl, count, blkSize;
	uint32_t uAudioStatus[8][8];
	uint8_t *srcPtr = (uint8_t *)src;
	ADIFSubFrame_T *dstData = (ADIFSubFrame_T *)dst;

	XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d]\n", __func__, __LINE__);
	XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "Src = 0x%llx, Dst = 0x%llx, Count = %d\n", src, dst, *size);

	if (*size % MIN_BLOCK_SIZE != 0 || *size <= 0) {
		XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "Audio data size is not 1536 bytes aligned or greater than 0.\n");
		return -1;
	}

	//init channel status data
	for (cl = 0; cl < xdx_hdmi_audio_ptr->channels; cl++) {//channel number (2,4,8)
		ADIFGetStatusData(cl, xdx_hdmi_audio_ptr->rate, &uAudioStatus[cl][0]);
		XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "Audio Status %d=[%x, %x, %x, %x,%x, %x]\n", cl, uAudioStatus[cl][0], uAudioStatus[cl][1],\
			uAudioStatus[cl][2], uAudioStatus[cl][3], uAudioStatus[cl][4], uAudioStatus[cl][5]);
	}


	blkSize = xdx_hdmi_audio_ptr->blkSize;
	count = *size / blkSize;

	for (i = 0; i < count; i++) {
		//packet data for XDMA transmit
		XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "Audio HDMI Data Block (%d)\n", i);
		for (frameCnt = 0; frameCnt < FRAME_CNT_PER_BLOCK; frameCnt++) {
			for (cl = 0; cl < xdx_hdmi_audio_ptr->channels; cl++) {
				//channel number (2,4,8)
				dstData->UnPacketFrame = 0;
				dstData->HdmiSubFrame.c = GET_BIT(uAudioStatus[cl][frameCnt / 32], frameCnt % 32);
				if (xdx_hdmi_audio_ptr->bit_width == 24)
					dstData->HdmiSubFrame.data2 = *srcPtr++;
				dstData->HdmiSubFrame.data1 = *srcPtr++;
				dstData->HdmiSubFrame.data0 = *srcPtr++;
				if (xdx_hdmi_audio_ptr->bit_width == 24)
					srcPtr++;   /// Pass 8 bits of dummy data
				else if (xdx_hdmi_audio_ptr->bit_width == 16) {
					/// Pass 16 bits of dummy data
					srcPtr++;
					srcPtr++;
				}

				dstData->HdmiSubFrame.p = AdifEvenCheck(dstData->UnPacketFrame, 30);
				if ((frameCnt == 0) && (cl & 0x1) == 0)
					dstData->HdmiSubFrame.b = 1;
				dstData++;
			}
		}
	}

	return 0;
}

static int xdx_hdmi_audio_prepareData(struct snd_xdx_hdmi_audio *xdx_hdmi_audio_ptr, uint64_t src, uint64_t dst, uint32_t *size)
{
	uint8_t *srcPtr = (uint8_t *)src;
	uint8_t *dstData = (uint8_t *)dst;
	uint32_t i;

	XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d]\n", __func__, __LINE__);
	XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "Src = 0x%llx, Dst = 0x%llx, Count = %d\n", src, dst, *size);

	if (*size % MIN_BLOCK_SIZE != 0 || *size <= 0) {
		XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "Audio data size is not 1536 bytes aligned or greater than 0.\n");
		return -1;
	}

	for (i = 0; i < 1536; i++) {
		*(dstData++) = *(srcPtr++);
	}

	return 0;
}

static void xdx_hdmi_audio_dma_one_period_a0(struct snd_xdx_hdmi_audio *xdx_hdmi_audio_ptr)
{
	struct snd_pcm_substream *substream = xdx_hdmi_audio_ptr->playback_substream;
	struct snd_pcm_runtime *runtime = substream->runtime;
	DMACmd_T dmaCmd;
	uint32_t period_bytes = snd_pcm_lib_period_bytes(substream);
	uint32_t hw_avail = 0;
	uint32_t tag = xdx_hdmi_audio_ptr->dmaTag;
	int i, pkt_num;
	uint32_t fifo_avail = 0;
	uint32_t dma_bytes = 0;
	uint32_t buffer_bytes;
	uint32_t temp_bytes = 0;
	uint32_t pos = 0;
	uint64_t srcVir = 0;
	unsigned long flags = 0;
	struct xdx_device *xdev = xdx_hdmi_audio_ptr->xdev;

	mutex_lock(xdx_hdmi_audio_ptr->transfer_mutex);

	if (xdx_hdmi_audio_ptr->dma_pending_cnt > 0 || (xdx_hdmi_audio_ptr->status != cPlayBackStart) || (xdx_hdmi_audio_ptr->transfer_flag != 1)) {
		xdx_hdmi_audio_ptr->transfer_flag = 0;
		mutex_unlock(xdx_hdmi_audio_ptr->transfer_mutex);
		return;
	}

	hw_avail = frames_to_bytes(runtime, snd_pcm_playback_hw_avail(runtime));
	fifo_avail = snd_pcm_lib_buffer_bytes(substream) - xdx_hdmi_audio_ptr->fifo_bytes;
	dma_bytes = min(period_bytes, hw_avail);
	dma_bytes = min(dma_bytes, fifo_avail);
	buffer_bytes = snd_pcm_lib_buffer_bytes(xdx_hdmi_audio_ptr->playback_substream);
	XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "%s fifo_bytes=0x%x pos=0x%x dma_bytes=0x%x\n", __func__, xdx_hdmi_audio_ptr->fifo_bytes, xdx_hdmi_audio_ptr->pos, dma_bytes);

	if ((dma_bytes + xdx_hdmi_audio_ptr->fifo_bytes) > snd_pcm_lib_buffer_bytes(substream)) {
		xdx_hdmi_audio_ptr->transfer_flag = 0;
		mutex_unlock(xdx_hdmi_audio_ptr->transfer_mutex);
		XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s -%d] Transfer dma_bytes and fifo_bytes are overflow.\n", __func__, __LINE__);
		return;
	}

	if (dma_bytes == 0) {
		XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d] fifo_bytes=%d pos=%d hw_avail=%d dma_bytes=%d \n", __func__, __LINE__, xdx_hdmi_audio_ptr->fifo_bytes, xdx_hdmi_audio_ptr->pos, hw_avail, dma_bytes);
		xdx_hdmi_audio_ptr->transfer_flag = 0;
		mutex_unlock(xdx_hdmi_audio_ptr->transfer_mutex);

		return;
	} else if ((dma_bytes & 0x7) != 0) {
		xdx_hdmi_audio_ptr->transfer_flag = 0;
		mutex_unlock(xdx_hdmi_audio_ptr->transfer_mutex);
		XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s -%d] Transfer size is not 4 bytes aligned.\n", __func__, __LINE__);

		return;
	}

	memset(&dmaCmd, 0, 24);
	pos = xdx_hdmi_audio_ptr->pos;
	temp_bytes = dma_bytes;
	pkt_num = (xdx_hdmi_audio_ptr->last_bytes + dma_bytes)/xdx_hdmi_audio_ptr->blkSize;
	if (xdx_hdmi_audio_ptr->pos + dma_bytes > MAX_BUFFER_BYTES_A0) {
		dma_bytes = MAX_BUFFER_BYTES_A0 - xdx_hdmi_audio_ptr->pos;
		if ((xdx_hdmi_audio_ptr->last_bytes > 0) && ((dma_bytes + xdx_hdmi_audio_ptr->last_bytes) >= xdx_hdmi_audio_ptr->blkSize)) {
			pkt_num = (xdx_hdmi_audio_ptr->last_bytes + dma_bytes)/xdx_hdmi_audio_ptr->blkSize;
			dma_bytes = pkt_num * xdx_hdmi_audio_ptr->blkSize - xdx_hdmi_audio_ptr->last_bytes;
			memcpy((void*)((uint64_t)xdx_hdmi_audio_ptr->tempBufPtr + xdx_hdmi_audio_ptr->last_bytes), (void*)((uint64_t)runtime->dma_area + xdx_hdmi_audio_ptr->pos), dma_bytes);
			xdx_hdmi_audio_ptr->last_bytes = pkt_num * xdx_hdmi_audio_ptr->blkSize;
		} else if ((dma_bytes + xdx_hdmi_audio_ptr->last_bytes) < xdx_hdmi_audio_ptr->blkSize) {
			memcpy((void*)((uint64_t)xdx_hdmi_audio_ptr->tempBufPtr + xdx_hdmi_audio_ptr->last_bytes), (void*)((uint64_t)runtime->dma_area + xdx_hdmi_audio_ptr->pos), dma_bytes);
			xdx_hdmi_audio_ptr->last_bytes += dma_bytes;
			pos += dma_bytes;
			if (pos >= buffer_bytes)
				pos -= buffer_bytes;
			xdx_hdmi_audio_ptr->pos = pos;

			xdx_hdmi_audio_ptr->transfer_flag = 0;
			mutex_unlock(xdx_hdmi_audio_ptr->transfer_mutex);

			snd_pcm_period_elapsed(xdx_hdmi_audio_ptr->playback_substream);
			return;
		}
	} else if (xdx_hdmi_audio_ptr->last_bytes > 0) {
		if (dma_bytes >= (xdx_hdmi_audio_ptr->blkSize - xdx_hdmi_audio_ptr->last_bytes)) {
			dma_bytes = pkt_num * xdx_hdmi_audio_ptr->blkSize - xdx_hdmi_audio_ptr->last_bytes;
			memcpy((void*)((uint64_t)xdx_hdmi_audio_ptr->tempBufPtr + xdx_hdmi_audio_ptr->last_bytes), (void*)((uint64_t)runtime->dma_area + xdx_hdmi_audio_ptr->pos), dma_bytes);
			xdx_hdmi_audio_ptr->last_bytes = pkt_num * xdx_hdmi_audio_ptr->blkSize;
		} else if ((dma_bytes + xdx_hdmi_audio_ptr->last_bytes) < xdx_hdmi_audio_ptr->blkSize) {
			memcpy((void*)((uint64_t)xdx_hdmi_audio_ptr->tempBufPtr + xdx_hdmi_audio_ptr->last_bytes), (void*)((uint64_t)runtime->dma_area + xdx_hdmi_audio_ptr->pos), dma_bytes);
			xdx_hdmi_audio_ptr->last_bytes += dma_bytes;
			pos += dma_bytes;
			if (pos >= buffer_bytes)
				pos -= buffer_bytes;
			xdx_hdmi_audio_ptr->pos = pos;

			xdx_hdmi_audio_ptr->transfer_flag = 0;
			mutex_unlock(xdx_hdmi_audio_ptr->transfer_mutex);

			snd_pcm_period_elapsed(xdx_hdmi_audio_ptr->playback_substream);
			return;
		}
	}

	if ((dma_bytes > MIN_BLOCK_SIZE) && (dma_bytes % MIN_BLOCK_SIZE !=0) && (xdx_hdmi_audio_ptr->last_bytes == 0))
		dma_bytes = pkt_num * MIN_BLOCK_SIZE;
	else if ((dma_bytes < MIN_BLOCK_SIZE) && (xdx_hdmi_audio_ptr->last_bytes == 0)) {
		memcpy((void*)((uint64_t)xdx_hdmi_audio_ptr->tempBufPtr + xdx_hdmi_audio_ptr->last_bytes), (void*)((uint64_t)runtime->dma_area + xdx_hdmi_audio_ptr->pos), dma_bytes);
		xdx_hdmi_audio_ptr->last_bytes += dma_bytes;
		pos += dma_bytes;
		if (pos >= buffer_bytes)
			pos -= buffer_bytes;
		xdx_hdmi_audio_ptr->pos = pos;

		xdx_hdmi_audio_ptr->transfer_flag = 0;
		mutex_unlock(xdx_hdmi_audio_ptr->transfer_mutex);

		snd_pcm_period_elapsed(xdx_hdmi_audio_ptr->playback_substream);
		return;
	}
	if (xdx_hdmi_audio_ptr->audioif_video_connect == XDX_AUDIO_MODULE_DP) {
		if (xdx_hdmi_audio_ptr->last_bytes == (pkt_num * xdx_hdmi_audio_ptr->blkSize)) {
			srcVir = (uint64_t)(xdx_hdmi_audio_ptr->tempBufPtr);

			temp_bytes = xdx_hdmi_audio_ptr->last_bytes;
			if (xdx_hdmi_audio_prepareDataForDp(xdx_hdmi_audio_ptr, srcVir, (uint64_t)(xdx_hdmi_audio_ptr->periodBufPtr), &temp_bytes) < 0) {
				xdx_hdmi_audio_ptr->transfer_flag = 0;
				mutex_unlock(xdx_hdmi_audio_ptr->transfer_mutex);
				return;
			}
			dmaCmd.dw0.srcAddr_47_3   = xdx_hdmi_audio_ptr->periodBufPhy >> 3;
		} else if (xdx_hdmi_audio_ptr->last_bytes == 0) {
			srcVir = (uint64_t)(runtime->dma_area + xdx_hdmi_audio_ptr->pos);

			if (xdx_hdmi_audio_prepareDataForDp(xdx_hdmi_audio_ptr, srcVir, (uint64_t)(xdx_hdmi_audio_ptr->periodBufPtr), &dma_bytes) < 0) {
				xdx_hdmi_audio_ptr->transfer_flag = 0;
				mutex_unlock(xdx_hdmi_audio_ptr->transfer_mutex);
				return;
			}
			dmaCmd.dw0.srcAddr_47_3   = xdx_hdmi_audio_ptr->periodBufPhy >> 3;
		} else {
			xdx_hdmi_audio_ptr->pos = pos;
			xdx_hdmi_audio_ptr->transfer_flag = 0;
			mutex_unlock(xdx_hdmi_audio_ptr->transfer_mutex);
			snd_pcm_period_elapsed(xdx_hdmi_audio_ptr->playback_substream);
			return;
		}
	} else {
		if (xdx_hdmi_audio_ptr->last_bytes == (pkt_num * xdx_hdmi_audio_ptr->blkSize)) {
			srcVir = (uint64_t)(xdx_hdmi_audio_ptr->tempBufPtr);

			temp_bytes = xdx_hdmi_audio_ptr->last_bytes;
			if (xdx_hdmi_audio_prepareDataForHdmi(xdx_hdmi_audio_ptr, srcVir, (uint64_t)(xdx_hdmi_audio_ptr->periodBufPtr), &temp_bytes) < 0) {
				xdx_hdmi_audio_ptr->transfer_flag = 0;
				mutex_unlock(xdx_hdmi_audio_ptr->transfer_mutex);
				return;
			}
			dmaCmd.dw0.srcAddr_47_3   = xdx_hdmi_audio_ptr->periodBufPhy >> 3;
		} else if (xdx_hdmi_audio_ptr->last_bytes == 0) {
			srcVir = (uint64_t)(runtime->dma_area + xdx_hdmi_audio_ptr->pos);

			if (xdx_hdmi_audio_prepareDataForHdmi(xdx_hdmi_audio_ptr, srcVir, (uint64_t)(xdx_hdmi_audio_ptr->periodBufPtr), &dma_bytes) < 0) {
			        xdx_hdmi_audio_ptr->transfer_flag = 0;
				mutex_unlock(xdx_hdmi_audio_ptr->transfer_mutex);
				return;
			}
			dmaCmd.dw0.srcAddr_47_3   = xdx_hdmi_audio_ptr->periodBufPhy >> 3;
		} else {
			xdx_hdmi_audio_ptr->pos = pos;
			xdx_hdmi_audio_ptr->transfer_flag = 0;
			mutex_unlock(xdx_hdmi_audio_ptr->transfer_mutex);
			snd_pcm_period_elapsed(xdx_hdmi_audio_ptr->playback_substream);
			return;
		}
	}

	dmaCmd.dw0.srcAddrMode  = 1;   // Source Address Mode:0-HOST_SVA, 1-HOST_SPA, 2-GPU_SVA, 3-GPU_GPA
	dmaCmd.dw0.intrEn = 1;         // Interrupt Enable: 1-set intsts when cmd cmpl
	dmaCmd.dw0.commandTag = tag;       // Command Tag
	dmaCmd.dw0.opCode = 0x80;           // Operation Code

	dmaCmd.dw1.dstAddr_47_3 = 0x92200000>>3;       // Destination Address[47:3]
	dmaCmd.dw1.dstAddrMode  = 3;   // Source Address Mode:0-HOST_SVA, 1-HOST_SPA, 2-GPU_SVA, 3-GPU_GPA
	if (xdx_hdmi_audio_ptr->last_bytes != 0)
		dmaCmd.dw1.transferLen = (xdx_hdmi_audio_ptr->last_bytes>>3) - 1;
	else
		dmaCmd.dw1.transferLen = (dma_bytes>>3) - 1;        // Transfer Length

	dmaCmd.dw2.pcieFuncNum = 1;

	++tag;
	tag &= 63;
	xdx_hdmi_audio_ptr->dmaTag = tag;
	xdx_hdmi_audio_ptr->dma_bytes = dma_bytes;
	spin_lock_irqsave(&xdx_hdmi_audio_ptr->lock, flags);
	if (xdx_hdmi_audio_ptr->last_bytes != 0) {
		xdx_hdmi_audio_ptr->fifo_bytes += xdx_hdmi_audio_ptr->last_bytes;
		xdx_hdmi_audio_ptr->last_bytes = 0;
	} else
		xdx_hdmi_audio_ptr->fifo_bytes += dma_bytes;
	xdx_hdmi_audio_ptr->fifo_empty = 0;
	spin_unlock_irqrestore(&xdx_hdmi_audio_ptr->lock, flags);
	xdx_hdmi_audio_ptr->dma_pending_cnt++;

	xdx_hdmi_audio_ptr->transfer_flag = 0;
	mutex_unlock(xdx_hdmi_audio_ptr->transfer_mutex);

	if (xdev->shareram_size == XDX_AUDIO_BAR4_16M) {
		for (i = 0; i < 6; ++i) {
			iowrite32(dmaCmd.cmd32[i], xdx_hdmi_audio_ptr->iobase + 0xD00000 + 0x2800 + 0x200 + i*4);
		}
		OSWriteHWReg64(xdx_hdmi_audio_ptr->iobase, 0xc00428, 0x1);
	} else if (xdev->shareram_size == XDX_AUDIO_BAR4_1M) {
		for (i = 0; i < 6; ++i) {
			iowrite32(dmaCmd.cmd32[i], xdx_hdmi_audio_ptr->iobase_vdma1 + 0x2800 + 0x200 + i*4);
		}
	}
	XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s -%d] Write VDMA register.\n", __func__, __LINE__);
}



static void xdx_hdmi_audio_dma_one_period_b0(struct snd_xdx_hdmi_audio *xdx_hdmi_audio_ptr)
{
	struct snd_pcm_substream *substream = xdx_hdmi_audio_ptr->playback_substream;
	struct snd_pcm_runtime *runtime = substream->runtime;
	VDMA_B0_CmdData_T dmaCmd;
	//uint32_t period_bytes = snd_pcm_lib_period_bytes(substream);
	uint32_t hw_avail = 0;
	uint32_t tag = xdx_hdmi_audio_ptr->dmaTag;
	int i;
	uint32_t fifo_avail = 0;
	uint32_t dma_bytes = 0;
	uint32_t buffer_bytes;
	uint32_t temp_bytes = 0;
	uint32_t pos = 0;
	uint64_t srcVir;

	mutex_lock(xdx_hdmi_audio_ptr->transfer_mutex);

	if (xdx_hdmi_audio_ptr->dma_pending_cnt > 0 || (xdx_hdmi_audio_ptr->status != cPlayBackStart) || (xdx_hdmi_audio_ptr->transfer_flag != 1)) {
		xdx_hdmi_audio_ptr->transfer_flag = 0;
		mutex_unlock(xdx_hdmi_audio_ptr->transfer_mutex);
		return;
	}

	//    spin_lock_irq(&xdx_hdmi_audio_ptr->lock);
	hw_avail = frames_to_bytes(runtime, snd_pcm_playback_hw_avail(runtime));
	//fifo_avail = snd_pcm_lib_buffer_bytes(substream) - xdx_hdmi_audio_ptr->fifo_bytes;
	fifo_avail = 32 * 1024 - xdx_hdmi_audio_ptr->fifo_bytes;
	//dma_bytes = min(period_bytes, hw_avail);
	dma_bytes = min(hw_avail, fifo_avail);
	if (dma_bytes > (12 * 1024))
		dma_bytes = 12 *1024;
	buffer_bytes = snd_pcm_lib_buffer_bytes(xdx_hdmi_audio_ptr->playback_substream);
	XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "%s fifo_bytes=0x%x pos=0x%x dma_bytes=0x%x\n", __func__, xdx_hdmi_audio_ptr->fifo_bytes, xdx_hdmi_audio_ptr->pos, dma_bytes);

	//if((dma_bytes + xdx_hdmi_audio_ptr->fifo_bytes) > snd_pcm_lib_buffer_bytes(substream))
	if ((dma_bytes + xdx_hdmi_audio_ptr->fifo_bytes) > 32 * 1024) {
		xdx_hdmi_audio_ptr->transfer_flag = 0;
		// spin_unlock_irq(&xdx_hdmi_audio_ptr->lock);
		mutex_unlock(xdx_hdmi_audio_ptr->transfer_mutex);
		XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s -%d] Transfer dma_bytes and fifo_bytes are overflow.\n", __func__, __LINE__);
		return;
	}

	if (dma_bytes == 0) {
		XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d] fifo_bytes=%d pos=%d hw_avail=%d dma_bytes=%d \n", __func__, __LINE__, xdx_hdmi_audio_ptr->fifo_bytes, xdx_hdmi_audio_ptr->pos, hw_avail, dma_bytes);
		xdx_hdmi_audio_ptr->transfer_flag = 0;
		//spin_unlock_irq(&xdx_hdmi_audio_ptr->lock);
        	mutex_unlock(xdx_hdmi_audio_ptr->transfer_mutex);

 		return;
	} else if ((dma_bytes & 0x7) != 0) {
		xdx_hdmi_audio_ptr->transfer_flag = 0;
		//spin_unlock_irq(&xdx_hdmi_audio_ptr->lock);
		mutex_unlock(xdx_hdmi_audio_ptr->transfer_mutex);
		XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s -%d] Transfer size is not 4 bytes aligned.\n", __func__, __LINE__);

		return;
	}

	memset(&dmaCmd, 0, 24);
	pos = xdx_hdmi_audio_ptr->pos;
	temp_bytes = dma_bytes;
	if (xdx_hdmi_audio_ptr->pos + dma_bytes > 36 * 1024) {
		dma_bytes = 36 * 1024 - xdx_hdmi_audio_ptr->pos;
#if 0
        if((xdx_hdmi_audio_ptr->last_bytes > 0) && ((dma_bytes + xdx_hdmi_audio_ptr->last_bytes) >= xdx_hdmi_audio_ptr->blkSize))
        {
            dma_bytes = xdx_hdmi_audio_ptr->blkSize - xdx_hdmi_audio_ptr->last_bytes;
            memcpy((void*)((uint64_t)xdx_hdmi_audio_ptr->tempBufPtr + xdx_hdmi_audio_ptr->last_bytes), (void*)((uint64_t)runtime->dma_area + xdx_hdmi_audio_ptr->pos), dma_bytes);
            xdx_hdmi_audio_ptr->last_bytes = xdx_hdmi_audio_ptr->blkSize;
        }
        else if((dma_bytes + xdx_hdmi_audio_ptr->last_bytes) < xdx_hdmi_audio_ptr->blkSize)
        {
            memcpy((void*)((uint64_t)xdx_hdmi_audio_ptr->tempBufPtr + xdx_hdmi_audio_ptr->last_bytes), (void*)((uint64_t)runtime->dma_area + xdx_hdmi_audio_ptr->pos), dma_bytes);
            xdx_hdmi_audio_ptr->last_bytes += dma_bytes;
            pos += dma_bytes;
            if(pos >= buffer_bytes)
                   pos -= buffer_bytes;
            xdx_hdmi_audio_ptr->pos = pos;

            xdx_hdmi_audio_ptr->transfer_flag = 0;
            spin_unlock_irq(&xdx_hdmi_audio_ptr->lock);
            mutex_unlock(xdx_hdmi_audio_ptr->transfer_mutex);

	    snd_pcm_period_elapsed(xdx_hdmi_audio_ptr->playback_substream);
            return;
        }
    }
    else if ((xdx_hdmi_audio_ptr->last_bytes < xdx_hdmi_audio_ptr->blkSize) && (xdx_hdmi_audio_ptr->last_bytes > 0))
    {
        if(dma_bytes >= (xdx_hdmi_audio_ptr->blkSize - xdx_hdmi_audio_ptr->last_bytes))
        {
            dma_bytes = xdx_hdmi_audio_ptr->blkSize - xdx_hdmi_audio_ptr->last_bytes;
            memcpy((void*)((uint64_t)xdx_hdmi_audio_ptr->tempBufPtr + xdx_hdmi_audio_ptr->last_bytes), (void*)((uint64_t)runtime->dma_area + xdx_hdmi_audio_ptr->pos), dma_bytes);
            xdx_hdmi_audio_ptr->last_bytes = xdx_hdmi_audio_ptr->blkSize;
        }
        else if((dma_bytes + xdx_hdmi_audio_ptr->last_bytes) < xdx_hdmi_audio_ptr->blkSize)
        {
            memcpy((void*)((uint64_t)xdx_hdmi_audio_ptr->tempBufPtr + xdx_hdmi_audio_ptr->last_bytes), (void*)((uint64_t)runtime->dma_area + xdx_hdmi_audio_ptr->pos), dma_bytes);
            xdx_hdmi_audio_ptr->last_bytes += dma_bytes;
            pos += dma_bytes;
            if(pos >= buffer_bytes)
                pos -= buffer_bytes;
            xdx_hdmi_audio_ptr->pos = pos;

            xdx_hdmi_audio_ptr->transfer_flag = 0;
            spin_unlock_irq(&xdx_hdmi_audio_ptr->lock);
            mutex_unlock(xdx_hdmi_audio_ptr->transfer_mutex);

            snd_pcm_period_elapsed(xdx_hdmi_audio_ptr->playback_substream);
            return;
        }
#endif
	}

	if ((dma_bytes > MIN_BLOCK_SIZE) && (dma_bytes % MIN_BLOCK_SIZE !=0)) {
		dma_bytes = MIN_BLOCK_SIZE;
	}  else if ((dma_bytes < MIN_BLOCK_SIZE) && (xdx_hdmi_audio_ptr->last_bytes == 0)) {

		memcpy((void*)((uint64_t)xdx_hdmi_audio_ptr->tempBufPtr + xdx_hdmi_audio_ptr->last_bytes), (void*)((uint64_t)runtime->dma_area + xdx_hdmi_audio_ptr->pos), dma_bytes);
		xdx_hdmi_audio_ptr->last_bytes += dma_bytes;
		pos += dma_bytes;
		if (pos >= buffer_bytes)
			pos -= buffer_bytes;
		xdx_hdmi_audio_ptr->pos = pos;

		xdx_hdmi_audio_ptr->transfer_flag = 0;
		//spin_unlock_irq(&xdx_hdmi_audio_ptr->lock);
		mutex_unlock(xdx_hdmi_audio_ptr->transfer_mutex);

		snd_pcm_period_elapsed(xdx_hdmi_audio_ptr->playback_substream);
		return;
	}

	if ((xdx_hdmi_audio_ptr->audioif_video_connect == XDX_AUDIO_MODULE_DP) || xdx_hdmi_audio_ptr->audioif_video_connect == XDX_AUDIO_MODULE_HDMI) {
#if 0
        if(xdx_hdmi_audio_ptr->last_bytes == xdx_hdmi_audio_ptr->blkSize)
        {
            srcVir = (uint64_t)(xdx_hdmi_audio_ptr->tempBufPtr);

            temp_bytes = xdx_hdmi_audio_ptr->blkSize;
            if (xdx_hdmi_audio_prepareDataForDp(xdx_hdmi_audio_ptr, srcVir, (uint64_t)(xdx_hdmi_audio_ptr->periodBufPtr), &temp_bytes) < 0)
            {
                xdx_hdmi_audio_ptr->transfer_flag = 0;
                spin_unlock_irq(&xdx_hdmi_audio_ptr->lock);
                mutex_unlock(xdx_hdmi_audio_ptr->transfer_mutex);
                return;
            }
            dmaCmd.dw0.srcAddr_47_3   = xdx_hdmi_audio_ptr->periodBufPhy >> 3;
        }
        else if(xdx_hdmi_audio_ptr->last_bytes == 0)
        {
           srcVir = (uint64_t)(runtime->dma_area + xdx_hdmi_audio_ptr->pos);

           if (xdx_hdmi_audio_prepareDataForDp(xdx_hdmi_audio_ptr, srcVir, (uint64_t)(xdx_hdmi_audio_ptr->periodBufPtr), &dma_bytes) < 0)
           {
                xdx_hdmi_audio_ptr->transfer_flag = 0;
                spin_unlock_irq(&xdx_hdmi_audio_ptr->lock);
                mutex_unlock(xdx_hdmi_audio_ptr->transfer_mutex);
                return;
           }
           dmaCmd.dw0.srcAddr_47_3   = xdx_hdmi_audio_ptr->periodBufPhy >> 3;
        }
        else
        {
            xdx_hdmi_audio_ptr->pos = pos;
            xdx_hdmi_audio_ptr->transfer_flag = 0;
            spin_unlock_irq(&xdx_hdmi_audio_ptr->lock);
            mutex_unlock(xdx_hdmi_audio_ptr->transfer_mutex);
            snd_pcm_period_elapsed(xdx_hdmi_audio_ptr->playback_substream);
            return;
        }
#endif
		srcVir = (uint64_t)(runtime->dma_area + xdx_hdmi_audio_ptr->pos);

		for (i = 0; i < dma_bytes/B0_AUDIOIF_SUB_FRAME_IN_BYTES; i++) {
			xdx_hdmi_audio_prepareData(xdx_hdmi_audio_ptr, srcVir + i * B0_AUDIOIF_SUB_FRAME_IN_BYTES, (uint64_t)(xdx_hdmi_audio_ptr->periodBufPtr) + i * B0_AUDIOIF_SUB_FRAME_OU_BYTES + 0x100, &dma_bytes);
		}
		dmaCmd.dw0.srcAddr_47_3   = xdx_hdmi_audio_ptr->periodBufPhy >> 3;
	}
#if 0
	} else {
        if(xdx_hdmi_audio_ptr->last_bytes == xdx_hdmi_audio_ptr->blkSize)
        {
            srcVir = (uint64_t)(xdx_hdmi_audio_ptr->tempBufPtr);

            temp_bytes = xdx_hdmi_audio_ptr->blkSize;
            if (xdx_hdmi_audio_prepareDataForHdmi(xdx_hdmi_audio_ptr, srcVir, (uint64_t)(xdx_hdmi_audio_ptr->periodBufPtr), &temp_bytes) < 0)
            {
                xdx_hdmi_audio_ptr->transfer_flag = 0;
                spin_unlock_irq(&xdx_hdmi_audio_ptr->lock);
                mutex_unlock(xdx_hdmi_audio_ptr->transfer_mutex);
                return;
            }
            dmaCmd.dw0.srcAddr_47_3   = xdx_hdmi_audio_ptr->periodBufPhy >> 3;
        }
        else if(xdx_hdmi_audio_ptr->last_bytes == 0)
        {
           srcVir = (uint64_t)(runtime->dma_area + xdx_hdmi_audio_ptr->pos);

           if (xdx_hdmi_audio_prepareDataForHdmi(xdx_hdmi_audio_ptr, srcVir, (uint64_t)(xdx_hdmi_audio_ptr->periodBufPtr), &dma_bytes) < 0)
           {
                xdx_hdmi_audio_ptr->transfer_flag = 0;
                spin_unlock_irq(&xdx_hdmi_audio_ptr->lock);
                mutex_unlock(xdx_hdmi_audio_ptr->transfer_mutex);
                return;
           }
           dmaCmd.dw0.srcAddr_47_3   = xdx_hdmi_audio_ptr->periodBufPhy >> 3;
        }
        else
        {
            xdx_hdmi_audio_ptr->pos = pos;
            xdx_hdmi_audio_ptr->transfer_flag = 0;
            spin_unlock_irq(&xdx_hdmi_audio_ptr->lock);
            mutex_unlock(xdx_hdmi_audio_ptr->transfer_mutex);
            snd_pcm_period_elapsed(xdx_hdmi_audio_ptr->playback_substream);
            return;
        }
		srcVir = (uint64_t)(runtime->dma_area + xdx_hdmi_audio_ptr->pos);

		for(i = 0; i < dma_bytes/B0_AUDIOIF_SUB_FRAME_IN_BYTES; i++)
		{
			xdx_hdmi_audio_prepareData(xdx_hdmi_audio_ptr, srcVir + i * B0_AUDIOIF_SUB_FRAME_IN_BYTES, (uint64_t)(xdx_hdmi_audio_ptr->periodBufPtr) + i * B0_AUDIOIF_SUB_FRAME_OU_BYTES + 0x100, &dma_bytes);
		}
		dmaCmd.dw0.srcAddr_47_3   = xdx_hdmi_audio_ptr->periodBufPhy >> 3;
	}
#endif

	dmaCmd.dw0.opCode = 0x80;           // Operation Code
	dmaCmd.dw0.LogicOp = 0;
	dmaCmd.dw0.Tag = tag;	// Command Tag
	dmaCmd.dw1.LenUnit = 0;
	dmaCmd.dw2.PcieFuncNum = 0;
	dmaCmd.dw2.PcieDontSnoop = 0;
	dmaCmd.dw2.IntrEn = 0x1;

	//before transfer, judge wheteher audio data lenght exceed the limit
	if (dma_bytes > AUDIO_RESERVE_VRAM_LIMIT) {
		dev_err(xdx_hdmi_audio_ptr->dev, "dma_bytes exceed AUDIO_RESERVE_VRAM_LIMIT!");
		return;
	}

	if (XDX_B0_AUDIOIF_BUF_PTR_1 == xdx_hdmi_audio_ptr->buf_ptr) {
		dmaCmd.dw1.DstAddr_47_3 = (XDX_GDDR0_BASE_ADDR + xdx_b0_audioif_buf_dst_addr_2)>>3;
		xdx_hdmi_audio_ptr->buf_ptr = XDX_B0_AUDIOIF_BUF_PTR_2;
	} else {
		dmaCmd.dw1.DstAddr_47_3 = (XDX_GDDR0_BASE_ADDR + xdx_b0_audioif_buf_dst_addr_1)>>3;
		xdx_hdmi_audio_ptr->buf_ptr = XDX_B0_AUDIOIF_BUF_PTR_1;
	}
	dmaCmd.dw2.DstAddrMode = 0xf;
	dmaCmd.dw2.SrcAddrMode = 0x0;
	XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev,"[%s +%d] (uint64_t)(dmaCmd.dw0.SrcAddr_47_3)=0x%llx, (uint64_t)(dmaCmd.dw1.DstAddr_47_3)=0x%llx\n", __func__, __LINE__, (uint64_t)(dmaCmd.dw0.srcAddr_47_3), (uint64_t)(dmaCmd.dw1.DstAddr_47_3));

	dmaCmd.dw1.Xstep_SrcDescNum = (dma_bytes * B0_AUDIOIF_SUB_FRAME_OU_BYTES/B0_AUDIOIF_SUB_FRAME_IN_BYTES/8) - 1;
	dmaCmd.dw2.Ystep_DstDescNum_DstXferLen_DWriteLenHigh = 0;
	dmaCmd.dw2.DstrideHigh_PatternLoc = 0;
	dmaCmd.dw2.DstrideLow_PatternLen = 0;
	dmaCmd.dw2.Sstride_DWriteLenLow = 0;

	++tag;
	tag &= 63;
	xdx_hdmi_audio_ptr->dmaTag = tag;
	xdx_hdmi_audio_ptr->dma_bytes = dma_bytes;
	if (xdx_hdmi_audio_ptr->last_bytes != 0) {
		xdx_hdmi_audio_ptr->fifo_bytes += xdx_hdmi_audio_ptr->blkSize;
		xdx_hdmi_audio_ptr->last_bytes = 0;
	} else
		xdx_hdmi_audio_ptr->fifo_bytes += 16 * 1024;
	xdx_hdmi_audio_ptr->fifo_empty = 0;
	xdx_hdmi_audio_ptr->dma_pending_cnt++;

	xdx_hdmi_audio_ptr->transfer_flag = 0;
//    spin_unlock_irq(&xdx_hdmi_audio_ptr->lock);
	mutex_unlock(xdx_hdmi_audio_ptr->transfer_mutex);


	for (i = 0; i < 6; ++i) {
		iowrite32(dmaCmd.cmd32[i], xdx_hdmi_audio_ptr->iobase_vdma1 + 0x2800 + 0x400 + i*4);
	}
	XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s -%d] Write VDMA register.\n", __func__, __LINE__);
}

static int snd_xdx_hdmi_audio_trigger(struct snd_pcm_substream *substream,
                               int cmd)
{
	struct snd_xdx_hdmi_audio *xdx_hdmi_audio_ptr = snd_pcm_substream_chip(substream);
	//struct snd_pcm_runtime *runtime = substream->runtime;
	uint32_t time_close, timeout_close,time_now;
	struct xdx_device *xdev = xdx_hdmi_audio_ptr->xdev;

	switch (cmd) {
		case SNDRV_PCM_TRIGGER_START:
			XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d] start\n", __func__, __LINE__);

			mutex_lock(xdx_hdmi_audio_ptr->transfer_mutex);
			xdx_hdmi_audio_ptr->status = cPlayBackStart;
			mutex_unlock(xdx_hdmi_audio_ptr->transfer_mutex);

			xdx_hdmi_audio_ptr->pos = 0;
			xdx_hdmi_audio_ptr->fifo_bytes = 0;
			xdx_hdmi_audio_ptr->fifo_empty = 1;
			xdx_hdmi_audio_ptr->dma_pending_cnt = 0;

			xdx_hdmi_audio_ptr->transfer_flag = 1;
			(void) queue_work(xdx_hdmi_audio_ptr->pWorkQueue, &xdx_hdmi_audio_ptr->audioIntWork);
			break;

		case SNDRV_PCM_TRIGGER_STOP:
			XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d] stop\n", __func__, __LINE__);

			mutex_lock(xdx_hdmi_audio_ptr->transfer_mutex);
			xdx_hdmi_audio_ptr->status = cPlayBackStop;
			mutex_unlock(xdx_hdmi_audio_ptr->transfer_mutex);

			if (xdev->asic_type == CHIP_PANGU_B0) {
				OSWriteHWReg32(xdx_hdmi_audio_ptr->iobase_audio0, 0x10, 0x0);

				timeout_close = msecs_to_jiffies(300);
				time_close = get_jiffies_64();
				while (!(xdx_hdmi_audio_ptr->fifo_empty) || xdx_hdmi_audio_ptr->dma_pending_cnt) {
					udelay(10);
					time_now = get_jiffies_64();
					if (time_now > time_close + timeout_close) {
						XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d]Wait stop fifo_empty timeout!\n", __func__, __LINE__);
						break;
					}
				}
			}

			break;

		case SNDRV_PCM_TRIGGER_SUSPEND:
			XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d] suspend\n", __func__, __LINE__);

			mutex_lock(xdx_hdmi_audio_ptr->transfer_mutex);
			xdx_hdmi_audio_ptr->status = cPlayBackStop;
			timeout_close = msecs_to_jiffies(300);
			time_close = get_jiffies_64();
			while (!(xdx_hdmi_audio_ptr->fifo_empty)) {
				udelay(10);
				time_now = get_jiffies_64();
				if (time_now > time_close + timeout_close) {
					XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d]Wait fifo_empty timeout!\n", __func__, __LINE__);
					break;
				}
			}
			mutex_unlock(xdx_hdmi_audio_ptr->transfer_mutex);
			XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d]\n", __func__, __LINE__);
			break;

		default:
			return -EINVAL;
	}

	return 0;
}

static snd_pcm_uframes_t snd_xdx_hdmi_audio_playback_pointer(struct snd_pcm_substream *substream)
{
	struct snd_xdx_hdmi_audio *xdx_hdmi_audio_ptr = snd_pcm_substream_chip(substream);
	struct snd_pcm_runtime *runtime = substream->runtime;
	snd_pcm_uframes_t pos;

	pos = bytes_to_frames(runtime, xdx_hdmi_audio_ptr->pos);

	XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "%s %d pos bytes=0x%x frames=0x%lx\n", __func__, __LINE__, xdx_hdmi_audio_ptr->pos, pos);

	if ((xdx_hdmi_audio_ptr->fifo_empty == 1) && (xdx_hdmi_audio_ptr->dma_pending_cnt == 0) && (xdx_hdmi_audio_ptr->transfer_flag == 0) && (xdx_hdmi_audio_ptr->status == cPlayBackStart)) {

		xdx_hdmi_audio_ptr->transfer_flag = 1;
		(void) queue_work(xdx_hdmi_audio_ptr->pWorkQueue, &xdx_hdmi_audio_ptr->audioIntWork);
		XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "%s %d Continuous transmission of data streams\n", __func__, __LINE__);
	}
	return pos;
}

static const struct snd_pcm_ops snd_xdx_hdmi_audio_playback_ops = {
	.open      =	snd_xdx_hdmi_audio_playback_open,
	.close     =	snd_xdx_hdmi_audio_playback_close,
	.ioctl     =	snd_pcm_lib_ioctl,
	.hw_params =	snd_xdx_hdmi_audio_hw_params,
	.hw_free   =	snd_xdx_hdmi_audio_hw_free,
	.prepare   =	snd_xdx_hdmi_audio_prepare,
	.trigger   =	snd_xdx_hdmi_audio_trigger,
	.pointer   =	snd_xdx_hdmi_audio_playback_pointer,
};

static void pangu_disable_irq(struct snd_xdx_hdmi_audio *xdx_hdmi_audio_ptr)
{
	int i = 0;
	int irqvec=0;
	for (i = 0; i < xdx_hdmi_audio_ptr->irq; i++) {
		irqvec = pci_irq_vector(xdx_hdmi_audio_ptr->pci, i); 
		free_irq(irqvec, xdx_hdmi_audio_ptr);
	}
	pci_free_irq_vectors(xdx_hdmi_audio_ptr->pci);
}

static int snd_xdx_hdmi_audio_free(struct snd_xdx_hdmi_audio *xdx_hdmi_audio_ptr)
{
	struct xdx_device *xdev = xdx_hdmi_audio_ptr->xdev;
	XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d]\n", __func__, __LINE__);

	destroy_workqueue(xdx_hdmi_audio_ptr->pWorkQueue);

	if (xdx_hdmi_audio_ptr->irq >= 0) {
		pangu_disable_irq(xdx_hdmi_audio_ptr);
		xdx_hdmi_audio_ptr->irq = -1;
	}

	snd_device_free(xdx_hdmi_audio_ptr->card, xdx_hdmi_audio_ptr);

	if (xdx_hdmi_audio_ptr->iobase != NULL) {
		iounmap(xdx_hdmi_audio_ptr->iobase);
		xdx_hdmi_audio_ptr->iobase = NULL;
	}

	if (xdx_hdmi_audio_ptr->dma_period.area) {
		snd_dma_free_pages(&xdx_hdmi_audio_ptr->dma_period);
		xdx_hdmi_audio_ptr->dma_period.area = NULL;
	}

	if (xdx_hdmi_audio_ptr->transfer_mutex != NULL)
		kfree(xdx_hdmi_audio_ptr->transfer_mutex);

	if (xdx_hdmi_audio_ptr->hw_mutex != NULL)
		kfree(xdx_hdmi_audio_ptr->hw_mutex);

	if (xdx_hdmi_audio_ptr->tempBuf.area) {
		snd_dma_free_pages(&xdx_hdmi_audio_ptr->tempBuf);
		xdx_hdmi_audio_ptr->tempBuf.area = NULL;
	}

	if (xdx_hdmi_audio_ptr->iobase_audio0 != NULL) {
		iounmap(xdx_hdmi_audio_ptr->iobase_audio0);
		xdx_hdmi_audio_ptr->iobase_audio0 = NULL;
	}

	if (xdx_hdmi_audio_ptr->iobase_vdma1 != NULL) {
		iounmap(xdx_hdmi_audio_ptr->iobase_vdma1);
		xdx_hdmi_audio_ptr->iobase_vdma1 = NULL;
	}

	if (xdx_hdmi_audio_ptr->iobase_acpu != NULL) {
		iounmap(xdx_hdmi_audio_ptr->iobase_acpu);
		xdx_hdmi_audio_ptr->iobase_acpu = NULL;
	}

	if (xdx_hdmi_audio_ptr->iobase_dp0 != NULL) {
		iounmap(xdx_hdmi_audio_ptr->iobase_dp0);
		xdx_hdmi_audio_ptr->iobase_dp0 = NULL;
	}

	if (xdev->shareram_size == XDX_AUDIO_BAR4_16M) {
		pci_release_regions(xdx_hdmi_audio_ptr->pci);
		pci_disable_device(xdx_hdmi_audio_ptr->pci);
	}
	kfree(xdx_hdmi_audio_ptr);

	return 0;
}

static int snd_xdx_hdmi_audio_dev_free(struct snd_device *device)
{
	struct snd_xdx_hdmi_audio *xdx_hdmi_audio_ptr = device->device_data;
	XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d] Freeing device\n", __func__, __LINE__);

	return snd_xdx_hdmi_audio_free(xdx_hdmi_audio_ptr);
}

static void xdx_hdmi_audio_int_work(struct work_struct *data)
{
	struct snd_xdx_hdmi_audio *xdx_hdmi_audio_ptr = container_of(data, struct snd_xdx_hdmi_audio, audioIntWork);
	struct xdx_device *xdev = xdx_hdmi_audio_ptr->xdev;

	if (xdev->asic_type == CHIP_PANGU_A0)
		xdx_hdmi_audio_dma_one_period_a0(xdx_hdmi_audio_ptr);
	else if (xdev->asic_type == CHIP_PANGU_B0)
	        xdx_hdmi_audio_dma_one_period_b0(xdx_hdmi_audio_ptr);
}

static irqreturn_t pangu_irq_handler(int irq, void *data)
{
	irqreturn_t ret = IRQ_NONE;
	struct snd_xdx_hdmi_audio *xdx_hdmi_audio_ptr = (struct snd_xdx_hdmi_audio *)data;
	uint64_t intStatus;
	uint32_t period_bytes;
	uint32_t buffer_bytes;
	DBCmd_a0f1_T dbCmd_a0f1;

	XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d] irq %d index %d\n", __func__, __LINE__, irq, pci_irq_vector(xdx_hdmi_audio_ptr->pci, PANGU_INTERRUPT_H1_START));

//	XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "%s %d intStatus=0x%llx\n", __func__, __LINE__, intStatus);

	if (!xdx_hdmi_audio_ptr->playback_substream->runtime) {
		/* workqueue to transfer data is destroyed during system poweroff */
		return IRQ_HANDLED;
	}

	period_bytes = snd_pcm_lib_period_bytes(xdx_hdmi_audio_ptr->playback_substream);
	buffer_bytes = snd_pcm_lib_buffer_bytes(xdx_hdmi_audio_ptr->playback_substream);

	if (irq == pci_irq_vector(xdx_hdmi_audio_ptr->pci, PANGU_INTERRUPT_ACPU)) {
		intStatus = OSReadHWReg64(xdx_hdmi_audio_ptr->iobase, 0x1002e8);
		dbCmd_a0f1.all = OSReadHWReg64(xdx_hdmi_audio_ptr->iobase, 0x100540);

		XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "%s %d dbCmd_a0f1.cmd=%d\n", __func__, __LINE__, dbCmd_a0f1.cmd);

		if (dbCmd_a0f1.cmd == 2) {
			//XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "aempty %s %d period_bytes=0x%x buffer_bytes=0x%x fifo_bytes=0x%x pos=0x%x status=%d pending_cnt=%d\n", __func__, __LINE__, period_bytes, buffer_bytes, xdx_hdmi_audio_ptr->fifo_bytes, xdx_hdmi_audio_ptr->pos, xdx_hdmi_audio_ptr->status, xdx_hdmi_audio_ptr->dma_pending_cnt);
			xdx_hdmi_audio_ptr->fifo_bytes = 16*1024;
		} else if (dbCmd_a0f1.cmd == 3) {
			XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "empty %s %d period_bytes=0x%x buffer_bytes=0x%x fifo_bytes=0x%x pos=0x%x status=%d peding_cnt=%d\n", __func__, __LINE__, period_bytes, buffer_bytes, xdx_hdmi_audio_ptr->fifo_bytes, xdx_hdmi_audio_ptr->pos, xdx_hdmi_audio_ptr->status, xdx_hdmi_audio_ptr->dma_pending_cnt);
			xdx_hdmi_audio_ptr->fifo_bytes = 0;
			xdx_hdmi_audio_ptr->fifo_empty = 1;
		} else if (dbCmd_a0f1.cmd == 4)
			XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s -%d] iis idle\n", __func__, __LINE__);
		else {
			XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s -%d] dbCmd_a0f1.cmd is error!\n", __func__, __LINE__);
			return IRQ_HANDLED;
		}

		OSWriteHWReg64(xdx_hdmi_audio_ptr->iobase, 0x1001c8, intStatus);
		dbCmd_a0f1.low32 |= 1;

		OSWriteHWReg32(xdx_hdmi_audio_ptr->iobase, 0x100540, dbCmd_a0f1.low32);
	}

	if (irq == pci_irq_vector(xdx_hdmi_audio_ptr->pci, PANGU_INTERRUPT_VD1MCU)) {
		intStatus = OSReadHWReg64(xdx_hdmi_audio_ptr->iobase, 0xc00328);
		dbCmd_a0f1.all = OSReadHWReg64(xdx_hdmi_audio_ptr->iobase, 0xc00430);

		XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "%s %d dbCmd_a0f0.cmd=%d\n", __func__, __LINE__, dbCmd_a0f1.cmd);

		if (dbCmd_a0f1.cmd==1) {
			xdx_hdmi_audio_ptr->pos += xdx_hdmi_audio_ptr->dma_bytes;
			if(xdx_hdmi_audio_ptr->pos >= buffer_bytes)
				xdx_hdmi_audio_ptr->pos -= buffer_bytes;

			snd_pcm_period_elapsed(xdx_hdmi_audio_ptr->playback_substream);
			if (xdx_hdmi_audio_ptr->dma_pending_cnt != 1)
				XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s -%d] Receive invalid interrupt.\n", __func__, __LINE__);
			else
				xdx_hdmi_audio_ptr->dma_pending_cnt = 0;

		} else
			XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s -%d] Receive DMA interrupt is not 1.\n", __func__, __LINE__);

		OSWriteHWReg64(xdx_hdmi_audio_ptr->iobase, 0xc00208, intStatus);
		dbCmd_a0f1.low32 |= 1;

		OSWriteHWReg32(xdx_hdmi_audio_ptr->iobase, 0xc00430, dbCmd_a0f1.low32);
	}

	if (((xdx_hdmi_audio_ptr->fifo_bytes + period_bytes) <= buffer_bytes) && (xdx_hdmi_audio_ptr->status == cPlayBackStart) && (xdx_hdmi_audio_ptr->dma_pending_cnt == 0)) {
		xdx_hdmi_audio_ptr->transfer_flag = 1;
		(void) queue_work(xdx_hdmi_audio_ptr->pWorkQueue, &xdx_hdmi_audio_ptr->audioIntWork);
	} else
		XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "%s %d irq %d period_bytes=0x%x buffer_bytes=0x%x fifo_bytes=0x%x pos=0x%x status=%d dma_pending_cnt=%d\n", __func__, __LINE__, irq, period_bytes, buffer_bytes, xdx_hdmi_audio_ptr->fifo_bytes, xdx_hdmi_audio_ptr->pos, xdx_hdmi_audio_ptr->status, xdx_hdmi_audio_ptr->dma_pending_cnt);


	ret = IRQ_HANDLED;

	return ret;
}


void pangu_a0_db_interrupt(struct snd_xdx_hdmi_audio *xdx_hdmi_audio_ptr)
{
	unsigned long flags;
	uint64_t intStatus;
	uint32_t period_bytes;
	uint32_t buffer_bytes;
	DBCmd_T dbCmd;

	XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d]  index %d\n", __func__, __LINE__,  pci_irq_vector(xdx_hdmi_audio_ptr->pci, PANGU_INTERRUPT_H1_START));
	spin_lock_irqsave(&xdx_hdmi_audio_ptr->lock_audio, flags);

	if (!xdx_hdmi_audio_ptr->playback_substream->runtime) {
		/* workqueue to transfer data is destroyed during system poweroff */
		spin_unlock_irqrestore(&xdx_hdmi_audio_ptr->lock_audio, flags);
		return;
	}

	period_bytes = snd_pcm_lib_period_bytes(xdx_hdmi_audio_ptr->playback_substream);
	buffer_bytes = snd_pcm_lib_buffer_bytes(xdx_hdmi_audio_ptr->playback_substream);

	intStatus = OSReadHWReg64(xdx_hdmi_audio_ptr->iobase_acpu, 0x1002e0);
	dbCmd.all = OSReadHWReg64(xdx_hdmi_audio_ptr->iobase_acpu, 0x100540);

	XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "%s %d dbCmd.cmd=%d\n", __func__, __LINE__, dbCmd.cmd);

	if (dbCmd.cmd == 2) {
		XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "aempty %s %d period_bytes=0x%x buffer_bytes=0x%x fifo_bytes=0x%x pos=0x%x status=%d peding_cnt=%d\n", __func__, __LINE__, period_bytes, buffer_bytes, xdx_hdmi_audio_ptr->fifo_bytes, xdx_hdmi_audio_ptr->pos, xdx_hdmi_audio_ptr->status, xdx_hdmi_audio_ptr->dma_pending_cnt);
		xdx_hdmi_audio_ptr->fifo_bytes = 16*1024;
	}
	else if (dbCmd.cmd == 3) {
		XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "empty %s %d period_bytes=0x%x buffer_bytes=0x%x fifo_bytes=0x%x pos=0x%x status=%d peding_cnt=%d\n", __func__, __LINE__, period_bytes, buffer_bytes, xdx_hdmi_audio_ptr->fifo_bytes, xdx_hdmi_audio_ptr->pos, xdx_hdmi_audio_ptr->status, xdx_hdmi_audio_ptr->dma_pending_cnt);
		mutex_lock(xdx_hdmi_audio_ptr->transfer_mutex);
		xdx_hdmi_audio_ptr->fifo_bytes = 0;
		xdx_hdmi_audio_ptr->fifo_empty = 1;
		mutex_unlock(xdx_hdmi_audio_ptr->transfer_mutex);
	} else if (dbCmd.cmd == 4) {
		XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s -%d] iis idle\n", __func__, __LINE__);
	} else {
		XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s -%d] dbCmd.cmd is error!\n", __func__, __LINE__);
		spin_unlock_irqrestore(&xdx_hdmi_audio_ptr->lock_audio, flags);
		return;
	}

	OSWriteHWReg64(xdx_hdmi_audio_ptr->iobase_acpu, 0x1001c0, intStatus);
	dbCmd.low32 |= 1;

	OSWriteHWReg32(xdx_hdmi_audio_ptr->iobase_acpu, 0x100540, dbCmd.low32);

	if (((xdx_hdmi_audio_ptr->fifo_bytes + period_bytes) <= buffer_bytes) && (xdx_hdmi_audio_ptr->status == cPlayBackStart) && (xdx_hdmi_audio_ptr->dma_pending_cnt == 0)) {
		xdx_hdmi_audio_ptr->transfer_flag = 1;
		(void) queue_work(xdx_hdmi_audio_ptr->pWorkQueue, &xdx_hdmi_audio_ptr->audioIntWork);
	} else {
		XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "%s %d period_bytes=0x%x buffer_bytes=0x%x fifo_bytes=0x%x pos=0x%x status=%d dma_pending_cnt=%d\n", __func__, __LINE__, period_bytes, buffer_bytes, xdx_hdmi_audio_ptr->fifo_bytes, xdx_hdmi_audio_ptr->pos, xdx_hdmi_audio_ptr->status, xdx_hdmi_audio_ptr->dma_pending_cnt);
	}

	spin_unlock_irqrestore(&xdx_hdmi_audio_ptr->lock_audio, flags);

}

static void pangu_a0_vdma1ch3_irq_handler(struct snd_xdx_hdmi_audio *xdx_hdmi_audio_ptr)
{
	unsigned long flags;
	uint64_t intStatus = 0x80090;
	uint32_t period_bytes;
	uint32_t buffer_bytes;
	XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d] \n", __func__, __LINE__);
	spin_lock_irqsave(&xdx_hdmi_audio_ptr->lock_vdma1ch3, flags);
	if (!xdx_hdmi_audio_ptr->playback_substream->runtime) {
		/* workqueue to transfer data is destroyed during system poweroff */
		spin_unlock_irqrestore(&xdx_hdmi_audio_ptr->lock_vdma1ch3, flags);
		return;
	}

	period_bytes = snd_pcm_lib_period_bytes(xdx_hdmi_audio_ptr->playback_substream);
	buffer_bytes = snd_pcm_lib_buffer_bytes(xdx_hdmi_audio_ptr->playback_substream);

	intStatus = OSReadHWReg64(xdx_hdmi_audio_ptr->iobase_vdma1, 0x2800 + 0x10);
	XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "%s %d intStatus=0x%llx\n", __func__, __LINE__, intStatus);

	if (intStatus != 0) {
		XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "%s %d \n", __func__, __LINE__);

		xdx_hdmi_audio_ptr->pos += xdx_hdmi_audio_ptr->dma_bytes;
		if (xdx_hdmi_audio_ptr->pos >= buffer_bytes)
			xdx_hdmi_audio_ptr->pos -= buffer_bytes;

		snd_pcm_period_elapsed(xdx_hdmi_audio_ptr->playback_substream);
		if (xdx_hdmi_audio_ptr->dma_pending_cnt != 1)
			XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s -%d] Receive invalid interrupt.\n", __func__, __LINE__);
		else {
			mutex_lock(xdx_hdmi_audio_ptr->transfer_mutex);
			xdx_hdmi_audio_ptr->dma_pending_cnt = 0;
			mutex_unlock(xdx_hdmi_audio_ptr->transfer_mutex);
		}

	}
	OSWriteHWReg64(xdx_hdmi_audio_ptr->iobase_vdma1, 0x2800 + 0x40, intStatus);

	if (((xdx_hdmi_audio_ptr->fifo_bytes + period_bytes) <= buffer_bytes) && (xdx_hdmi_audio_ptr->status == cPlayBackStart) && (xdx_hdmi_audio_ptr->dma_pending_cnt) == 0) {
		xdx_hdmi_audio_ptr->transfer_flag = 1;
		(void) queue_work(xdx_hdmi_audio_ptr->pWorkQueue, &xdx_hdmi_audio_ptr->audioIntWork);
	} else
		XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "%s %d  period_bytes=0x%x buffer_bytes=0x%x fifo_bytes=0x%x pos=0x%x status=%d atomic_read(&xdx_hdmi_audio_ptr->dma_pending_cnt)=%d\n", __func__, __LINE__,  period_bytes, buffer_bytes, xdx_hdmi_audio_ptr->fifo_bytes, xdx_hdmi_audio_ptr->pos, xdx_hdmi_audio_ptr->status, xdx_hdmi_audio_ptr->dma_pending_cnt);

	spin_unlock_irqrestore(&xdx_hdmi_audio_ptr->lock_vdma1ch3, flags);
}


static void pangu_audio0_irq_handler(struct xdx_irq_src *src, void *data)
{
	struct snd_xdx_hdmi_audio *xdx_hdmi_audio_ptr = (struct snd_xdx_hdmi_audio *)data;
	uint32_t period_bytes;
	uint32_t buffer_bytes;
	unsigned long flags;
	uint32_t *pReg = NULL;


	spin_lock_irqsave(&xdx_hdmi_audio_ptr->lock_audio, flags);
	XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev,"[%s %d]\n", __func__, __LINE__);

	period_bytes = snd_pcm_lib_period_bytes(xdx_hdmi_audio_ptr->playback_substream);
	buffer_bytes = snd_pcm_lib_buffer_bytes(xdx_hdmi_audio_ptr->playback_substream);
	

	pReg = (uint32_t *)(xdx_hdmi_audio_ptr->iobase_audio0 + XDX_AUDIOIF0_INTR_STS);

	if (*pReg & 4) {
		if (1 == xdx_hdmi_audio_ptr->dma_pending_cnt) {
			mutex_lock(xdx_hdmi_audio_ptr->transfer_mutex);
			xdx_hdmi_audio_ptr->dma_pending_cnt = 0;
			mutex_unlock(xdx_hdmi_audio_ptr->transfer_mutex);
		}
		xdx_hdmi_audio_ptr->fifo_bytes = 16*1024;
		pReg = (uint32_t *)(xdx_hdmi_audio_ptr->iobase_audio0 + XDX_AUDIOIF0_INTR_STS_CLR);
		*pReg |= 4;
		pReg = (uint32_t *)(xdx_hdmi_audio_ptr->iobase_audio0 + XDX_AUDIOIF0_INTR_STS);

		if (*pReg & 2) {
			pReg = (uint32_t *)(xdx_hdmi_audio_ptr->iobase_audio0 + XDX_AUDIOIF0_INTR_STS_CLR);
			*pReg |= 2;
			XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "empty %s %d period_bytes=0x%x buffer_bytes=0x%x fifo_bytes=0x%x pos=0x%x status=%d peding_cnt=%d\n", __func__, __LINE__, period_bytes, buffer_bytes, xdx_hdmi_audio_ptr->fifo_bytes, xdx_hdmi_audio_ptr->pos, xdx_hdmi_audio_ptr->status, xdx_hdmi_audio_ptr->dma_pending_cnt);
		}

	} else if (*pReg & 2) {
		pReg = (uint32_t *)(xdx_hdmi_audio_ptr->iobase_audio0 + XDX_AUDIOIF0_INTR_STS_CLR);
		*pReg |= 2;
		XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "empty %s %d period_bytes=0x%x buffer_bytes=0x%x fifo_bytes=0x%x pos=0x%x status=%d peding_cnt=%d\n", __func__, __LINE__, period_bytes, buffer_bytes, xdx_hdmi_audio_ptr->fifo_bytes, xdx_hdmi_audio_ptr->pos, xdx_hdmi_audio_ptr->status, xdx_hdmi_audio_ptr->dma_pending_cnt);

	} else if (*pReg & 1) {
		pReg = (uint32_t *)(xdx_hdmi_audio_ptr->iobase_audio0 + XDX_AUDIOIF0_INTR_STS_CLR);
		*pReg |= 1;

		mutex_lock(xdx_hdmi_audio_ptr->transfer_mutex);
		xdx_hdmi_audio_ptr->fifo_bytes = 0;
		xdx_hdmi_audio_ptr->fifo_empty = 1;
		mutex_unlock(xdx_hdmi_audio_ptr->transfer_mutex);

		XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s +%d] iis idle\n", __func__, __LINE__);
	}
//		pReg = (uint32_t *)(xdx_hdmi_audio_ptr->iobase_audio0 + XDX_AUDIOIF0_INTR_STS);


	if (((xdx_hdmi_audio_ptr->fifo_bytes + period_bytes) <= buffer_bytes) && (xdx_hdmi_audio_ptr->status == cPlayBackStart) && (xdx_hdmi_audio_ptr->dma_pending_cnt == 0) && (xdx_hdmi_audio_ptr->vdma1ch3_finish)) {
		xdx_hdmi_audio_ptr->transfer_flag = 1;
		xdx_hdmi_audio_ptr->vdma1ch3_finish = 0;
	        (void) queue_work(xdx_hdmi_audio_ptr->pWorkQueue, &xdx_hdmi_audio_ptr->audioIntWork);
	} else {
		if (xdx_hdmi_audio_ptr->status == cPlayBackStart)
			XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "%s %d  period_bytes=0x%x buffer_bytes=0x%x fifo_bytes=0x%x pos=0x%x status=%d dma_pending_cnt=%d\n", __func__, __LINE__,  period_bytes, buffer_bytes, xdx_hdmi_audio_ptr->fifo_bytes, xdx_hdmi_audio_ptr->pos, xdx_hdmi_audio_ptr->status, xdx_hdmi_audio_ptr->dma_pending_cnt);
		else
			xdx_hdmi_audio_ptr->vdma1ch3_finish = 0;
	}

	spin_unlock_irqrestore(&xdx_hdmi_audio_ptr->lock_audio, flags);

}

static void pangu_vdma1ch3_irq_handler(struct xdx_irq_src *src, void *data)
{
	struct snd_xdx_hdmi_audio *xdx_hdmi_audio_ptr = (struct snd_xdx_hdmi_audio *)data;
	uint32_t period_bytes;
	uint32_t buffer_bytes;
	unsigned long flags;
	uint64_t intStatus = 0;

	spin_lock_irqsave(&xdx_hdmi_audio_ptr->lock_vdma1ch3, flags);
	XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev,"[%s %d]\n", __func__, __LINE__);

	period_bytes = snd_pcm_lib_period_bytes(xdx_hdmi_audio_ptr->playback_substream);
	buffer_bytes = snd_pcm_lib_buffer_bytes(xdx_hdmi_audio_ptr->playback_substream);
	intStatus = OSReadHWReg64(xdx_hdmi_audio_ptr->iobase_vdma1, 0x2800 + 0x10);

	// then check waitfor
	if (0 != intStatus) {
		if (xdx_hdmi_audio_ptr->dma_pending_cnt == 1) {
                        //xdx_hdmi_audio_ptr->pos += xdx_hdmi_audio_ptr->dma_bytes;
			xdx_hdmi_audio_ptr->pos += 12*1024;
			if (xdx_hdmi_audio_ptr->pos >= buffer_bytes) {
				xdx_hdmi_audio_ptr->pos -= buffer_bytes;
			}
			snd_pcm_period_elapsed(xdx_hdmi_audio_ptr->playback_substream);

			mutex_lock(xdx_hdmi_audio_ptr->transfer_mutex);
			xdx_hdmi_audio_ptr->dma_pending_cnt = 0;
			mutex_unlock(xdx_hdmi_audio_ptr->transfer_mutex);

			XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s +%d] xdx_hdmi_audio_ptr->buf_ptr=0x%x\n", __func__, __LINE__, xdx_hdmi_audio_ptr->buf_ptr);

			OSWriteHWReg32(xdx_hdmi_audio_ptr->iobase_audio0, 0x8, xdx_hdmi_audio_ptr->buf_ptr);   //HDMI0 DP0

			OSWriteHWReg32(xdx_hdmi_audio_ptr->iobase_audio0, 0xc, 0x1);
		} else
                        XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s +%d] Recieved double isr\n", __func__, __LINE__);

                OSWriteHWReg64(xdx_hdmi_audio_ptr->iobase_vdma1, 0x2800 + 0x40, intStatus);
	}

	if (!(OSReadHWReg32(xdx_hdmi_audio_ptr->iobase_audio0, 0x10) & 0x1))
		OSWriteHWReg32(xdx_hdmi_audio_ptr->iobase_audio0, 0x10, 0x1);

	xdx_hdmi_audio_ptr->vdma1ch3_finish = 1;

	if (((xdx_hdmi_audio_ptr->fifo_bytes + period_bytes) <= buffer_bytes) && (xdx_hdmi_audio_ptr->status == cPlayBackStart) && (xdx_hdmi_audio_ptr->dma_pending_cnt == 0) && (xdx_hdmi_audio_ptr->vdma1ch3_finish)) {
		xdx_hdmi_audio_ptr->transfer_flag = 1;
		xdx_hdmi_audio_ptr->vdma1ch3_finish = 0;
		(void) queue_work(xdx_hdmi_audio_ptr->pWorkQueue, &xdx_hdmi_audio_ptr->audioIntWork);
	} else {
		if (xdx_hdmi_audio_ptr->status == cPlayBackStart)
			XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "%s %d  period_bytes=0x%x buffer_bytes=0x%x fifo_bytes=0x%x pos=0x%x status=%d dma_pending_cnt=%d\n", __func__, __LINE__,  period_bytes, buffer_bytes, xdx_hdmi_audio_ptr->fifo_bytes, xdx_hdmi_audio_ptr->pos, xdx_hdmi_audio_ptr->status, xdx_hdmi_audio_ptr->dma_pending_cnt);
		else
			xdx_hdmi_audio_ptr->vdma1ch3_finish = 0;

	}

	spin_unlock_irqrestore(&xdx_hdmi_audio_ptr->lock_vdma1ch3, flags);
}

static int xdx_hdmi_audio_createVolumeControl(struct snd_xdx_hdmi_audio *xdx_hdmi_audio_ptr)
{
	int i;
	int err;

	for (i = 0; i < ARRAY_SIZE(snd_xdx_hdmi_control); i++) {
		err = snd_ctl_add(xdx_hdmi_audio_ptr->card, snd_ctl_new1(&snd_xdx_hdmi_control[i], xdx_hdmi_audio_ptr));
		if (err < 0)
			return err;
	}

	return 1;
}




int snd_xdx_hdmi_audio_create(struct snd_card *card, struct xdx_device *xdev, struct snd_xdx_hdmi_audio **rchip)

{
	int err;
	struct snd_xdx_hdmi_audio *xdx_hdmi_audio_ptr;
	static struct snd_device_ops ops =
	{
		.dev_free = snd_xdx_hdmi_audio_dev_free,
	};
	void *pointer;
	void *hw_pointer;
	unsigned iomem_size;
	int i = 0,irq_num = 0,irqvec = 0;
	struct pci_dev *pci = NULL;
    
	* rchip = NULL;

	xdx_hdmi_audio_ptr = kzalloc(sizeof(struct snd_xdx_hdmi_audio), GFP_KERNEL);
	if (xdx_hdmi_audio_ptr == NULL) {
		err = -ENOMEM;
		goto ERROR_XDX_HDMI_AUDIO_CREATE_1;
	}

	if (xdev->shareram_size == XDX_AUDIO_BAR4_16M) {
		pci = xdev->audio_pdev;
		err = pci_enable_device(pci);
	        if (err < 0)
		        return err;
		pci_set_master(pci);
		err = pci_request_regions(pci, "xdx_hdmi_audio");
		if (err<0)
			return err;

	} else if (xdev->shareram_size == XDX_AUDIO_BAR4_1M)
		pci = xdev->pdev;

	xdx_hdmi_audio_ptr->dev = &pci->dev;
	xdx_hdmi_audio_ptr->card = card;
	xdx_hdmi_audio_ptr->pci = pci;

/*
 * uint32_t buffer_size = 32768 * 4;
    xdx_hdmi_audio_ptr->pcmbufferData = (u8 *)kzalloc(buffer_size, GFP_KERNEL);
    if (xdx_hdmi_audio_ptr->pcmbufferData == NULL) {
		err = -ENOMEM;
		goto ERROR_XDX_HDMI_AUDIO_CREATE_3;
    }
*/
	xdx_hdmi_audio_ptr->periodBufSize = AUDIO_FIFO_SIZE;
	if (snd_dma_alloc_pages(SNDRV_DMA_TYPE_DEV, &pci->dev, AUDIO_FIFO_SIZE, &xdx_hdmi_audio_ptr->dma_period) < 0) {
		err = -ENOMEM;
		goto ERROR_XDX_HDMI_AUDIO_CREATE_2;
	}
	xdx_hdmi_audio_ptr->periodBufPtr = (uint8_t*)(xdx_hdmi_audio_ptr->dma_period.area);
	xdx_hdmi_audio_ptr->periodBufPhy = xdx_hdmi_audio_ptr->dma_period.addr;

	if (snd_dma_alloc_pages(SNDRV_DMA_TYPE_DEV, &pci->dev, 6144, &xdx_hdmi_audio_ptr->tempBuf) < 0) {
		err = -ENOMEM;
		goto ERROR_XDX_HDMI_AUDIO_CREATE_3;
	}
	xdx_hdmi_audio_ptr->tempBufPtr = (uint8_t*)(xdx_hdmi_audio_ptr->tempBuf.area);


	xdx_hdmi_audio_ptr->pWorkQueue = create_singlethread_workqueue("xdx_hdmi_audio_misr");
	INIT_WORK(&(xdx_hdmi_audio_ptr->audioIntWork), xdx_hdmi_audio_int_work);

	err = snd_device_new(card, SNDRV_DEV_LOWLEVEL, xdx_hdmi_audio_ptr, &ops);
	if (err < 0) {
		XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s +%d] snd_device_new err=%d\n", __func__, __LINE__, err);
		goto ERROR_XDX_HDMI_AUDIO_CREATE_4;
	}

	err = snd_pcm_new(xdx_hdmi_audio_ptr->card, "xdx hdmi pcm", 0, 1, 0, &xdx_hdmi_audio_ptr->pcm);
	if (err < 0) {
		XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s +%d] snd_pcm_new err=%d\n", __func__, __LINE__, err);
		//goto ERROR_XDX_HDMI_AUDIO_CREATE_5;
		return err;
	}

	pointer = kzalloc(sizeof(struct mutex), GFP_KERNEL);
	if (!pointer) {
		XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s +%d] Allocate memory failed.\n", __func__, __LINE__);
		goto ERROR_XDX_HDMI_AUDIO_CREATE_5;
	}
	xdx_hdmi_audio_ptr->transfer_mutex = pointer;
	mutex_init((struct mutex *)xdx_hdmi_audio_ptr->transfer_mutex);

	hw_pointer = kzalloc(sizeof(struct mutex), GFP_KERNEL);
	if (!hw_pointer) {
		XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s +%d] Allocate memory failed.\n", __func__, __LINE__);
		goto ERROR_XDX_HDMI_AUDIO_CREATE_6;
	}
	xdx_hdmi_audio_ptr->hw_mutex = hw_pointer;
	mutex_init((struct mutex *)xdx_hdmi_audio_ptr->hw_mutex);

	xdx_hdmi_audio_ptr->pcm->private_data = xdx_hdmi_audio_ptr;
	xdx_hdmi_audio_ptr->pcm->private_free = snd_xdx_hdmi_audio_free_pcm;
	strcpy(xdx_hdmi_audio_ptr->pcm->name, "xdx_hdmi_pcm");

	snd_pcm_set_ops(xdx_hdmi_audio_ptr->pcm, SNDRV_PCM_STREAM_PLAYBACK, &snd_xdx_hdmi_audio_playback_ops);

	snd_pcm_lib_preallocate_pages_for_all(xdx_hdmi_audio_ptr->pcm, SNDRV_DMA_TYPE_DEV, &((xdx_hdmi_audio_ptr->pci)->dev), 64*1024, 64*1024);

	if (xdx_hdmi_audio_createVolumeControl(xdx_hdmi_audio_ptr))
		XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s +%d] Create control done.\n", __func__, __LINE__);
	else
		XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s +%d] Create control fail.\n", __func__, __LINE__);

	xdx_hdmi_audio_ptr->audioif_video_connect = XDX_AUDIO_MODULE_NULL;

	//we reserved 64M for ACPU, FW use high 16M, audio use at most 16M, the space is enough
	audioif_reg_val[1][1] = ((xdev->real_vram_size + 0x100000000)>>21);
	xdx_b0_audioif_buf_dst_addr_1 = xdev->real_vram_size;
	xdx_b0_audioif_buf_dst_addr_2 = xdx_b0_audioif_buf_dst_addr_1 + 0x4000;


	if (xdev->shareram_size == XDX_AUDIO_BAR4_16M) {
		xdx_hdmi_audio_ptr->iomem = pci_resource_start(xdx_hdmi_audio_ptr->pci, 0);

		iomem_size = pci_resource_len(xdx_hdmi_audio_ptr->pci, 0);

		pr_info("[%s +%d] resources: iomem = 0x%lx size=0x%x\n", __func__, __LINE__, xdx_hdmi_audio_ptr->iomem, iomem_size);

		xdx_hdmi_audio_ptr->iobase = ioremap(xdx_hdmi_audio_ptr->iomem, iomem_size);
		if (!xdx_hdmi_audio_ptr->iobase) {
			XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s +%d] ioremap failed\n", __func__, __LINE__);
			goto ERROR_XDX_HDMI_AUDIO_CREATE_7;
		}

		irq_num = pci_alloc_irq_vectors(pci, PANGU_INTERRUPT_MAX, PANGU_INTERRUPT_MAX, PCI_IRQ_MSIX);
		XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s +%d] irq_num=%d\n", __func__, __LINE__, irq_num);
		if (irq_num < 0) {
			XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s +%d] Fail to enable MSI/MSIX interrupt ret=%d. PANGU_INTERRUPT_MAX=%d\n", __func__, __LINE__, xdx_hdmi_audio_ptr->irq, PANGU_INTERRUPT_MAX);
			err = irq_num;
			return err;
		}
		for (i = 0; i < irq_num; ++i) {
			irqvec = pci_irq_vector(pci, i);

			XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "%s pci_irq_vector irq=%d, irq_nr=%d\n",__func__, i, irqvec);

			err = request_irq(irqvec, pangu_irq_handler, IRQF_SHARED, XDX_HDMI_AUDIO_SYSTEM_NAME"-msix", xdx_hdmi_audio_ptr);
			if (err < 0) {
				XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s +%d] Fail to request_irq(%d) ret=%#x\n", __func__, __LINE__, i, err);
				return err;
			}
		}

		xdx_hdmi_audio_ptr->irq = irq_num;
	}

	* rchip = xdx_hdmi_audio_ptr;
	return 0;

ERROR_XDX_HDMI_AUDIO_CREATE_7:
	kfree(xdx_hdmi_audio_ptr->hw_mutex);
ERROR_XDX_HDMI_AUDIO_CREATE_6:
	kfree(xdx_hdmi_audio_ptr->transfer_mutex);
ERROR_XDX_HDMI_AUDIO_CREATE_5:
	snd_device_free(card, xdx_hdmi_audio_ptr);
ERROR_XDX_HDMI_AUDIO_CREATE_4:
	snd_dma_free_pages(&xdx_hdmi_audio_ptr->tempBuf);
ERROR_XDX_HDMI_AUDIO_CREATE_3:
	snd_dma_free_pages(&xdx_hdmi_audio_ptr->dma_period);
ERROR_XDX_HDMI_AUDIO_CREATE_2:
	kfree(xdx_hdmi_audio_ptr);
ERROR_XDX_HDMI_AUDIO_CREATE_1:
	return err;

}
/*Audio volume info*/
static int xdx_hdmi_audio_volumeInfo(struct snd_kcontrol *kcontrol, struct snd_ctl_elem_info *uinfo)
{
	uinfo->value.integer.step = 1;
	uinfo->type = SNDRV_CTL_ELEM_TYPE_INTEGER;
	uinfo->count = 1;

	uinfo->value.integer.max = DEFAULT_MAX_VOLUME;
	uinfo->value.integer.min = DEFAULT_MIN_VOLUME;

	return 0;
}

static int  xdx_hdmi_audio_volumeGet(struct snd_kcontrol *kcontrol, struct snd_ctl_elem_value *u)
{
	struct snd_xdx_hdmi_audio *xdx_hdmi_audio_ptr = snd_kcontrol_chip(kcontrol);
	uint8_t hwVolume = 0;

#if 0
    if (xdx_ddcci_writectrl(xdx_hdmi_audio_ptr->hdmi, DEFAULT_DDCCI_ADDR, SPEAKER_SELECT_OFFSET, 0x0))
    {
        printk("[%s %d] Volume get fail[SPEAKER_SELECT].\n", __func__, __LINE__);
        u->value.integer.value[0] = DEFAULT_MIN_VOLUME;

        return 0;
    }
#endif
	if (xdx_hdmi_audio_ptr->audioif_video_connect == XDX_AUDIO_MODULE_HDMI) {
		if (xdx_hdmi_ddcci_dumpctrl(xdx_hdmi_audio_ptr->hdmi, DEFAULT_DDCCI_ADDR, SPEAKER_VOLUME_ADJUST_OFFSET, &hwVolume, NULL) != 0) {
			XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d]HDMI volume get fail[VOLUME_ADJUST].\n", __func__, __LINE__);
			u->value.integer.value[0] = DEFAULT_MIN_VOLUME;

		return 0;
		}
		XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d]HDMI volume get value: %d \n", __func__, __LINE__, hwVolume);
	} else if (xdx_hdmi_audio_ptr->audioif_video_connect == XDX_AUDIO_MODULE_DP) {
		if (xdx_dp_ddcci_dumpctrl(xdx_hdmi_audio_ptr->dp, DEFAULT_DDCCI_ADDR, SPEAKER_VOLUME_ADJUST_OFFSET, &hwVolume, NULL) != 0) {
			XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d]DP volume get fail[VOLUME_ADJUST].\n", __func__, __LINE__);
			u->value.integer.value[0] = DEFAULT_MIN_VOLUME;
			return 0;
		}
		XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d]DP Volume get value: %d \n", __func__, __LINE__, hwVolume);
	} else
		XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d]The current connect video is error!\n", __func__, __LINE__);

	u->value.integer.value[0] = hwVolume;

	return 0;
}
static int xdx_hdmi_audio_volumePut(struct snd_kcontrol *kcontrol, struct snd_ctl_elem_value *u)
{
	struct snd_xdx_hdmi_audio *xdx_hdmi_audio_ptr = snd_kcontrol_chip(kcontrol);
	int change = 0;
	int val = 0;

	val = u->value.integer.value[0];
	XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d] Volume change value: %d \n", __func__, __LINE__, val);
#if 0
	/// 1. select left/right speaker to be adjusted
	if (xdx_ddcci_writectrl(xdx_hdmi_audio_ptr->hdmi, DEFAULT_DDCCI_ADDR, SPEAKER_SELECT_OFFSET, 0x0))
	{
		printk("[%s %d] Volume change fail[SPEAKER_SELECT].\n", __func__, __LINE__);

		return 0;
	}
#endif
	/// 2. Start to adjust
	if (xdx_hdmi_audio_ptr->audioif_video_connect == XDX_AUDIO_MODULE_HDMI) {
		if (!(xdx_hdmi_ddcci_writectrl(xdx_hdmi_audio_ptr->hdmi, DEFAULT_DDCCI_ADDR, SPEAKER_VOLUME_ADJUST_OFFSET, val))) {
			XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d] Volume change fail[VOLUME_ADJUST].\n", __func__, __LINE__);

			return 0;
		}
		XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d]HDMI volume change success: %d \n", __func__, __LINE__, val);
	} else if (xdx_hdmi_audio_ptr->audioif_video_connect == XDX_AUDIO_MODULE_DP) {
		if (!(xdx_dp_ddcci_writectrl(xdx_hdmi_audio_ptr->dp, DEFAULT_DDCCI_ADDR, SPEAKER_VOLUME_ADJUST_OFFSET, val)))
		{
			XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d]DP volume change fail[VOLUME_ADJUST].\n", __func__, __LINE__);

			return 0;
		}
		XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d]DP volume change success: %d \n", __func__, __LINE__, val);
	} else
		XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d]The current connect video is error!\n", __func__, __LINE__);

	change = 1;

	return change;
}

int xdx_audio_drv_hotPlug(struct xdx_device *xdev, PlayDrv module, PlayCon status)
{
	struct snd_xdx_hdmi_audio *xdx_hdmi_audio_ptr = xdev->xdx_hdmi_audio_hotplug_ptr;
	struct xdx_dm *dm = xdev_to_dm(xdev);
	struct xdx_output *output = dm_to_output(dm);
	struct hdmi_dev *hdmi = output->hdmi[0];
	struct dp_dev *dp = output->dp[0];
	int change = 0;   //0 do nothing 1-dp/hdmi 2-hdmi/hdmi
	uint32_t time_close, timeout_close, time_now;
	uint32_t temp_status;
	uint32_t *pReg = NULL;
	Audio_Regs_T *pAudioReg = (Audio_Regs_T *)(xdx_hdmi_audio_ptr->iobase_audio0);

	if (dm->dc_num ==2 ) {
		//todo, add contorl for dc1
	}

	if (xdx_hdmi_audio_ptr->flag == 0) {
		xdx_hdmi_audio_ptr->audioif_hdmi_disconnect = XDX_AUDIO_MODULE_DRV_CONNECT;  //default value hdmi_disconnect
		xdx_hdmi_audio_ptr->audioif_dp_disconnect = XDX_AUDIO_MODULE_DRV_CONNECT;    //default value dp_disconnect
	}

	if (XDX_AUDIO_MODULE_DISCONNECT == status) {
		if (module == XDX_AUDIO_MODULE_HDMI) {
			xdx_hdmi_audio_ptr->audioif_hdmi_connect = XDX_AUDIO_MODULE_NULL;
			xdx_hdmi_audio_ptr->audioif_hdmi_disconnect = XDX_AUDIO_MODULE_DRV_DISCONNECT;
			xdx_hdmi_audio_ptr->flag++;
		} else if (module == XDX_AUDIO_MODULE_DP) {
			xdx_hdmi_audio_ptr->audioif_dp_connect = XDX_AUDIO_MODULE_NULL;
			xdx_hdmi_audio_ptr->audioif_dp_disconnect = XDX_AUDIO_MODULE_DRV_DISCONNECT;
			xdx_hdmi_audio_ptr->flag++;
		}

		xdx_hdmi_audio_ptr->audioif_video_connect = XDX_AUDIO_MODULE_NULL;

		if (xdx_hdmi_audio_ptr->audioif_double_connect) {
			if ((xdx_hdmi_audio_ptr->audioif_hdmi_connect == XDX_AUDIO_MODULE_NULL) && (xdx_hdmi_audio_ptr->audioif_dp_connect)) {
				change = 1;//hdmi->dp
				xdx_hdmi_audio_ptr->audioif_video_connect = XDX_AUDIO_MODULE_DP;
			}
			if ((xdx_hdmi_audio_ptr->audioif_dp_connect == XDX_AUDIO_MODULE_NULL) && (xdx_hdmi_audio_ptr->audioif_hdmi_connect)) {
				change = 0;
				xdx_hdmi_audio_ptr->audioif_video_connect = XDX_AUDIO_MODULE_HDMI;
			}
		}
		XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d]The display status is unplug, the video is disconnect! If double connect is %d, the video connect is %d. \n", __func__, __LINE__, xdx_hdmi_audio_ptr->audioif_double_connect, xdx_hdmi_audio_ptr->audioif_video_connect);
	} else if (XDX_AUDIO_MODULE_CONNECT == status) {
		XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d]The display status is plug! \n", __func__, __LINE__);

		if (xdx_hdmi_audio_ptr->audioif_video_connect == XDX_AUDIO_MODULE_NULL) {

			if (module == XDX_AUDIO_MODULE_HDMI) {
				xdx_hdmi_audio_ptr->audioif_hdmi_connect = XDX_AUDIO_MODULE_HDMI;
				xdx_hdmi_audio_ptr->audioif_video_connect = XDX_AUDIO_MODULE_HDMI;
				xdx_hdmi_audio_ptr->audioif_dp_connect = XDX_AUDIO_MODULE_NULL;
				change = 2;  //hdmi->hdmi

				if (xdx_hdmi_audio_ptr->audioif_dp_disconnect == XDX_AUDIO_MODULE_DRV_DISCONNECT) {
					change =1;   //dp->hdmi
					xdx_hdmi_audio_ptr->audioif_dp_disconnect = XDX_AUDIO_MODULE_DRV_CONNECT;
				}
				XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d]The video connect is %d, the hdmi connect is %d, the change is %d. \n", __func__, __LINE__, xdx_hdmi_audio_ptr->audioif_video_connect, xdx_hdmi_audio_ptr->audioif_hdmi_connect, change);
			} else if (module == XDX_AUDIO_MODULE_DP) {
				xdx_hdmi_audio_ptr->audioif_dp_connect = XDX_AUDIO_MODULE_DP;
				xdx_hdmi_audio_ptr->audioif_video_connect = XDX_AUDIO_MODULE_DP;
				xdx_hdmi_audio_ptr->audioif_hdmi_connect = XDX_AUDIO_MODULE_NULL;
				change = 2;  //dp->dp

				if (xdx_hdmi_audio_ptr->audioif_hdmi_disconnect == XDX_AUDIO_MODULE_DRV_DISCONNECT) {
					change =1;   //hdmi->dp
					xdx_hdmi_audio_ptr->audioif_hdmi_disconnect = XDX_AUDIO_MODULE_DRV_CONNECT;
				}
				XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d]The video connect is %d, the dp connect is %d, the change is %d. \n", __func__, __LINE__, xdx_hdmi_audio_ptr->audioif_video_connect, xdx_hdmi_audio_ptr->audioif_dp_connect, change);
			} else {
				XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d]The module is error! \n", __func__, __LINE__);
				return -1;
			}
		} else if ((xdx_hdmi_audio_ptr->audioif_video_connect == XDX_AUDIO_MODULE_HDMI) || (xdx_hdmi_audio_ptr->audioif_video_connect == XDX_AUDIO_MODULE_DP)) {
			if ((module == XDX_AUDIO_MODULE_DP) && (xdx_hdmi_audio_ptr->audioif_hdmi_connect)) {
				xdx_hdmi_audio_ptr->audioif_video_connect = XDX_AUDIO_MODULE_HDMI;
				xdx_hdmi_audio_ptr->audioif_dp_connect = XDX_AUDIO_MODULE_DP;
				change = 0;
			}
			if ((module == XDX_AUDIO_MODULE_HDMI) && (xdx_hdmi_audio_ptr->audioif_dp_connect)) {
				xdx_hdmi_audio_ptr->audioif_video_connect = XDX_AUDIO_MODULE_HDMI;
				xdx_hdmi_audio_ptr->audioif_hdmi_connect = XDX_AUDIO_MODULE_HDMI;
				change = 1; //dp->hdmi
			} else if ((module == XDX_AUDIO_MODULE_HDMI) && (!xdx_hdmi_audio_ptr->audioif_dp_connect)) {
				xdx_hdmi_audio_ptr->audioif_video_connect = XDX_AUDIO_MODULE_HDMI;
				change = 2;
			}
			if ((xdx_hdmi_audio_ptr->audioif_hdmi_connect) && (xdx_hdmi_audio_ptr->audioif_dp_connect)) {
				xdx_hdmi_audio_ptr->audioif_double_connect = XDX_AUDIO_MODULE_DOUBLE;
				XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d]Double connect!\n", __func__, __LINE__);
			}
		}
	} else {
		XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d]The status is error! \n", __func__, __LINE__);
		return -1;
	}

	if (change == 1) {
		mutex_lock(xdx_hdmi_audio_ptr->transfer_mutex);
		temp_status = xdx_hdmi_audio_ptr->status;
		xdx_hdmi_audio_ptr->status = cPlayBackStop;
		mutex_unlock(xdx_hdmi_audio_ptr->transfer_mutex);

		if (xdev->asic_type == CHIP_PANGU_B0) {
			mutex_lock(xdx_hdmi_audio_ptr->hw_mutex);
			pReg = (uint32_t *)(xdx_hdmi_audio_ptr->iobase_audio0 + 0x10);
			*pReg = 0;
			mutex_unlock(xdx_hdmi_audio_ptr->hw_mutex);
		}

		timeout_close = msecs_to_jiffies(200);
		time_close = get_jiffies_64();
		while (!xdx_hdmi_audio_ptr->fifo_empty || xdx_hdmi_audio_ptr->dma_pending_cnt) {
			udelay(10);
			time_now = get_jiffies_64();
			if (time_now > time_close + timeout_close) {
				XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d]Wait fifo_empty timeout!\n", __func__, __LINE__);
				break;
			}
		}
		if (xdev->asic_type == CHIP_PANGU_A0) {
			mutex_lock(xdx_hdmi_audio_ptr->hw_mutex);
			pAudioReg->TXEN.iis_en = 0;
			pAudioReg->CKEN.hdmi_bclk_en = 0;
			mutex_unlock(xdx_hdmi_audio_ptr->hw_mutex);
		}

		if ((xdx_hdmi_audio_ptr->audioif_video_connect == XDX_AUDIO_MODULE_DP) && (xdx_hdmi_audio_ptr->audioif_dp_connect == XDX_AUDIO_MODULE_DP)) {
			//hdmi=0 dp=1 hdmi->dp
			if (xdev->asic_type == CHIP_PANGU_A0) {
				mutex_lock(xdx_hdmi_audio_ptr->hw_mutex);
				pAudioReg->CNTL.hdmi_cdiv = 0;
				pAudioReg->CNTL.dp_iis_sel = 1;
				pAudioReg->CNTL.pl_shift = 0;
				pAudioReg->CNTL.aempty_thrd = 0;
				pAudioReg->CKEN.hdmi_bclk_en = 1;
				pAudioReg->IE.all = 0x6;
				pAudioReg->TXEN.iis_en = 1;
				mutex_unlock(xdx_hdmi_audio_ptr->hw_mutex);
			} else if (xdev->asic_type == CHIP_PANGU_B0) {
				mutex_lock(xdx_hdmi_audio_ptr->hw_mutex);
				OSWriteHWReg32(xdx_hdmi_audio_ptr->iobase_audio0, 0x4, 0x402031a);
				mutex_unlock(xdx_hdmi_audio_ptr->hw_mutex);
			}
			ADIFOutputSelConfig(xdx_hdmi_audio_ptr);

			XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d]The video connect from HDMI to DP. \n", __func__, __LINE__);
		} else if ((xdx_hdmi_audio_ptr->audioif_video_connect == XDX_AUDIO_MODULE_HDMI) && (xdx_hdmi_audio_ptr->audioif_hdmi_connect == XDX_AUDIO_MODULE_HDMI)) {
			//hdmi=1 dp=0 dp->hdmi
			if (xdev->asic_type == CHIP_PANGU_A0) {
				mutex_lock(xdx_hdmi_audio_ptr->hw_mutex);
				pAudioReg->CNTL.hdmi_cdiv = 1;
				pAudioReg->CNTL.dp_iis_sel = 0;
				pAudioReg->CNTL.pl_shift = 0;
				pAudioReg->CNTL.aempty_thrd = 0;
				pAudioReg->CKEN.hdmi_bclk_en = 1;
				pAudioReg->IE.all = 0x6;
				pAudioReg->TXEN.iis_en = 1;
				mutex_unlock(xdx_hdmi_audio_ptr->hw_mutex);
			} else if (xdev->asic_type == CHIP_PANGU_B0) {
				mutex_lock(xdx_hdmi_audio_ptr->hw_mutex);
				OSWriteHWReg32(xdx_hdmi_audio_ptr->iobase_audio0, 0x4, 0x4020318);
				mutex_unlock(xdx_hdmi_audio_ptr->hw_mutex);
			}

			if (hdmi) {
				xdx_hdmi_audio_ptr->hdmi = hdmi;
				xdx_hdmi_audio_setRate(xdx_hdmi_audio_ptr);
			}

 			XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d]The video connect from DP to HDMI. \n", __func__, __LINE__);
		}

		mutex_lock(xdx_hdmi_audio_ptr->transfer_mutex);
		xdx_hdmi_audio_ptr->status = temp_status;
		mutex_unlock(xdx_hdmi_audio_ptr->transfer_mutex);
	}

	if (change == 2) {
		if (xdx_hdmi_audio_ptr->audioif_hdmi_connect) {
			//0->hdmi
			if (xdev->asic_type == CHIP_PANGU_A0) {
				mutex_lock(xdx_hdmi_audio_ptr->hw_mutex);
				pAudioReg->CNTL.hdmi_cdiv = 1;
				pAudioReg->CNTL.dp_iis_sel = 0;
				pAudioReg->CNTL.pl_shift = 0;
				pAudioReg->CNTL.aempty_thrd = 0;
				pAudioReg->CKEN.hdmi_bclk_en = 1;
				pAudioReg->IE.all = 0x6;
				pAudioReg->TXEN.iis_en = 1;
				mutex_unlock(xdx_hdmi_audio_ptr->hw_mutex);
			}
			if (hdmi) {
				xdx_hdmi_audio_ptr->hdmi = hdmi;
				xdx_hdmi_audio_setRate(xdx_hdmi_audio_ptr);
			}
			XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d]The video connect from NULL to HDMI. \n", __func__, __LINE__);
		} else if (xdx_hdmi_audio_ptr->audioif_dp_connect) {
			//0->dp
			if (xdev->asic_type == CHIP_PANGU_A0) {
				mutex_lock(xdx_hdmi_audio_ptr->hw_mutex);
				pAudioReg->CNTL.hdmi_cdiv = 0;
				pAudioReg->CNTL.dp_iis_sel = 1;
				pAudioReg->CNTL.pl_shift = 0;
				pAudioReg->CNTL.aempty_thrd = 0;
				pAudioReg->CKEN.hdmi_bclk_en = 1;
				pAudioReg->IE.all = 0x6;
				pAudioReg->TXEN.iis_en = 1;
				mutex_unlock(xdx_hdmi_audio_ptr->hw_mutex);
			}
			if (dp) {
				xdx_hdmi_audio_ptr->dp = dp;
				ADIFOutputSelConfig(xdx_hdmi_audio_ptr);
			}

			XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d]The video connect from NULL to DP. \n", __func__, __LINE__);
		}
	}
	///init hdmi and dp
	if ((xdx_hdmi_audio_ptr->audioif_video_connect==XDX_AUDIO_MODULE_HDMI) && hdmi)
		xdx_hdmi_audio_ptr->hdmi = hdmi;
	else if ((xdx_hdmi_audio_ptr->audioif_video_connect==XDX_AUDIO_MODULE_DP) && dp)
		xdx_hdmi_audio_ptr->dp = dp;
	else if ((xdx_hdmi_audio_ptr->audioif_video_connect==XDX_AUDIO_MODULE_DOUBLE) && (hdmi)) {
		xdx_hdmi_audio_ptr->hdmi = hdmi;
		xdx_hdmi_audio_ptr->dp = dp;
	}

	return 0;
}
int xdx_audio_resume(struct xdx_device *xdev)
{
	struct snd_xdx_hdmi_audio *xdx_hdmi_audio_ptr = xdev->xdx_hdmi_audio_hotplug_ptr;
	struct snd_pcm_runtime *runtime = xdx_hdmi_audio_ptr->playback_substream->runtime;
	DBCmd_T dbCmd;
	DBCmd_a0f1_T dbCmd_a0f1;

	if (xdev->shareram_size == XDX_AUDIO_BAR4_16M) {
		switch (xdx_hdmi_audio_ptr->rate) {
			case 44100:
				dbCmd_a0f1.data = 44100;
				break;
			case 48000:
				dbCmd_a0f1.data = 48000;
				break;
			default:
				break;
		}
		XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d] dbCmd_a0f1.all=0x%llx\n", __func__, __LINE__, dbCmd_a0f1.all);
		OSWriteHWReg64(xdx_hdmi_audio_ptr->iobase, 0x100588, dbCmd_a0f1.all);
	} else if (xdev->shareram_size == XDX_AUDIO_BAR4_1M) {
		dbCmd.pixclock = xdx_hdmi_audio_ptr->rate;
		dbCmd.subCmd = cH02AAudioSubCmdPllConfig;
		XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d] dbCmd.all=0x%llx\n", __func__, __LINE__, dbCmd.all);
		OSWriteHWReg64(xdx_hdmi_audio_ptr->iobase_acpu, 0x100588, dbCmd.all);
	}

	if ((runtime != NULL) && (runtime->status->state != runtime->status->suspended_state))
		runtime->status->state = SNDRV_PCM_STATE_SETUP;

	XDX_HDMI_AUDIO_PRINT(xdx_hdmi_audio_ptr->dev, "[%s %d]\n", __func__, __LINE__);
	return 0;
}

static int xdxaudio_irq_process(struct xdx_device *xdev,
			       struct xdx_irq_src *src,
			       struct xdx_irq_entry *entry)
{
	pangu_audio0_irq_handler(src, xdev->xdx_hdmi_audio_hotplug_ptr);
	return 0;
}


static struct xdx_irq_src_funcs xdxaudio_src_funcs = {
	.process = xdxaudio_irq_process,
};

static int xdxvdma1_irq_process(struct xdx_device *xdev,
                               struct xdx_irq_src *src,
                               struct xdx_irq_entry *entry)
{
	if (xdev->asic_type == CHIP_PANGU_A0)
		pangu_a0_vdma1ch3_irq_handler(xdev->xdx_hdmi_audio_hotplug_ptr);
	else if (xdev->asic_type == CHIP_PANGU_B0)
		pangu_vdma1ch3_irq_handler(src, xdev->xdx_hdmi_audio_hotplug_ptr);
	return 0;
}


static struct xdx_irq_src_funcs xdxvdma1_src_funcs = {
	.process = xdxvdma1_irq_process,
};

int xdxaudio_isr_install(struct xdx_device *xdev)
{
	struct xdx_irq_src *irq_src_audio;
	struct xdx_irq_src *irq_src_vdma1;

	if (xdev->asic_type == CHIP_PANGU_B0) {
		irq_src_audio = kmalloc(sizeof(*irq_src_audio), GFP_KERNEL);
		if (!irq_src_audio)
			return -ENOMEM;

		irq_src_audio->id = XDX_IRQ_CLIENTID_AUDIO;
		irq_src_audio->funcs = &xdxaudio_src_funcs;

		xdx_irq_register_src(xdev, irq_src_audio);
		xdev->xdx_hdmi_audio_hotplug_ptr->irq_src_audio = irq_src_audio;
	}

	irq_src_vdma1 = kmalloc(sizeof(*irq_src_vdma1), GFP_KERNEL);
	if (!irq_src_vdma1)
		return -ENOMEM;

	irq_src_vdma1->id = XDX_IRQ_CLIENTID_AUDIO_DMA;
	irq_src_vdma1->funcs = &xdxvdma1_src_funcs;

	xdx_irq_register_src(xdev, irq_src_vdma1);
	xdev->xdx_hdmi_audio_hotplug_ptr->irq_src_vdma1 = irq_src_vdma1;
	return 0;
}

int xdxaudio_isr_uninstall(struct xdx_device *xdev)
{
	int ret;
	struct snd_xdx_hdmi_audio *xdx_hdmi_audio_ptr = xdev->xdx_hdmi_audio_hotplug_ptr;
	if (xdev->asic_type == CHIP_PANGU_B0) {
		ret = xdx_irq_unregister_src(xdev, xdx_hdmi_audio_ptr->irq_src_audio);
		if (ret) {
			dev_err(xdev->dev, "failed to unregister irq source\n");
			return ret;
		}

		kfree(xdx_hdmi_audio_ptr->irq_src_audio);
		xdx_hdmi_audio_ptr->irq_src_audio = NULL;
	}

	ret = xdx_irq_unregister_src(xdev, xdx_hdmi_audio_ptr->irq_src_vdma1);
	if (ret) {
		dev_err(xdev->dev, "failed to unregister irq source\n");
		return ret;
	}

	kfree(xdx_hdmi_audio_ptr->irq_src_vdma1);
	xdx_hdmi_audio_ptr->irq_src_vdma1 = NULL;

	return 0;
}
