#include <linux/module.h>
#include <linux/version.h>

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 5, 0)
#include <drm/drm_fourcc.h>
#else
#include <drm/drmP.h>
#endif
#include <drm/drm_gem.h>
#include <drm/drm_crtc.h>
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0)
#include <drm/drm_damage_helper.h>
#endif
#include <drm/drm_fb_helper.h>
#include <drm/drm_crtc_helper.h>
#include <drm/drm_gem_framebuffer_helper.h>
#include <drm/drm_vblank.h>
#include <linux/vga_switcheroo.h>

#include "dm_fb.h"
#include "dm_dc.h"
#include "xdx_gem.h"
#include "dm_plane.h"
#include "xdx_vmm.h"

static struct drm_framebuffer_funcs dm_fb_funcs = {
	.create_handle = drm_gem_fb_create_handle,
	.destroy = drm_gem_fb_destroy,
};

static struct drm_framebuffer *
dm_fb_alloc(struct drm_device *dev, const struct drm_mode_fb_cmd2 *mode_cmd,
		 struct xdx_bo **bos, unsigned int num_planes)
{
	struct drm_framebuffer *fb;
	int ret;
	unsigned int i;

	fb = kzalloc(sizeof(*fb), GFP_KERNEL);
	if (!fb)
		return ERR_PTR(-ENOMEM);

	drm_helper_mode_fill_fb_struct(dev, fb, mode_cmd);

	for (i = 0; i < num_planes; i++)
		fb->obj[i] = &bos[i]->base;

	ret = drm_framebuffer_init(dev, fb, &dm_fb_funcs);
	if (ret) {
		dev_err(dev->dev, "Failed to initialize framebuffer: %d\n",
			ret);
		kfree(fb);
		return ERR_PTR(ret);
	}

	return fb;
}

static struct drm_framebuffer *
xdx_dm_fb_create(struct drm_device *dev, struct drm_file *file_priv,
		  const struct drm_mode_fb_cmd2 *mode_cmd)
{
	struct drm_framebuffer *fb;
	const struct drm_format_info *info;
	struct xdx_bo *bos[MAX_NUM_PLANES];
	struct drm_gem_object *obj;
	unsigned int height, size;
	unsigned char i, num_planes;
	int ret = 0;

	info = drm_get_format_info(dev, mode_cmd);
	if (!info)
		return ERR_PTR(-EINVAL);

	num_planes = info->num_planes;
	if (num_planes > MAX_NUM_PLANES)
		return ERR_PTR(-EINVAL);

	for (i = 0; i < num_planes; i++) {
		struct xdx_bo *bo;

		obj = drm_gem_object_lookup(file_priv, mode_cmd->handles[i]);
		if (!obj) {
			dev_err(dev->dev, "Failed to lookup GEM object.\n");
			ret = -ENXIO;
			goto err;
		}
		bo = to_xdx_gem(obj);

		if (bo->pro.placement != XDXGPU_PL_VRAM) {
			dev_err(dev->dev,
				"bo of plane[%d]: wrong placement %d\n", i,
				bo->pro.placement);
			ret = -EINVAL;
			goto err;
		}

		#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 3, 0))
		height =
			drm_format_info_plane_height(info, mode_cmd->height, i);
	    #else
		height =
			drm_format_plane_height(mode_cmd->height, info->format, i);
	    #endif

		size = height * mode_cmd->pitches[i] + mode_cmd->offsets[i];

		if (obj->size < size) {
			dev_err(dev->dev,
				"bo of plane[%d]: size doesn't match (%d, %ld)\n",
				i, size, obj->size);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 9, 0)
			drm_gem_object_put(obj);
#else
			drm_gem_object_put_unlocked(obj);
#endif
			ret = -EINVAL;
			goto err;
		}

		bos[i] = bo;
	}

	fb = dm_fb_alloc(dev, mode_cmd, bos, i);
	if (IS_ERR(fb)) {
		ret = PTR_ERR(fb);
		goto err;
	}

	return fb;

err:
	for (; i > 0; i--)
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 9, 0)
		drm_gem_object_put(&bos[i - 1]->base);
#else
		drm_gem_object_put_unlocked(&bos[i - 1]->base);
#endif

	return ERR_PTR(ret);
}
 #if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0))
static const struct drm_format_info xdx_formats[] = {
	{ .format = DRM_FORMAT_NV12,
	  .depth = 0,
	  .num_planes = 2,
	  .char_per_block = { 20, 40, 0 },
	  .block_w = { 4, 4, 0 },
	  .block_h = { 4, 4, 0 },
	  .hsub = 2,
	  .vsub = 2,
	  .is_yuv = true },
	{ .format = DRM_FORMAT_YUV444,
	  .depth = 0,
	  .num_planes = 3,
	  .char_per_block = { 20, 20, 20 },
	  .block_w = { 4, 4, 4 },
	  .block_h = { 4, 4, 4 },
	  .hsub = 1,
	  .vsub = 1,
	  .is_yuv = true },
};
#else
static const struct drm_format_info xdx_formats[] = {
	{ .format = DRM_FORMAT_NV12,
	  .depth = 0,
	  .num_planes = 2,	  
	  .hsub = 2,
	  .vsub = 2,
	  .is_yuv = true },
	{ .format = DRM_FORMAT_YUV444,
	  .depth = 0,
	  .num_planes = 3,
	  .hsub = 1,
	  .vsub = 1,
	  .is_yuv = true },
};
#endif
// #if (LINUX_VERSION_CODE < KERNEL_VERSION(5, 0, 0))
// static const struct drm_format_info_a *
// lookup_format_info(const struct drm_format_info_a formats[], int num_formats,
// 				u32 format)
// #else
static const struct drm_format_info *
lookup_format_info(const struct drm_format_info formats[], int num_formats,
				u32 format)
{
	int i;
	for (i = 0; i < num_formats; i++) {
		if (formats[i].format == format)
			return &formats[i];
	}
	return NULL;
}

static const struct drm_format_info *
xdx_dm_get_format_info(const struct drm_mode_fb_cmd2 *cmd)
{
	if (fourcc_mod_xdx_get_type(cmd->modifier[0]) ==
		DRM_FORMAT_MOD_XDX_TYPE_CUSTOM_10BIT)
		return lookup_format_info(xdx_formats,
						ARRAY_SIZE(xdx_formats),
						cmd->pixel_format);
	else
		return NULL;
}

static int xdx_dm_atomic_commit(struct drm_device *dev,
				   struct drm_atomic_state *state,
				   bool nonblock)
{
	struct drm_crtc *crtc;
	struct drm_crtc_state *old_crtc_state, *new_crtc_state;
	int i;

	for_each_oldnew_crtc_in_state(state, crtc, old_crtc_state, new_crtc_state, i) {
		if (drm_atomic_crtc_needs_modeset(new_crtc_state)) {
			drm_crtc_vblank_off(crtc);
		}
	}

	return drm_atomic_helper_commit(dev, state, nonblock);
}

static void xdx_dm_atomic_commit_tail_rpm(struct drm_atomic_state *state)
{
	struct drm_crtc *crtc;
	struct drm_crtc_state *old_crtc_state, *new_crtc_state;
	int i;

	for_each_oldnew_crtc_in_state(state, crtc, old_crtc_state, new_crtc_state, i) {
		if (drm_atomic_crtc_needs_modeset(new_crtc_state)) {
			drm_crtc_vblank_on(crtc);
		}
	}

	return drm_atomic_helper_commit_tail_rpm(state);
}

static const struct drm_mode_config_funcs dm_mode_config_funcs = {
	.fb_create = xdx_dm_fb_create,
	.get_format_info = xdx_dm_get_format_info,
	.output_poll_changed = drm_fb_helper_output_poll_changed,
	.atomic_check = drm_atomic_helper_check,
	.atomic_commit = xdx_dm_atomic_commit,
};

static struct drm_mode_config_helper_funcs dm_mode_config_helpers = {
	.atomic_commit_tail = xdx_dm_atomic_commit_tail_rpm,
};

void xdx_dm_mode_config_init(struct drm_device *ddev)
{
	#if (LINUX_VERSION_CODE < KERNEL_VERSION(5,18,0)) 
	ddev->mode_config.allow_fb_modifiers = true;
	#else
	ddev->mode_config.fb_modifiers_not_supported = false;
	#endif
	ddev->mode_config.async_page_flip = true;

	if (ddev->mode_config.max_width == 0 ||
		ddev->mode_config.max_height == 0) {
		ddev->mode_config.min_width = 0;
		ddev->mode_config.min_height = 0;
		ddev->mode_config.max_width = 8192;
		ddev->mode_config.max_height = 8192;
	}
	ddev->mode_config.funcs = &dm_mode_config_funcs;
	ddev->mode_config.helper_private = &dm_mode_config_helpers;
}

/* define the interface of fbdev to control framebuffer and config parameter */
static struct fb_ops xdx_fbdev_fb_ops = {
	.owner = THIS_MODULE,
	.fb_check_var = drm_fb_helper_check_var,
	.fb_set_par = drm_fb_helper_set_par,
	.fb_fillrect = cfb_fillrect,
	.fb_copyarea = cfb_copyarea,
	.fb_imageblit = cfb_imageblit,
	.fb_pan_display = drm_fb_helper_pan_display,
	.fb_blank = drm_fb_helper_blank,
	.fb_setcmap = drm_fb_helper_setcmap,
	.fb_debug_enter = drm_fb_helper_debug_enter,
	.fb_debug_leave = drm_fb_helper_debug_leave,
	.fb_ioctl = drm_fb_helper_ioctl
};

/* func xdx_fbdev_fb_fill - fill fbdev framebuffer with specific color
 */
 /*
static void xdx_fbdev_fb_fill(struct drm_device *dev, struct fb_info *info, unsigned int color)
{
	struct fb_fillrect rect;
	rect.dx = rect.dy = 0;
	rect.width = info->var.xres_virtual;
	rect.height = info->var.yres_virtual;
	rect.color = color;
	rect.rop = ROP_COPY;
	info->fbops->fb_fillrect(info, &rect);
}
*/

/* func xdx_fbdev_fb_init - initialize a drm_framebuffer var for fbdev
 */
static int xdx_fbdev_fb_init(struct drm_device *dev,
			struct drm_framebuffer *fb,
			const struct drm_mode_fb_cmd2 *mode_cmd,
			struct drm_gem_object *obj)
{
	int ret;
	fb->obj[0] = obj;
	drm_helper_mode_fill_fb_struct(dev, fb, mode_cmd);
	ret = drm_framebuffer_init(dev, fb, &dm_fb_funcs);
	if (ret) {
		fb->obj[0] = NULL;
		return ret;
	}
	return 0;
}

/* func xdx_fbdev_fb_create
 * (1) create fb for fbdev
 * (2) initialize fb_info and operation callbacks
 */
#if (LINUX_VERSION_CODE < KERNEL_VERSION(5, 2, 0))
void drm_fb_helper_fill_info_below(struct fb_info *info,
			     struct drm_fb_helper *fb_helper,
			     struct drm_fb_helper_surface_size *sizes)
{
	struct drm_framebuffer *fb = fb_helper->fb;

	drm_fb_helper_fill_fix(info, fb->pitches[0], fb->format->depth);
	drm_fb_helper_fill_var(info, fb_helper,
			       sizes->fb_width, sizes->fb_height);
    // printk("fill fix complete!\n");
	info->par = fb_helper;
	snprintf(info->fix.id, sizeof(info->fix.id), "%sdrmfb",
		 fb_helper->dev->driver->name);

}
#endif
static int xdx_fbdev_fb_create(struct drm_fb_helper *helper,
				struct drm_fb_helper_surface_size *sizes)
{
	int err, size;
	struct xdx_fbdev *fbdev = container_of(helper, struct xdx_fbdev, helper);
	struct xdx_device *xdev = (struct xdx_device*)fbdev->priv;
	struct drm_device *dev = helper->dev;
	struct fb_info *info;
	struct drm_mode_fb_cmd2 mode_cmd;
	struct xdx_bo* bo = NULL;
	struct drm_framebuffer *fb = NULL;
	struct drm_gem_object *obj = NULL;
	void *vaddr = NULL;
	u64 paddr = 0;

	memset(&mode_cmd, 0, sizeof(mode_cmd));
	mode_cmd.width = sizes->surface_width;
	mode_cmd.height = sizes->surface_height;
	mode_cmd.pitches[0] = sizes->surface_width * DIV_ROUND_UP(sizes->surface_bpp, 8);
	mode_cmd.pixel_format = drm_mode_legacy_fb_format(sizes->surface_bpp, sizes->surface_depth);
	size = mode_cmd.height * mode_cmd.pitches[0];
	size = PAGE_ALIGN(size);

#ifdef CONFIG_MMU_1K_16M
	err = xdx_bo_create_reserved(xdev, round_up(size, 1 << 24), PAGE_SIZE << 12, XDXGPU_PL_VRAM,
								 &bo, &paddr, (void **)&vaddr);
#else
	err = xdx_bo_create_reserved(xdev, size, PAGE_SIZE, XDXGPU_PL_VRAM,
								 &bo, &paddr, (void **)&vaddr);
#endif
	if(err || !bo || !vaddr || !paddr)
	{
		dev_dbg(dev->dev,"Line%d xdx_bo_create_reserved() alloc buffer for fbdev failed.\n", __LINE__);
		goto on_error;
	}

	paddr = paddr - xdev->vram_start + xdev->aper_base;
	obj = &bo->base;

	err = xdx_bo_vmap(bo);
	if (err)
	{
		dev_dbg(dev->dev,"Line%d dc_bo_map failed.\n", __LINE__);
		goto on_error;
	}

	err = xdx_fbdev_fb_init(dev, &fbdev->fb, &mode_cmd, obj);

	/* Create a fb_info */
	info  = drm_fb_helper_alloc_fbi(helper);
	if (!info) {
		dev_dbg(dev->dev,"Line%d drm_fb_helper_alloc_fb failed.\n", __LINE__);
		goto on_error;
	}

	fb = &fbdev->fb;
	fbdev->helper.fb = &fbdev->fb;
	fbdev->helper.fbdev = info;

	info->apertures->ranges[0].base = xdev->aper_base;;
	info->apertures->ranges[0].size = xdev->aper_size;
	info->fix.smem_start = paddr;
	info->fix.smem_len = size;
	info->screen_base = vaddr;
	info->screen_size = size;
	info->fbops = &xdx_fbdev_fb_ops;
	/* initlize fbdev buffer  */
	memset(vaddr, 0x0, size);
	//xdx_fbdev_fb_fill(dev, info, 0x88888888);
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 2, 0))
	drm_fb_helper_fill_info(info, &fbdev->helper, sizes);
#else
    drm_fb_helper_fill_info_below(info, &fbdev->helper, sizes);
#endif
    // printk("fb->format->depth is %d,sizes->fb_width is %d,sizes->fb_height is %d\n",fb->format->depth,sizes->fb_width,sizes->fb_height);
	// printk(" &fbdev->helper aaaaa is %p,&fbdev->helper is %p\n",&fbdev->helper,&fbdev->helper);
    
	vga_switcheroo_client_fb_set(xdev->pdev, info);

	return 0;
on_error:
	dev_dbg(dev->dev,"line: %d, return failed.\n", __LINE__);
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 1, 0))	
	if (obj->funcs->vunmap)
		obj->funcs->vunmap(obj, NULL);
#else
	if (obj->dev->driver->gem_prime_vunmap)
		obj->dev->driver->gem_prime_vunmap(obj, NULL);
#endif
	return err;
}

static const struct drm_fb_helper_funcs xdx_fbdev_helper_funcs = {
	.fb_probe = xdx_fbdev_fb_create,
};

/* func xdx_fbdev_init - initialize drm fbdev and register
 * xdx_fbdev_init need be called after call drm_dev_register
 * there are 3steps mainly for a new fbdev:
 * step1: drm_fb_helper_prepare
 * step2: drm_fb_helper_init
 * step3: drm_fb_helper_initial_config
 * return 0 on success, nonzero otherwise
 */
int xdx_fbdev_init(struct drm_device *dev)
{
	struct xdx_device *xdev = drm_to_xdev(dev);
	struct xdx_fbdev *fbdev;
	int ret;

	fbdev = kzalloc(sizeof (*fbdev), GFP_KERNEL);
	if (!fbdev)
		return -ENOMEM;

	xdev->fbdev = (void*)fbdev;
	fbdev->priv = (void*)xdev;

	mutex_init(&fbdev->hpd_lock);
	fbdev->preferred_bpp = 32;

	drm_fb_helper_prepare(dev, &fbdev->helper, &xdx_fbdev_helper_funcs);

#if (LINUX_VERSION_CODE < KERNEL_VERSION(4, 11, 0))
	ret = drm_fb_helper_init(dev, &fbdev->helper, 1, 1);
#elif (LINUX_VERSION_CODE < KERNEL_VERSION(5, 7, 0))
	ret = drm_fb_helper_init(dev, &fbdev->helper, 1);
#else
	ret = drm_fb_helper_init(dev, &fbdev->helper);
#endif
	if(ret) {
		goto fini;
	}
    // printk("drm_fb_helper_init complete!\n");
	#if (LINUX_VERSION_CODE < KERNEL_VERSION(5, 3, 0))
    // (&fbdev->helper)->connector_count = dev->mode_config.num_connector;
	// printk("fbdev->helper.connector_count is %d\n",fbdev->helper.connector_count);
	ret = drm_fb_helper_single_add_all_connectors(&fbdev->helper);
	// printk("ret fb helper is %d\n",ret);
	if (ret < 0) {
		DRM_DEV_ERROR(dev->dev, "Failed to add connectors\n");
		goto err_drm_fb_helper_fini;
	}
#endif
	ret = drm_fb_helper_initial_config(&fbdev->helper, fbdev->preferred_bpp);
	if (ret) {
		goto fini;
	}
	// printk("&fbdev->helper ccc is %p\n",&fbdev->helper);
 	//printk("drm_fb_helper_initial_config complete!\n");
	dev_dbg(dev->dev,"fbdev init done\n");
	return 0;

fini:
	dev_dbg(dev->dev,"fbdev init failed\n");
	xdx_fbdev_fini(fbdev);
#if (LINUX_VERSION_CODE < KERNEL_VERSION(5, 3, 0))
err_drm_fb_helper_fini:
	drm_fb_helper_fbdev_teardown(xdev->ddev);
#endif

	return ret;
}

static void  xdx_fbdev_destroy(struct xdx_fbdev *fbdev)
{
	struct xdx_device *priv;
	struct drm_framebuffer *fb;
	if (!fbdev)
		return;

	fb = &fbdev->fb;
	priv = (struct xdx_device*)fbdev->priv;
#if !defined XDX_WA_ARM_POWEROFF
	if (fb->obj[0]) {
		drm_gem_object_put(fb->obj[0]);
		fb->obj[0] = NULL;
		drm_framebuffer_unregister_private(fb);
		drm_framebuffer_cleanup(fb);
	}
	drm_fb_helper_fini(&fbdev->helper);
#endif
	mutex_destroy(&fbdev->hpd_lock);

	kfree(fbdev);
	priv->fbdev = NULL;
}

/* func xdx_fbdev_fini - interface to deinit fbdev  whicn created by xdx_fbdev_init
 */
void xdx_fbdev_fini(struct xdx_fbdev *fbdev)
{
	if (!fbdev)
		return;

	/* xdx_fbdev unregister */
	drm_fb_helper_unregister_fbi(&fbdev->helper);
	xdx_fbdev_destroy(fbdev);
}
