/*
 * Copyright (c) 2016 Synopsys, Inc.
 *
 * Synopsys DP TX Linux Software Driver and documentation (hereinafter,
 * "Software") is an Unsupported proprietary work of Synopsys, Inc. unless
 * otherwise expressly agreed to in writing between Synopsys and you.
 *
 * The Software IS NOT an item of Licensed Software or Licensed Product under
 * any End User Software License Agreement or Agreement for Licensed Product
 * with Synopsys or any supplement thereto. You are permitted to use and
 * redistribute this Software in source and binary forms, with or without
 * modification, provided that redistributions of source code must retain this
 * notice. You may not view, use, disclose, copy or distribute this file or
 * any information contained herein except pursuant to this license grant from
 * Synopsys. If you do not agree with this notice, including the disclaimer
 * below, then you are not authorized to use the Software.
 *
 * THIS SOFTWARE IS BEING DISTRIBUTED BY SYNOPSYS SOLELY ON AN "AS IS" BASIS
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE HEREBY DISCLAIMED. IN NO EVENT SHALL SYNOPSYS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */
#if((LINUX_VERSION_CODE>=KERNEL_VERSION(5,18,0))&&(LINUX_VERSION_CODE<KERNEL_VERSION(5,19,0)))
#include <drm/dp/drm_dp_helper.h>
#else
#include <drm/drm_dp_helper.h>
#endif
#include "xdx_dp.h"
#include "xdx_dp_accel.h"
#include "core.h"
#include "dm.h"
#include "xdx_device.h"
#include "xdx_tile_reg.h"

/**
 * dptx_intr_en() - Enables interrupts
 * @dptx: The dptx struct
 * @bits: The interrupts to enable
 *
 * This function enables (unmasks) all interrupts in the INTERRUPT
 * register specified by @bits.
 */
static void dptx_intr_en(struct dptx *dptx, u32 bits)
{
	struct xdx_output *output = dptx->output;
	struct xdx_dm *dm = output_to_dm(output);
	struct xdx_device *xdev = dm_to_xdev(dm);
	u32 ien;

	dev_dbg(xdev->dev, "%s:\n", __func__);
	ien = dptx_readl(dptx, DPTX_IEN);
	ien |= bits;
	dptx_writel(dptx, DPTX_IEN, ien);

	ien = dptx_readl(dptx, DPTX_IEN);
	dev_dbg(xdev->dev, "DPTX_IEN(0x%x): 0x%x\n", DPTX_IEN, ien);
}

/**
 * dptx_intr_dis() - Disables interrupts
 * @dptx: The dptx struct
 * @bits: The interrupts to disable
 *
 * This function disables (masks) all interrupts in the INTERRUPT
 * register specified by @bits.
 */
static void dptx_intr_dis(struct dptx *dptx, u32 bits)
{
	struct xdx_output *output = dptx->output;
	struct xdx_dm *dm = output_to_dm(output);
	struct xdx_device *xdev = dm_to_xdev(dm);
	u32 ien;

	dev_dbg(xdev->dev, "%s:\n", __func__);
	ien = dptx_readl(dptx, DPTX_IEN);
	ien &= ~bits;
	dptx_writel(dptx, DPTX_IEN, ien);
}

/**
 * dptx_global_intr_en() - Enables top-level interrupts
 * @dptx: The dptx struct
 *
 * Enables (unmasks) all top-level interrupts.
 */
void dptx_global_intr_en(struct dptx *dptx)
{
	//dptx_intr_en(dptx, DPTX_IEN_ALL_INTR & ~(DPTX_ISTS_AUX_REPLY | DPTX_ISTS_AUX_CMD_INVALID));
	dptx_intr_en(dptx, DPTX_IEN_HPD & ~(DPTX_ISTS_AUX_REPLY | DPTX_ISTS_AUX_CMD_INVALID));
}

/**
 * dptx_global_intr_dis() - Disables top-level interrupts
 * @dptx: The dptx struct
 *
 * Disables (masks) all top-level interrupts.
 */
void dptx_global_intr_dis(struct dptx *dptx)
{
	dptx_intr_dis(dptx, DPTX_IEN_ALL_INTR);
}

// HPD IEN
void dptx_hpd_intr_dis(struct dptx *dptx)
{
	u32 hpd_ien;
	/* disable all HPD interrupts */
	hpd_ien = dptx_readl(dptx, DPTX_HPD_IEN);
	hpd_ien &= (~(DPTX_HPD_IEN_IRQ_EN |
				DPTX_HPD_IEN_HOT_PLUG_EN |
				DPTX_HPD_IEN_HOT_UNPLUG_EN));

	dptx_writel(dptx, DPTX_HPD_IEN, hpd_ien);
}

void dptx_hpd_intr_en(struct dptx *dptx)
{
	struct xdx_output *output = dptx->output;
	struct xdx_dm *dm = output_to_dm(output);
	struct xdx_device *xdev = dm_to_xdev(dm);
	u32 hpd_ien;
	u32 hpd_sts;

	hpd_sts = dptx_readl(dptx, DPTX_HPDSTS);
	if (hpd_sts & 0x7) {
		dptx_writel(dptx, DPTX_HPDSTS, hpd_sts);
	}
	hpd_sts = dptx_readl(dptx, DPTX_HPDSTS);

	hpd_ien = dptx_readl(dptx, DPTX_HPD_IEN);
	hpd_ien |= (DPTX_HPD_IEN_IRQ_EN |
				DPTX_HPD_IEN_HOT_PLUG_EN |
				DPTX_HPD_IEN_HOT_UNPLUG_EN);
	dptx_writel(dptx, DPTX_HPD_IEN, hpd_ien);
	hpd_ien = dptx_readl(dptx, DPTX_HPD_IEN);
	dev_dbg(xdev->dev, "HPD_IEN(0x%x): 0x%x\n", DPTX_HPD_IEN, hpd_ien);
}

/**
 * dptx_soft_reset() - Performs a core soft reset
 * @dptx: The dptx struct
 * @bits: The components to reset
 *
 * Resets specified parts of the core by writing @bits into the core
 * soft reset control register and clearing them 10-20 microseconds
 * later.
 */
void dptx_soft_reset(struct dptx *dptx, u32 bits)
{
	u32 rst;

	bits &= (DPTX_SRST_CTRL_ALL);

	/* Set reset bits */
	rst = dptx_readl(dptx, DPTX_SRST_CTRL);
	rst |= bits;
	dptx_writel(dptx, DPTX_SRST_CTRL, rst);

	usleep_range(10, 20);

	/* Clear reset bits */
	rst = dptx_readl(dptx, DPTX_SRST_CTRL);
	rst &= ~bits;
	dptx_writel(dptx, DPTX_SRST_CTRL, rst);
}

/**
 * dptx_soft_reset_all() - Reset all core modules
 * @dptx: The dptx struct
 */
void dptx_soft_reset_all(struct dptx *dptx)
{
	dptx_soft_reset(dptx, DPTX_SRST_CTRL_ALL);
}

void dptx_phy_soft_reset(struct dptx *dptx)
{
	dptx_soft_reset(dptx, DPTX_SRST_CTRL_PHY);
}

/**
 * dptx_core_init_phy() - Initializes the DP TX PHY module
 * @dptx: The dptx struct
 *
 * Initializes the PHY layer of the core. This needs to be called
 * whenever the PHY layer is reset.
 */
void dptx_core_init_phy(struct dptx *dptx)
{
	u32 phyifctrl;

	if (dptx->dp_pcs_width == 20) {
		phyifctrl = dptx_readl(dptx, DPTX_PHYIF_CTRL);
		phyifctrl &= ~DPTX_PHYIF_CTRL_WIDTH;
		dptx_writel(dptx, DPTX_PHYIF_CTRL, phyifctrl);
	} else {
		phyifctrl = dptx_readl(dptx, DPTX_PHYIF_CTRL);
		phyifctrl |= DPTX_PHYIF_CTRL_WIDTH;
		dptx_writel(dptx, DPTX_PHYIF_CTRL, phyifctrl);
	}
}


/**
* dptx_sink_enabled_ssc() - Returns true, if sink is enabled ssc
* @dptx: The dptx struct
*
*/
bool dptx_sink_enabled_ssc(struct dptx *dptx)
{
	u8 byte;

	dptx_read_dpcd(dptx, DP_MAX_DOWNSPREAD, &byte);

	return byte & 1;
}

/**
 * dptx_core_program_ssc() - Move phy to P3 state and programs SSC
 * @dptx: The dptx struct
 *
 * Enables SSC should be called during hot plug.
 *
 */
int dptx_core_program_ssc(struct dptx *dptx, bool sink_ssc)
{
	struct xdx_output *output = dptx->output;
	struct xdx_dm *dm = output_to_dm(output);
	struct xdx_device *xdev = dm_to_xdev(dm);
	u32 tile_data;
	u32 phyifctrl;
	u8  retval;

	// tile
	tile_data = dptx_tilereadl(dptx, DP_PCS_CTRL);
	dev_dbg(xdev->dev, "read fomr  DP_PCS_CTRL: 0x%x\n", tile_data);

	/* Enable 4 lanes, before programming SSC */
	dptx_phy_set_lanes(dptx, 4);

	// Move PHY to P3 to program SSC
	phyifctrl = dptx_readl(dptx, DPTX_PHYIF_CTRL);
	phyifctrl |= (3 << DPTX_PHYIF_CTRL_LANE_PWRDOWN_SHIFT);// move phy to P3 state
	dptx_writel(dptx, DPTX_PHYIF_CTRL, phyifctrl);

	retval = dptx_phy_wait_busy(dptx, DPTX_MAX_LINK_LANES);
	if (retval) {
		dev_err(xdev->dev, "timed out waiting for phy busy\n");
		return retval;
	}

	phyifctrl = dptx_readl(dptx, DPTX_PHYIF_CTRL);
	if (dptx->ssc_en && sink_ssc) {
		dev_dbg(xdev->dev, "enable ssc\n");
		phyifctrl &= ~DPTX_PHYIF_CTRL_SSC_DIS;
		tile_data |= 0x8;
	} else {
		dev_dbg(xdev->dev, "disable ssc\n");
		phyifctrl |= DPTX_PHYIF_CTRL_SSC_DIS;
		tile_data &= ~0x8;
	}

	dev_dbg(xdev->dev, "write to DP_PCS_CTRL: 0x%x\n", tile_data);
	dptx_tilewritel(dptx, DP_PCS_CTRL, tile_data);

	dptx_writel(dptx, DPTX_PHYIF_CTRL, phyifctrl);

	retval = dptx_phy_wait_busy(dptx, DPTX_MAX_LINK_LANES);
	if (retval) {
			dev_err(xdev->dev, "Timed out waiting for PHY BUSY\n");
			return retval;
	 }

	// Move PHY to P0 to program SSC
	phyifctrl = dptx_readl(dptx, DPTX_PHYIF_CTRL);
	phyifctrl |= (0 << DPTX_PHYIF_CTRL_LANE_PWRDOWN_SHIFT);// move phy to P0 state
	dptx_writel(dptx, DPTX_PHYIF_CTRL, phyifctrl);

	retval = dptx_phy_wait_busy(dptx, DPTX_MAX_LINK_LANES);
	if (retval) {
		dev_err(xdev->dev, "Timed out waiting for PHY BUSY\n");
		return retval;
	}
	dev_dbg(xdev->dev, "SSC program done\n");
	return 0;
}

/**
 * dptx_check_dptx_id() - Check value of DPTX_ID register
 * @dptx: The dptx struct
 *
 * Returns True if DPTX core correctly identifyed.
 */
bool dptx_check_dptx_id(struct dptx *dptx)
{
	u32 dptx_id;

	dptx_id = dptx_readl(dptx, DPTX_ID);
	if (dptx_id != (((u32)DPTX_ID_DEVICE_ID << DPTX_ID_DEVICE_ID_SHIFT) | DPTX_ID_VENDOR_ID))
		return false;

	return true;
}

/**
* dptx_enable_ssc() - Enables SSC based on automation request, 
*			  if DPTX controller enables ssc
* @dptx: The dptx struct
*
*/
void dptx_enable_ssc(struct dptx *dptx)
{
	struct xdx_output *output = dptx->output;
	struct xdx_dm *dm = output_to_dm(output);
	struct xdx_device *xdev = dm_to_xdev(dm);
	bool sink_ssc = dptx_sink_enabled_ssc(dptx);

	if(sink_ssc)
		dev_dbg(xdev->dev, "%s: SSC enable on the sink side\n", __func__);
	else
		dev_dbg(xdev->dev, "%s: SSC disabled on the sink side\n", __func__);

	dptx_core_program_ssc(dptx, sink_ssc);
}

void dptx_init_hwparams(struct dptx *dptx)
{
	struct xdx_output *output = dptx->output;
	struct xdx_dm *dm = output_to_dm(output);
	struct xdx_device *xdev = dm_to_xdev(dm);
	u32 reg;

	reg = dptx_readl(dptx, DPTX_CONFIG1);

	/* Num MST streams */
	dptx->streams = (reg & DPTX_CONFIG1_NUM_STREAMS_MASK) >>
		DPTX_CONFIG1_NUM_STREAMS_SHIFT;

	dev_dbg(xdev->dev, "dptx->streams: %d\n", dptx->streams);

	/* Combo PHY */
	dptx->hwparams.gen2phy = !!(reg & DPTX_CONFIG1_GEN2_PHY);

	/* DSC */
	dptx->hwparams.dsc = !!(reg & DPTX_CONFIG1_DSC_EN);

	/* Multi pixel mode */
	switch ((reg & DPTX_CONFIG1_MP_MODE_MASK) >> DPTX_CONFIG1_MP_MODE_SHIFT) {
		case DPTX_CONFIG1_MP_MODE_SINGLE:
			dptx->hwparams.multipixel = DPTX_MP_SINGLE_PIXEL;
			break;
		case DPTX_CONFIG1_MP_MODE_DUAL:
			dptx->hwparams.multipixel = DPTX_MP_DUAL_PIXEL;
			break;
		case DPTX_CONFIG1_MP_MODE_QUAD:
			dptx->hwparams.multipixel = DPTX_MP_QUAD_PIXEL;
			break;
		default:
			dptx->hwparams.multipixel = DPTX_MP_SINGLE_PIXEL;
			break;
	}
}

/**
 * dptx_core_init() - Initializes the DP TX core
 * @dptx: The dptx struct
 *
 * Initialize the DP TX core and put it in a known state.
 */
int dptx_core_init(struct dptx *dptx)
{
	struct xdx_output *output = dptx->output;
	struct xdx_dm *dm = output_to_dm(output);
	struct xdx_device *xdev = dm_to_xdev(dm);
	char str[15];
	u32 version;
	u32 reg; //sahakyan
	u32 retval;
	/* Reset the core */
	//dptx_soft_reset_all(dptx);

	/* Enable MST */
	dptx_writel(dptx, DPTX_CCTL,
				DPTX_CCTL_ENH_FRAME_EN
				| (dptx->mst ? DPTX_CCTL_ENABLE_MST_MODE : 0)
		);

	/* Check the core version */
	memset(str, 0, sizeof(str));
	version = dptx_readl(dptx, DPTX_VER_NUMBER);
	str[0] = (version >> 24) & 0xff;
	str[1] = '.';
	str[2] = (version >> 16) & 0xff;
	str[3] = (version >> 8) & 0xff;
	str[4] = version & 0xff;

	version = dptx_readl(dptx, DPTX_VER_TYPE);
	str[5] = '-';
	str[6] = (version >> 24) & 0xff;
	str[7] = (version >> 16) & 0xff;
	str[8] = (version >> 8) & 0xff;
	str[9] = version & 0xff;

	dev_dbg(xdev->dev, "Core version: %s\n", str);
	dptx->version = version;
	dptx_core_init_phy(dptx);

	dptx_hpd_intr_en(dptx);
	/* Mask interrupt related to HDCP22 GPIO output status */
	reg = dptx_readl(dptx, DPTX_HDCP_API_INT_MSK);  ///sahakyan
	reg |= DPTX_HDCP22_GPIOINT;
	dptx_writel(dptx, DPTX_HDCP_API_INT_MSK, reg);

	/* Enable all top-level interrupts */
	dptx_global_intr_en(dptx);

	dptx_writel(dptx, DPTX_PHYIF_PWRDOWN_CRTL, 0);

	retval = dptx_phy_wait_busy(dptx, DPTX_MAX_LINK_LANES);
	if (retval) {
		dev_err(xdev->dev, "timed out waiting for phy busy\n");
			return retval;
	}
	dev_dbg(xdev->dev, "core init done\n");
	return 0;
}

void dptx_interrupt_en(struct dptx *dptx)
{
	dptx_hpd_intr_en(dptx);
	dptx_global_intr_en(dptx);
}

/**
 * dptx_core_deinit() - Deinitialize the core
 * @dptx: The dptx struct
 *
 * Disable the core in preparation for module shutdown.
 */
int dptx_core_deinit(struct dptx *dptx)
{
	dptx_global_intr_dis(dptx);
	dptx_soft_reset_all(dptx);
	return 0;
}

/*
 * PHYIF core access functions
 */

unsigned int dptx_phy_get_lanes(struct dptx *dptx)
{
	struct xdx_output *output = dptx->output;
	struct xdx_dm *dm = output_to_dm(output);
	struct xdx_device *xdev = dm_to_xdev(dm);
	u32 phyifctrl;
	u32 val;

	dev_dbg(xdev->dev, "%s:\n", __func__);

	phyifctrl = dptx_readl(dptx, DPTX_PHYIF_CTRL);
	val = (phyifctrl & DPTX_PHYIF_CTRL_LANES_MASK) >>
		DPTX_PHYIF_CTRL_LANES_SHIFT;

	return (1 << val);
}

void dptx_phy_set_lanes(struct dptx *dptx, unsigned int lanes)
{
	struct xdx_output *output = dptx->output;
	struct xdx_dm *dm = output_to_dm(output);
	struct xdx_device *xdev = dm_to_xdev(dm);
	u32 phyifctrl;
	u32 val;

	dev_dbg(xdev->dev, "%s: lanes=%d\n", __func__, lanes);

	switch (lanes) {
	case 1:
		val = 0;
		break;
	case 2:
		val = 1;
		break;
	case 4:
		val = 2;
		break;
	default:
		WARN(1, "Invalid number of lanes %d\n", lanes);
		return;
	}

	phyifctrl = 0;
	phyifctrl = dptx_readl(dptx, DPTX_PHYIF_CTRL);
	phyifctrl &= ~DPTX_PHYIF_CTRL_LANES_MASK;
	phyifctrl |= (val << DPTX_PHYIF_CTRL_LANES_SHIFT);
	dptx_writel(dptx, DPTX_PHYIF_CTRL, phyifctrl);
}

void dptx_phy_set_powerdown(struct dptx *dptx, unsigned int lanes)
{
	u32 val;
	switch (lanes) {
	case 1:
		val = 0x13330;
		break;
	case 2:
		val = 0x13300;
		break;
	case 4:
		val = 0x0000;
		break;
	default:
		WARN(1,"pd Invalid number of lanes %d\n", lanes);
		return;
	}

	dptx_writel(dptx,DPTX_PHYIF_PWRDOWN_CRTL, val);
}


void dptx_phy_set_rate(struct dptx *dptx, unsigned int rate)
{
	struct xdx_output *output = dptx->output;
	struct xdx_dm *dm = output_to_dm(output);
	struct xdx_device *xdev = dm_to_xdev(dm);
	u32 phyifctrl;

	dev_dbg(xdev->dev, "%s: rate=%d\n", __func__, rate);

	phyifctrl = dptx_readl(dptx, DPTX_PHYIF_CTRL);

//#ifdef DPTX_COMBO_PHY
//	switch (rate) {
//	case DPTX_PHYIF_CTRL_RATE_RBR:
//	case DPTX_PHYIF_CTRL_RATE_HBR:
//		/* Set 20-bit PHY width */
//		phyifctrl = dptx_readl(dptx, DPTX_PHYIF_CTRL);
//		phyifctrl &= ~DPTX_PHYIF_CTRL_WIDTH;
//		dptx_writel(dptx, DPTX_PHYIF_CTRL, phyifctrl);
//		break;
//	case DPTX_PHYIF_CTRL_RATE_HBR2:
//	case DPTX_PHYIF_CTRL_RATE_HBR3:
//		/* Set 40-bit PHY width */
//		phyifctrl = dptx_readl(dptx, DPTX_PHYIF_CTRL);
//		phyifctrl |= DPTX_PHYIF_CTRL_WIDTH;
//		dptx_writel(dptx, DPTX_PHYIF_CTRL, phyifctrl);
//		break;
//	default:
//		WARN(1, "Invalid PHY rate %d\n", rate);
//		break;
//	}
//#endif

	phyifctrl = dptx_readl(dptx, DPTX_PHYIF_CTRL);
	phyifctrl &= ~DPTX_PHYIF_CTRL_RATE_MASK;
	phyifctrl |= rate << DPTX_PHYIF_CTRL_RATE_SHIFT;
	dptx_writel(dptx, DPTX_PHYIF_CTRL, phyifctrl);

	
	//tile register phy mode config
	//phyifctrl = dptx_tilereadl(dptx,DP_PCS_CTRL);
	//phyifctrl = (phyifctrl | (rate << 6));
	//dptx_tilewritel(dptx,DP_PCS_CTRL,phyifctrl);
}

unsigned int dwc_phy_get_rate(struct dptx *dptx)
{
	struct xdx_output *output = dptx->output;
	struct xdx_dm *dm = output_to_dm(output);
	struct xdx_device *xdev = dm_to_xdev(dm);
	u32 phyifctrl;
	u32 rate;

	dev_dbg(xdev->dev, "%s:\n", __func__);

	phyifctrl = dptx_readl(dptx, DPTX_PHYIF_CTRL);
	rate = (phyifctrl & DPTX_PHYIF_CTRL_RATE_MASK) >>
		DPTX_PHYIF_CTRL_RATE_SHIFT;

	return rate;
}

int dptx_phy_wait_busy(struct dptx *dptx, unsigned int lanes)
{
	struct xdx_output *output = dptx->output;
	struct xdx_dm *dm = output_to_dm(output);
	struct xdx_device *xdev = dm_to_xdev(dm);
	unsigned int count;
	u32 phyifctrl;
	u32 mask = 0;

	dev_dbg(xdev->dev, "%s: lanes=%d\n", __func__, lanes);
#ifndef FALLTHROUGH_INTENDED
#define FALLTHROUGH_INTENDED __fallthrough
#endif

	switch (lanes) {
	case 4:
		mask |= DPTX_PHYIF_CTRL_BUSY(3);
		mask |= DPTX_PHYIF_CTRL_BUSY(2);
		FALLTHROUGH_INTENDED;
	case 2:
		mask |= DPTX_PHYIF_CTRL_BUSY(1);
		FALLTHROUGH_INTENDED;
	case 1:
		mask |= DPTX_PHYIF_CTRL_BUSY(0);
		break;
	default:
		WARN(1, "Invalid number of lanes %d\n", lanes);
		break;
	}

	count = 0;

	while (1) {
		phyifctrl = dptx_readl(dptx, DPTX_PHYIF_CTRL);

		if (!(phyifctrl & mask))
			break;

		count++;
		if (count > 200) { // was 50
			dev_dbg(xdev->dev, "%s: PHY BUSY timed out\n", __func__);
			//return -EBUSY;
			return 0;
		}
		mdelay(1);
	}

	return 0;
}

void dptx_phy_set_pre_emphasis(struct dptx *dptx,
		unsigned int lane,
		unsigned int level)
{
	struct xdx_output *output = dptx->output;
	struct xdx_dm *dm = output_to_dm(output);
	struct xdx_device *xdev = dm_to_xdev(dm);
	u32 phytxeq;

	dev_dbg(xdev->dev, "%s: lane=%d, level=0x%x\n", __func__, lane, level);

	if (WARN(lane > 3, "Invalid lane %d", lane))
		return;

	if (WARN(level > 3, "Invalid pre-emphasis level %d, using 3", level))
		level = 3;

	phytxeq = dptx_readl(dptx, DPTX_PHY_TX_EQ);
	phytxeq &= ~DPTX_PHY_TX_EQ_PREEMP_MASK(lane);
	phytxeq |= (level << DPTX_PHY_TX_EQ_PREEMP_SHIFT(lane)) &
				DPTX_PHY_TX_EQ_PREEMP_MASK(lane);

	dptx_writel(dptx, DPTX_PHY_TX_EQ, phytxeq);
}

void dptx_phy_set_vswing(struct dptx *dptx,
			 unsigned int lane,
			 unsigned int level)
{
	struct xdx_output *output = dptx->output;
	struct xdx_dm *dm = output_to_dm(output);
	struct xdx_device *xdev = dm_to_xdev(dm);
	u32 phytxeq;

	dev_dbg(xdev->dev, "%s: lane=%d, level=0x%x\n", __func__, lane, level);

	if (WARN(lane > 3, "Invalid lane %d", lane))
		return;

	if (WARN(level > 3, "Invalid vswing level %d, using 3", level))
		level = 3;

	phytxeq = dptx_readl(dptx, DPTX_PHY_TX_EQ);
	phytxeq &= ~DPTX_PHY_TX_EQ_VSWING_MASK(lane);
	phytxeq |= (level << DPTX_PHY_TX_EQ_VSWING_SHIFT(lane)) &
				DPTX_PHY_TX_EQ_VSWING_MASK(lane);

	dptx_writel(dptx, DPTX_PHY_TX_EQ, phytxeq);
}

void dptx_phy_set_pattern(struct dptx *dptx,
			  unsigned int pattern)
{
	struct xdx_output *output = dptx->output;
	struct xdx_dm *dm = output_to_dm(output);
	struct xdx_device *xdev = dm_to_xdev(dm);
	u32 phyifctrl = 0;

	dev_dbg(xdev->dev, "%s: Setting PHY pattern=0x%x\n", __func__, pattern);

	phyifctrl = dptx_readl(dptx, DPTX_PHYIF_CTRL);
	phyifctrl &= ~DPTX_PHYIF_CTRL_TPS_SEL_MASK;
	phyifctrl |= ((pattern << DPTX_PHYIF_CTRL_TPS_SEL_SHIFT) &
				DPTX_PHYIF_CTRL_TPS_SEL_MASK);
	dptx_writel(dptx, DPTX_PHYIF_CTRL, phyifctrl);
}

void dptx_phy_enable_xmit(struct dptx *dptx, unsigned int lanes, bool enable)
{
	struct xdx_output *output = dptx->output;
	struct xdx_dm *dm = output_to_dm(output);
	struct xdx_device *xdev = dm_to_xdev(dm);
	u32 phyifctrl;
	u32 mask = 0;

	dev_dbg(xdev->dev, "%s: lanes=%d, enable=%d\n", __func__, lanes, enable);

	phyifctrl = dptx_readl(dptx, DPTX_PHYIF_CTRL);
#ifndef FALLTHROUGH_INTENDED
#define FALLTHROUGH_INTENDED __fallthrough
#endif

	switch (lanes) {
	case 4:
		mask |= DPTX_PHYIF_CTRL_XMIT_EN(3);
		mask |= DPTX_PHYIF_CTRL_XMIT_EN(2);
		FALLTHROUGH_INTENDED;
	case 2:
		mask |= DPTX_PHYIF_CTRL_XMIT_EN(1);
		FALLTHROUGH_INTENDED;
	case 1:
		mask |= DPTX_PHYIF_CTRL_XMIT_EN(0);
		break;
	default:
		WARN(1, "Invalid number of lanes %d\n", lanes);
		break;
	}

	if (enable)
		phyifctrl |= mask;
	else
		phyifctrl &= ~mask;

	dptx_writel(dptx, DPTX_PHYIF_CTRL, phyifctrl);
}

int dptx_phy_rate_to_bw(unsigned int rate)
{
	switch (rate) {
	case DPTX_PHYIF_CTRL_RATE_RBR:
		return DP_LINK_BW_1_62;
	case DPTX_PHYIF_CTRL_RATE_HBR:
		return DP_LINK_BW_2_7;
	case DPTX_PHYIF_CTRL_RATE_HBR2:
		return DP_LINK_BW_5_4;
	case DPTX_PHYIF_CTRL_RATE_HBR3:
		return DP_LINK_BW_8_1;
	default:
		WARN(1, "Invalid rate 0x%x\n", rate);
		return -EINVAL;
	}
}

int dptx_bw_to_phy_rate(unsigned int bw)
{
	switch (bw) {
	case DP_LINK_BW_1_62:
		return DPTX_PHYIF_CTRL_RATE_RBR;
	case DP_LINK_BW_2_7:
		return DPTX_PHYIF_CTRL_RATE_HBR;
	case DP_LINK_BW_5_4:
		return DPTX_PHYIF_CTRL_RATE_HBR2;
	case DP_LINK_BW_8_1:
		return DPTX_PHYIF_CTRL_RATE_HBR3;
	default:
		WARN(1, "Invalid bw 0x%x\n", bw);
		return -EINVAL;
	}
}

void xdx_tile_init(struct dptx *dptx)
{
	struct xdx_output *output = dptx->output;
	struct xdx_dm *dm = output_to_dm(output);
	struct xdx_device *xdev = dm_to_xdev(dm);
	u32 data;

	//tile init
	data = dptx_tilereadl(dptx,PHYREF_PLL_EN);
	data = data & 0xfffffffe;
	dptx_tilewritel(dptx,PHYREF_PLL_EN,data);

	dptx_tilewritel(dptx,PHYREF_PLL_EN,1);

	while(1){
		data = dptx_tilereadl(dptx,PHYREF_PLL_LOCK);
		if(data & 1)
			break;
	}
	dev_dbg(xdev->dev, "PHYREF PLL LOCK %x\n",data);
	//SRAM bypass
	data = dptx_tilereadl(dptx,DP_PHY_CTRL);
	dptx_tilewritel(dptx,DP_PHY_CTRL,data | 0x2);
	dev_dbg(xdev->dev, "PHYREF DP_PHY_CTRL %x\n",data);

	// video mapping
	if(dptx->vparams.pix_enc == RGB){
		if(dptx->vparams.bpc == 8){
			data = dptx_tilereadl(dptx,DP_CTRL);
			data = data | 0x2 | (1<<2);
			dptx_tilewritel(dptx,DP_CTRL,data);
			dev_dbg(xdev->dev, "PHYREF DP_CTRL %x\n",data);
		}else if(dptx->vparams.bpc == 10){
			data = dptx_tilereadl(dptx,DP_CTRL);
			data = data | 0x2 | (2<<2);
			dptx_tilewritel(dptx,DP_CTRL,data);
			dev_dbg(xdev->dev, "PHYREF DP_CTRL %x\n",data);
		}
	}else if(dptx->vparams.pix_enc == YCBCR422){
		if(dptx->vparams.bpc == 8){
			data = dptx_tilereadl(dptx,DP_CTRL);
			data = data | 0x2 | (9<<2);
			dptx_tilewritel(dptx,DP_CTRL,data);
			dev_dbg(xdev->dev, "PHYREF DP_CTRL %x\n",data);
		}else if(dptx->vparams.bpc == 10){
			data = dptx_tilereadl(dptx,DP_CTRL);
			data = data | 0x2 | (10<<2);
			dptx_tilewritel(dptx,DP_CTRL,data);
			dev_dbg(xdev->dev, "PHYREF DP_CTRL %x\n",data);
		}
	}else if(dptx->vparams.pix_enc == YCBCR444){
		if(dptx->vparams.bpc == 8){
			data = dptx_tilereadl(dptx,DP_CTRL);
			data = data | 0x2 | (5<<2);
			dptx_tilewritel(dptx,DP_CTRL,data);
			dev_dbg(xdev->dev, "PHYREF DP_CTRL %x\n",data);
		}else if(dptx->vparams.bpc == 10){
			data = dptx_tilereadl(dptx,DP_CTRL);
			data = data | 0x2 | (6<<2);
			dptx_tilewritel(dptx,DP_CTRL,data);
			dev_dbg(xdev->dev, "PHYREF DP_CTRL %x\n",data);
		}
	}else if(dptx->vparams.pix_enc == YCBCR420){
		if(dptx->vparams.bpc == 8){
			data = dptx_tilereadl(dptx,DP_CTRL);
			data = data | 0x2 | (13<<2);
			dptx_tilewritel(dptx,DP_CTRL,data);
			dev_dbg(xdev->dev, "PHYREF DP_CTRL %x\n",data);
		}else if(dptx->vparams.bpc == 10){
			data = dptx_tilereadl(dptx,DP_CTRL);
			data = data | 0x2 | (14<<2);
			dptx_tilewritel(dptx,DP_CTRL,data);
			dev_dbg(xdev->dev, "PHYREF DP_CTRL %x\n",data);
		}
	}
	//phy mode config
	data = dptx_tilereadl(dptx,DP_PCS_CTRL);
	data = data | (0x3<<20 | (3<<6));
	if(dptx->dp_pcs_width == 20){
		data = data | 1 << 4;
	}else {
		data = data | 2 << 4;
	}
	dptx_tilewritel(dptx,DP_PCS_CTRL,data);
	dev_dbg(xdev->dev, "PHYREF DP_PCS_CTRL %x\n",data);
}

