/*
 * Copyright (c) 2016 Synopsys, Inc.
 *
 * Synopsys DP TX Linux Software Driver and documentation (hereinafter,
 * "Software") is an Unsupported proprietary work of Synopsys, Inc. unless
 * otherwise expressly agreed to in writing between Synopsys and you.
 *
 * The Software IS NOT an item of Licensed Software or Licensed Product under
 * any End User Software License Agreement or Agreement for Licensed Product
 * with Synopsys or any supplement thereto. You are permitted to use and
 * redistribute this Software in source and binary forms, with or without
 * modification, provided that redistributions of source code must retain this
 * notice. You may not view, use, disclose, copy or distribute this file or
 * any information contained herein except pursuant to this license grant from
 * Synopsys. If you do not agree with this notice, including the disclaimer
 * below, then you are not authorized to use the Software.
 *
 * THIS SOFTWARE IS BEING DISTRIBUTED BY SYNOPSYS SOLELY ON AN "AS IS" BASIS
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE HEREBY DISCLAIMED. IN NO EVENT SHALL SYNOPSYS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 */
#if((LINUX_VERSION_CODE>=KERNEL_VERSION(5,18,0))&&(LINUX_VERSION_CODE<KERNEL_VERSION(5,19,0)))
#include <drm/dp/drm_dp_helper.h>
#else
#include <drm/drm_dp_helper.h>
#endif
#include "link.h"
#include "dm.h"
#include "xdx_device.h"
#include "core.h"
#include "xdx_dp_accel.h"

static int dptx_link_read_status(struct dptx *dptx)
{
	return dptx_read_bytes_from_dpcd(dptx, DP_LANE0_1_STATUS,
					 dptx->link.status,
					// DP_LINK_STATUS_SIZE);
					4);
}

static int dptx_link_check_cr_done(struct dptx *dptx, bool *out_done, bool eq)
{
	struct xdx_output *output = dptx->output;
	struct xdx_dm *dm = output_to_dm(output);
	struct xdx_device *xdev = dm_to_xdev(dm);
	int retval;
	u8 byte;
	u32 reg;

	dev_dbg(xdev->dev, "%s:\n", __func__);

	dev_dbg(xdev->dev,"%s: link lanes:%x\n",__func__,dptx->link.lanes);
	if (WARN_ON(!out_done))
		return -EINVAL;

	*out_done = false;

	retval = dptx_read_dpcd(dptx, DP_TRAINING_AUX_RD_INTERVAL, &byte);
	if (retval)
		return retval;

	if(eq == 1){
		switch(byte & 0x7F)
		{
			case 0:
				reg = 400;
				break;
			case 1:
				reg = 4 * 1000;
				break;
			case 2:
				reg = 8 * 1000;
				break;
			case 3:
				reg = 12 * 1000;
				break;
			case 4:
				reg = 16 * 1000;
				break;
			default:
				reg = 400;
				break;
		}
	}else{
		reg = 100;
	}
	udelay(reg);

	dev_dbg(xdev->dev,"%s: link lanes:%x\n",__func__, dptx->link.lanes);
	retval = dptx_link_read_status(dptx);
	if (retval)
		return retval;

	dev_dbg(xdev->dev,"%s: link status %x lanes %x\n",__func__,*(u32 *)dptx->link.status,dptx->link.lanes);
	*out_done = drm_dp_clock_recovery_ok(dptx->link.status, dptx->link.lanes);

	dev_dbg(xdev->dev, "%s: CR_DONE = %d\n", __func__, *out_done);
	return 0;
}

static int dptx_link_check_ch_eq_done(struct dptx *dptx,
					bool *out_cr_done,
					bool *out_ch_eq_done)
{
	struct xdx_output *output = dptx->output;
	struct xdx_dm *dm = output_to_dm(output);
	struct xdx_device *xdev = dm_to_xdev(dm);
	int retval;
	bool done;

	dev_dbg(xdev->dev, "%s:\n", __func__);

	if (WARN_ON(!out_cr_done || !out_ch_eq_done))
		return -EINVAL;

	retval = dptx_link_check_cr_done(dptx, &done, 1);
	if (retval)
		return retval;

	*out_cr_done = false;
	*out_ch_eq_done = false;

	if (!done)
		return 0;

	*out_cr_done = true;
	dev_dbg(xdev->dev,"TDB eq check link.status:%x link.lanes%d\n",*(u32 *)dptx->link.status,dptx->link.lanes);
	*out_ch_eq_done = drm_dp_channel_eq_ok(dptx->link.status,
						dptx->link.lanes);

	dev_dbg(xdev->dev, "%s: CH_EQ_DONE = %d\n", __func__, *out_ch_eq_done);

	return 0;
}

void dptx_link_set_preemp_vswing(struct dptx *dptx)
{
	struct xdx_output *output = dptx->output;
	struct xdx_dm *dm = output_to_dm(output);
	struct xdx_device *xdev = dm_to_xdev(dm);
	unsigned int i;
	u32 pes = 0,vss = 0;

	for (i = 0; i < dptx->link.lanes; i++) {
		u8 pe;
		u8 vs;

		pe = dptx->link.preemp_level[i];
		vs = dptx->link.vswing_level[i];
		if(i > 3)
			break;

		pes |= (pe << DPTX_PHY_TX_EQ_PREEMP_SHIFT(i)) &
		DPTX_PHY_TX_EQ_PREEMP_MASK(i);
		vss |= (vs << DPTX_PHY_TX_EQ_VSWING_SHIFT(i)) &
		DPTX_PHY_TX_EQ_VSWING_MASK(i);
		dev_dbg(xdev->dev, "loop:%d pe:%x vs:%x\n",i,pe,vs);
	}
	dptx_writel(dptx, DPTX_PHY_TX_EQ, pes | vss);
}

int dptx_link_training_lanes_set(struct dptx *dptx)
{
	int retval;
	unsigned int i;
	u8 bytes[4] = { 0xff, 0xff, 0xff, 0xff };

	for (i = 0; i < dptx->link.lanes; i++) {
		u8 byte = 0;

		byte |= ((dptx->link.vswing_level[i] <<
				DP_TRAIN_VOLTAGE_SWING_SHIFT) &
				DP_TRAIN_VOLTAGE_SWING_MASK);

		if (dptx->link.vswing_level[i] == 3)
			byte |= DP_TRAIN_MAX_SWING_REACHED;

		byte |= ((dptx->link.preemp_level[i] <<
			DP_TRAIN_PRE_EMPHASIS_SHIFT) &
			DP_TRAIN_PRE_EMPHASIS_MASK);

		if (dptx->link.preemp_level[i] == 3)
			byte |= DP_TRAIN_MAX_PRE_EMPHASIS_REACHED;

		bytes[i] = byte;
	}

	retval = dptx_write_bytes_to_dpcd(dptx, DP_TRAINING_LANE0_SET,
									bytes, dptx->link.lanes);
	if (retval)
		return retval;

	return 0;
}

int dptx_link_adjust_drive_settings(struct dptx *dptx, int *out_changed, bool try_our_adjust)
{
	struct xdx_output *output = dptx->output;
	struct xdx_dm *dm = output_to_dm(output);
	struct xdx_device *xdev = dm_to_xdev(dm);
	int retval;
	unsigned int lanes;
	unsigned int i;
	u8 byte;
	u8 adj[4] = { 0, };
	int changed = false;

	lanes = dptx->link.lanes;
#ifndef FALLTHROUGH_INTENDED
#define FALLTHROUGH_INTENDED __fallthrough
#endif

	switch (lanes) {
		case 4:
			retval = dptx_read_dpcd(dptx, DP_ADJUST_REQUEST_LANE2_3, &byte);
			if (retval)
				return retval;

			adj[2] = byte & 0x0f;
			adj[3] = (byte & 0xf0) >> 4;
			FALLTHROUGH_INTENDED;
		case 2:
		case 1:
			retval = dptx_read_dpcd(dptx, DP_ADJUST_REQUEST_LANE0_1, &byte);
			if (retval)
				return retval;

			adj[0] = byte & 0x0f;
			adj[1] = (byte & 0xf0) >> 4;
			break;
		default:
			WARN(1, "Invalid number of lanes %d\n", lanes);
			return -EINVAL;
	}
	dev_dbg(xdev->dev, "adjust lane0:%x lane1:%x lane2:%x lane3:%x\n",adj[0],adj[1],adj[2],adj[3]);
	/* Save the drive settings */
	for (i = 0; i < lanes; i++) {
		u8 vs = adj[i] & 0x3;
		u8 pe = (adj[i] & 0xc) >> 2;

		if (dptx->link.vswing_level[i] != vs)
			changed = true;
		if(try_our_adjust){
				dptx->link.preemp_level[i] = pe;
				if(dptx->link.cr_vswing_level[i] == 0){
						dptx->link.vswing_level[i] = vs;
				}else{
					dptx->link.vswing_level[i] = dptx->link.cr_vswing_level[i];
				}
		}else{
				dptx->link.vswing_level[i] = vs;
				dptx->link.preemp_level[i] = pe;
		}
		dev_dbg(xdev->dev, "lane:%d vswing:%x preemp:%x\n",i,dptx->link.vswing_level[i],dptx->link.preemp_level[i] );
	}

	dptx_link_set_preemp_vswing(dptx);

	retval = dptx_link_training_lanes_set(dptx);
	if (retval)
		return retval;

	if (out_changed)
		*out_changed = changed;

	return 0;
}

static int dptx_link_training_init(struct dptx *dptx,
					u8 rate,
					u8 lanes)
{
	struct xdx_output *output = dptx->output;
	struct xdx_dm *dm = output_to_dm(output);
	struct xdx_device *xdev = dm_to_xdev(dm);
	u8 sink_max_rate;

	dev_dbg(xdev->dev, "%s: lanes=%d, rate=%d\n", __func__, lanes, rate);

	if (WARN(rate > DPTX_PHYIF_CTRL_RATE_HBR3,
		 "Invalid rate %d\n", rate))
		rate = DPTX_PHYIF_CTRL_RATE_RBR;

	if (WARN(!lanes || (lanes == 3) || (lanes > 4),
		 "Invalid lanes %d\n", lanes))
		lanes = 1;

	/* Initialize link parameters */
	memset(dptx->link.preemp_level, 0, sizeof(u8) * 4);
	memset(dptx->link.vswing_level, 0, sizeof(u8) * 4);
	memset(dptx->link.status, 0, 4);

	sink_max_rate = DPTX_PHYIF_CTRL_RATE_HBR3;
	if (rate > sink_max_rate)
	rate = sink_max_rate;

	dptx->link.lanes = lanes;
	dptx->link.rate = rate;
	dptx->link.trained = false;

	return 0;
}

int dptx_link_training_pattern_set(struct dptx *dptx, u8 pattern)
{
	int retval;

	retval = dptx_write_dpcd(dptx, DP_TRAINING_PATTERN_SET, pattern);
	if (retval)
		return retval;

	return 0;
}

static int dptx_link_training_start(struct dptx *dptx)
{
	struct xdx_output *output = dptx->output;
	struct xdx_dm *dm = output_to_dm(output);
	struct xdx_device *xdev = dm_to_xdev(dm);
	int retval;
	u32 cctl,phyifctrl;
	u8 byte;

	/* Initialize PHY */

	// Move PHY to P3 to program SSC
	phyifctrl = dptx_readl(dptx, DPTX_PHYIF_CTRL);

	phyifctrl |= (2 << DPTX_PHYIF_CTRL_LANE_PWRDOWN_SHIFT);// move phy to P3 state

	dptx_writel(dptx, DPTX_PHYIF_CTRL, phyifctrl);


	retval = dptx_phy_wait_busy(dptx, DPTX_MAX_LINK_LANES);
        if (retval) {
		dev_err(xdev->dev, "Timed out waiting for PHY BUSY\n");
 		return retval;
	}

	dev_dbg(xdev->dev,"link lanes:%x\n",dptx->link.lanes);
#if defined SUPPORT_PANGU_EMU_DP
	dptx_phy_set_rate(dptx, 0); //emu workaround
#else
	dptx_phy_set_rate(dptx, dptx->link.rate); //emu workaround
#endif
	
	retval = dptx_phy_wait_busy(dptx, dptx->link.lanes);
	if (retval) {
		dev_err(xdev->dev, "Timed out waiting for PHY BUSY\n");
		return retval;
	}

	dev_dbg(xdev->dev, "set rate done got po state\n");
	dptx_phy_set_lanes(dptx, dptx->link.lanes);

#if !defined SUPPORT_PANGU_EMU_DP
	dev_dbg(xdev->dev, "xdx p platrom\n");
	dptx_phy_set_powerdown(dptx, dptx->link.lanes);
#endif
	//move the phy to p0
	phyifctrl = dptx_readl(dptx, DPTX_PHYIF_CTRL);

	phyifctrl &= ~DPTX_PHYIF_CTRL_LANE_PWRDOWN_MASK;// P0

	dptx_writel(dptx, DPTX_PHYIF_CTRL, phyifctrl);


	/* Wait for PHY busy */
	retval = dptx_phy_wait_busy(dptx, dptx->link.lanes);
	if (retval) {
		dev_err(xdev->dev, "Timed out waiting for PHY BUSY\n");
		return retval;
	}

	/* Set PHY_TX_EQ */

	dptx_writel(dptx, DPTX_PHY_TX_EQ,0);
	dev_dbg(xdev->dev, "---training start---\n");
	dptx_phy_set_pattern(dptx, DPTX_PHYIF_CTRL_TPS_NONE);
	retval = dptx_link_training_pattern_set(dptx,
						DP_TRAINING_PATTERN_DISABLE);
	if (retval){
		dev_err(xdev->dev,"dp pattern set error %x\n",retval);
		return retval;
	}
	//emu workaround
#if defined SUPPORT_PANGU_EMU_DP
	dptx_phy_enable_xmit(dptx, dptx->link.lanes, false);
#else
	dptx_phy_enable_xmit(dptx, dptx->link.lanes, true);
#endif

	retval = dptx_phy_rate_to_bw(dptx->link.rate);
	if (retval < 0){
		dev_err(xdev->dev,"dptx rate to bw fail:%x\n",retval);
		return retval;
	}

	byte = retval;
	retval = dptx_write_dpcd(dptx, DP_LINK_BW_SET, byte);
	if (retval){
		dev_err(xdev->dev,"dp link bw set fail %x\n",retval);
		return retval;
	}

	byte = dptx->link.lanes;
	cctl = dptx_readl(dptx, DPTX_CCTL);

	if (drm_dp_enhanced_frame_cap(dptx->rx_caps)) {
		byte |= DP_ENHANCED_FRAME_CAP;
		cctl |= DPTX_CCTL_ENH_FRAME_EN;
	} else {
		cctl &= ~DPTX_CCTL_ENH_FRAME_EN;
	}

	if(dptx->scramble_dis == 1){
		cctl |= BIT(0);
	}

	dptx_writel(dptx, DPTX_CCTL, cctl);

	retval = dptx_write_dpcd(dptx, DP_LANE_COUNT_SET, byte);
	if (retval){
		dev_err(xdev->dev,"lane count set fail %x\n",retval);
		return retval;
	}
	/* C10 PHY ... check if SSC is enabled and program DPCD*/
	if(dptx->ssc_en && dptx_sink_enabled_ssc(dptx))
		byte = DP_SPREAD_AMP_0_5;
	else
		byte = 0;

	retval = dptx_write_dpcd(dptx, DP_DOWNSPREAD_CTRL, byte);
	if (retval){
		dev_err(xdev->dev,"DP_DOWNSPREAD_CTRL fail %x\n",retval);
		return retval;
	}

	byte = 1;
	retval = dptx_write_dpcd(dptx, DP_MAIN_LINK_CHANNEL_CODING_SET, byte);
	if (retval){
		dev_err(xdev->dev,"DP_MAIN_LINK_CHANNEL_CODING_SET fail %x\n",retval);
		return retval;
	}
	return 0;
}

int dptx_link_wait_cr_and_adjust(struct dptx *dptx, bool ch_eq)
{
	struct xdx_output *output = dptx->output;
	struct xdx_dm *dm = output_to_dm(output);
	struct xdx_device *xdev = dm_to_xdev(dm);
	int i, j;
	int retval;
	int changed = 0;
	bool done = false;

	dev_dbg(xdev->dev, "%s:\n", __func__);

	dev_dbg(xdev->dev,"%s: link lanes:%x\n",__func__,dptx->link.lanes);
	retval = dptx_link_check_cr_done(dptx, &done, 0);
	dev_dbg(xdev->dev,"%s: link lanes:%x\n",__func__,dptx->link.lanes);
	if (retval)
		return retval;

	if (done)
		return 0;

	/* Try each adjustment setting 5 times */
	for (i = 0; i < 10; i++) {
		retval = dptx_link_adjust_drive_settings(dptx, &changed, false);
		if (retval)
			return retval;

		/* Reset iteration count if we changed settings */
		if (changed)
			i = 0;

		retval = dptx_link_check_cr_done(dptx, &done, 0);
		if (retval)
			return retval;

		if (done)
		{
			dev_dbg(xdev->dev,"%s: store cr done value vs = %d, pe = %d \n",__func__, dptx->link.vswing_level[0], dptx->link.preemp_level[0]);
			for (j = 0; j < dptx->link.lanes; j++) {
				dptx->link.cr_vswing_level[j] = dptx->link.vswing_level[j];
				dptx->link.cr_preemp_level[j] = dptx->link.preemp_level[j];
			}

			return 0;
		}
		/* TODO check for all lanes? */
		/* Failed and reached the maximum voltage swing */
		if (dptx->link.vswing_level[0] == 3)
			return -EPROTO;
	}

	return -EPROTO;
}

void emu_workaround(struct dptx *dptx)
{
	int retval;

	retval = dptx_write_dpcd(dptx, 0x106, 3);
	retval = dptx_write_dpcd(dptx, 0x105, 3);
	retval = dptx_write_dpcd(dptx, 0x104, 3);
	retval = dptx_write_dpcd(dptx, 0x103, 3);   
}

int dptx_link_cr(struct dptx *dptx)
{
	struct xdx_output *output = dptx->output;
	struct xdx_dm *dm = output_to_dm(output);
	struct xdx_device *xdev = dm_to_xdev(dm);
	int retval;

	dev_dbg(xdev->dev, "%s:  link_cr start\n", __func__);
#if !defined SUPPORT_PANGU_EMU_DP
	dptx_phy_set_pattern(dptx, DPTX_PHYIF_CTRL_TPS_1);
#endif
	if(dptx->scramble_dis == 0){
	    retval = dptx_link_training_pattern_set(dptx, DP_TRAINING_PATTERN_1);
	}else{
		retval = dptx_link_training_pattern_set(dptx, DP_TRAINING_PATTERN_1 | 0x20);
	}
	if (retval)
		return retval;

#if defined SUPPORT_PANGU_EMU_DP
	dptx_phy_enable_xmit(dptx, dptx->link.lanes, true);//emu workaround ++
	dptx_phy_set_pattern(dptx, DPTX_PHYIF_CTRL_TPS_1); //emu workaround ++
#endif
	dptx->link.vswing_level[0] = dptx->link.vswing_level[1] = 0;
	dptx->link.vswing_level[2] = dptx->link.vswing_level[3] = 0;
	dptx->link.preemp_level[0] = dptx->link.preemp_level[1] = 0;
	dptx->link.preemp_level[2] = dptx->link.preemp_level[3] = 0;
	dptx_link_training_lanes_set(dptx);
	dev_dbg(xdev->dev,"%s: link lanes:%x\n",__func__,dptx->link.lanes);
	return dptx_link_wait_cr_and_adjust(dptx, false);
}

int dptx_link_ch_eq(struct dptx *dptx)
{
	struct xdx_output *output = dptx->output;
	struct xdx_dm *dm = output_to_dm(output);
	struct xdx_device *xdev = dm_to_xdev(dm);
	int retval;
	bool cr_done;
	bool ch_eq_done;
	unsigned int pattern;
	unsigned int i;
	u8 dp_pattern;
	bool try_our_adjust = false;

	dev_dbg(xdev->dev, "%s:\n", __func__);

	switch(dptx->link.rate) {
	case DPTX_PHYIF_CTRL_RATE_HBR3:
		pattern = DPTX_PHYIF_CTRL_TPS_4;
		dp_pattern = DP_TRAINING_PATTERN_4;
		break;
		/* Fall through */
	case DPTX_PHYIF_CTRL_RATE_HBR2:
		pattern = DPTX_PHYIF_CTRL_TPS_3;
		dp_pattern = DP_TRAINING_PATTERN_3;
		break;
		/* Fall through */
	case DPTX_PHYIF_CTRL_RATE_RBR:
	case DPTX_PHYIF_CTRL_RATE_HBR:
		pattern = DPTX_PHYIF_CTRL_TPS_2;
		dp_pattern = DP_TRAINING_PATTERN_2;
		break;
	default:
		WARN(1, "Invalid rate %d\n", dptx->link.rate);
		return -EINVAL;
	}
#if !defined SUPPORT_PANGU_EMU_DP
    //emu workaround --
	dptx_phy_set_pattern(dptx, pattern);
#endif
	/* TODO this needs to be different for other versions of
	 * DPRX
	 */
	if(dptx->scramble_dis == 0){
		dev_dbg(xdev->dev, "scrambling enable\n");
		retval = dptx_link_training_pattern_set(dptx, dp_pattern);
	}else{
		if(dp_pattern != DP_TRAINING_PATTERN_4) {
			retval = dptx_link_training_pattern_set(dptx, dp_pattern | 0x20);
		} else {
			retval = dptx_link_training_pattern_set(dptx, dp_pattern);

			dev_dbg(xdev->dev, "%s:  Enabling scrambling for TPS4\n", __func__);
		}
	}
#if defined SUPPORT_PANGU_EMU_DP
	emu_workaround(dptx);//emu workaround ++
	dptx_phy_set_pattern(dptx, pattern);//emu workaround ++
#endif
	if (retval)
		return retval;

	for (i = 0; i < 10; i++) {
		retval = dptx_link_check_ch_eq_done(dptx,
							&cr_done,
							&ch_eq_done);

		if (retval)
			return retval;

		dptx->cr_fail = false;

		if ((!cr_done) && (try_our_adjust == true)) {
			dptx->cr_fail = true;
			return -EPROTO;
		}

		if (ch_eq_done)
			return 0;
		if((i >= 10) || !cr_done)
		{
			dev_dbg(xdev->dev, "eq fail, we try to use cr vswing value to try again \n");
			try_our_adjust = true;
			i = 0;
		}
		retval = dptx_link_adjust_drive_settings(dptx, NULL, try_our_adjust);
		if (retval)
			return retval;
	}

	return -EPROTO;
}

int dptx_link_reduce_rate(struct dptx *dptx)
{
	struct xdx_output *output = dptx->output;
	struct xdx_dm *dm = output_to_dm(output);
	struct xdx_device *xdev = dm_to_xdev(dm);
	unsigned int rate = dptx->link.rate;

	switch (rate) {
	case DPTX_PHYIF_CTRL_RATE_RBR:
		return -EPROTO;
	case DPTX_PHYIF_CTRL_RATE_HBR:
		rate = DPTX_PHYIF_CTRL_RATE_RBR;
		break;
	case DPTX_PHYIF_CTRL_RATE_HBR2:
		rate = DPTX_PHYIF_CTRL_RATE_HBR;
		break;
	case DPTX_PHYIF_CTRL_RATE_HBR3:
		rate = DPTX_PHYIF_CTRL_RATE_HBR2;
		break;
	}

	dev_dbg(xdev->dev, "%s: Reducing rate from %d to %d\n",
		 __func__, dptx->link.rate, rate);
	dptx->link.rate = rate;
	return 0;
}

int dptx_link_reduce_lanes(struct dptx *dptx)
{
	struct xdx_output *output = dptx->output;
	struct xdx_dm *dm = output_to_dm(output);
	struct xdx_device *xdev = dm_to_xdev(dm);
	unsigned int lanes;

	switch (dptx->link.lanes) {
	case 4:
		lanes = 2;
		break;
	case 2:
		lanes = 1;
		break;
	case 1:
	default:
		return -EPROTO;
	}

	dev_dbg(xdev->dev, "%s: Reducing lanes from %d to %d\n",
		 __func__, dptx->link.lanes, lanes);
	dptx->link.lanes = lanes;
	return 0;
}

int dptx_link_training(struct dptx *dptx, u8 rate, u8 lanes)
{
	struct xdx_output *output = dptx->output;
	struct xdx_dm *dm = output_to_dm(output);
	struct xdx_device *xdev = dm_to_xdev(dm);
	int retval,retval1;
	u8 byte;
	u32 hpd_sts;

	retval = dptx_link_training_init(dptx, rate, lanes);
	if (retval)
		goto fail;

again:
	retval = dptx_link_training_start(dptx);
	if (retval)
		goto fail;
	retval = dptx_link_cr(dptx);
	if (retval) {
		if (retval == -EPROTO) {
			if (dptx_link_reduce_rate(dptx)) {
				/* TODO If CR_DONE bits for some lanes
				 * are set, we should reduce lanes to
				 * those lanes.
				 */
				if (dptx_link_reduce_lanes(dptx)) {
					retval = -EPROTO;
					goto fail;
				}
				else {
					/*
					 * Check clock recovery status (LANE0_CR_DONE bit) in LANE0_1_STATUS DPCD register
					 * and fail training, if clock recovery failed for Lane 0
					 */
					if (!(dptx->link.status[0] & 1))
						goto fail;
				}
			}

			dptx_link_training_init(dptx,
						dptx->link.rate,
						dptx->link.lanes);
			goto again;
		} else {
        			goto fail;
		}
	}

	retval = dptx_link_ch_eq(dptx);
	if (retval) {
		if (retval == -EPROTO) {
			if(!dptx->cr_fail) { 
				dev_dbg(xdev->dev, "harutk ----- link training failure %0x, %0x", retval, -EPROTO);
				if(dptx->link.lanes == 1) {
					if(dptx_link_reduce_rate(dptx))  {
						goto fail;
					}
					dptx->link.lanes = dptx->max_lanes;
				} else {
					dptx_link_reduce_lanes(dptx);
				} 
			} else {
				if (dptx_link_reduce_rate(dptx)) {
					if (dptx_link_reduce_lanes(dptx)) {
						retval = -EPROTO;
						goto fail;
					}
				}
			}

			dptx_link_training_init(dptx,
						dptx->link.rate,
						dptx->link.lanes);
			goto again;
		} else {
			goto fail;
		}
	}

	dptx_phy_set_pattern(dptx, DPTX_PHYIF_CTRL_TPS_NONE);

	retval = dptx_link_training_pattern_set(dptx,
						DP_TRAINING_PATTERN_DISABLE);
	if (retval){
		goto fail;
	}
	dptx_phy_enable_xmit(dptx, dptx->link.lanes, true);
	dptx->link.trained = true;

	/* Branch device detection */
	retval = dptx_read_dpcd(dptx, DP_SINK_COUNT, &byte);
	if (retval)
		return retval;

	retval = dptx_read_dpcd(dptx, 0x2002, &byte);
	if (retval)
		return retval;

	dptx_video_ts_change(dptx, 0);
	dev_dbg(xdev->dev, "Link training succeeded rate=%d lanes=%d\n",
		 dptx->link.rate, dptx->link.lanes);

	return 0;

fail:
	hpd_sts = dptx_readl(dptx, DPTX_HPDSTS);

	if(hpd_sts & DPTX_HPDSTS_STATUS) 
	{
		dptx_phy_set_pattern(dptx, DPTX_PHYIF_CTRL_TPS_NONE);
		retval1 = dptx_link_training_pattern_set(dptx, DP_TRAINING_PATTERN_DISABLE);         
		if (retval1)
			return retval1;
		dev_err(xdev->dev, "Link training failed %d\n",retval);
	} else {
		dev_err(xdev->dev, "Link training failed  as sink is disconnected %d\n", retval);
	}
	return retval;
}

int dptx_link_check_status(struct dptx *dptx)
{
	struct xdx_output *output = dptx->output;
	struct xdx_dm *dm = output_to_dm(output);
	struct xdx_device *xdev = dm_to_xdev(dm);
	int retval;
	u8 byte;
	u8 bytes[2];

	retval = dptx_read_bytes_from_dpcd(dptx, DP_SINK_COUNT,
						bytes, 2);
	if (retval)
		return retval;

	retval = dptx_link_read_status(dptx);
	if (retval)
		return retval;

	byte = dptx->link.status[DP_LANE_ALIGN_STATUS_UPDATED -
				 DP_LANE0_1_STATUS];

	if (!(byte & DP_LINK_STATUS_UPDATED))
		return 0;

	/* Check if need to retrain link */
	if ((!drm_dp_channel_eq_ok(dptx->link.status, dptx->link.lanes) ||
		!drm_dp_clock_recovery_ok(dptx->link.status, dptx->link.lanes))) {
		dev_dbg(xdev->dev, "%s: Retraining link\n", __func__);
		return true;
	}

	return false;
}
