#include <linux/moduleparam.h>
#include <linux/version.h>
#include <linux/component.h>
#include <linux/module.h>
#include <linux/delay.h>

#include <drm/drm_crtc.h>
#include <drm/drm_crtc_helper.h>
#include <drm/drm_of.h>
#include <drm/drm_edid.h>

#include <drm/drm_atomic_helper.h>
#if((LINUX_VERSION_CODE>=KERNEL_VERSION(5,18,0))&&(LINUX_VERSION_CODE<KERNEL_VERSION(5,19,0)))
#include <drm/dp/drm_dp_helper.h>
#else
#include <drm/drm_dp_helper.h>
#endif
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 1, 0))
#include <drm/drm_probe_helper.h>
#endif

#include "dm.h"
#include "xdx_aux.h"
#include "link.h"
#include "core.h"
#include "xdx_dp_accel.h"
#include "dtd.h"
#include "xdx_dp_debugfs.h"
#include "xdx_device.h"
#include "dm.h"
#include "doorbell.h"
#include "../../../audio/xdxaudio_hal.h"
#include "rgxfwutils.h"
#include "pvrsrv_error.h"
#include "pvr_drv.h"

#define CLOCK_4K_TIMING 594000
#define EDID_SIZE(ext) (ext ? 256 : 128)
#define SUPPORT_DP_CTS 0

static const struct drm_display_mode default_mode = {
	DRM_MODE("1920x1080", DRM_MODE_TYPE_DRIVER, 148500, 1920, 2008, 2052,
		 2200, 0, 1080, 1084, 1089, 1125, 0,
		 DRM_MODE_FLAG_PHSYNC | DRM_MODE_FLAG_PVSYNC),
	.picture_aspect_ratio = HDMI_PICTURE_ASPECT_16_9,
};
static bool force_connected = false;

static void xdx_dp_hpd_handler(struct dp_dev *DisplayPort);
static void xdx_dp_hpd_delayed_handler(struct work_struct *work);

static enum drm_connector_status
xdx_dp_helper_connector_detect(struct drm_connector *connector, bool force)
{
	xdx_connector_to_dp(connector);
	enum drm_connector_status status;
	u32 hpd_status;
	struct dptx *dptx;
	dptx = dp->dptx;

	hpd_status = dptx_readl(dptx, DPTX_HPDSTS);
	if (hpd_status & (DPTX_HPDSTS_STATUS)) {
		status = connector_status_connected;
	} else {	
		status = connector_status_disconnected;
	}

	return status;

}

static void xdx_dp_helper_connector_force(struct drm_connector *connector)
{
	struct drm_device *dev = connector->dev;
	if (force_connected && connector) {
#if defined(DP_DEBUG)
		xdx_connector_to_dp(connector);
		WARN(dp == NULL, "container dp_dev err %s", __FUNCTION__);
		dev_dbg(dev->dev, "%s: skip EDID, force status connected\n",
			__func__);
#endif
		connector->status = connector_status_connected;
	}
}

static void
xdx_dp_helper_connector_destroy(struct drm_connector *connector)
{
	xdx_connector_to_dp(connector);
	WARN(dp == NULL, "container dp_dev err %s", __FUNCTION__);
	/* Disable all interrupts */
	drm_connector_unregister(connector);
	drm_connector_cleanup(connector);
}

static struct drm_display_mode *
xdx_dp_connector_native_mode(struct drm_connector *connector)
{
	xdx_connector_to_dp(connector);
	struct drm_device *dev = connector->dev;
	struct drm_display_mode *mode, *largest = NULL;
	int high_w = 0, high_h = 0; 
	unsigned int high_v = 0;
	const int vrefresh_default = 60;

	/* loop for all probed mode added from drm_add_edid_modes */
	list_for_each_entry(mode, &dp->connector.probed_modes, head) {
		dp->vrefresh = drm_mode_vrefresh(mode);
		if (!dp->vrefresh)
		{
			dp->vrefresh = vrefresh_default;
			DRM_INFO_ONCE(
				"vertical refresh rate is zero, defaulting to %d\n",
				dp->vrefresh);
		}

		/* Use preferred mode if there is one.. */
		if (mode->type & DRM_MODE_TYPE_PREFERRED) {
			dev_dbg(dev->dev, "Retrieving native mode from preferred\n");
			return drm_mode_duplicate(dev, mode);
		}

		/*
		 * Otherwise, take the resolution with the largest width, then
		 * height, then vertical refresh
		 */
		if (mode->hdisplay < high_w)
			continue;

		if (mode->hdisplay == high_w && mode->vdisplay < high_h)
			continue;

		if (mode->hdisplay == high_w && mode->vdisplay == high_h &&
			dp->vrefresh < high_v)
			continue;

		high_w = mode->hdisplay;
		high_h = mode->vdisplay;
		high_v = dp->vrefresh;
		largest = mode;
	}

	dev_dbg(dev->dev, "native mode from largest: %dx%d@%u\n",
			high_w, high_h, high_v);

	return largest ? drm_mode_duplicate(dev, largest) : NULL;
}

/* PLL list supported by FW */
static const u32 pll_clk[] = {
	25175,  25200,   27000,   33750,   46970,  65000,  74250,  79500,  296000, \
	83500,  99000,   101000,  106500, 108000, 121750, 131481, 146250, \
	148500, 156750,  162000,  193250,  204750, 218250, 234000, 268500, 348500, \
	443250, 556744,  594000,  241500,  497750, 592000, 533250, 666610, 755480, \
	888810, 1066610, 1188000, 1331070, 94500,  122500, 135000, 157000, 198000, \
	229500, 281250,  297000,  296703,  148352, 74176,  27027,  30240,  28320,  \
	31500, 296000,   174500,  156000,  40000,  36000,  78750,  75000,  50000,  \
	57284, 49500, 319750, 543500, 325080, 571000
};

/* PHILIPS 276B9: 2560x1440; pixel clock: 0x5e56=241500 */
/*
static u8 default_edid[256] = {
	0x00, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0x00,
	0x41, 0x0c, 0x5d, 0x09, 0xf8, 0x04, 0x00, 0x00,
	0x2b, 0x1e, 0x01, 0x04, 0xa5, 0x3c, 0x22, 0x78,
	0x3a, 0xb3, 0x75, 0xac, 0x52, 0x4e, 0xa0, 0x26,
	0x0f, 0x50, 0x54, 0xbf, 0xef, 0x00, 0xd1, 0xc0,
	0xb3, 0x00, 0x95, 0x00, 0x81, 0x80, 0x81, 0x40,
	0x81, 0xc0, 0x01, 0x01, 0x01, 0x01, 0x56, 0x5e,
	0x00, 0xa0, 0xa0, 0xa0, 0x29, 0x50, 0x30, 0x20,
	0x35, 0x00, 0x55, 0x50, 0x21, 0x00, 0x00, 0x1e,
	0x00, 0x00, 0x00, 0xff, 0x00, 0x55, 0x48, 0x42,
	0x32, 0x30, 0x34, 0x33, 0x30, 0x30, 0x31, 0x32,
	0x37, 0x32, 0x00, 0x00, 0x00, 0xfc, 0x00, 0x50,
	0x48, 0x4c, 0x20, 0x32, 0x37, 0x36, 0x42, 0x39,
	0x0a, 0x20, 0x20, 0x20, 0x00, 0x00, 0x00, 0xfd,
	0x00, 0x30, 0x4b, 0x1e, 0x72, 0x1e, 0x01, 0x0a,
	0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x01, 0x1a,

	0x02, 0x03, 0x18, 0xf1, 0x4b, 0x01, 0x03, 0x05,
	0x14, 0x04, 0x13, 0x1f, 0x12, 0x02, 0x11, 0x90,
	0x23, 0x09, 0x07, 0x07, 0x83, 0x01, 0x00, 0x00,
	0xa0, 0x73, 0x00, 0x6a, 0xa0, 0xa0, 0x29, 0x50,
	0x08, 0x20, 0x35, 0x00, 0x55, 0x50, 0x21, 0x00,
	0x00, 0x1a, 0x2a, 0x44, 0x80, 0xa0, 0x70, 0x38,
	0x27, 0x40, 0x30, 0x20, 0x35, 0x00, 0x55, 0x50,
	0x21, 0x00, 0x00, 0x1a, 0x02, 0x3a, 0x80, 0x18,
	0x71, 0x38, 0x2d, 0x40, 0x58, 0x2c, 0x45, 0x00,
	0x55, 0x50, 0x21, 0x00, 0x00, 0x1e, 0x01, 0x1d,
	0x00, 0x72, 0x51, 0xd0, 0x1e, 0x20, 0x6e, 0x28,
	0x55, 0x00, 0x55, 0x50, 0x21, 0x00, 0x00, 0x1e,
	0xf0, 0x3c, 0x00, 0xd0, 0x51, 0xa0, 0x35, 0x50,
	0x60, 0x88, 0x3a, 0x00, 0x55, 0x50, 0x21, 0x00,
	0x00, 0x1c, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xa1,
};
*/
#if 0
/* SAMSUNG 4K monitor, edid read from DisplayPort
 * 3840x2160; pixel clock: 0xd04d=533250
 */
static u8 default_edid[256] = {
	0x00, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0x00,
	0x4c, 0x2d, 0xf0, 0x71, 0x33, 0x37, 0x4c, 0x30,
	0x1c, 0x20, 0x01, 0x04, 0xb5, 0x46, 0x28, 0x78,
	0x3a, 0x56, 0x05, 0xb1, 0x4d, 0x3d, 0xae, 0x26,
	0x0d, 0x50, 0x54, 0xbf, 0xef, 0x80, 0x71, 0x4f,
	0x81, 0x00, 0x81, 0xc0, 0x81, 0x80, 0xa9, 0xc0,
	0xb3, 0x00, 0x95, 0x00, 0x01, 0x01, 0x4d, 0xd0,
	0x00, 0xa0, 0xf0, 0x70, 0x3e, 0x80, 0x30, 0x20,
	0x35, 0x00, 0xba, 0x89, 0x21, 0x00, 0x00, 0x1a,
	0x00, 0x00, 0x00, 0xfd, 0x00, 0x1e, 0x4b, 0x1e,
	0x87, 0x3c, 0x00, 0x0a, 0x20, 0x20, 0x20, 0x20,
	0x20, 0x20, 0x00, 0x00, 0x00, 0xfc, 0x00, 0x53,
	0x33, 0x32, 0x42, 0x38, 0x30, 0x50, 0x0a, 0x20,
	0x20, 0x20, 0x20, 0x20, 0x00, 0x00, 0x00, 0xff,
	0x00, 0x48, 0x4e, 0x54, 0x54, 0x37, 0x30, 0x30,
	0x30, 0x32, 0x32, 0x0a, 0x20, 0x20, 0x01, 0xb7,

	0x02, 0x03, 0x1f, 0xf0, 0x47, 0x5f, 0x10, 0x1f,
	0x04, 0x13, 0x03, 0x12, 0x23, 0x09, 0x07, 0x07,
	0x83, 0x01, 0x00, 0x00, 0xe3, 0x05, 0xc0, 0x00,
	0xe6, 0x06, 0x05, 0x01, 0x73, 0x5a, 0x00, 0x02,
	0x3a, 0x80, 0x18, 0x71, 0x38, 0x2d, 0x40, 0x58,
	0x2c, 0x45, 0x00, 0xba, 0x89, 0x21, 0x00, 0x00,
	0x1e, 0x56, 0x5e, 0x00, 0xa0, 0xa0, 0xa0, 0x29,
	0x50, 0x30, 0x20, 0x35, 0x00, 0xba, 0x89, 0x21,
	0x00, 0x00, 0x1a, 0x04, 0x74, 0x00, 0x30, 0xf2,
	0x70, 0x5a, 0x80, 0xb0, 0x58, 0x8a, 0x00, 0xba,
	0x89, 0x21, 0x00, 0x00, 0x1e, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x89,
};
#endif

/* query for valid clk supported by fw */
static bool is_pll_clk_supported (u32 clk)
{
	u32 i;
	for (i = 0; i < ARRAY_SIZE(pll_clk); i++)
	{
		if (pll_clk[i] == clk)
			return true;
	}
	return false;
}

static void xdx_dp_fix_bpc(struct drm_connector *connector)
{
	xdx_connector_to_dp(connector);
	struct drm_device *dev = connector->dev;
	struct dptx *dptx;
	u8 type;

	dptx = dp->dptx;

	dev_dbg(dev->dev, "%s(): bpc & color_format error, fix it", __func__);

	// DRM_COLOR_FORMAT\_ defines: (RGB444 | YCRCB444 | YCRCB422)
	dp->dp_data.color_formats = 7;

	if (dptx->rx_caps[DP_DPCD_REV] >= 0x11) {
		type = dptx->downstream_ports[0] & DP_DS_PORT_TYPE_MASK;
		if (type == DP_DS_PORT_TYPE_VGA) {
			dev_dbg(dev->dev, "downstream is vga port");

			if ((dptx->rx_caps[DP_DOWNSTREAMPORT_PRESENT] & DP_DETAILED_CAP_INFO_AVAILABLE) == 0) {
				dev_dbg(dev->dev, "set bpc to 8 directly");
				dp->dp_data.bpc = 8;
				return;
			}

			dev_dbg(dev->dev, "downstream_port[2]: 0x%x", dptx->downstream_ports[2]);
			switch (dptx->downstream_ports[2] & DP_DS_MAX_BPC_MASK) {
			case DP_DS_8BPC:
				dp->dp_data.bpc = 8;
				break;
			case DP_DS_10BPC:
				dp->dp_data.bpc = 10;
				break;
			case DP_DS_12BPC:
				dp->dp_data.bpc = 12;
				break;
			case DP_DS_16BPC:
				dp->dp_data.bpc = 16;
				break;
			default:
				dp->dp_data.bpc = 8;
				break;
				}
			}
	}
}

static int xdx_dp_connector_get_modes(struct drm_connector *connector)
{
	xdx_connector_to_dp(connector);
	struct drm_device *dev = connector->dev;
	int count = 0;
	struct drm_display_mode *mode;
	struct dptx *dptx;
	u16 i;

	dptx = dp->dptx;

	if (!dp->edid || dp->hpd_changed == 1 || (connector->force && connector->override_edid)) {
		if (dp->edid)
			kfree(dp->edid);

		dp->edid = drm_get_edid(connector, &dp->aux->ddc);
		if (!dp->edid) {
			dev_err(dev->dev, "Read EDID failed, will using same common low resolution timing.\n");
			return 0;
		}
		dp->hpd_changed = 0;
	}

	for (i = 0; i < EDID_SIZE(dp->edid->extensions); i += 8) {
		dev_dbg(dev->dev,"%d: %02x %02x %02x %02x %02x %02x %02x %02x\n", \
			i, *(u8 *)((u8 *)dp->edid + i),*(u8 *)((u8 *)dp->edid + i + 1), \
			*(u8 *)((u8 *)dp->edid + i + 2),*(u8 *)((u8 *)dp->edid + i +3), \
			*(u8 *)((u8 *)dp->edid + i+ 4),*(u8 *)((u8 *)dp->edid + i + 5), \
			*(u8 *)((u8 *)dp->edid + i + 6),*(u8 *)((u8 *)dp->edid + i +7));
	}

	// should use "drm_dp_read_dpcd_caps()"
	for (i = 0; i < DP_RECEIVER_CAP_SIZE; i++) {
		dptx_read_dpcd(dptx, i , &(dptx->rx_caps[i]));
	}

	drm_dp_dpcd_read(dp->aux, DP_DOWNSTREAM_PORT_0, dptx->downstream_ports, 0x10);

	drm_connector_update_edid_property(connector, dp->edid);
	count = drm_add_edid_modes(connector, dp->edid);

	dp->dp_data = connector->display_info;
	dp->dp_data.bpc = dp->dp_data.bpc > 10 ? 10 : dp->dp_data.bpc;
	dev_dbg(dev->dev,"bpc %d format %d\n", dp->dp_data.bpc, dp->dp_data.color_formats);

	/*
	 * in case of get the wrong bpc value from the EDID info of Analog Monitor
	 * (DP to VGA Convertor), fix it by fetch the right one from the
	 * DPCD_DOWNSTREAM_PORT capability register.
	 */
	if (dp->dp_data.bpc == 0) {
		xdx_dp_fix_bpc(connector);
	}

	dev_dbg(dev->dev,"bpc %d format %d\n", dp->dp_data.bpc, dp->dp_data.color_formats);

	/*
	 * Destroy the native mode, the attached monitor could have changed.
	 */
	if (dp->native_mode) {
		drm_mode_destroy(dev, dp->native_mode);
		dp->native_mode = NULL;
	}
	dp->native_mode = xdx_dp_connector_native_mode(connector);

	if (count == 0 && dp->native_mode) {
		/*
		 * Find the native mode if this is a digital panel, if we didn't
		 * find any modes through DDC previously add the native mode to
		 * the list of modes.
		 */
		dev_dbg(dev->dev,"cp native mode\n");
		mode = drm_mode_duplicate(dev, dp->native_mode);
		drm_mode_probed_add(connector, mode);
		dev_dbg(dev->dev,"mode add done\n");
		count = 1;
	}

	dev_dbg(dev->dev, "\nDP:Version: %d\nRevision: %d\nExtensions: %d\n# of modes: %d\n",
			dp->edid->version, dp->edid->revision,
			dp->edid->extensions, count);

	return count;
}

static int xdx_dp_mode_valid(struct drm_connector *connector,
						struct drm_display_mode *mode)
{
	struct drm_device *dev = connector->dev;

	if (!is_pll_clk_supported(mode->clock) || (mode->hdisplay >= 3840 && (drm_mode_vrefresh(mode) > 60)))
	{
		dev_dbg(dev->dev, "display mode:%dx%d vrefresh:%d pix_clock:%d un-supported", mode->hdisplay, mode->vdisplay, drm_mode_vrefresh(mode), mode->clock);
		return MODE_ERROR;
	} else {
		dev_dbg(dev->dev, "display mode:%dx%d vrefresh:%d pix_clock:%d	  supported", mode->hdisplay, mode->vdisplay, drm_mode_vrefresh(mode), mode->clock);
		return MODE_OK;
	}
}

static struct drm_encoder *xdx_dp_connector_best_encoder(
				struct drm_connector *connector)
{
	xdx_connector_to_dp(connector);
	return &dp->encoder;
}

static void xdx_dp_mode_setup_sink(struct dp_dev *dp, struct drm_display_mode *mode)
{
	struct xdx_output* output = dp_to_output(dp);
	struct xdx_dm* dm = output_to_dm(output);
	struct xdx_device* xdev = dm_to_xdev(dm);
	struct dptx *dptx;
	int err;
	u8 downspread;

	dptx = dp->dptx;

	dev_info(xdev->dev, "process SSC before link training\n");
	dptx_read_dpcd(dptx, 0x3, &downspread);
	dev_info(xdev->dev, "dpcd 0x3 SSC: 0x%x\n", downspread);
	if (downspread & 0x1) {
		dptx->ssc_en = true;
		dptx_enable_ssc(dptx);
	}

	dev_info(xdev->dev, "xdx_dp_mode_setup!\n");
	err = xdx_dp_phy_startup_fw(dptx, mode);
	if(err){
		dev_err(xdev->dev, "mode setup fail\n");
	}
	return;
}

static void dp_pll_config(struct xdx_device *xdev, u32 clock)
{
        u32 TimeOut, DBValue;
        DPDBCmd_T_NOSINK PixClockConf;
        PixClockConf.cmd = cDPCmdPLLConfig;
        PixClockConf.dcpixclock = clock;
        xdx_mmio_writel(xdev, SYS_PANGU_REG_HDMIDOORBELL,
                        (u64)PixClockConf.all & 0xffffffff);
        xdx_mmio_writel(xdev, SYS_PANGU_REG_HDMIDOORBELL + 4,
                        ((u64)PixClockConf.all >> 32) & 0xffffffff);
        TimeOut = 10000;
        while(1){
                DBValue = xdx_mmio_readl(xdev, SYS_PANGU_REG_HDMIDOORBELL);
                if((DBValue & 1) == 1)
                        break;
                TimeOut--;
                if(TimeOut == 0){
                        dev_dbg(xdev->dev,"powerup fail\n");
                        return;
                }
                udelay(1000);
        }
        return;
}

static void xdx_dp_mode_setup(struct dp_dev *dp, struct drm_display_mode *mode)
{
	struct xdx_output* output = dp_to_output(dp);
	struct xdx_dm* dm = output_to_dm(output);
	struct xdx_device* xdev = dm_to_xdev(dm);
	struct dptx *dptx;
	int retval;
	struct video_params *vparams;
	u8 rate,rate1;
	u8 lanes;
	struct dtd *mdtd;
	dptx = dp->dptx;

	vparams = &dptx->vparams;
	mdtd = &vparams->mdtd;
#if 0
	mode->hdisplay = 3840;
	mode->htotal = 4400;
	mode->hsync_start = 4016;
	mode->hsync_end = 4104;
	mode->vdisplay = 2160;
	mode->vtotal = 2250;
	mode->vsync_start = 2168;
	mode->vsync_end = 2178;
	mode->clock =  594000;
#endif

	mdtd->pixel_repetition_input = 0;
	mdtd->pixel_clock = mode->clock;
	mdtd->interlaced = 0;
	mdtd->h_active = mode->hdisplay;
	mdtd->h_blanking = mode->htotal - mode->hdisplay;
	mdtd->h_sync_offset = mode->hsync_start - mode->hdisplay;
	mdtd->h_sync_pulse_width = mode->hsync_end - mode->hsync_start;
	mdtd->h_sync_polarity = mode->flags & DRM_MODE_FLAG_PHSYNC ? 1 : 0;
	mdtd->v_active = mode->vdisplay;
	mdtd->v_blanking = mode->vtotal - mode->vdisplay;
	mdtd->v_sync_offset = mode->vsync_start - mode->vdisplay;
	mdtd->v_sync_pulse_width = mode->vsync_end - mode->vsync_start;
	mdtd->v_sync_polarity = mode->flags & DRM_MODE_FLAG_PVSYNC ? 1 : 0;
	switch(mode->picture_aspect_ratio) {
		case HDMI_PICTURE_ASPECT_4_3:
			mdtd->h_image_size = 4;
			mdtd->v_image_size = 3;
			break;

		case HDMI_PICTURE_ASPECT_16_9:
			mdtd->h_image_size = 16;
			mdtd->v_image_size = 9;
			break;

		case HDMI_PICTURE_ASPECT_64_27:
			mdtd->h_image_size = 64;
			mdtd->v_image_size = 27;
			break;

		case HDMI_PICTURE_ASPECT_256_135:
			mdtd->h_image_size = 256;
			mdtd->v_image_size = 135;
			break;

		default:
			mdtd->h_image_size = 4;
			mdtd->v_image_size = 3;
			break;
	}

	dev_dbg(xdev->dev, "mdtd->pixel_clock: %d\n", mdtd->pixel_clock);
	dev_dbg(xdev->dev, "mdtd->h_active: %d\n", mdtd->h_active);
	dev_dbg(xdev->dev, "mdtd->h_blanking: %d\n", mdtd->h_blanking);
	dev_dbg(xdev->dev, "mdtd->h_sync_offset: %d\n", mdtd->h_sync_offset);
	dev_dbg(xdev->dev, "mdtd->h_sync_pulse_width: %d\n", mdtd->h_sync_pulse_width);
	dev_dbg(xdev->dev, "mdtd->h_sync_pol: %d\n", mdtd->h_sync_polarity);
	dev_dbg(xdev->dev, "mdtd->v_active: %d\n", mdtd->v_active);
	dev_dbg(xdev->dev, "mdtd->v_blanking: %d\n", mdtd->v_blanking);
	dev_dbg(xdev->dev, "mdtd->v_sync_offset: %d\n", mdtd->v_sync_offset);
	dev_dbg(xdev->dev, "mdtd->v_sync_pulse_width: %d\n", mdtd->v_sync_pulse_width);
	dev_dbg(xdev->dev, "mdtd->v_sync_pol: %d\n", mdtd->v_sync_polarity);

#if 0
	vparams->bpc = dp->dp_data.bpc;

	if (dp->dp_data.color_formats & DRM_COLOR_FORMAT_RGB444)
		vparams->pix_enc = RGB;
	else if (dp->dp_data.color_formats & DRM_COLOR_FORMAT_YCRCB444)
		vparams->pix_enc = YCBCR444;
	else if (dp->dp_data.color_formats & DRM_COLOR_FORMAT_YCRCB422)
		vparams->pix_enc = YCBCR422;
	else
		vparams->pix_enc = YCBCR420;
#endif
	dev_dbg(xdev->dev,"bpc:%d  pix_enc:%d pixel_clock:%d mcode:%d video_format:%d\n",vparams->bpc,vparams->pix_enc, \
			mdtd->pixel_clock,vparams->mode,vparams->video_format);

	dp_pll_config(xdev, mdtd->pixel_clock);

#if 0
	xdx_tile_init(dptx);
	dptx_global_intr_dis(dptx);
	retval = dptx_core_init(dptx);
	if (retval){
		dev_err(xdev->dev,"DP core init fail\n");
		return;
	}
#endif

	dev_dbg(xdev->dev,"wait hdp status\n");
	{
		u32 count = 0;
		u32 hpdsts, intsts;
		while (1) {
			hpdsts = dptx_readl(dptx, DPTX_HPDSTS);
			if (hpdsts & DPTX_HPDSTS_STATUS) {
				dev_dbg(xdev->dev,"done hpd_status: %x\n", hpdsts);
				dptx_writel(dptx,DPTX_HPDSTS, hpdsts);
				intsts = dptx_readl(dptx, DPTX_ISTS);
				dptx_writel(dptx, DPTX_ISTS, intsts);
				break;
			}

			count++;
			mdelay(1);
			if (count >= 100) {
				dev_dbg(xdev->dev,"DP not connected, hpd_status: %x\n", hpdsts);
				return;
			}
		}
	}
	dptx->ssc_en = 0;
	dptx_enable_ssc(dptx);

	//get real lanes and rates number from dpcd of sink monitor
	dptx_read_dpcd(dptx, 2, &lanes);
	lanes = lanes & 0x1f;

	dptx_read_dpcd(dptx, 1, &rate1);
	switch( rate1 ) {
		case 0x6:
			rate = DPTX_PHYIF_CTRL_RATE_RBR;
			break;
		case 0xa:
			rate = DPTX_PHYIF_CTRL_RATE_HBR;
			break;
		case 0x14:
			rate = DPTX_PHYIF_CTRL_RATE_HBR2;
			break;
		case 0x1e:
			rate = DPTX_PHYIF_CTRL_RATE_HBR3;
			break;
		default:
			rate = DPTX_PHYIF_CTRL_RATE_RBR;
			dev_dbg(xdev->dev, "invalid dp rate\n");
	}
	dev_dbg(xdev->dev, "DP sink max rate:%d, max lanes:%d\n", rate, lanes);

	retval = dptx_video_ts_calculate(dptx, lanes, rate, vparams->bpc,
			 vparams->pix_enc, mdtd->pixel_clock);

	if (retval) {
		dev_dbg(xdev->dev, "DP tx video ts calculate error, retval:%d\n", retval);
                return;
        }

	retval = dptx_link_training(dptx, rate, lanes);

	if (retval){
		dev_dbg(xdev->dev, "DP tx link trainning error, retval:%d\n", retval);
		return;
	}

	dptx_video_config(dptx, 0);
	dptx_video_ts_change(dptx, 0);

	dptx_enable_default_video_stream(dptx, 0);
}

int xdx_dp_phy_power_down(struct dp_dev *dp)
{
	/* For dptx phy power down by placing PHY in reset */
	//need add code
	struct xdx_output *output = dp_to_output(dp);
	struct xdx_dm *dm = output_to_dm(output);
	struct xdx_device *xdev = dm_to_xdev(dm);
	u32 phyifctl;

	dev_dbg(xdev->dev,"%s\n",__func__);
	phyifctl = dptx_readl(dp->dptx, DPTX_PHYIF_CTRL);
	phyifctl = phyifctl | (3 << DPTX_PHYIF_CTRL_LANE_PWRDOWN_SHIFT);
	dptx_writel(dp->dptx, DPTX_PHYIF_CTRL, phyifctl);
	return 0;
}

static void
xdx_dp_encoder_destroy(struct drm_encoder *encoder)
{
	drm_encoder_cleanup(encoder);
}

static void xdx_dp_poweron(struct dp_dev *dp)
{
#if 1
#ifdef SUPPORT_AUDIO
	struct xdx_output *output = dp_to_output(dp);
	struct xdx_dm *dm = output_to_dm(output);
	struct xdx_device *xdev = dm_to_xdev(dm);
#endif
#endif

	if (xdev->asic_type == CHIP_PANGU_B0) {
		xdx_dp_mode_setup_sink(dp, dp->dp_mode);
	} else {
		xdx_dp_mode_setup(dp, dp->dp_mode);
	}

#if 1
#ifdef SUPPORT_AUDIO
	if(xdx_audio_drv_hotPlug(xdev, XDX_AUDIO_MODULE_DP, XDX_AUDIO_MODULE_CONNECT))
	{
		dev_dbg(xdev->dev, "%s: Call audio DP plug error!\n", __func__);
	}
#endif
#endif
}

static void xdx_dp_poweroff(struct dp_dev *dp)
{
#ifdef SUPPORT_AUDIO
	struct xdx_output *output = dp_to_output(dp);
	struct xdx_dm *dm = output_to_dm(output);
	struct xdx_device *xdev = dm_to_xdev(dm);
#endif

#ifdef SUPPORT_AUDIO
	if(xdx_audio_drv_hotPlug(xdev, XDX_AUDIO_MODULE_DP, XDX_AUDIO_MODULE_DISCONNECT))
	{
		dev_dbg(xdev->dev, "%s: Call audio DP unplug error!\n", __func__);
	}
#endif

	xdx_dp_phy_power_down(dp);
}

static void xdx_dp_encoder_dpms(struct drm_encoder *encoder, int mode)
{
	xdx_encoder_to_dp(encoder);

	if (mode)
		xdx_dp_poweroff(dp);
	else
		xdx_dp_poweron(dp);
}

static void xdx_dp_encoder_mode_set(struct drm_encoder *encoder,
			struct drm_display_mode *mode,
			struct drm_display_mode *adjusted_mode)
{
	struct dp_dev *dp = container_of(encoder, struct dp_dev, encoder);
	struct xdx_output* output = dp_to_output(dp);
	struct xdx_dm* dm = output_to_dm(output);
	struct xdx_device* xdev = dm_to_xdev(dm);
	struct drm_device *ddev = xdev_to_drm(xdev);
	struct drm_crtc* crtc;
	struct dm_crtc_state* dm_state;
	u8 rate, lanes;
	int link_rate, color_dep;
	int ts, tu;
	struct video_params *vparams;
	drm_for_each_crtc (crtc, ddev)
	{
		if (drm_crtc_index(crtc) == drm_encoder_index(encoder))
			break;
	}

	dm_state = to_dm_crtc_state(crtc->state);
	dm_state->output_fmt = dp->dp_data.color_formats;

	vparams = &dp->dptx->vparams;
	#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 18, 0)
	if (dp->dp_data.color_formats & DRM_COLOR_FORMAT_RGB444)
                vparams->pix_enc = RGB;
        else if (dp->dp_data.color_formats & DRM_COLOR_FORMAT_YCRCB444)
                vparams->pix_enc = YCBCR444;
        else if (dp->dp_data.color_formats & DRM_COLOR_FORMAT_YCRCB422)
                vparams->pix_enc = YCBCR422;
        else
                vparams->pix_enc = YCBCR420;
	#else
	if (dp->dp_data.color_formats & DRM_COLOR_FORMAT_RGB444)
                vparams->pix_enc = RGB;
        else if (dp->dp_data.color_formats & DRM_COLOR_FORMAT_YCBCR444)
                vparams->pix_enc = YCBCR444;
        else if (dp->dp_data.color_formats & DRM_COLOR_FORMAT_YCBCR422)
                vparams->pix_enc = YCBCR422;
        else
                vparams->pix_enc = YCBCR420;
	#endif	

	//get real lanes and rates number from dpcd of sink monitor
	dptx_read_dpcd(dp->dptx, 2, &lanes);
	lanes = lanes & 0x1f;
	if (!lanes) {
		dev_dbg(xdev->dev, "invalid dp lanes, return\n");
		return;
	}

        dptx_read_dpcd(dp->dptx, 1, &rate);
        switch( rate ) {
                case 0x6:
                        link_rate = 162;
                        break;
                case 0xa:
                        link_rate = 270;
                        break;
                case 0x14:
                        link_rate = 540;
                        break;
                case 0x1e:
                        link_rate = 810;
                        break;
                default:
                        link_rate = 162;
                        dev_dbg(xdev->dev, "invalid dp rate\n");
			break;
        }

	vparams->bpc = dp->dp_data.bpc;
	while(1){
		switch (vparams->bpc) {
	                case COLOR_DEPTH_6:
	                        color_dep = 18;
	                        break;
	                case COLOR_DEPTH_8:
	                        if (vparams->pix_enc == YCBCR420)
	                                color_dep  = 12;
	                        else if (vparams->pix_enc == YCBCR422)
	                                color_dep = 16;
	                        else if (vparams->pix_enc == YONLY)
	                                color_dep = 8;
	                        else
	                                color_dep = 24;
	                        break;
	                case COLOR_DEPTH_10:
	                        if (vparams->pix_enc == YCBCR420)
	                                color_dep = 15;
	                        else if (vparams->pix_enc == YCBCR422)
	                                color_dep = 20;
	                        else if (vparams->pix_enc  == YONLY)
	                                color_dep = 10;
	                        else
	                                color_dep = 30;
	                        break;
	
	                case COLOR_DEPTH_12:
	                        if (vparams->pix_enc == YCBCR420)
	                                color_dep = 18;
	                        else if (vparams->pix_enc == YCBCR422)
	                                color_dep = 24;
	                        else if (vparams->pix_enc == YONLY)
	                                color_dep = 12;
	                        else
	                                color_dep = 36;
	                        break;
	
	                case COLOR_DEPTH_16:
	                        if (vparams->pix_enc == YCBCR420)
	                                color_dep = 24;
	                        else if (vparams->pix_enc == YCBCR422)
	                                color_dep = 32;
	                        else if (vparams->pix_enc == YONLY)
	                                color_dep = 16;
	                        else
	                                color_dep = 48;
	                        break;
	                default:
	                        color_dep = 18;
	                        break;
	        }
		if (dp->dptx->dsc)
	                color_dep = dp->dptx->vparams.dsc_bpp;
	        dev_dbg(xdev->dev, "color_dep:%d clock:%d lane_num:%d link_rate:%d\n",color_dep, mode->clock, lanes, link_rate);
	        ts = (8 * color_dep * mode->clock) / (lanes * link_rate);
	
	        tu  = ts / 1000;
	        if (tu >= 65) {
			if(vparams->bpc <= 8){
	                	dev_dbg(xdev->dev, "%s: tu(%d) > 65", __func__, tu);
	                	return;
			}else{
				vparams->bpc = 8;
	                	dev_dbg(xdev->dev, "%s: tu(%d) > 65, change to 8 bit\n", __func__, tu);
				continue;
			}
	        }

		dm_state->bpc = vparams->bpc;
		break;
	}

	if(dp->dp_mode)
	{
		drm_mode_destroy(ddev, dp->dp_mode);
		dp->dp_mode = NULL;
	}
	dp->dp_mode = drm_mode_duplicate(ddev, mode);
}

static void xdx_dp_encoder_commit(struct drm_encoder *encoder)
{
	xdx_encoder_to_dp(encoder);
	/* If forced, the driver has always got a modeline. Either
	 * via EDID, or via VESA fallback. The encoder has already
	 * been updated (stolen) in ->mode_set(), never attempt to
	 * re-initialise the dp status.
	 */
	if (force_connected)
		return;

	xdx_dp_poweron(dp);
}

static bool xdx_dp_encoder_mode_fixup(struct drm_encoder *encoder,
			const struct drm_display_mode *mode,
			struct drm_display_mode *adjusted_mode)
{
	return true;
}

static void xdx_dp_encoder_prepare(struct drm_encoder *encoder)
{
	xdx_encoder_to_dp(encoder);

	xdx_dp_poweroff(dp);
}


static const struct drm_encoder_funcs xdx_dp_encoder_funcs = {
	.destroy = xdx_dp_encoder_destroy,
};

static const struct drm_encoder_helper_funcs xdx_dp_encoder_helper_funcs = {
	.dpms = xdx_dp_encoder_dpms,
	.mode_set = xdx_dp_encoder_mode_set,
	.commit = xdx_dp_encoder_commit,
	.mode_fixup = xdx_dp_encoder_mode_fixup,
	.prepare = xdx_dp_encoder_prepare,
};

static const struct drm_connector_funcs xdx_dp_connector_funcs = {
	.dpms = drm_helper_connector_dpms,
	.reset = drm_atomic_helper_connector_reset,
	.detect = xdx_dp_helper_connector_detect,
	.force = xdx_dp_helper_connector_force,
	.fill_modes = drm_helper_probe_single_connector_modes,
	.destroy = xdx_dp_helper_connector_destroy,
	//.set_property = xdx_hdmi_helper_connector_set_property,
	.atomic_duplicate_state = drm_atomic_helper_connector_duplicate_state,
	.atomic_destroy_state = drm_atomic_helper_connector_destroy_state,
};

static const struct drm_connector_helper_funcs xdx_dp_connector_helper_funcs = {
	.get_modes = xdx_dp_connector_get_modes,
	.mode_valid = xdx_dp_mode_valid,
	/*
	 * For atomic, don't set atomic_best_encoder or best_encoder. This will
	 * cause the DRM core to fallback to drm_atomic_helper_best_encoder().
	 * This is fine as we only have a single connector and encoder.
	 */
	.best_encoder = xdx_dp_connector_best_encoder,
};

void xdx_dp_trans_para_fill(struct dptx *dptx, struct dp_trans_para *pass_para, struct drm_display_mode *mode, void* para_sram_vaddr)
{
	struct xdx_output *output = dptx->output;
	struct xdx_dm *dm = output_to_dm(output);
	struct xdx_device *xdev = dm_to_xdev(dm);

	struct video_params *vparams;
	vparams = &dptx->vparams;
	pass_para->mode = vparams->mode;
	pass_para->bpc = vparams->bpc;
	pass_para->multipixel = dptx->multipixel;
	pass_para->rate = dptx->link.rate;
	pass_para->dc_idx = dptx->dp->id;
	pass_para->clock = mode->clock;
	pass_para->hdisplay = mode->hdisplay;
	pass_para->htotal = mode->htotal;
	pass_para->hsync_start = mode->hsync_start;
	pass_para->hsync_end = mode->hsync_end;
	pass_para->flag = mode->flags;
	pass_para->vdisplay = mode->vdisplay;
	pass_para->vtotal = mode->vtotal;
	pass_para->vsync_start = mode->vsync_start;
	pass_para->vsync_end = mode->vsync_end;
	pass_para->picture_aspect_ratio = mode->picture_aspect_ratio;
	pass_para->mode_flag_pvsync = DRM_MODE_FLAG_PVSYNC;
	pass_para->dc_idx = dptx->dp->id;

	dev_info(xdev->dev,"clock=%d, hdisplay=%d, htotal=%d, hsync_start=%d, hsync_end=%d, flag=%d, vdisplay=%d, vtotal=%d, vsync_start=%d, vsync_end=%d, picture_aspect_ratio=%d\n",
		pass_para->clock, pass_para->hdisplay, pass_para->htotal, pass_para->hsync_start, pass_para->hsync_end, pass_para->flag, pass_para->vdisplay, pass_para->vtotal,
		pass_para->vsync_start, pass_para->vsync_end, pass_para->picture_aspect_ratio);

	memcpy_toio(para_sram_vaddr, pass_para, sizeof(struct dp_trans_para));

	(void)(pass_para->dc_idx);

}

int xdx_dp_phy_startup_fw(struct dptx *dptx, struct drm_display_mode *mode)
{
	struct xdx_output *output = dptx->output;
	struct xdx_dm *dm = output_to_dm(output);
	struct xdx_device *xdev = dm_to_xdev(dm);
	int ret;

	DPDBCmd_T_SINK PhyConfig;
	u32 TimeOut, DBValue;
	struct dp_trans_para *pass_para = kmalloc(sizeof(struct dp_trans_para), GFP_KERNEL);
	if(output->psDoorbellShareMemDesc[2 + dptx->dp->id] == NULL){
		ret = DPDoorbellShareMemAllocate(output,2 + dptx->dp->id);
		if(ret){
			dev_err(xdev->dev,"DP share ram allocate fail\n");
			return ret;
		}
	}
	xdx_dp_trans_para_fill(dptx, pass_para, mode, output->psDoorbellVirAddr[2 + dptx->dp->id]);

	dev_info(xdev->dev,"%s entry!\n",__func__);
	PhyConfig.all = 0;
	PhyConfig.cmd = cH02ADisplayDBCmdDP;
	PhyConfig.subCmd = cH02ADPSubCmdCNTLConfig;
	PhyConfig.dword = output->psDoorbellPhyAddr[2 + dptx->dp->id].uiAddr;
	xdx_dp_h2a_doorbell_w(dptx,(u64)PhyConfig.all);

	TimeOut = 10000;
	while(1)
	{
		DBValue = xdx_dp_h2a_doorbell_r(dptx);
		if((DBValue & 1) == 1)
			break;
		TimeOut--;
		if(TimeOut == 0){
			kfree(pass_para);
			return -1;
		}
		udelay(1000);
	}
	dev_info(xdev->dev,"%s end!\n",__func__);
	kfree(pass_para);
	return 0;
}

static int xdx_dp_register(struct dp_dev *dp, u32 crtc_mask)
{
	struct xdx_output *output = dp_to_output(dp);
	struct xdx_dm *dm = output_to_dm(output);
	struct xdx_device *xdev = dm_to_xdev(dm);
	struct drm_device *ddev = xdev_to_drm(xdev);
	int ret;

	INIT_WORK(&dp->hpd_work, xdx_dp_hpd_delayed_handler);
	dp->connector.polled = DRM_CONNECTOR_POLL_HPD;
	dp->connector.interlace_allowed = 0;
	dp->connector.doublescan_allowed = 0;

	drm_connector_init(ddev, &dp->connector, &xdx_dp_connector_funcs,
			   DRM_MODE_CONNECTOR_DisplayPort);

	drm_connector_helper_add(&dp->connector,
				 &xdx_dp_connector_helper_funcs);

	//dp->encoder.possible_crtcs = crtc_mask & BIT(1);
	//dp->encoder.possible_clones = 1 << 1;

    if (!dp->id) {
        dp->encoder.possible_clones = 1 << 1;
        dp->encoder.possible_crtcs = BIT(1);
    } else {
        dp->encoder.possible_clones = 1 << 3;
        dp->encoder.possible_crtcs = BIT(3);
    }

	ret = drm_encoder_init(ddev, &dp->encoder, &xdx_dp_encoder_funcs,
					DRM_MODE_ENCODER_TMDS, NULL);
	if (ret) {
		dev_err(xdev->dev, "failed to init drm encoder\n");
		return ret;
	}

	drm_encoder_helper_add(&dp->encoder, &xdx_dp_encoder_helper_funcs);
	dev_dbg(xdev->dev, "DP_Connector_ID=%d\n",dp->connector.base.id);
	ret = drm_connector_attach_encoder(&dp->connector, &dp->encoder);
	if (ret) {
		dev_err(xdev->dev,
			"can't attach [ENCODER:%d:%s] to [CONNECTOR:%d:%s] (err=%d)\n",
			dp->encoder.base.id, dp->encoder.name,
			dp->connector.base.id, dp->connector.name, ret);
	}

	output->connector = &dp->connector;
	output->encoder = &dp->encoder;

	return 0;
}

void dp_video_params_init(struct dp_dev *dp)
{
	struct dptx *dptx;
	struct video_params *params;

	dptx = dp->dptx;
	params = &dptx->vparams;
	
	dptx->cr_fail = false;
	 /* Should be disabled for HDCP. */
	dptx->mst = false;
	dptx->ssc_en = false;
	dptx->fec = false;
	dptx->dsc = false;
	dptx->streams = 1;
	dptx->multipixel = DPTX_MP_SINGLE_PIXEL;
	dptx->dummy_dtds_present = false;
	dptx->selected_est_timing = None;
	dptx->dp_pcs_width = 20;

	/* dptx_video_params_reset(); */
	params->pix_enc = RGB;
	params->mode = 1;
	dptx->scramble_dis = 0;

#if 0
	if (dptx->mst || dptx->dsc) {
		params->bpc = COLOR_DEPTH_6;
	} else {
		params->bpc = COLOR_DEPTH_8;
	}
#endif

	params->bpc = COLOR_DEPTH_8;
	params->colorimetry = ITU601;
	params->dynamic_range = VESA;
	params->aver_bytes_per_tu = 30;
	params->aver_bytes_per_tu_frac = 0;
	params->init_threshold = 15;
	params->pattern_mode = RAMP;
	params->refresh_rate = 60000;
	params->video_format = DMT;
}

#ifdef SUPPORT_DP_CTS
static int dptx_set_custom_pattern(struct dptx *dptx)
{
	int retval;
	u8 pattern0 , pattern1, pattern2, pattern3, pattern4, pattern5,
       pattern6, pattern7, pattern8, pattern9;

	u32 custompat0;
	u32 custompat1;
	u32 custompat2;

	retval = dptx_read_dpcd(dptx, DP_TEST_80BIT_CUSTOM_PATTERN_0, &pattern0);
	if (retval)
		return retval;

    retval = dptx_read_dpcd(dptx, DP_TEST_80BIT_CUSTOM_PATTERN_1, &pattern1);
    if (retval)
        return retval;

    retval = dptx_read_dpcd(dptx, DP_TEST_80BIT_CUSTOM_PATTERN_2, &pattern2);
    if (retval)
        return retval;

    retval = dptx_read_dpcd(dptx, DP_TEST_80BIT_CUSTOM_PATTERN_3, &pattern3);
    if (retval)
        return retval;

    retval = dptx_read_dpcd(dptx, DP_TEST_80BIT_CUSTOM_PATTERN_4, &pattern4);
    if (retval)
        return retval;

    retval = dptx_read_dpcd(dptx, DP_TEST_80BIT_CUSTOM_PATTERN_5, &pattern5);
    if (retval)
        return retval;

    retval = dptx_read_dpcd(dptx, DP_TEST_80BIT_CUSTOM_PATTERN_6, &pattern6);
    if (retval)
        return retval;

    retval = dptx_read_dpcd(dptx, DP_TEST_80BIT_CUSTOM_PATTERN_7, &pattern7);
    if (retval)
        return retval;

    retval = dptx_read_dpcd(dptx, DP_TEST_80BIT_CUSTOM_PATTERN_8, &pattern8);
    if (retval)
        return retval;

    retval = dptx_read_dpcd(dptx, DP_TEST_80BIT_CUSTOM_PATTERN_9, &pattern9);
    if (retval)
        return retval;

    /*
     *  Calculate 30,30 and 20 bits custom patterns depending on TEST_80BIT_CUSTOM_PATTERN sequence
     */
    custompat0 = ((((((pattern3 & (0xff >> 2)) << 8) | pattern2) << 8) | pattern1) << 8) | pattern0;
    custompat1 = ((((((((pattern7 & (0xf)) << 8) | pattern6) << 8) | pattern5) << 8) | pattern4) << 2) | ((pattern3 >> 6) & 0x3);
    custompat2 = (((pattern9 << 8) | pattern8) << 4) | ((pattern7 >> 4) & 0xf);

    dptx_writel(dptx, DPTX_CUSTOMPAT0, custompat0);
    dptx_writel(dptx, DPTX_CUSTOMPAT1, custompat1);
    dptx_writel(dptx, DPTX_CUSTOMPAT2, custompat2);

    return 0;
}

static void adjust_pcs_tx_eq(struct dptx *dptx)
{
	struct xdx_output *output = dptx->output;
    struct xdx_dm   *dm = output_to_dm(output);
    struct xdx_device *xdev = dm_to_xdev(dm);
	int retval;
	int i;
	u32 test_link_rate = 0x219;
	u8 rate;
    u8 lane_01;
    u8 lane_23;
	u32 reg;
	u32 read_val = 0;
	u16 val_0_2 = 0;
	u16 val_1_3 = 0;

	dev_dbg(xdev->dev, "%s()\n", __func__);

	// get rate
	dptx_read_dpcd(dptx, test_link_rate, &rate);

	if (rate >= 0x14) {
		dev_dbg(xdev->dev, "HBR2 or HBR3\n");
		return;
	}

	dev_dbg(xdev->dev, "RBR or HBR1\n");

	retval = dptx_read_dpcd(dptx, DP_ADJUST_REQUEST_LANE0_1, &lane_01);
    if (retval)
        return;
	dev_dbg(xdev->dev, "DP_ADJUST_REQUEST_LANE0_1: 0x%x\n", lane_01);

    retval = dptx_read_dpcd(dptx, DP_ADJUST_REQUEST_LANE2_3, &lane_23);
    if (retval)
        return;
	dev_dbg(xdev->dev, "DP_ADJUST_REQUEST_LANE2_3: 0x%x\n", lane_23);

	// pcs global register
	dptx_pcs_writel(dptx, 0xa, 0x400);

	// loop number should be real test lane number
    for (i = 0; i < dptx->link.lanes; i++) {
        u8 pe = 0;
        u8 vs = 0;

        switch (i) {
        case 0:
            pe = (lane_01 &  DP_ADJUST_PRE_EMPHASIS_LANE0_MASK)
                >> DP_ADJUST_PRE_EMPHASIS_LANE0_SHIFT;
            vs = (lane_01 &  DP_ADJUST_VOLTAGE_SWING_LANE0_MASK)
                >> DP_ADJUST_VOLTAGE_SWING_LANE0_SHIFT;
            break;
        case 1:
            pe = (lane_01 & DP_ADJUST_PRE_EMPHASIS_LANE1_MASK)
                >> DP_ADJUST_PRE_EMPHASIS_LANE1_SHIFT;
            vs = (lane_01 & DP_ADJUST_VOLTAGE_SWING_LANE1_MASK)
                >> DP_ADJUST_VOLTAGE_SWING_LANE1_SHIFT;
            break;
        case 2:
            pe = (lane_23 & DP_ADJUST_PRE_EMPHASIS_LANE0_MASK)
                >> DP_ADJUST_PRE_EMPHASIS_LANE0_SHIFT;
            vs = (lane_23 & DP_ADJUST_VOLTAGE_SWING_LANE0_MASK)
                >> DP_ADJUST_VOLTAGE_SWING_LANE0_SHIFT;
            break;
        case 3:
            pe = (lane_23 & DP_ADJUST_PRE_EMPHASIS_LANE1_MASK)
                >> DP_ADJUST_PRE_EMPHASIS_LANE1_SHIFT;
            vs = (lane_23 & DP_ADJUST_VOLTAGE_SWING_LANE1_MASK)
                >> DP_ADJUST_VOLTAGE_SWING_LANE1_SHIFT;
            break;
        default:
            break;
        }

        dev_dbg(xdev->dev, "vs: 0x%x\n", vs);
		dev_dbg(xdev->dev, "pe: 0x%x\n", pe);

		if (0x0 == vs) {
			if (0x0 == pe) {
				val_0_2 = 0x708;
				val_1_3 = 0x0;
			}
			if (0x1 == pe) {
				val_0_2 = 0x70a;
				val_1_3 = 0x8;
			}
			if (0x2 == pe) {
				val_0_2 = 0x70c;
				val_1_3 = 0x10;
			}
			if (0x3 == pe) {
				val_0_2 = 0x710;
				val_1_3 = 0x20;
			}
		}

		if (0x1 == vs) {
			if (0x0 == pe) {
				val_0_2 = 0x70c;
				val_1_3 = 0x0;
			}
			if (0x1 == pe) {
				val_0_2 = 0x70f;
				val_1_3 = 0xc;
			}
			if (0x2 == pe) {
				val_0_2 = 0x712;
				val_1_3 = 0x18;
			}
		}

		if (0x2 == vs) {
			if (0x0 == pe) {
				val_0_2 = 0x710;
				val_1_3 = 0x0;
			}
			if (0x1 == pe) {
				val_0_2 = 0x714;
				val_1_3 = 0x10;
			}
		}

        dev_dbg(xdev->dev, "val_0_2: 0x%x\n", val_0_2);
		dev_dbg(xdev->dev, "val_1_3: 0x%x\n", val_1_3);

		// set pcs eq
		reg = 0x5000 + i*0x100;
		dptx_pcs_writel(dptx, (reg+0), val_0_2);
		read_val = dptx_pcs_readl(dptx, (reg+0));
		dev_dbg(xdev->dev, "pcs write test: 0x%x\n", read_val);

		dptx_pcs_writel(dptx, (reg+1), val_1_3);
		dptx_pcs_writel(dptx, (reg+2), val_0_2);
		dptx_pcs_writel(dptx, (reg+3), val_1_3);

		// set pcs deskew enable
		reg = 0x1010 + i*0x100;
		dptx_pcs_writel(dptx, reg, 0xc);
	}
}

static int adjust_vswing_and_preemphasis(struct dptx *dptx)
{
    struct xdx_output *output = dptx->output;
    struct xdx_dm   *dm = output_to_dm(output);
    struct xdx_device *xdev = dm_to_xdev(dm);
    int retval;
    int i;
    u8 lane_01;
    u8 lane_23;
	u8 post_cursor2;
	u32 phytxeq;

    retval = dptx_read_dpcd(dptx, DP_ADJUST_REQUEST_LANE0_1, &lane_01);
    if (retval)
        return retval;
	dev_dbg(xdev->dev, "DP_ADJUST_REQUEST_LANE0_1: 0x%x\n", lane_01);

    retval = dptx_read_dpcd(dptx, DP_ADJUST_REQUEST_LANE2_3, &lane_23);
    if (retval)
        return retval;
	dev_dbg(xdev->dev, "DP_ADJUST_REQUEST_LANE2_3: 0x%x\n", lane_23);

	retval = dptx_read_dpcd(dptx, DP_ADJUST_REQUEST_POST_CURSOR2, &post_cursor2);
    if (retval)
        return retval;
	dev_dbg(xdev->dev, "DP_ADJUST_REQUEST_POST_CURSOR2: 0x%x\n", post_cursor2);

    for (i = 0; i < dptx->link.lanes; i++) {
        u8 pe = 0;
        u8 vs = 0;
		//u8 cursor2;

        switch (i) {
        case 0:
            pe = (lane_01 &  DP_ADJUST_PRE_EMPHASIS_LANE0_MASK)
                >> DP_ADJUST_PRE_EMPHASIS_LANE0_SHIFT;
            vs = (lane_01 &  DP_ADJUST_VOLTAGE_SWING_LANE0_MASK)
                >> DP_ADJUST_VOLTAGE_SWING_LANE0_SHIFT;
            //cursor2 = (post_cursor2 & DP_ADJUST_POST_CURSOR2_LANE0_MASK) >> DP_ADJUST_POST_CURSOR2_LANE0_SHIFT;
            break;
        case 1:
            pe = (lane_01 & DP_ADJUST_PRE_EMPHASIS_LANE1_MASK)
                >> DP_ADJUST_PRE_EMPHASIS_LANE1_SHIFT;
            vs = (lane_01 & DP_ADJUST_VOLTAGE_SWING_LANE1_MASK)
                >> DP_ADJUST_VOLTAGE_SWING_LANE1_SHIFT;
            //cursor2 = (post_cursor2 & DP_ADJUST_POST_CURSOR2_LANE1_MASK) >> DP_ADJUST_POST_CURSOR2_LANE1_SHIFT;
            break;
        case 2:
            pe = (lane_23 & DP_ADJUST_PRE_EMPHASIS_LANE0_MASK)
                >> DP_ADJUST_PRE_EMPHASIS_LANE0_SHIFT;
            vs = (lane_23 & DP_ADJUST_VOLTAGE_SWING_LANE0_MASK)
                >> DP_ADJUST_VOLTAGE_SWING_LANE0_SHIFT;
            //cursor2 = (post_cursor2 & DP_ADJUST_POST_CURSOR2_LANE2_MASK) >> DP_ADJUST_POST_CURSOR2_LANE2_SHIFT;
            break;
        case 3:
            pe = (lane_23 & DP_ADJUST_PRE_EMPHASIS_LANE1_MASK)
                >> DP_ADJUST_PRE_EMPHASIS_LANE1_SHIFT;
            vs = (lane_23 & DP_ADJUST_VOLTAGE_SWING_LANE1_MASK)
                >> DP_ADJUST_VOLTAGE_SWING_LANE1_SHIFT;
            //cursor2 = (post_cursor2 & DP_ADJUST_POST_CURSOR2_LANE3_MASK) >> DP_ADJUST_POST_CURSOR2_LANE3_SHIFT;
            break;
        default:
            break;
        }

        dptx_phy_set_pre_emphasis(dptx, i, pe);
        dptx_phy_set_vswing(dptx, i, vs);
		//dptx_phy_set_post_cursor2(dptx, i, cursor2);
    }

	phytxeq = dptx_readl(dptx, DPTX_PHY_TX_EQ);
	dev_dbg(xdev->dev, "PHY_TX_EQ: 0x%x\n", phytxeq);

    return 0;
}

static int handle_test_phy_pattern(struct dptx *dptx)
{
    struct xdx_output *output = dptx->output;
    struct xdx_dm   *dm = output_to_dm(output);
    struct xdx_device *xdev = dm_to_xdev(dm);
    u8 pattern;
	u32 test_link_rate = 0x219;
	u32 test_lane_count = 0x220;
	u8 rate1, lanes;
	u8 rate;
	//u8 downspread;
	u8 rev;
	u32 phyifctrl;
    int retval;

	// get pattern
    //retval = dptx_read_dpcd(dptx, DP_TEST_PHY_PATTERN, &pattern);
    retval = dptx_read_dpcd(dptx, 0x248, &pattern);
    if (retval)
        return retval;

    pattern &= DP_TEST_PHY_PATTERN_SEL_MASK;
	dev_dbg(xdev->dev, "PATTERN: %d\n", pattern);

	// get dpcd version
	dptx_read_dpcd(dptx, 0x0, &rev);
	dev_dbg(xdev->dev, "DPCD_REV: 0x%x\n", rev);

	// get lane
	dptx_read_dpcd(dptx, test_lane_count, &lanes);
	lanes = lanes & 0x1f;
	dev_dbg(xdev->dev, "lanes: %d\n", lanes);

	// get rate
	dptx_read_dpcd(dptx, test_link_rate, &rate1);
	switch (rate1) {
    // 1.62Gbps
	case 0x6:
		rate = DPTX_PHYIF_CTRL_RATE_RBR;
		break;

    // 2.7Gbps
    case 0xa:
        rate = DPTX_PHYIF_CTRL_RATE_HBR;
        break;

    // 5.4Gbps
    case 0x14:
        rate = DPTX_PHYIF_CTRL_RATE_HBR2;
        break;

    // 8.1Gbps
    case 0x1e:
        rate = DPTX_PHYIF_CTRL_RATE_HBR3;
        break;
    default:
	rate = DPTX_PHYIF_CTRL_RATE_RBR;
	dev_dbg(xdev->dev, "invalid dp rate\n");

	}

	dev_dbg(xdev->dev, "rate: %d\n", rate);

    // set pattern
    switch (pattern) {
    case DP_TEST_PHY_PATTERN_NONE:
        retval = adjust_vswing_and_preemphasis(dptx);
        if (retval)
            return retval;
        dev_dbg(xdev->dev, "No test pattern selected\n");
        dptx_phy_set_pattern(dptx, DPTX_PHYIF_CTRL_TPS_NONE);
        break;

    case DP_TEST_PHY_PATTERN_D10:
        retval = adjust_vswing_and_preemphasis(dptx);
        if (retval)
            return retval;
        dev_dbg(xdev->dev, "D10.2 without scrambling test phy pattern\n");
        dptx_phy_set_pattern(dptx, DPTX_PHYIF_CTRL_TPS_1);
        break;

    case DP_TEST_PHY_PATTERN_SEMC:
        retval = adjust_vswing_and_preemphasis(dptx);
        if (retval)
            return retval;
        dev_dbg(xdev->dev, "Symbol error measurement count test phy pattern\n");
        dptx_phy_set_pattern(dptx, DPTX_PHYIF_CTRL_TPS_SYM_ERM);
        break;

    case DP_TEST_PHY_PATTERN_PRBS7:
		adjust_pcs_tx_eq(dptx);
        retval = adjust_vswing_and_preemphasis(dptx);
        if (retval)
            return retval;
        dev_dbg(xdev->dev, "PRBS7 test phy pattern\n");
        dptx_phy_set_pattern(dptx, DPTX_PHYIF_CTRL_TPS_PRBS7);
        break;

    case DP_TEST_PHY_PATTERN_CUSTOM:
        retval = adjust_vswing_and_preemphasis(dptx);
        if (retval)
            return retval;
        dev_dbg(xdev->dev, "80-bit custom pattern transmitted test phy pattern\n");
        //dptx_writel(dptx, DPTX_CUSTOMPAT0, 0x3E0F83E0);
        //dptx_writel(dptx, DPTX_CUSTOMPAT1, 0x3E0F83E0);
        //dptx_writel(dptx, DPTX_CUSTOMPAT2, 0xF83E0);
		retval = dptx_set_custom_pattern(dptx);
        if (retval)
            return retval;
        dptx_phy_set_pattern(dptx, DPTX_PHYIF_CTRL_TPS_CUSTOM80);
        break;

    case DP_TEST_PHY_PATTERN_CP2520_1:
        retval = adjust_vswing_and_preemphasis(dptx);
        if (retval)
            return retval;
        dev_dbg(xdev->dev, "CP2520_1 - HBR2 Compliance EYE pattern\n");
        dptx_phy_set_pattern(dptx, DPTX_PHYIF_CTRL_TPS_CP2520_1);
        break;

    case DP_TEST_PHY_PATTERN_CP2520_2:
        retval = adjust_vswing_and_preemphasis(dptx);
        if (retval)
            return retval;
        dev_dbg(xdev->dev, "CP2520_2 - pattern\n");
        dptx_phy_set_pattern(dptx, DPTX_PHYIF_CTRL_TPS_CP2520_2);
        break;

    case DP_TEST_PHY_PATTERN_CP2520_3_TPS4:
        retval = adjust_vswing_and_preemphasis(dptx);
        if (retval)
            return retval;
        dev_dbg(xdev->dev, "DP_TEST_PHY_PATTERN_CP2520_3_TPS4 - pattern\n");
        dptx_phy_set_pattern(dptx, DPTX_PHYIF_CTRL_TPS_4);
        break;

    default:
        dev_dbg(xdev->dev, "Invalid TEST_PHY_PATTERN\n");
        return -EINVAL;
    }

    // Move PHY to P3
	phyifctrl = dptx_readl(dptx, DPTX_PHYIF_CTRL);
	phyifctrl |= (2 << DPTX_PHYIF_CTRL_LANE_PWRDOWN_SHIFT);// move phy to P3 state
	dptx_writel(dptx, DPTX_PHYIF_CTRL, phyifctrl);

	retval = dptx_phy_wait_busy(dptx, DPTX_MAX_LINK_LANES);
    if (retval) {
		dev_err(xdev->dev, "Timed out waiting for PHY BUSY\n");
		return retval;
	}

	// set rate
	dptx_phy_set_rate(dptx, rate);

	retval = dptx_phy_wait_busy(dptx, dptx->link.lanes);
	if (retval) {
		dev_err(xdev->dev, "Timed out waiting for PHY BUSY\n");
		return retval;
	}

	// set lanes
	dptx_phy_set_lanes(dptx, lanes);

	// Move phy to P0
	phyifctrl = dptx_readl(dptx, DPTX_PHYIF_CTRL);
	phyifctrl &= ~DPTX_PHYIF_CTRL_LANE_PWRDOWN_MASK;// P0
	dptx_writel(dptx, DPTX_PHYIF_CTRL, phyifctrl);

	/* Wait for PHY busy */
	retval = dptx_phy_wait_busy(dptx, dptx->link.lanes);
	if (retval) {
		dev_err(xdev->dev, "Timed out waiting for PHY BUSY\n");
		return retval;
	}

	dptx_phy_enable_xmit(dptx, lanes, true);

    return retval;
}

static int handle_automated_test_request(struct dptx *dptx)
{
    struct xdx_output *output = dptx->output;
    struct xdx_dm   *dm = output_to_dm(output);
    struct xdx_device *xdev = dm_to_xdev(dm);
    int retval;
    u8 test;

    retval = dptx_read_dpcd(dptx, DP_TEST_REQUEST, &test);
    if (retval)
        return retval;

    if (test & DP_TEST_LINK_TRAINING) {
        dev_dbg(xdev->dev, "%s(): DP_TEST_LINK_TRAINING\n", __func__);
#if 0
        retval = dptx_write_dpcd(dptx, DP_TEST_RESPONSE, DP_TEST_ACK);
        if (retval)
            return retval;

        retval = handle_test_link_training(dptx);
        if (retval)
            return retval;
#endif
    }

    if (test & DP_TEST_LINK_VIDEO_PATTERN) {
        dev_dbg(xdev->dev, "%s(): DP_TEST_LINK_VIDEO_PATTERN\n", __func__);
#if 0
        retval = dptx_write_dpcd(dptx, DP_TEST_RESPONSE, DP_TEST_ACK);
        if (retval)
            return retval;

        retval = handle_test_link_video_pattern(dptx, 0);
        if (retval)
            return retval;
#endif
    }
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 1, 0))
    if (test & DP_TEST_LINK_AUDIO_PATTERN) {
        dev_dbg(xdev->dev, "%s(): DP_TEST_LINK_AUDIO_PATTERN\n", __func__);
#if 0
        retval = dptx_write_dpcd(dptx, DP_TEST_RESPONSE, DP_TEST_ACK);
        if (retval)
            return retval;

        retval = handle_test_link_audio_pattern(dptx);
        if (retval)
            return retval;
#endif
    }
#endif
    if (test & DP_TEST_LINK_EDID_READ) {
        /* Invalid, this should happen on HOTPLUG */
		dev_dbg(xdev->dev, "%s(): DP_TEST_LINK_EDID_READ\n", __func__);
        return -ENOTSUPP;
    }

    if (test & DP_TEST_LINK_PHY_TEST_PATTERN) {
        dev_dbg(xdev->dev, "DP_TEST_LINK_PHY_TEST_PATTERN\n");
        retval = handle_test_phy_pattern(dptx);
        if (retval)
            return retval;
    }

    return 0;
}

static int handle_sink_request(struct dptx *dptx)
{
    struct xdx_output *output = dptx->output;
    struct xdx_dm   *dm = output_to_dm(output);
    struct xdx_device *xdev = dm_to_xdev(dm);
	int retval;
	u8 vector;
	u8 bytes[1];

	retval = dptx_link_check_status(dptx);
	if (retval)
		return retval;

	retval = dptx_read_bytes_from_dpcd(dptx, DP_DEVICE_SERVICE_IRQ_VECTOR_ESI0, bytes, sizeof(bytes)); //sahakyan
	if (retval)
		return retval;

	retval = dptx_read_dpcd(dptx, DP_DEVICE_SERVICE_IRQ_VECTOR, &vector);
	if (retval)
		return retval;

	dev_dbg(xdev->dev, "%s: IRQ_VECTOR: 0x%02x\n", __func__, vector);

	if (!vector)
		return 0;

	if ((vector & DP_REMOTE_CONTROL_COMMAND_PENDING) || (bytes[0] & DP_REMOTE_CONTROL_COMMAND_PENDING)) {
	/* TODO */
		dev_dbg(xdev->dev, "DP_REMOTE_CONTROL_COMMAND_PENDING: Not yet implemented\n");
	}

	if ((vector & DP_AUTOMATED_TEST_REQUEST) || (bytes[0] & DP_AUTOMATED_TEST_REQUEST)) {
		dev_dbg(xdev->dev, "DP_AUTOMATED_TEST_REQUEST\n");
		retval = handle_automated_test_request(dptx);
		if (retval) {
			dev_dbg(xdev->dev, "Automated test request failed\n");
			if (retval == -ENOTSUPP) {
				retval = dptx_write_dpcd(dptx, DP_TEST_RESPONSE, DP_TEST_NAK);
				if (retval)
					return retval;
			}
		}
	}

	return retval;
}
#endif // SUPPORT_DP_CTS

static void xdx_dp_hpd_handler(struct dp_dev *DisplayPort)
{
	struct xdx_output *output = dp_to_output(DisplayPort);
	struct xdx_dm *dm = output_to_dm(output);
	struct xdx_device *xdev = dm_to_xdev(dm);
	u32 ists;
	struct dptx *dptx;
    int retval;
	dptx = DisplayPort->dptx;

	ists = dptx_readl(dptx, DPTX_ISTS);
	dev_dbg(xdev->dev, "%s: Interrupt status=0x%08x\n", __func__, ists);

	if (!(ists & DPTX_ISTS_ALL_INTR)) {
		dev_dbg(xdev->dev, "%s: IRQ_NONE\n", __func__);
		return;
	}

	if (ists & DPTX_ISTS_TYPE_C) {
		dptx_writel(dptx, DPTX_ISTS, DPTX_ISTS_TYPE_C);
		ists = dptx_readl(dptx, DPTX_ISTS);
	}

	if (ists & DPTX_ISTS_HPD) {
		u32 hpdsts;
		hpdsts = dptx_readl(dptx, DPTX_HPDSTS);
		dev_dbg(xdev->dev, "%s: HPD status=0x%08x\n", __func__, hpdsts);

        if (hpdsts & DPTX_HPDSTS_IRQ) {
			dptx_writel(dptx, DPTX_HPDSTS, DPTX_HPDSTS_IRQ);

#ifdef SUPPORT_DP_CTS
            retval = handle_sink_request(dptx);
            if (retval) {
                dev_dbg(xdev->dev, "unable to handle sink request：%d\n", retval);
            }
#endif

		}

		if (hpdsts & DPTX_HPDSTS_HOT_PLUG) {
			dptx_writel(dptx, DPTX_HPDSTS, DPTX_HPDSTS_HOT_PLUG);
			DisplayPort->hpd_detect = 1;
			DisplayPort->hpd_changed = 1;
			queue_work(DisplayPort->workq, &DisplayPort->hpd_work);
		}

		if (hpdsts & DPTX_HPDSTS_HOT_UNPLUG) {
			dptx_writel(dptx, DPTX_HPDSTS, DPTX_HPDSTS_HOT_UNPLUG);
			dptx->link.trained = false;
			DisplayPort->hpd_detect = 0;
			queue_work(DisplayPort->workq, &DisplayPort->hpd_work);
		}

		if (hpdsts & 0x80)
			dptx_writel(dptx, DPTX_HPDSTS, 0x80 | DPTX_HPDSTS_HOT_UNPLUG);

		dptx_writel(dptx, DPTX_ISTS, ists & ~DPTX_ISTS_HPD);
	}
}
static int handle_hotunplug(struct xdx_dm *dm,struct dptx *dptx)
{
	struct xdx_device *xdev = dm_to_xdev(dm);
	u32 phyifctrl;
	u8 retval;

	/* clear xmit enables */
	phyifctrl = dptx_readl(dptx, DPTX_PHYIF_CTRL);
	phyifctrl &= ~DPTX_PHYIF_CTRL_XMIT_EN_MASK;

	/* move PHY to P3 state*/
	phyifctrl |= (3 << DPTX_PHYIF_CTRL_LANE_PWRDOWN_SHIFT);
	dptx_writel(dptx, DPTX_PHYIF_CTRL, phyifctrl);

	retval = dptx_phy_wait_busy(dptx, DPTX_MAX_LINK_LANES);
	if (retval) {
		dev_dbg(xdev->dev, "Timed out waiting for PHY BUSY\n");
		return retval;
	}

	/* Power down all lanes*/

	dptx->link.trained = false;

	return 0;
}

static void xdx_dp_hpd_delayed_handler(struct work_struct *work)
{
	struct dp_dev *dp = container_of(work, struct dp_dev, hpd_work);
	struct xdx_output *output = dp_to_output(dp);
	struct xdx_dm *dm = output_to_dm(output);
	struct xdx_device *xdev = dm_to_xdev(dm);
	struct drm_device *ddev = xdev_to_drm(xdev);
	u32 hpdsts = 0;
	struct dptx *dptx;
	dptx = dp->dptx;

	hpdsts = dptx_readl(dptx, DPTX_HPDSTS);

	if (ddev) {
		if (hpdsts & DPTX_HPDSTS_STATUS) {
			dev_dbg(xdev->dev,"handle DP plug in\n");
			//drm_helper_connector_dpms(&dp->connector,DRM_MODE_DPMS_OFF);
			//drm_helper_connector_dpms(&dp->connector,DRM_MODE_DPMS_ON);
			dptx_link_check_status(dptx);
		} else {
			dev_dbg(xdev->dev,"handle DP plug out\n");
			handle_hotunplug(dm,dptx);
		}
	}
	drm_kms_helper_hotplug_event(ddev);
}

static int dp_irq_set(struct xdx_device *xdev, struct xdx_irq_src *src,
			unsigned type, enum xdx_interrupt_state state)
{
	// TODO: enable/disable IRQ
	return -1;
}

static int dp_irq_process(struct xdx_device *xdev, struct xdx_irq_src *src,
				struct xdx_irq_entry *entry)
{
	struct xdx_dm *dm = xdev_to_dm(xdev);
	struct xdx_output *output = dm_to_output(dm);
	struct dp_dev *DisplayPort;

	if (src->id == XDX_IRQ_CLIENTID_DP0)
		DisplayPort = output->dp[0];
	else if (src->id == XDX_IRQ_CLIENTID_DP1 && dm->dc_num ==2 )
		DisplayPort = output->dp[1];
	else
		return -EINVAL;

	xdx_dp_hpd_handler(DisplayPort);

	return 0;
}

static struct xdx_irq_src_funcs dp_src_funcs = {
	.set = dp_irq_set,
	.process = dp_irq_process,
};

static int dp_isr_install(struct dp_dev *dp)
{
	struct xdx_output* output = dp_to_output(dp);
	struct xdx_dm* dm = output_to_dm(output);
	struct xdx_device* xdev = dm_to_xdev(dm);
	struct xdx_irq_src *irq_src;

	irq_src = kmalloc(sizeof(*irq_src), GFP_KERNEL);
	if (!irq_src)
		return -ENOMEM;

	if (dp->id == 0)
		irq_src->id = XDX_IRQ_CLIENTID_DP0;
	else if (dp->id == 1 && xdev->asic_type == CHIP_PANGU_B0)
		irq_src->id = XDX_IRQ_CLIENTID_DP1;
	else
		return -EINVAL;

	irq_src->funcs = &dp_src_funcs;

	xdx_irq_register_src(xdev, irq_src);

	dp->irq_src = irq_src;
	return 0;
}

static int dp_isr_uninstall(struct dp_dev *dp)
{
	struct xdx_output* output = dp_to_output(dp);
	struct xdx_dm* dm = output_to_dm(output);
	struct xdx_device* xdev = dm_to_xdev(dm);
	int ret;

	ret = xdx_irq_unregister_src(xdev, dp->irq_src);
	if (ret) {
		dev_err(xdev->dev, "failed to unregister irq source\n");
		return ret;
	}

	kfree(dp->irq_src);
	dp->irq_src = NULL;

	return 0;
}

int xdx_dp_preinit(struct dptx *dptx)
{
	struct xdx_output *output = dptx->output;
	struct xdx_dm *dm = output_to_dm(output);
	struct xdx_device *xdev = dm_to_xdev(dm);
	DPDBCmd_T_SINK PhyConfig;
	u32 TimeOut, DBValue;

	dev_info(xdev->dev,"%s entry!\n",__func__);
	PhyConfig.all = 0;
	PhyConfig.cmd = cH02ADisplayDBCmdDP;
	PhyConfig.subCmd = cH02ADPSubCmdCNTLPreConfig;
	PhyConfig.args0 = dptx->dp->id;
	xdx_dp_h2a_doorbell_w(dptx,(u64)PhyConfig.all);

	TimeOut = 10000;
	while(1)
	{
		DBValue = xdx_dp_h2a_doorbell_r(dptx);
		if((DBValue & 1) == 1)
			break;
		TimeOut--;
		if(TimeOut == 0){
			return -1;
		}
		udelay(1000);
	}
	dev_info(xdev->dev,"%s end!\n",__func__);
	return 0;
}

int xdx_dp_init(struct xdx_dm *dm, u32 id)
{
	struct xdx_device *xdev = dm_to_xdev(dm);
	struct xdx_output *output = dm_to_output(dm);
	struct drm_device* ddev = xdev_to_drm(xdev);
	struct dp_dev *DisplayPort;
	int ret;
	DisplayPort = kzalloc(sizeof(*DisplayPort), GFP_KERNEL);
	if (!DisplayPort) {
		ret = -ENOMEM;
		goto failed_alloc;
	}

	DisplayPort->dptx = kzalloc(sizeof(struct dptx), GFP_KERNEL);
	if (!DisplayPort->dptx) {
		ret = -ENOMEM;
		goto failed_alloc;
	}

	DisplayPort->workq = alloc_ordered_workqueue("xdxdc_dp_hpdq", 0);
	if (!DisplayPort->workq) {
		dev_err(xdev->dev, "failed to alloc ordered workqueue\n");
		ret = -ENOMEM;
		goto failed_alloc;
	}
	output->psDoorbellShareMemDesc[2 + id] =NULL;
	output->psDoorbellVirAddr[2 + id] = NULL;
	output->dp[id] = DisplayPort;
	DisplayPort->id = id;
	DisplayPort->output = output;
	DisplayPort->dptx->output = output;
	DisplayPort->dptx->dp = DisplayPort;
	DisplayPort->native_mode = drm_mode_duplicate(ddev, &default_mode);

	dp_video_params_init(DisplayPort);

	dptx_init_hwparams(DisplayPort->dptx);

	xdx_dp_register(DisplayPort, dm_dc_get_crtc_mask(dm->dc[id]));

	xdx_dp_aux_init(DisplayPort);

	if (xdev->asic_type == CHIP_PANGU_B0){
		xdx_dp_preinit(DisplayPort->dptx);
		dptx_global_intr_dis(DisplayPort->dptx);
		dp_isr_install(DisplayPort);
		dptx_interrupt_en(DisplayPort->dptx);
	} else {
		xdx_tile_init(DisplayPort->dptx);
		dptx_global_intr_dis(DisplayPort->dptx);
		dp_isr_install(DisplayPort);
		dptx_core_init(DisplayPort->dptx);
	}

#ifdef CONFIG_DEBUG_FS
	xdx_connector_debugfs_init(DisplayPort);
#endif
	return 0;

failed_alloc:
	if (DisplayPort) {
		if (DisplayPort->dptx)
			kfree(DisplayPort->dptx);

		if (DisplayPort->workq)
			destroy_workqueue(DisplayPort->workq);

		kfree(DisplayPort);

		output->dp[id] = NULL;
	}

	return ret;
}

int xdx_dp_deinit(struct xdx_dm *dm, u32 id)
{
	struct xdx_device *xdev = dm_to_xdev(dm);
	struct xdx_output *output = dm_to_output(dm);
	struct dp_dev *dp = output->dp[id];

#ifdef CONFIG_DEBUG_FS
	xdx_connector_debugfs_fini(dp);
#endif

	dptx_hpd_intr_dis(dp->dptx);

	dptx_global_intr_dis(dp->dptx);

	dp_isr_uninstall(dp);

	DPDoorbellShareMemFree(output,2 + dp->id);
	
	if (dp->workq)
		destroy_workqueue(dp->workq);

	if (dp->edid)
		kfree(dp->edid);

	if (dp->native_mode)
	{
		drm_mode_destroy(xdev_to_drm(xdev), dp->native_mode);
		dp->native_mode = NULL;
	}

	if (dp->dp_mode)
	{
		drm_mode_destroy(xdev_to_drm(xdev), dp->dp_mode);
		dp->dp_mode = NULL;
	}

	if (dp->dptx)
		kfree(dp->dptx);

	if (dp->aux)
	{
		drm_dp_aux_unregister(dp->aux);
		kfree(dp->aux);
	}
	kfree(dp);

	output->dp[id] = NULL;
	return 0;
}

int xdx_dp_hw_init(struct xdx_dm *dm, u32 id)
{
	struct xdx_output *output = dm_to_output(dm);
	struct dptx *dptx = output->dp[id]->dptx;

	xdx_tile_init(dptx);
	dptx_global_intr_dis(dptx);
	return dptx_core_init(dptx);
}

