/*-*- linux-c -*-
 *  linux/drivers/video/i810_accel.c -- Hardware Acceleration
 *
 *      Copyright (C) 2001 Antonino Daplas<adaplas@pol.net>
 *      All Rights Reserved      
 *
 *  This file is subject to the terms and conditions of the GNU General Public
 *  License. See the file COPYING in the main directory of this archive for
 *  more details.
 */
#include <linux/kernel.h>
#include <linux/clk.h>
#include <linux/debugfs.h>
#include <linux/version.h>

#include <drm/drm_atomic.h>
#include <drm/drm_atomic_helper.h>
#include <drm/drm_crtc.h>

#include "xdx_dp.h"
#include "xdx_aux.h"
#include "xdx_device.h"
#include "xdx_dp_accel.h"

#define xdx_container_of(ptr, type, member) \
	(type *)((char *)(ptr) - (char *) &((type *)0)->member)

u32 dptx_readl(struct dptx *dptx, u32 offset)
{
	struct xdx_output *output = dptx->output;
	struct xdx_dm *dm = output_to_dm(output);
	struct xdx_device *xdev = dm_to_xdev(dm);
	struct dp_dev *dp = dptx->dp;

	return xdx_mmio_readl(xdev, xdev->hw_info.dp_regs[dp->id].base + offset);
}

void dptx_writel(struct dptx *dptx, u32 offset, u32 value)
{
	struct xdx_output *output = dptx->output;
	struct xdx_dm *dm = output_to_dm(output);
	struct xdx_device *xdev = dm_to_xdev(dm);
	struct dp_dev *dp = dptx->dp;

	xdx_mmio_writel(xdev, xdev->hw_info.dp_regs[dp->id].base + offset, value);

}

int dptx_read_dpcd(struct dptx *dptx, u32 addr, u8 *byteptr)
{
	int ret;
	struct dp_dev *dp = dptx->dp;

	ret = drm_dp_dpcd_read(dp->aux, addr, byteptr, 1);
	if (ret != 1)
		return ret;

	return 0;
}

int dptx_write_dpcd(struct dptx *dptx, u32 addr, u8 byte)
{
	int ret;
	struct dp_dev *dp = dptx->dp;

	ret = drm_dp_dpcd_write(dp->aux, addr, &byte, 1);
	if (ret != 1)
		return ret;

	return 0;
}

int dptx_write_bytes_to_dpcd(struct dptx *dptx, unsigned int reg_addr,
				u8 *bytes, u32 len)
{
	struct xdx_output *output = dptx->output;
	struct xdx_dm *dm = output_to_dm(output);
	struct xdx_device *xdev = dm_to_xdev(dm);
	u8 i;
	int ret;

	for (i = 0; i < len; i++) {
		dev_dbg(xdev->dev, "TDB: DPCD Write bytes 0x%03x = 0x%02x\n",
			 reg_addr + i, *(bytes + i));
	}

	ret = __dptx_write_bytes_to_dpcd(dptx, reg_addr, bytes, len);

	return ret;
}

int dptx_read_bytes_from_dpcd(struct dptx *dptx, unsigned int reg_addr,
				u8 *bytes, u32 len)
{
	struct xdx_output *output = dptx->output;
	struct xdx_dm *dm = output_to_dm(output);
	struct xdx_device *xdev = dm_to_xdev(dm);
	u8 i;
	int ret;

	ret = __dptx_read_bytes_from_dpcd(dptx, reg_addr, bytes, len);

	for (i = 0; i < len; i++) {
		dev_dbg(xdev->dev, "TDB: DPCD Read  bytes 0x%03x = 0x%02x\n",
			 reg_addr + i, *(bytes + i));
	}

	return ret;
}

u32 dptx_tilereadl(struct dptx *dptx, u32 offset)
{
	struct xdx_output *output = dptx->output;
	struct xdx_dm *dm = output_to_dm(output);
	struct xdx_device *xdev = dm_to_xdev(dm);
	struct dp_dev *dp = dptx->dp;

	return xdx_mmio_readl(xdev, xdev->hw_info.hdmi_tile_regs[dp->id].base + offset);
}

void dptx_tilewritel(struct dptx *dptx, u32 offset, u32 data)
{
	struct xdx_output *output = dptx->output;
	struct xdx_dm *dm = output_to_dm(output);
	struct xdx_device *xdev = dm_to_xdev(dm);
	struct dp_dev *dp = dptx->dp;

	xdx_mmio_writel(xdev, xdev->hw_info.hdmi_tile_regs[dp->id].base + offset, data);

}

u32 dptx_pcs_readl(struct dptx *dptx, u32 offset)
{
	struct xdx_output *output = dptx->output;
	struct xdx_dm *dm = output_to_dm(output);
	struct xdx_device *xdev = dm_to_xdev(dm);
	struct dp_dev *dp = dptx->dp;

	return xdx_mmio_readl(xdev, (xdev->hw_info.dp_pcs_regs[dp->id].base + (offset << 2)));
}

void dptx_pcs_writel(struct dptx *dptx, u32 offset, u16 val)
{
	struct xdx_output *output = dptx->output;
	struct xdx_dm *dm = output_to_dm(output);
	struct xdx_device *xdev = dm_to_xdev(dm);
	struct dp_dev *dp = dptx->dp;
	u32 data;

	data = (val << 16) | val;

	xdx_mmio_writel(xdev, (xdev->hw_info.dp_pcs_regs[dp->id].base + (offset << 2)), data);

}

void xdx_dp_h2a_doorbell_w(struct dptx *dptx, u64 value)
{
	struct xdx_output *output = dptx->output;
	struct xdx_dm *dm = output_to_dm(output);
	struct xdx_device *xdev = dm_to_xdev(dm);

	xdx_mmio_writel(xdev,
			SYS_PANGU_REG_HDMIDOORBELL + 8,
			value & 0xffffffff);

	xdx_mmio_writel(xdev,
			SYS_PANGU_REG_HDMIDOORBELL + 0xc,
			(value >> 32) & 0xffffffff);
}

u32 xdx_dp_h2a_doorbell_r(struct dptx *dptx)
{
	struct xdx_output *output = dptx->output;
	struct xdx_dm *dm = output_to_dm(output);
	struct xdx_device *xdev = dm_to_xdev(dm);

	return xdx_mmio_readl(xdev,
			SYS_PANGU_REG_HDMIDOORBELL + 8);
}


