#include <linux/uaccess.h>
#include <drm/drm_debugfs.h>

#include "xdx_dp.h"
#include "xdx_dp_accel.h"
#include "core.h"
#include "xdx_device.h"
#include "dm.h"

static ssize_t dp_dpcd_data_read(struct file *f, char __user *buf,
				 size_t size, loff_t *pos)
{
	int r;
	char *data;
	struct dp_dev *dpdev = file_inode(f)->i_private;
	struct xdx_output *output = dp_to_output(dpdev);
	struct xdx_dm *dm = output_to_dm(output);
	struct xdx_device *xdev = dm_to_xdev(dm);
	uint32_t read_size = dpdev->debugfs_dpcd_size;

	if (size < read_size)
		return 0;

	data = kzalloc(read_size, GFP_KERNEL);
	if (!data)
		return 0;

	dptx_read_dpcd(dpdev->dptx, dpdev->debugfs_dpcd_address, (u8 *)data);
	dev_dbg(xdev->dev,"dp debugfs read value:%x",*data);
	r = copy_to_user(buf, data, read_size);

	kfree(data);
	return r;
}

static ssize_t dp_dpcd_data_write(struct file *f, const char __user *buf,
				size_t size, loff_t *pos)
{
	char *data;
	struct dp_dev *dpdev = file_inode(f)->i_private;
	struct xdx_output *output = dp_to_output(dpdev);
	struct xdx_dm *dm = output_to_dm(output);
	struct xdx_device *xdev = dm_to_xdev(dm);
	uint32_t write_size = dpdev->debugfs_dpcd_size;

	if (size < write_size)
		return 0;

	data = kzalloc(write_size, GFP_KERNEL);
	if (!data)
		return 0;

	if(sscanf(buf,"%hhx",data) != 1){
		dev_dbg(xdev->dev,"The number of arguments is not equal to 1!\n");
		return size;
	}
	//r = copy_from_user(data, buf, &data);

	dev_dbg(xdev->dev,"dp debugfs write value:%x",*data);
	dptx_write_dpcd(dpdev->dptx,dpdev->debugfs_dpcd_address, *(u8 *)data);
	kfree(data);
	return size;
}

static ssize_t dp_dpcd_address_write(struct file *f, const char __user *buf,
				 size_t size, loff_t *pos)
{
	struct dp_dev *dpdev = file_inode(f)->i_private;
	struct xdx_output *output = dp_to_output(dpdev);
	struct xdx_dm *dm = output_to_dm(output);
	struct xdx_device *xdev = dm_to_xdev(dm);
	dev_dbg(xdev->dev,"dpcd address write size:%lx",size);
	//if (size < sizeof(dpdev->debugfs_dpcd_address))
	//	return 0;
	//r = copy_from_user(&dpdev->debugfs_dpcd_address,
	//		buf, sizeof(dpdev->debugfs_dpcd_address));
	if(sscanf(buf,"%x",&dpdev->debugfs_dpcd_address) != 1){
		dev_dbg(xdev->dev,"The number of arguments is not equal to 1!\n");
		return size;
	}
	dev_dbg(xdev->dev,"dp debugfs write dpcd addr:%x\n",dpdev->debugfs_dpcd_address);
	return size;
}

static ssize_t dp_dpcd_size_write(struct file *f, const char __user *buf,
				size_t size, loff_t *pos)
{
	int r;
	struct dp_dev *dpdev = file_inode(f)->i_private;
	struct xdx_output *output = dp_to_output(dpdev);
	struct xdx_dm *dm = output_to_dm(output);
	struct xdx_device *xdev = dm_to_xdev(dm);

	if (size < sizeof(dpdev->debugfs_dpcd_size))
		return 0;

	r = copy_from_user(&dpdev->debugfs_dpcd_size,
			buf, sizeof(dpdev->debugfs_dpcd_size));

	dev_dbg(xdev->dev,"dp debugfs write dpcd addr size:%x \n",dpdev->debugfs_dpcd_size);
	if (dpdev->debugfs_dpcd_size > 256)
		dpdev->debugfs_dpcd_size = 0;

	return size - r;
}


static const struct file_operations dp_dpcd_address_debugfs_fops = {
	.owner = THIS_MODULE,
	.write = dp_dpcd_address_write,
	.llseek = default_llseek
};

static const struct file_operations dp_dpcd_size_debugfs_fops = {
	.owner = THIS_MODULE,
	.write = dp_dpcd_size_write,
	.llseek = default_llseek
};


static const struct file_operations dp_dpcd_data_debugfs_fops = {
	.owner = THIS_MODULE,
	.read = dp_dpcd_data_read,
	.write = dp_dpcd_data_write,
	.llseek = default_llseek
};

static const struct {
	char *name;
	const struct file_operations *fops;
} dp_debugfs_entries[] = {
	//{"link_settings", &dp_link_settings_debugfs_fops},
	//{"phy_settings", &dp_phy_settings_debugfs_fop},
	//{"test_pattern", &dp_phy_test_pattern_fops},
	//{"output_bpc", &output_bpc_fops},
	//{"vrr_range", &vrr_range_fops},
	//{"sdp_message", &sdp_message_fops},
	{"aux_dpcd_address", &dp_dpcd_address_debugfs_fops},
	{"aux_dpcd_size", &dp_dpcd_size_debugfs_fops},
	{"aux_dpcd_data", &dp_dpcd_data_debugfs_fops}
};

#ifdef CONFIG_DEBUG_FS
void xdx_connector_debugfs_init(struct dp_dev *dpdev)
{
	unsigned int i;
	struct dentry *dir = dpdev->connector.debugfs_entry;
	struct xdx_output *output = dp_to_output(dpdev);
	struct xdx_dm *dm = output_to_dm(output);
	struct xdx_device *xdev = dm_to_xdev(dm);
	dev_dbg(xdev->dev,"dp debugfs init\n");
	dpdev->dirs = (struct dentry**)kzalloc(sizeof(struct dentry*) * ARRAY_SIZE(dp_debugfs_entries), GFP_KERNEL);
	if (!dpdev->dirs)
		return;
	if (dpdev->connector.connector_type == DRM_MODE_CONNECTOR_DisplayPort) {
		for (i = 0; i < ARRAY_SIZE(dp_debugfs_entries); i++) {
			dpdev->dirs[i] = debugfs_create_file(dp_debugfs_entries[i].name,
						0644, dir, dpdev,
						dp_debugfs_entries[i].fops);
			dpdev->dirs_num ++;
		}
	}
	dpdev->debugfs_dpcd_address = 0;
	dpdev->debugfs_dpcd_size = sizeof(u8);
}


void xdx_connector_debugfs_fini(struct dp_dev *dpdev)
{
	u8 i;
	if (!dpdev->dirs)
		return;

	for (i = 0; i < dpdev->dirs_num; i++)
	{
		debugfs_remove(dpdev->dirs[i]);
	}
	kfree(dpdev->dirs);
	dpdev->dirs_num = 0;
	dpdev->dirs = NULL;
}
#endif
