// SPDX-License-Identifier: GPL-2.0
/*
 * Copyright (C) 2020 VeriSilicon Holdings Co., Ltd.
 */

#include <linux/moduleparam.h>
#include <linux/version.h>
#include <linux/component.h>
#include <linux/module.h>
#include <linux/delay.h>
#include <linux/hdmi.h>

#include <drm/drm_crtc.h>
#include <drm/drm_crtc_helper.h>
#include <drm/drm_of.h>
#include <drm/drm_edid.h>

#include <drm/drm_atomic_helper.h>
#include <drm/drm_scdc_helper.h>

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 1, 0))
#include <drm/drm_probe_helper.h>
#endif

#include "xdx_hdmi.h"
#include "xdx_hdmi_accel.h"
#include "xdx_hdmi_reg.h"
#include "xdx_drv.h"
#include "pangu.h"
#include "xdx_tile_reg.h"
#include "xdx_device.h"
#include "phyconfig.h"
#include "main_unit.h"
#include "syscommon.h"
#include "dm.h"
#include "doorbell.h"
#include "dtd.h"
#include "../../../audio/xdxaudio_hal.h"

static bool force_connected = false;

static const struct drm_display_mode default_mode = {
	DRM_MODE("1920x1080", DRM_MODE_TYPE_DRIVER, 148500, 1920, 2008, 2052,
		 2200, 0, 1080, 1084, 1089, 1125, 0,
		 DRM_MODE_FLAG_PHSYNC | DRM_MODE_FLAG_PVSYNC),
	.picture_aspect_ratio = HDMI_PICTURE_ASPECT_16_9,
};

static enum drm_connector_status
xdx_hdmi_helper_connector_detect(struct drm_connector *connector, bool force)
{
	struct hdmi_dev *hdmi =
		container_of(connector, struct hdmi_dev, connector);
	enum drm_connector_status status;
	u32 phy_status;
	phy_status = xdx_hdmitile_read(hdmi, HDMI_HPD_INT_STATUS);
	if (phy_status & HPD_STATUS_MASE)
		status = connector_status_connected;
	else
		status = connector_status_disconnected;
	return status;
}

static void xdx_hdmi_helper_connector_force(struct drm_connector *connector)
{
	if (force_connected && connector) {
#if defined(HDMI_DEBUG)
	struct drm_device *dev = connector->dev;
	dev_dbg(dev->dev, "%s: skip EDID, force status connected\n",
		  __func__);
#endif
		connector->status = connector_status_connected;
	}
}

static void xdx_hdmi_helper_connector_destroy(struct drm_connector *connector)
{
	/* Disable all interrupts */
	//kfree(hdmi->edid);
	drm_connector_unregister(connector);
	drm_connector_cleanup(connector);
}

static struct drm_display_mode *
xdx_hdmi_connector_native_mode(struct drm_connector *connector)
{
	struct hdmi_dev *hdmi =
		container_of(connector, struct hdmi_dev, connector);
	const struct drm_connector_helper_funcs *helper =
		connector->helper_private;
	struct drm_device *dev = connector->dev;
	struct drm_display_mode *mode, *largest = NULL;
	int high_w = 0, high_h = 0;
	uint32_t high_v = 0;
	const int vrefresh_default = 60;

	list_for_each_entry (mode, &hdmi->connector.probed_modes, head) {
		hdmi->vrefresh = drm_mode_vrefresh(mode);
		if (!hdmi->vrefresh) {
			hdmi->vrefresh = vrefresh_default;
			DRM_INFO_ONCE(
				"vertical refresh rate is zero, defaulting to %d\n",
				hdmi->vrefresh);
		}

		if (helper->mode_valid(connector, mode) != MODE_OK ||
			(mode->flags & DRM_MODE_FLAG_INTERLACE)){
			dev_dbg(dev->dev,"drm mode flag interlace\n");
			continue;
		}

		/* Use preferred mode if there is one.. */
		if (mode->type & DRM_MODE_TYPE_PREFERRED) {
			dev_dbg(dev->dev,
				  "Retrieving native mode from preferred\n");
			return drm_mode_duplicate(dev, mode);
		}

		/*
		 * Otherwise, take the resolution with the largest width, then
		 * height, then vertical refresh
		 */
		if (mode->hdisplay < high_w)
			continue;

		if (mode->hdisplay == high_w && mode->vdisplay < high_h)
			continue;

		if (mode->hdisplay == high_w && mode->vdisplay == high_h &&
		    hdmi->vrefresh < high_v)
			continue;

		high_w = mode->hdisplay;
		high_h = mode->vdisplay;
		high_v = hdmi->vrefresh;
		largest = mode;
	}

	dev_dbg(dev->dev, "native mode from largest: %dx%d@%d\n", high_w, high_h,
		  high_v);

	return largest ? drm_mode_duplicate(dev, largest) : NULL;
}
/*
static u8 edid_demo[256] = {
	0x00, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0x00, 0x50, 0x21, 0x32, 0x27,
	0x00, 0x00, 0x00, 0x00, 0x22, 0x1a, 0x01, 0x03, 0x80, 0x30, 0x1b, 0x78,
	0x0f, 0xee, 0x91, 0xa3, 0x54, 0x4c, 0x99, 0x26, 0x0f, 0x50, 0x54, 0x2f,
	0xcf, 0x00, 0x31, 0x59, 0x45, 0x59, 0x61, 0x59, 0x81, 0x99, 0x01, 0x01,
	0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x02, 0x3a, 0x80, 0x18, 0x71, 0x38,
	0x2d, 0x40, 0x58, 0x2c, 0x46, 0x00, 0xe0, 0x0e, 0x11, 0x00, 0x00, 0x1e,
	0x00, 0x00, 0x00, 0xfd, 0x00, 0x18, 0x55, 0x18, 0x5e, 0x11, 0x00, 0x0a,
	0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x00, 0x00, 0x00, 0xfc, 0x00, 0x63,
	0x6f, 0x62, 0x61, 0x6c, 0x74, 0x0a, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20,
	0x00, 0x00, 0x00, 0x10, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x9c,

	0x02, 0x03, 0x1f, 0xf0, 0x4a, 0x90, 0x1f, 0x04, 0x13, 0x22, 0x21, 0x20,
	0x02, 0x11, 0x01, 0x23, 0x09, 0x07, 0x07, 0x68, 0x03, 0x0c, 0x00, 0x10,
	0x00, 0x00, 0x22, 0x0f, 0xe2, 0x00, 0xea, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xa7,
};
*/
static int xdx_hdmi_connector_get_modes(struct drm_connector *connector)
{
	struct hdmi_dev *hdmi =
		container_of(connector, struct hdmi_dev, connector);
	struct drm_device *dev = connector->dev;
	struct xdx_device *xdev = drm_to_xdev(dev);
	int count = 0;
	u16 i;

	if (xdev->flags.fake_mode) {
		count = drm_add_modes_noedid(connector, 800, 600);
		drm_set_preferred_mode(connector, 800, 600);

		if (hdmi->native_mode) {
			drm_mode_destroy(dev, hdmi->native_mode);
			hdmi->native_mode = NULL;
		}
		hdmi->native_mode = xdx_hdmi_connector_native_mode(connector);
		hdmi->hdmi_data = connector->display_info;
		return count;
	}

	if (!hdmi->edid || hdmi->hpd_changed || (connector->force && connector->override_edid)) {
		if (hdmi->edid)
			kfree(hdmi->edid);
		hdmi->hpd_changed = 0;
		hdmi->edid = drm_get_edid(connector, &hdmi->i2c->adap);
		if (!hdmi->edid) {
			dev_err(dev->dev,"failed to get edid\n");
			return 0;
		}
	}
	//hdmi->edid = (struct edid *)edid_demo;
	for(i=0; i< 0xff ;i=i+8){
	    dev_dbg(dev->dev,"%d: %02x %02x %02x %02x %02x %02x %02x %02x\n", i, *(u8 *)((u8 *)hdmi->edid + i),*(u8 *)((u8 *)hdmi->edid + i + 1),*(u8 *)((u8 *)hdmi->edid + i + 2),*(u8 *)((u8 *)hdmi->edid + i +3),\
	            *(u8 *)((u8 *)hdmi->edid + i+ 4),*(u8 *)((u8 *)hdmi->edid + i + 5),*(u8 *)((u8 *)hdmi->edid + i + 6),*(u8 *)((u8 *)hdmi->edid + i +7));
	}
	drm_connector_update_edid_property(connector, hdmi->edid);

	/*
	 * Destroy the native mode, the attached monitor could have changed.
	 */
	if (hdmi->native_mode) {
		drm_mode_destroy(dev, hdmi->native_mode);
		hdmi->native_mode = NULL;
	}

	count = drm_add_edid_modes(connector, hdmi->edid);
	dev_dbg(dev->dev, "display mode count %d", count);
	hdmi->native_mode = xdx_hdmi_connector_native_mode(connector);
	if (count == 0 && hdmi->native_mode) {
		/*
		 * Find the native mode if this is a digital panel, if we didn't
		 * find any modes through DDC previously add the native mode to
		 * the list of modes.
		 */
		struct drm_display_mode *mode;

		mode = drm_mode_duplicate(dev, hdmi->native_mode);
		drm_mode_probed_add(connector, mode);
		count = 1;
	}
	hdmi->hdmi_data = connector->display_info;
	if(hdmi->hdmi_data.bpc > 10)
		hdmi->hdmi_data.bpc = 8;
	dev_dbg(dev->dev, "bpc %d format %d\n", hdmi->hdmi_data.bpc,
		  hdmi->hdmi_data.color_formats);
#if (LINUX_VERSION_CODE < KERNEL_VERSION(5, 9, 0))
	dev_dbg(dev->dev, "Native mode: %dx%d@%d\n", hdmi->native_mode->hdisplay,
		hdmi->native_mode->vdisplay, hdmi->native_mode->vrefresh);
#else
    dev_dbg(dev->dev, "Native mode: %dx%d\n", hdmi->native_mode->hdisplay,
		hdmi->native_mode->vdisplay);
#endif
	dev_dbg(dev->dev, "\nHDMI:Version: %d\nRevision: %d\nExtensions: %d\n# of modes: %d\n",
		hdmi->edid->version, hdmi->edid->revision,
		hdmi->edid->extensions, count);

	return count;
}

static int xdx_hdmi_mode_valid(struct drm_connector *connector,
			       struct drm_display_mode *mode)
{
	struct drm_device *dev = connector->dev;
	if (!mode->hdisplay || !mode->vdisplay ||!drm_mode_vrefresh(mode) )
	{
		dev_dbg(dev->dev,"[DRM]par:mode:%dx%d vrefresh:%d pix_clock:%d un-supported\n",
			mode->hdisplay,mode->vdisplay,drm_mode_vrefresh(mode),mode->clock);
		return MODE_ERROR;
	}else{
		dev_dbg(dev->dev,"[DRM]par:mode:%dx%d vrefresh:%d pix_clock:%d    supported\n",
			mode->hdisplay,mode->vdisplay,drm_mode_vrefresh(mode),mode->clock);
		return MODE_OK;
	}
}

static struct drm_encoder *
xdx_hdmi_connector_best_encoder(struct drm_connector *connector)
{
	struct hdmi_dev *hdmi =
		container_of(connector, struct hdmi_dev, connector);
	return &hdmi->encoder;
}

void xdx_hdmi_trans_para_fill(struct hdmi_dev *hdmi, struct hdmi_trans_para *pass_para, void* para_sram_vaddr)
{
	u32 PixClock = hdmi->hdmi_mode->clock;
	u32 ColorDeep = hdmi->hdmi_data.bpc;
	u32 Format = 1;//todo, please change usage of format
	if(Format == 2){
		pass_para->colordeep = 8;
	}else{
		pass_para->colordeep = ColorDeep;
	}

	pass_para->pixclock = PixClock;
	pass_para->colorformat = Format;
	pass_para->picture_aspect_ratio = hdmi->connector.state->picture_aspect_ratio;
	pass_para->vic_code = get_mcode(hdmi->hdmi_mode->clock, hdmi->hdmi_mode->hdisplay);
	pass_para->dc_idx = hdmi->id;

	memcpy_toio(para_sram_vaddr, pass_para, sizeof(struct hdmi_trans_para));
}

int xdx_hdmi_phy_init(struct hdmi_dev *hdmi)
{
	struct xdx_output *output = hdmi_to_output(hdmi);
	struct xdx_dm *dm = output_to_dm(output);
	struct xdx_device *xdev = dm_to_xdev(dm);
	HDMIDBCmd_T_SINK    PhyConfig;
	u32 TimeOut, DBValue;
	int ret;

	struct hdmi_trans_para *pass_para = kmalloc(sizeof(struct hdmi_trans_para), GFP_KERNEL);
	if(output->psDoorbellShareMemDesc[hdmi->id] == NULL){
		ret = DPDoorbellShareMemAllocate(output,hdmi->id);
		if(ret){
			dev_err(xdev->dev,"HDMI share ram allocate fail\n");
			return ret;
		}
	}
	xdx_hdmi_trans_para_fill(hdmi, pass_para, output->psDoorbellVirAddr[hdmi->id]);
	PhyConfig.all = 0;
	PhyConfig.cmd = cH02ADisplayDBCmdHDMI;
	PhyConfig.subCmd = cH02AHDMISubCmdPhyInit;
	PhyConfig.dword = output->psDoorbellPhyAddr[hdmi->id].uiAddr;
	xdx_hdmi_h2a_doorbell_w(hdmi,(u64)PhyConfig.all);

	TimeOut = 10000;
	while(1){
		DBValue = xdx_hdmi_h2a_doorbell_r(hdmi);
		if((DBValue & 1) == 1)
			break;
		TimeOut--;
		if(TimeOut == 0){
			dev_dbg(xdev->dev,"HDMI phyinit timeout\n");
			return -1;
		}
		udelay(1000);
	}
	dev_info(xdev->dev,"HDMI phyinit exit!\n");

	PhyConfig.all = DBValue;
	kfree(pass_para);
	return PhyConfig.ret;
}

static void xdx_hdmi_mode_setup(struct hdmi_dev *hdmi,
				struct drm_display_mode *mode)
{
	struct xdx_output *output = hdmi_to_output(hdmi);
	struct xdx_dm *dm = output_to_dm(output);
	struct xdx_device *xdev = dm_to_xdev(dm);
	int err;
#if (LINUX_VERSION_CODE < KERNEL_VERSION(5, 9, 0))
	dev_dbg(xdev->dev, "%s: Pixel clock: %d, Mode: %dx%d@%d\n", __func__,
		 hdmi->hdmi_mode->clock, hdmi->hdmi_mode->hdisplay, hdmi->hdmi_mode->vdisplay, hdmi->hdmi_mode->vrefresh);
#else
    dev_dbg(xdev->dev, "%s: Pixel clock: %d, Mode: %dx%d\n", __func__,
		 hdmi->hdmi_mode->clock, hdmi->hdmi_mode->hdisplay, hdmi->hdmi_mode->vdisplay);
#endif

	xdx_hdmi_set_ratio_scrambling(hdmi);
        spin_lock(&hdmi->hdmi_reg_lock);
	if (xdev->asic_type == CHIP_PANGU_B0)
		err = xdx_hdmi_phy_init(hdmi);
	else {
		err = xdx_hdmi_phy_startup(hdmi);
		avp_configure(hdmi);
	}
	if (err) {
		dev_err(xdev->dev, "mode setup fail\n");
	}
	spin_unlock(&hdmi->hdmi_reg_lock);
#if (LINUX_VERSION_CODE < KERNEL_VERSION(5, 9, 0))
	dev_dbg(xdev->dev,"%s: Final HDMI timing configuration:\nPixel clock: %d\nMode: %dx%d@%d\n\n",
		__func__, mode->clock, mode->hdisplay, mode->vdisplay, mode->vrefresh);
#else
    	dev_dbg(xdev->dev, "%s: Pixel clock: %d, Mode: %dx%d\n", __func__,
		 hdmi->hdmi_mode->clock, hdmi->hdmi_mode->hdisplay, hdmi->hdmi_mode->vdisplay);
#endif
}

static void xdx_hdmi_encoder_destroy(struct drm_encoder *encoder)
{
	drm_encoder_cleanup(encoder);
}

static void xdx_hdmi_poweron(struct hdmi_dev *hdmi)
{
	//mdelay(100);
#ifdef SUPPORT_AUDIO
	struct xdx_output * output = hdmi_to_output(hdmi);
	struct xdx_dm* dm = output_to_dm(output);
	struct xdx_device* xdev = dm_to_xdev(dm);
#endif

	if (xdev->asic_type == CHIP_PANGU_A0) {
		spin_lock(&hdmi->hdmi_reg_lock);
		xdx_phy_power_up(hdmi);
		//mdelay(100);
		spin_unlock(&hdmi->hdmi_reg_lock);
	}

	xdx_hdmi_mode_setup(hdmi, hdmi->hdmi_mode);
#ifdef SUPPORT_AUDIO
	if (xdev->xdx_hdmi_audio_hotplug_ptr != NULL) {
		if (xdx_audio_drv_hotPlug(xdev, XDX_AUDIO_MODULE_HDMI, XDX_AUDIO_MODULE_CONNECT)) {
			dev_dbg(xdev->dev, "%s: Call audio HDMI plug error!\n", __func__);
		}
	}
#endif
}

static void xdx_hdmi_poweroff(struct hdmi_dev *hdmi)
{
#ifdef SUPPORT_AUDIO
	//xdx_phy_power_down(hdmi);
	struct xdx_output * output = hdmi_to_output(hdmi);
	struct xdx_dm* dm = output_to_dm(output);
	struct xdx_device* xdev = dm_to_xdev(dm);
#endif

	spin_lock(&hdmi->hdmi_reg_lock);
	if (xdev->asic_type == CHIP_PANGU_B0)
		xdx_phy_power_down_sink(hdmi);
	else
		xdx_phy_power_down(hdmi);
	spin_unlock(&hdmi->hdmi_reg_lock);
#ifdef SUPPORT_AUDIO
	if (xdev->xdx_hdmi_audio_hotplug_ptr != NULL) {
		if (xdx_audio_drv_hotPlug(xdev, XDX_AUDIO_MODULE_HDMI, XDX_AUDIO_MODULE_DISCONNECT)) {
			dev_dbg(xdev->dev, "%s: Call audio HDMI unplug error!\n", __func__);
		}
	}
#endif
}

static void xdx_hdmi_encoder_dpms(struct drm_encoder *encoder, int mode)
{
	xdx_encoder_to_hdmi(encoder);
	if (mode)
		xdx_hdmi_poweroff(hdmi);
	else
		xdx_hdmi_poweron(hdmi);
}

static void xdx_hdmi_encoder_mode_set(struct drm_encoder *encoder,
				      struct drm_display_mode *mode,
				      struct drm_display_mode *adjusted_mode)
{
	xdx_encoder_to_hdmi(encoder);
	struct drm_device *ddev = encoder->dev;
	struct drm_crtc* crtc;
	struct dm_crtc_state* dm_state;
	drm_for_each_crtc (crtc, ddev)
	{
		if (drm_crtc_index(crtc) == drm_encoder_index(encoder))
			break;
	}

	dm_state = to_dm_crtc_state(crtc->state);
	dm_state->output_fmt = hdmi->hdmi_data.color_formats;
	dm_state->bpc = hdmi->hdmi_data.bpc;

	if (hdmi->hdmi_mode)
	{
		drm_mode_destroy(encoder->dev, hdmi->hdmi_mode);
		hdmi->hdmi_mode = NULL;
	}

	hdmi->hdmi_mode = drm_mode_duplicate(encoder->dev, mode);
}

static void xdx_hdmi_encoder_commit(struct drm_encoder *encoder)
{
	xdx_encoder_to_hdmi(encoder);
	xdx_hdmi_poweron(hdmi);
}

static bool xdx_hdmi_encoder_mode_fixup(struct drm_encoder *encoder,
					const struct drm_display_mode *mode,
					struct drm_display_mode *adjusted_mode)
{
	return true;
}

static void xdx_hdmi_encoder_prepare(struct drm_encoder *encoder)
{
	xdx_encoder_to_hdmi(encoder);
	xdx_hdmi_poweroff(hdmi);
}

static const struct drm_encoder_funcs xdx_hdmi_encoder_funcs = {
	.destroy = xdx_hdmi_encoder_destroy,
};

static const struct drm_encoder_helper_funcs xdx_hdmi_encoder_helper_funcs = {
	.dpms = xdx_hdmi_encoder_dpms,
	.mode_set = xdx_hdmi_encoder_mode_set,
	.commit = xdx_hdmi_encoder_commit,
	.mode_fixup = xdx_hdmi_encoder_mode_fixup,
	.prepare = xdx_hdmi_encoder_prepare,
};

static const struct drm_connector_funcs xdx_hdmi_connector_funcs = {
	.dpms = drm_helper_connector_dpms,
	.reset = drm_atomic_helper_connector_reset,
	.detect = xdx_hdmi_helper_connector_detect,
	.force = xdx_hdmi_helper_connector_force,
	.fill_modes = drm_helper_probe_single_connector_modes,
	.destroy = xdx_hdmi_helper_connector_destroy,
	//.set_property = xdx_hdmi_helper_connector_set_property,
	.atomic_duplicate_state = drm_atomic_helper_connector_duplicate_state,
	.atomic_destroy_state = drm_atomic_helper_connector_destroy_state,
};

static const struct drm_connector_helper_funcs xdx_hdmi_connector_helper_funcs = {
	.get_modes = xdx_hdmi_connector_get_modes,
	.mode_valid = xdx_hdmi_mode_valid,
	/*
	 * For atomic, don't set atomic_best_encoder or best_encoder. This will
	 * cause the DRM core to fallback to drm_atomic_helper_best_encoder().
	 * This is fine as we only have a single connector and encoder.
	 */
	.best_encoder = xdx_hdmi_connector_best_encoder,
};

static void xdxdc_hdmi_delayed_hpd(struct work_struct *work)
{
	struct hdmi_dev *hdmi = container_of(work, struct hdmi_dev, hpd_work);
	struct xdx_output *output = hdmi_to_output(hdmi);
	struct xdx_dm *dm = output_to_dm(output);
	struct xdx_device *xdev = dm_to_xdev(dm);
	struct drm_device *ddev = xdev_to_drm(xdev);

	if(hdmi->hpd_detect){
		if(hdmi->hdmi_mode->clock > 340000){
			/* HDMI2.0 says when unplug , sink will reset the config
			 * before,  and should not send scramble data prior to
			 * configure the sink ,  so we should power off hdmi
			 * tmds and subsequently do a full hdmi setup when system
			 * not calling the setup in replug in the same monitor
			 */
			xdx_hdmi_poweroff(hdmi);
			xdx_hdmi_poweron(hdmi);
		}
	}

	drm_helper_hpd_irq_event(ddev);
}

static int xdx_hdmi_register(struct hdmi_dev *hdmi, u32 crtc_mask)
{
	struct xdx_output *output = hdmi_to_output(hdmi);
	struct xdx_dm *dm = output_to_dm(output);
	struct xdx_device *xdev = dm_to_xdev(dm);
	struct drm_device *ddev = xdev_to_drm(xdev);
	int ret;

	// we should make our own container connector object and put queue into that.
	INIT_WORK(&hdmi->hpd_work, xdxdc_hdmi_delayed_hpd);
	hdmi->connector.polled = DRM_CONNECTOR_POLL_HPD;
	hdmi->connector.interlace_allowed = 0;
	hdmi->connector.doublescan_allowed = 0;

	drm_connector_init(ddev, &hdmi->connector,
			   &xdx_hdmi_connector_funcs, DRM_MODE_CONNECTOR_HDMIA);

	drm_connector_helper_add(&hdmi->connector,
				 &xdx_hdmi_connector_helper_funcs);

	//hdmi->encoder.possible_crtcs = crtc_mask & BIT(0);
	//hdmi->encoder.possible_clones = 1 << 0;

    if (!hdmi->id) {
        hdmi->encoder.possible_clones = 1 << 0;
        hdmi->encoder.possible_crtcs = BIT(0);
    } else {
        hdmi->encoder.possible_clones = 1 << 2;
        hdmi->encoder.possible_crtcs = BIT(2);
    }

	ret = drm_encoder_init(ddev, &hdmi->encoder,
			       &xdx_hdmi_encoder_funcs, DRM_MODE_ENCODER_TMDS,
			       NULL);

	if (ret)
		return ret;

	drm_encoder_helper_add(&hdmi->encoder, &xdx_hdmi_encoder_helper_funcs);

	ret = drm_connector_attach_encoder(&hdmi->connector, &hdmi->encoder);
	if (ret) {
		dev_dbg(xdev->dev,"can't attach [ENCODER:%d:%s] to [CONNECTOR:%d:%s] (err=%d)\n",
			hdmi->encoder.base.id, hdmi->encoder.name,
			hdmi->connector.base.id, hdmi->connector.name, ret);
	}

	return 0;
}

static void video_parms_init(struct hdmi_dev *hdmi)
{
	u8 packet_3d = 0;

	hdmi->mode->pVideo.mEncoding = RGB;
	hdmi->mode->pVideo.mPixelRepetitionInput = 0;
	hdmi->mode->pVideo.mPixelRepetitionOutput = 0;
	//hdmi->mode->pVideo.mRgbQuantizationRange = 0;
	hdmi->mode->pVideo.mRgbQuantizationRange = 2;
	hdmi->mode->pVideo.mPixelPackingDefaultPhase = 0;
	hdmi->mode->pVideo.mColorimetry = ITU709 ;
	//hdmi->mode->pVideo.mColorimetry = NO_COLORIMETRY;
	hdmi->mode->pVideo.mScanInfo = 0;
	hdmi->mode->pVideo.mActiveFormatAspectRatio = 8;
	//hdmi->mode->pVideo.mActiveFormatAspectRatio = 0xb;
	hdmi->mode->pVideo.mNonUniformScaling = 0;
	//hdmi->mode->pVideo.mNonUniformScaling = 3;
	//hdmi->mode->pVideo.mExtColorimetry = 0;
	hdmi->mode->pVideo.mExtColorimetry = 1;
	//hdmi->mode->pVideo.mItContent = 0;
	hdmi->mode->pVideo.mItContent = 2;
	hdmi->mode->pVideo.mEndTopBar = (u16)-1;
	hdmi->mode->pVideo.mStartBottomBar = (u16)-1;
	hdmi->mode->pVideo.mEndLeftBar = (u16)-1;
	hdmi->mode->pVideo.mStartRightBar = (u16)-1;
	if (packet_3d == 0) {
		hdmi->mode->pVideo.mHdmiVideoFormat = HDMI_NORMAL_FORMAT;
		//app->mode.pVideo.mHdmiVideoFormat = HDMI_EXT_RES_FORMAT;
		hdmi->mode->pVideo.m3dStructure = UNDEFINED_3D;
		hdmi->mode->pVideo.m3dExtData = UNDEFINED_EXTDATA_3D;
	} else {
		hdmi->mode->pVideo.mHdmiVideoFormat = HDMI_3D_FORMAT;
		hdmi->mode->pVideo.m3dStructure = SIDE_BY_SIDE_3D;
		hdmi->mode->pVideo.m3dExtData = TOP_AND_BOTTOM_3D;
	}
}

void EnableHDMIHPDInterrupt(struct hdmi_dev *hdmi)
{
	xdx_hdmitile_write(hdmi, ALL_HDMI_HPD_CTRL, 0x1f);
	xdx_hdmitile_write(hdmi, ALL_HDMI_HPD_CLR, 0x1f);
}

void DisableHDMIHPDInterrupt(struct hdmi_dev *hdmi)
{
	xdx_hdmitile_write(hdmi, ALL_HDMI_HPD_CTRL, 0x0);
	xdx_hdmitile_write(hdmi, ALL_HDMI_HPD_CLR, 0x1f);
}

u32 ClearHDMIHPDInterrupt(struct hdmi_dev *hdmi)
{
	xdx_hdmitile_write(hdmi, ALL_HDMI_HPD_CLR, 0x1f);
	mb();
	return 0;
}

static void XDXHDMIInterruptHandler(void *Data)
{
	u32 irq_status;
	struct hdmi_dev *hdmi = (struct hdmi_dev *)Data;

	irq_status = xdx_hdmitile_read(hdmi, ALL_HDMI_HPD_INT_STATUS);

	if (irq_status & (HPD_HOT_PLUG | HPD_HOT_UNPLUG)) {
		if ((irq_status & HPD_HOT_PLUG) && (irq_status & HPD_STATUS) &&
		    (hdmi->hpd_detect == 0)) {
			hdmi->hpd_detect = 1;
			ClearHDMIHPDInterrupt(hdmi);
			hdmi->hpd_changed = 1;
			queue_work(hdmi->workq, &hdmi->hpd_work);
		}
		if ((irq_status & HPD_UNPLUG_RAW) &&
		    (!(irq_status & HPD_STATUS))) {
			hdmi->hpd_detect = 0;
			ClearHDMIHPDInterrupt(hdmi);
			queue_work(hdmi->workq, &hdmi->hpd_work);
		}
	}
	ClearHDMIHPDInterrupt(hdmi);
}

static int hdmi_irq_set(struct xdx_device *xdev, struct xdx_irq_src *src,
			unsigned type, enum xdx_interrupt_state state)
{
	// TODO: enable/disable IRQ
	return -1;
}

static int hdmi_irq_process(struct xdx_device *xdev, struct xdx_irq_src *src,
			    struct xdx_irq_entry *entry)
{
	struct xdx_dm *dm = xdev_to_dm(xdev);
	struct xdx_output *output = dm_to_output(dm);
	struct hdmi_dev *hdmi;

	if (src->id == XDX_IRQ_CLIENTID_HDMI0)
		hdmi = output->hdmi[0];
	else if (src->id == XDX_IRQ_CLIENTID_HDMI1 && dm->dc_num ==2 )
		hdmi = output->hdmi[1];
	else
		return -EINVAL;

	XDXHDMIInterruptHandler(hdmi);

	return 0;
}

static struct xdx_irq_src_funcs hdmi_src_funcs = {
	.set = hdmi_irq_set,
	.process = hdmi_irq_process,
};

static int hdmi_isr_install(struct hdmi_dev *hdmi)
{
	struct xdx_output* output = hdmi_to_output(hdmi);
	struct xdx_dm* dm = output_to_dm(output);
	struct xdx_device* xdev = dm_to_xdev(dm);

	struct xdx_irq_src *irq_src;

	irq_src = kmalloc(sizeof(*irq_src), GFP_KERNEL);
	if (!irq_src)
		return -ENOMEM;

	if (hdmi->id == 0)
		irq_src->id = XDX_IRQ_CLIENTID_HDMI0;
	else if (hdmi->id == 1 && xdev->asic_type == CHIP_PANGU_B0)
		irq_src->id = XDX_IRQ_CLIENTID_HDMI1;
	else
		return -EINVAL;

	irq_src->funcs = &hdmi_src_funcs;

	xdx_irq_register_src(xdev, irq_src);

	hdmi->irq_src = irq_src;

	return 0;
}

static int hdmi_isr_uninstall(struct hdmi_dev *hdmi)
{
	struct xdx_output* output = hdmi_to_output(hdmi);
	struct xdx_dm* dm = output_to_dm(output);
	struct xdx_device* xdev = dm_to_xdev(dm);
	int ret;

	ret = xdx_irq_unregister_src(xdev, hdmi->irq_src);
	if (ret) {
		dev_err(xdev->dev, "failed to unregister irq source\n");
		return ret;
	}

	kfree(hdmi->irq_src);
	hdmi->irq_src = NULL;

	return 0;
}

int xdx_hdmi_init(struct xdx_dm *dm, u32 id)
{
	struct xdx_device *xdev = dm_to_xdev(dm);
	struct drm_device *ddev = xdev_to_drm(xdev);
	struct xdx_output *output = dm_to_output(dm);
	struct hdmi_dev *hdmi = NULL;
	int ret;

	hdmi = kzalloc(sizeof(*hdmi), GFP_KERNEL);
	if (!hdmi) {
		ret = -ENOMEM;
		goto failed_alloc;
	}

	hdmi->i2c = kzalloc(sizeof(struct xdx_hdmi_i2c), GFP_KERNEL);
	if (!hdmi->i2c) {
		ret = -ENOMEM;
		goto failed_alloc;
	}

	hdmi->mode = kzalloc(sizeof(struct hdmi_params), GFP_KERNEL);
	if (!hdmi->mode) {
		ret = -ENOMEM;
		goto failed_alloc;
	}

	hdmi->workq = alloc_ordered_workqueue("dm_hdmi_hpdq", 0);
	if (!hdmi->workq) {
		ret = -ENOMEM;
		goto failed_alloc;
	}
	output->psDoorbellShareMemDesc[id] =NULL;
	output->psDoorbellVirAddr[id] = NULL;
	hdmi->id = id;
	output->hdmi[id] = hdmi;
	hdmi->output = output;
	hdmi->hdmi_mode = NULL;

	video_parms_init(hdmi);

	xdx_hdmi_i2c_init(hdmi);

	xdx_hdmi_register(hdmi, dm_dc_get_crtc_mask(dm->dc[id]));
	hdmi_isr_install(hdmi);
	hdmi->native_mode = drm_mode_duplicate(ddev, &default_mode);
	hdmi->hdmi_mode = drm_mode_duplicate(ddev, &default_mode);
	EnableHDMIHPDInterrupt(hdmi);

	return 0;

failed_alloc:
	if (hdmi) {
		if (hdmi->workq)
			destroy_workqueue(hdmi->workq);

		if (hdmi->mode)
			kfree(hdmi->mode);

		if (hdmi->i2c)
			kfree(hdmi->i2c);

		kfree(hdmi);
	}
	return ret;
}

int xdx_hdmi_deinit(struct xdx_dm *dm, u32 id)
{
	struct xdx_device *xdev = dm_to_xdev(dm);
	struct xdx_output *output = dm_to_output(dm);
	struct hdmi_dev *hdmi = output->hdmi[id];

	DisableHDMIHPDInterrupt(hdmi);
	hdmi_isr_uninstall(hdmi);
	xdx_hdmi_i2c_deinit(hdmi);
	DPDoorbellShareMemFree(output,hdmi->id);
	if (hdmi->workq)
		destroy_workqueue(hdmi->workq);

	if (hdmi->mode)
		kfree(hdmi->mode);

	if (hdmi->native_mode)
	{
		drm_mode_destroy(xdev_to_drm(xdev), hdmi->native_mode);
		hdmi->native_mode = NULL;
	}

	if (hdmi->hdmi_mode)
	{
		drm_mode_destroy(xdev_to_drm(xdev), hdmi->hdmi_mode);
		hdmi->hdmi_mode = NULL;
	}

	if (hdmi->i2c)
		kfree(hdmi->i2c);

	kfree(hdmi);

	output->hdmi[id] = NULL;

	return 0;
}

int xdx_hdmi_hw_init(struct xdx_dm *dm, u32 id)
{
	struct xdx_output *output = dm_to_output(dm);
	struct hdmi_dev *hdmi = output->hdmi[id];
	EnableHDMIHPDInterrupt(hdmi);
	return 0;
}


