/* SPDX-License-Identifier: GPL-2.0 */
/*
 * Copyright (C) 2020 VeriSilicon Holdings Co., Ltd.
 */

#ifndef __XDX_HDMI_DRV_H__
#define __XDX_HDMI_DRV_H__

#include <linux/module.h>
#include <linux/platform_device.h>
#include <linux/version.h>

#include <drm/drm_gem.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 5, 0)
#include <drm/drmP.h>
#endif
#include <drm/drm_crtc.h>
#include <drm/drm_crtc_helper.h>
#include <drm/drm_of.h>
#include <drm/drm_edid.h>

#include <drm/drm_atomic_helper.h>
#include <img_types.h>

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 1, 0))
#include <drm/drm_probe_helper.h>
#endif
#include "pix.h"
#include "dm.h"

#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif

#define PANGU_DEVICE_NAME_HDMI			"xdx_hdmi"
#define HDMI_DEBUG						1
#define REFCLK_HZ						300000000
//#define REFCLK_HZ						80000000


struct xdx_hdmi_i2c {
	struct i2c_adapter adap;

	u32 ddc_addr;
	u32 segment_ptr;

	struct mutex lock;
	struct completion comp;

	u32	fast_speed;
	u32	slow_speed;
	bool	i2cm_fast_mode;
	bool	scdc_fast_mode;
};

typedef enum {
	MODE_UNDEFINED = -1,
	DVI = 0,
	HDMI
} video_mode_t;

typedef enum {
	LINK_UNDEFINED = -1,
	TMDS = 0,
	FRL_3G3L,
	FRL_6G3L,
	FRL_6G4L,
	FRL_8G4L,
	FRL_10G4L,
	FRL_12G4L
} video_link_t;

typedef enum {
	XV_YCC601 = 0,
	XV_YCC709,
	S_YCC601,
	ADOBE_YCC601,
	ADOBE_RGB,
	BT2020_YCC_C,
	BT2020,
	ADDITIONAL_COLORIMETRY
} ext_colorimetry_t;

/**
 * @file
 * For detailed handling of this structure, refer to documentation of the functions
 */
typedef struct {
	/** VIC code */
	u32 mCode;

	/** Identifies modes that ONLY can be displayed in YCC 4:2:0 */
	u8 mLimitedToYcc420;

	/** Identifies modes that can also be displayed in YCC 4:2:0 */
	u8 mYcc420;

	u16 mPixelRepetitionFactor;

	/** In units of 1MHz */
	double mPixelClock;

	/** 1 for interlaced, 0 progressive */
	u8 mInterlaced;

	u16 mHActive;

	u16 mHBlanking;

	u16 mHBorder;

	u16 mHImageSize;

	u16 mHSyncOffset;

	u16 mHSyncPulseWidth;

	/** 0 for Active low, 1 active high */
	u8 mHSyncPolarity;

	u16 mVActive;

	u16 mVBlanking;

	u16 mVBorder;

	u16 mVImageSize;

	u16 mVSyncOffset;

	u16 mVSyncPulseWidth;

	/** 0 for Active low, 1 active high */
	u8 mVSyncPolarity;

} dtd_t;


typedef struct {
	double refresh_rate;
	dtd_t dtd;
} display_mode_t;

typedef enum interfaceType_t {
	PA = 0,
	I2S,
	SPDIF,
} interfaceType_t;
typedef enum {
	STREAM = 0,
	LPCM,
	DOLBY_DIGITAL,		// AC3
	MPEG1,
	MP3,
	MPEG2,
	AAC,
	DTS,
	ATRAC,
	ONE_BIT_AUDIO,
	DOLBY_DIGITAL_PLUS,	// Enhanced AC3
	DTS_HD,
	DOLBY_TRUE_HD,		// MAT
	DST,
	WMAPRO,
	EXTCODINGTYPE
} codingType_t;


typedef enum {
	XCT_NO_DATA = 0,
	HE_AAC,			// Deprecated
	HE_AAV_V2,		// Deprecated
	MPEG_SURROUND,		// Deprecated
	MPEG4_HE_AAC,
	MPEG4_HE_AAC_V2,
	MPEG4_AAC_LC,
	DRA,
	MPEG4_GE_AAC_SURROUND,
	XCT_RESERVED,
	MPEG4_AAC_LC_SURROUND,
	MPEG_H_3D,
	AC4,
	LPCM_3D
} extCodingType_t;
	
/* Supplementary Audio type, table 8-14 HDMI 2.0 Spec. pg 79 */
typedef enum {
	RESERVED = 0,
	AUDIO_FOR_VIS_IMP_NARR,
	AUDIO_FOR_VIS_IMP_SPOKEN,
	AUDIO_FOR_HEAR_IMPAIRED,
	ADDITIONAL_AUDIO
} suppl_A_Type_t;

/* Audio Metadata Packet Header, table 8-4 HDMI 2.0 Spec. pg 71 */
typedef struct {
	u8 m3dAudio;
	u8 mNumViews;
	u8 mNumAudioStreams;
} audioMetaDataHeader_t;

/* Audio Metadata Descriptor, table 8-13 HDMI 2.0 Spec. pg 78 */
typedef struct {
	u8 mMultiviewRightLeft;
	u8 mLC_Valid;
	u8 mSuppl_A_Valid;
	u8 mSuppl_A_Mixed;
	suppl_A_Type_t mSuppl_A_Type;
	u8 mLanguage_Code[3];	/*ISO 639.2 alpha-3 code, examples: eng,fre,spa,por,jpn,chi */

} audioMetaDataDescriptor_t;

typedef struct {
	audioMetaDataHeader_t mAudioMetaDataHeader;
	audioMetaDataDescriptor_t mAudioMetaDataDescriptor[4];
} audioMetaDataPacket_t;


/**
 * For detailed handling of this structure,
 * refer to documentation of the functions
 */
typedef struct {
	interfaceType_t mInterfaceType;	/* Audio interface to be configured */
	codingType_t mCodingType;	/* Coding type */
	extCodingType_t mExtCodingType;	/* Extended Coding Type */
	//packetType_t mPacketType;	/* Packet type */
	int mChannelAllocation;		/* Channel allocation */
	u8 mChannelCount;		/* Channel count */
	u8 mSampleSize;			/* Sample size (audioParams_t *params, 16 to 24) */
	double mSamplingFrequency;	/* Sampling frequency (audioParams_t *params, kHz) */
	u8 mLFEPlayback;		/* LFE Playback vaue (audioParams_t *params, 
					 * see InfoFrame) */
	u8 mLevelShiftValue;		/* Level shift value (audioParams_t *params, 
					 * see InfoFrame) */
	u8 mDownMixInhibitFlag;		/* Down-mix inhibit flag (audioParams_t *params, 
					 * see InfoFrame) */
	u8 mIecCopyright;		/* IEC copyright */
	u8 mIecCgmsA;			/* IEC CGMS-A */
	u8 mIecPcmMode;			/* IEC PCM mode */
	u8 mIecCategoryCode;		/* IEC category code */
	u8 mIecSourceNumber;		/* IEC source number */
	u8 mIecClockAccuracy;		/* IEC clock accuracy */
	u8 mI2SInsertBpcuv;		/* Use the BPCUV data (Parity, Channel status,
					 * User bit, Valid and B bit) or configured them
					 * in the controller instead */
	audioMetaDataPacket_t mAudioMetaDataPacket; /** Audio Multistream variables, to be written to the Audio Metadata Packet */
} audioParams_t;

typedef struct {
	video_mode_t mHdmi;
	video_link_t mLink;
	enum pixel_enc_type mEncoding;
	//encoding_t mEncoding;
	u8 mColorResolution;
	u8 mPixelRepetitionInput;
	u8 mPixelRepetitionOutput;
	display_mode_t mDisplayMode;
	u8 mRgbQuantizationRange;
	u8 mPixelPackingDefaultPhase;
	u8 mColorimetry;
	u8 mScanInfo;
	u8 mActiveFormatAspectRatio;
	u8 mNonUniformScaling;
	ext_colorimetry_t mExtColorimetry;
	u8 mColorimetryDataBlock;
	u8 mItContent;
	u16 mEndTopBar;
	u16 mStartBottomBar;
	u16 mEndLeftBar;
	u16 mStartRightBar;
	u8 mHdmiVideoFormat;
	u8 m3dStructure;
	u8 m3dExtData;
} videoParams_t;

struct hdmi_params {
	videoParams_t	pVideo;
	audioParams_t	pAudio;
};

/**
 * @short HDMI TX controller status information
 *
 * Initialize @b user fields (set status to zero).
 * After opening this data is for internal use only.
 */
typedef struct hdmi_tx_dev {
	u32		version_number;
	u32		version_type;
	u8		avi_fieldrate;	/* Used to configure AVI rate */

	int		hpd;		/* Used to signalize if a cable is connected */
	int		avp_state;	/* Used to register the AVP config state */

	/* System data */
	double		pixel_clock;
	double		link_clock;
	u8		pixel_repetition;
	u8		color_resolution;
	u8		encoding;
	u8		audio_on;
	u8		hdmi_on;
	u8		frl_on;
	u8		hdcp;
	//cec_device_t	cec;
	//i2c_params_t	i2c;
	//earc_params_t	earc;
} hdmi_tx_dev_t;

struct hdmi_dev {
	struct drm_connector connector;
	struct drm_encoder encoder;
	struct xdx_hdmi_i2c *i2c;

	/* Memory Resources */
	//void __iomem		*mem[MAX_MEM_IDX];
	//void __iomem		*mb;

	struct xdx_output  *output;
	struct workqueue_struct *workq;
	struct work_struct hpd_work;

	/* Video parameters */
	struct drm_display_info hdmi_data;
	struct edid *edid;
	struct drm_display_mode *hdmi_mode;
	struct drm_display_mode *native_mode;
	struct hdmi_params		*mode;
	unsigned int vrefresh;
	IMG_HANDLE irq_handle;

	bool scdc_support;

	bool hpd_detect;
	bool hpd_changed;
	bool phy_enabled;
	u32 id;
	struct xdx_irq_src *irq_src;
	spinlock_t hdmi_reg_lock;
	unsigned long irq_flags;
};

static inline struct xdx_output *hdmi_to_output(struct hdmi_dev *hdmi)
{
	return hdmi->output;
}

/**
 * @short Main structures to instantiate the driver
 */
struct hdmiqptx_dev {
	/* Device node */
	struct device		*dev;
	/* Interrupts */
	//int			irq[MAX_IRQ_IDX];

	/* Device Tree Information */
	char			*device_name;

	/* Memory Resources */
	//void __iomem		*mem[MAX_MEM_IDX];

	int			hpd;
};


typedef enum irq_sources {
	MU_INT0 = 1,
	MU_INT1,
	AVP_INT0,
	AVP_INT1,
	AVP_INT2,
	AVP_INT3,
	AVP_INT4,
	AVP_INT5,
	AVP_INT6,
	CEC_INT,
	EARC_INT0,
	EARC_INT1,
	PHYIF_INT,
} irq_sources_t;

typedef enum {
	HDMI_NORMAL_FORMAT = 0,
	HDMI_EXT_RES_FORMAT = 1,
	HDMI_3D_FORMAT = 2,
} hdmi_video_format_t;

typedef enum {
	UNDEFINED_3D = 0xFF,
	FRAME_PACKING_3D = 0,
	TOP_AND_BOTTOM_3D = 6,
	SIDE_BY_SIDE_3D = 8
} hdmi_3d_structure_t;

typedef enum {
	UNDEFINED_EXTDATA_3D = 0,
} hdmi_3d_extdata_t;

struct hdmi_trans_para{
	int clock;
	int hdisplay;
	int hsync_start;
	int hsync_end;
	int htotal;
	int hskew;
	int vdisplay;
	int vsync_start;
	int vsync_end;
	int vtotal;
	int vscan;
	int vrefresh;
	u32 flag;

	u32 colordeep;
	u32 colorformat;
	u32 vic_code;
	u32 pixclock;

	u32 mode;
	u32 bpc;
	u32 multipixel;
	u32 mode_flag_pvsync;
	u32 lanes;
	u32 rate;
	u32 dc_idx;	
	enum hdmi_picture_aspect picture_aspect_ratio;
};

#define PACKET_HDMIVIDEOFORMAT_MASK 0xE0
#define PACKET_3D_STRUCTURE_MASK    0xF0
#define PACKET_3D_EXT_DATA_MASK     0xF0
#define PACKET_HDMI_VIC_MASK   	    0xFF
	
	// HDMI Licensing, LLC
#define HDMI_LICENSING_LLC_OUI	0x000C03

#define xdx_encoder_to_hdmi(e) \
	struct hdmi_dev *hdmi = container_of((e), struct hdmi_dev, encoder)

u32 ClearHDMIHPDInterrupt(struct hdmi_dev *hdmi);
void avp_configure(struct hdmi_dev *hdmi);
int xdx_hdmi_i2c_init(struct hdmi_dev *hdmi);
void xdx_hdmi_i2c_deinit(struct hdmi_dev *hdmi);
void avp_reset(struct hdmi_dev *dev);
bool xdx_hdmi_support_scdc(struct hdmi_dev *hdmi);
void xdx_hdmi_set_ratio_scrambling(struct hdmi_dev *hdmi);
void EnableHDMIHPDInterrupt(struct hdmi_dev *hdmi);
void DisableHDMIHPDInterrupt(struct hdmi_dev *hdmi);
static inline struct hdmi_dev *connector_to_hdmi(struct drm_connector *connector) {
	return container_of(connector, struct hdmi_dev, connector);
}
int xdx_hdmi_phy_init(struct hdmi_dev *hdmi);
void xdx_hdmi_trans_para_fill(struct hdmi_dev *hdmi, struct hdmi_trans_para *pass_para, void* para_sram_vaddr);
/************************************/
//I2C config
#define I2CDDC_TIMEOUT 			5000
int xdx_hdmi_init(struct xdx_dm *dm, u32 id);
int xdx_hdmi_deinit(struct xdx_dm *dm, u32 id);
int xdx_hdmi_hw_init(struct xdx_dm *dm, u32 id);
#endif /* __XDX_HDMI_DRV_H__ */
