#include <linux/delay.h>
#include <drm/drm_scdc_helper.h>
#include "xdx_hdmi.h"
#include "xdx_hdmi_reg.h"
#include "xdx_hdmi_accel.h"
#include "main_unit.h"
#include "fc_packets.h"
#include "xdx_device.h"
#include "dm.h"
/**
 * Check if AVP is supported by core in hardware
 * @return TRUE if hardware supports AVP
 */
int id_avp_support(struct hdmi_dev *hdmi);
static void vi_PixelRepetitionFactor(struct hdmi_dev *dev, u8 value);
static void vp_PixelPackingDefaultPhase(struct hdmi_dev *dev, u8 bit);
static void xdx_FifoInit(struct hdmi_dev *dev);
int video_Configure(struct hdmi_dev *hdmi);
void link_Configure(struct hdmi_dev *dev);

int id_avp_support(struct hdmi_dev *hdmi)
{
	if (dev_read_mask(hdmi, CONFIG_REG, CONFIG_REG_HDCP14_MASK))
		return TRUE;
	else
		return FALSE;
}

static void vi_PixelRepetitionFactor(struct hdmi_dev *dev, u8 value)
{
	/* desired factor */
	dev_write_mask(dev, VIDEO_INTERFACE_CONFIG1, VIDEO_INTERFACE_CONFIG1_PREP_MASK, value);
}

static void vp_PixelPackingDefaultPhase(struct hdmi_dev *dev, u8 bit)
{
	dev_write_mask(dev, VIDEO_PACKING_CONFIG,
		       VIDEO_PACKING_CONFIG_GCP_PHASE_MASK, bit);
}

static void xdx_FifoInit(struct hdmi_dev *dev)
{
	dev_write_mask_shifted(dev, VIDEO_INTERFACE_CONFIG2,
			       VIDEO_INTERFACE_CONFIG2_IPI_DESERIALIZER_MASK, 1);
}


int video_Configure(struct hdmi_dev *hdmi)
{
	struct xdx_output *output = hdmi_to_output(hdmi);
	struct xdx_dm *dm = output_to_dm(output);
	struct xdx_device* xdev = dm_to_xdev(dm);
	dev_dbg(xdev->dev, "PixelRepetitionFactor:%x  PixelPackingDefaultPhase:%x\n",
		hdmi->mode->pVideo.mPixelRepetitionOutput,hdmi->mode->pVideo.mPixelPackingDefaultPhase);
	vi_PixelRepetitionFactor(hdmi, hdmi->mode->pVideo.mPixelRepetitionOutput);
	xdx_FifoInit(hdmi);	
	vp_PixelPackingDefaultPhase(hdmi, hdmi->mode->pVideo.mPixelPackingDefaultPhase);
	return TRUE;
}

static void avp_clear_avmute(struct hdmi_dev *dev)
{
	/* Sleep some time to allow sink to synchonize to the stream.
	 * CTA-861-H recommendation: 5 Video Fields
	 * Lowest refresh rate is 24Hz -> Topping up this value */
	mdelay(21);

	packets_AvMute(dev, FALSE);
}
enum opmode_frl_lanes {
	THREE_LANES = 0,
	FOUR_LANES
};

enum opmode_encoder {
	OPMODE_TMDS = 0,
	OPMODE_FRL
};

enum opmode_transport {
	OPMODE_HDMI = 0,
	OPMODE_DVI
};

void link_Configure(struct hdmi_dev *dev)
{
	// Configure DVI or HDMI
	dev_write_mask_shifted(dev, LINK_CONFIG0, LINK_CONFIG0_OPMODE_DVI_MASK,OPMODE_HDMI);

	// Configure TMDS or FRL
	dev_write_mask_shifted(dev, LINK_CONFIG0, LINK_CONFIG0_OPMODE_FRL_MASK,OPMODE_TMDS);

	// Configure FRL number of lanes
	dev_write_mask_shifted(dev, LINK_CONFIG0, LINK_CONFIG0_OPMODE_FRL_4LANES_MASK,THREE_LANES);

	dev_write_mask_shifted(dev, TMDS_FIFO_CONFIG0,
				       TMDS_FIFO_CONFIG0_FIFO_INIT_ON_OVF_MASK, 1);
}



bool xdx_hdmi_support_scdc(struct hdmi_dev *hdmi)
{
	struct xdx_output *output = hdmi_to_output(hdmi);
	struct xdx_dm *dm = output_to_dm(output);
	struct xdx_device *xdev = dm_to_xdev(dm);
	dev_dbg(xdev->dev,"scdc supported:%d scrambling.supported %d max_tmds_clock %d",
			hdmi->hdmi_data.hdmi.scdc.supported,hdmi->hdmi_data.hdmi.scdc.scrambling.supported,hdmi->hdmi_data.max_tmds_clock);
	if(!hdmi->hdmi_data.hdmi.scdc.supported || !hdmi->hdmi_data.hdmi.scdc.scrambling.supported)
		return false;
	if(hdmi->hdmi_data.max_tmds_clock <= 340000)
		 return false;
	return true; 
}

void xdx_hdmi_set_ratio_scrambling(struct hdmi_dev *hdmi)
{
	if(xdx_hdmi_support_scdc(hdmi)){
		if(hdmi->hdmi_mode->clock > 340000){
			drm_scdc_set_high_tmds_clock_ratio(&hdmi->i2c->adap,1);
			drm_scdc_set_scrambling(&hdmi->i2c->adap,1);
		}else{
			drm_scdc_set_high_tmds_clock_ratio(&hdmi->i2c->adap,0);
			drm_scdc_set_scrambling(&hdmi->i2c->adap,0);
		}
	}
}

void avp_configure(struct hdmi_dev *hdmi)
{
	struct xdx_output *output = hdmi_to_output(hdmi);
	struct xdx_dm *dm = output_to_dm(output);
	struct xdx_device* xdev = dm_to_xdev(dm);
//	struct xdx_device* xdev = container_of(dm, struct xdx_device, dm);
	bool ready;
	int tries;

	if (!id_avp_support(hdmi)) {
		dev_dbg(xdev->dev, "%s: HW doesn't support AVP", __func__);
		return;
	}  

	dev_dbg(xdev->dev,"hdmi avp reset\n");
	xdx_reset_avp(hdmi);
	tries = 10;
	ready = FALSE;
	while(!ready && tries){
		udelay(1000);
		ready = xdx_hdmicore_read(hdmi,MU_INT0_STATUS);
		ready = (ready >> 31) & 1;
		tries--;
	}
	mu_reset_avp(hdmi);
	
	mu_enable(hdmi);

	mu_enable_tmds(hdmi);
	//configure_audio_clk(app, app->mode.pAudio.mSamplingFrequency)//config audo
	tries = 1000;
	ready = FALSE;
	while (!ready && tries) {
		udelay(1000);
		ready = mu_check_avp_clocks_status(hdmi);
		tries--;
	}
	if (!tries) {
		dev_err(xdev->dev, "%s: Failed to have AVP clocks locked", __func__);
		goto exit;
	}
	dev_dbg(xdev->dev,"hdmi link config\n");
	link_Configure(hdmi);
	if (video_Configure(hdmi) != TRUE) {
		dev_err(xdev->dev, "%s: video_Configure error", __func__);
		goto exit;
	}

	dev_dbg(xdev->dev,"hdmi packet config\n");
	if (packets_Configure(hdmi) != TRUE) {
		dev_err(xdev->dev, "%s: Could not configure packets", __func__);
		goto exit;
	}
	//audio config
	//if (audio_Configure(dev, &app->mode.pAudio, p_vsdb) != TRUE) {
	//	LOGGER(ERROR, "%s: Audio not configured", __func__);
	//	goto exit;
	//}
	//hdcp config
	//if (hdcp_configure(dev, &app->pHdcp, &app->mode.pVideo) != TRUE) {
	//	LOGGER(ERROR, "%s: Could not configure HDCP", __func__);
	//	goto exit;
	//}

	avp_clear_avmute(hdmi);
#if 0
	/// Fix audio probelm
	xdx_hdmicore_write(hdmi, 0x44, 0x0);
	xdx_hdmicore_write(hdmi, 0xa8, 0x2002);
	xdx_hdmicore_write(hdmi, 0x820, 0x11);
	xdx_hdmicore_write(hdmi, 0xaa8, 0xf00e);
	xdx_hdmicore_write(hdmi, 0xbc4, 0xc035d);
	xdx_hdmicore_write(hdmi, 0xc24, 0x16f);
	xdx_hdmicore_write(hdmi, 0xc28, 0x100);
	xdx_hdmicore_write(hdmi, 0xe40, 0x1880);
	xdx_hdmicore_write(hdmi, 0xe44, 0x6d602);
#endif
	if(xdx_hdmi_support_scdc(hdmi) && hdmi->hdmi_mode->clock > 340000){
		dev_dbg(xdev->dev,"enable scramb\n");
		xdx_enable_scramb(hdmi,1);
	}else{
		dev_dbg(xdev->dev,"disable scramb\n");
		xdx_enable_scramb(hdmi,0);
	}

	return;
exit:
	return;
	//irq_unmute_all(dev);
}

