/*************************************************************************/ /*!
@File           dma_km.c
@Title          kernel side of dma transfer scheduling
@Copyright      Copyright (c) Imagination Technologies Ltd. All Rights Reserved
@Description    Implements server side code for allowing DMA transfers between
                cpu and device memory.
@License        Dual MIT/GPLv2

The contents of this file are subject to the MIT license as set out below.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

Alternatively, the contents of this file may be used under the terms of
the GNU General Public License Version 2 ("GPL") in which case the provisions
of GPL are applicable instead of those above.

If you wish to allow use of your version of this file only under the terms of
GPL, and not to allow others to use your version of this file under the terms
of the MIT license, indicate your decision by deleting the provisions above
and replace them with the notice and other provisions required by GPL as set
out in the file called "GPL-COPYING" included in this distribution. If you do
not delete the provisions above, a recipient may use your version of this file
under the terms of either the MIT license or GPL.

This License is also included in this distribution in the file called
"MIT-COPYING".

EXCEPT AS OTHERWISE STATED IN A NEGOTIATED AGREEMENT: (A) THE SOFTWARE IS
PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT; AND (B) IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/ /**************************************************************************/
#if defined(__linux__)
#include <linux/version.h>
#include <linux/uaccess.h>
#include <asm/current.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/highmem.h>
#endif

#include "pmr.h"
#include "log2.h"
#include "device.h"
#include "pvrsrv.h"
#include "osfunc.h"
#include "dma_km.h"
#include "pvr_debug.h"
#include "lock_types.h"
#include "allocmem.h"
#include "process_stats.h"
#if defined(PVRSRV_ENABLE_GPU_MEMORY_INFO) && defined(DEBUG)
#include "ri_server.h"
#endif
#include "devicemem.h"
#include "pvrsrv_apphint.h"
#include "pvrsrv_sync_server.h"
#include "km_apphint_defs.h"
#include "di_server.h"
#include "dma_flags.h"
#include "rgx_fwif_km.h"
#include "dma_drv.h"
#include "hash.h"
#include "xdxacpu.h"
#include "xdx_gfx_convert.h"
#include "physmem_lma.h"

/* This header must always be included last */
#if defined(__linux__)
#include "kernel_compatibility.h"
#endif
#define DMA_ALIGN 32
#define LEN_ALIGN 8
#define DMA_MIN_LEN 0x0

#define DMA_MAX_TRANSLATION_STACK_ALLOC				(8)

typedef struct _SERVER_CLEANUP_DATA_
{
	PVRSRV_DEVICE_NODE *psDevNode;
	CONNECTION_DATA *psConnection;
	IMG_UINT32 uiCount;
	SYNC_TIMELINE_OBJ sTimelineObject;
	PMR** ppsPMR;
} SERVER_CLEANUP_DATA;

#if !defined(NO_HARDWARE)
static void Cleanup(void* pvCleanupData, IMG_BOOL bAdvanceTimeline)
{
	PVRSRV_ERROR eError;
	SERVER_CLEANUP_DATA* psCleanupData = (SERVER_CLEANUP_DATA*) pvCleanupData;

#if defined(DMA_VERBOSE)
	PVR_DPF((PVR_DBG_ERROR, "Server Cleanup thread entry (%p)", pvCleanupData));
#endif

	/* Advance timeline */
	if (psCleanupData->sTimelineObject.pvTlObj && bAdvanceTimeline)
	{
		eError = SyncSWTimelineAdvanceKM(psCleanupData->psDevNode, &psCleanupData->sTimelineObject);
		PVR_LOG_IF_ERROR(eError, "SyncSWTimelineAdvanceKM");
		eError = SyncSWTimelineReleaseKM(&psCleanupData->sTimelineObject);
		PVR_LOG_IF_ERROR(eError, "SyncSWTimelineReleaseKM");
	}

	OSAtomicDecrement(&psCleanupData->psConnection->ui32NumDmaTransfersInFlight);
#if defined(DMA_VERBOSE)
	PVR_DPF((PVR_DBG_ERROR, "Decremented to %d", OSAtomicRead(&psCleanupData->psConnection->ui32NumDmaTransfersInFlight)));
#endif
	eError = OSEventObjectSignal(psCleanupData->psConnection->hDmaEventObject);
	if (eError != PVRSRV_OK)
	{
		PVR_DPF((PVR_DBG_ERROR, "%s: OSEventObjectSignal failed: %s",
		        __func__, PVRSRVGetErrorString(eError)));
	}
}
#endif /* !defined(NO_HARDWARE) */


IMG_EXPORT PVRSRV_ERROR
PVRSRVInitialiseDMA(PVRSRV_DEVICE_NODE *psDeviceNode)
{
	PVRSRV_DEVICE_CONFIG *psDevConfig = psDeviceNode->psDevConfig;
	struct xdx_device *xdev = gfx_to_xdev(psDeviceNode);

	if (psDevConfig->bHasDma)
	{
		if (!xdev->vdma.dma_dev)
		{
			PVR_DPF((PVR_DBG_ERROR,
				"%s: Dma device is not ready or some error happed.", __func__));
			return PVRSRV_ERROR_NOT_READY;
		}

		psDeviceNode->hDmaTxChan = (IMG_HANDLE)get_dma_channel(xdev->vdma.dma_dev);
		if (!psDeviceNode->hDmaTxChan)
		{
			return PVRSRV_ERROR_RESOURCE_UNAVAILABLE;
		}
		psDeviceNode->bHasSystemDMA = true;
		psDeviceNode->cache = KMEM_CACHE(sg_table, 0);
		if (!psDeviceNode->cache)
		{
			return PVRSRV_ERROR_RESOURCE_UNAVAILABLE;
		}

		psDeviceNode->sg_pool = mempool_create_slab_pool(1, psDeviceNode->cache);
		if (!psDeviceNode->sg_pool)
		{
			return PVRSRV_ERROR_RESOURCE_UNAVAILABLE;
		}
	}

	return PVRSRV_OK;
}

IMG_EXPORT void
PVRSRVDeInitialiseDMA(PVRSRV_DEVICE_NODE *psDeviceNode)
{
	if (psDeviceNode->bHasSystemDMA)
	{
		PVR_ASSERT(psDeviceNode->hDmaTxChan != NULL);
		mempool_destroy(psDeviceNode->sg_pool);
		kmem_cache_destroy(psDeviceNode->cache);
		release_dma_channel((struct dma_chan *)psDeviceNode->hDmaTxChan);
	}
}

IMG_EXPORT PVRSRV_ERROR
DmaDeviceParams(CONNECTION_DATA *psConnection,
				PVRSRV_DEVICE_NODE *psDevNode,
				IMG_UINT32 *ui32DmaBuffAlign,
				IMG_UINT32 *ui32DmaTransferMult)
				//IMG_UINT32 *ui32DmaTransferMult,
				//IMG_BOOL   *iommuEnable)
{
	PVRSRV_DEVICE_CONFIG *psDevConfig = psDevNode->psDevConfig;

	*ui32DmaBuffAlign = psDevConfig->ui32DmaAlignment;
	*ui32DmaTransferMult = psDevConfig->ui32DmaTransferUnit;
	//*iommuEnable = psDevConfig->bIommuEnable;

	 return PVRSRV_OK;
}

IMG_EXPORT PVRSRV_ERROR
DmaSparseMappingTable(PMR *psPMR,
					  IMG_DEVMEM_OFFSET_T uiOffset,
					  IMG_UINT32 ui32SizeInPages,
					  IMG_BOOL *pbTable)
{
		PVRSRV_ERROR eError = PVRSRV_OK;
		IMG_DEV_PHYADDR *psDevPhyAddr;
		IMG_BOOL *pbValid;

		psDevPhyAddr = OSAllocZMem(ui32SizeInPages * sizeof(IMG_CPU_PHYADDR));
		PVR_LOG_GOTO_IF_NOMEM(psDevPhyAddr, eError, err1);

		pbValid = OSAllocZMem(ui32SizeInPages * sizeof(IMG_BOOL));
		PVR_LOG_GOTO_IF_NOMEM(pbValid, eError, err2);

		eError = PMRLockSysPhysAddresses(psPMR);
		PVR_LOG_GOTO_IF_ERROR(eError, "PMRLockSysPhysAddresses", err3);

		eError = PMR_DevPhysAddr(psPMR,
								 OSGetPageShift(),
								 ui32SizeInPages,
								 uiOffset,
								 psDevPhyAddr,
								 pbValid);
		PVR_LOG_GOTO_IF_ERROR(eError, "PMR_DevPhysAddr", err3);

		PMRUnlockSysPhysAddresses(psPMR);

		memcpy(pbTable, pbValid, ui32SizeInPages * sizeof(IMG_BOOL));

err3:
		OSFreeMem(pbValid);
err2:
		OSFreeMem(psDevPhyAddr);
err1:
		return eError;
}

#if defined(SUPPORT_PANGU_ACPU) && defined(SUPPORT_DMA_TRANSFER)
FORCE_INLINE
void Sharemem_ACPUTINYCOPY(AcpuMemCopy_T *AcpuMemCopyInfo, IMG_UINT64 devPAddr, IMG_UINT32 size, IMG_UINT32 dir)
{
	AcpuMemCopyInfo->DstAddr = devPAddr;
	AcpuMemCopyInfo->dataSize = size;
	AcpuMemCopyInfo->copyType = dir;
    AcpuMemCopyInfo->cmdType = acpu_dbcmd_memop_subtype_tiny_copy;
	PVR_ASSERT(size <= MAX_TINYBUFFER_MAX_SIZE);
}
#if 0
FORCE_INLINE
void Sharemem_ACPUTINYCOPY_D2D(AcpuMemCopy_T *AcpuMemCopyInfo, IMG_UINT64 devPAddr, IMG_UINT64 srcPAddr, IMG_UINT32 size)
{
	AcpuMemCopyInfo->DstAddr = devPAddr;
	AcpuMemCopyInfo->dataSize = size;
	AcpuMemCopyInfo->copyType = 2;
    AcpuMemCopyInfo->cmdType = acpu_dbcmd_memop_subtype_tiny_copy;
	AcpuMemCopyInfo->SrcAddr = srcPAddr;
}
#endif
FORCE_INLINE
void Sharemem_ACPUREGCOPY(AcpuMemCopy_T *AcpuMemCopyInfo,
									IMG_UINT64 srcPADDR, IMG_UINT64 devPAddr,
										IMG_UINT32 mask, IMG_UINT32 value)
{
    AcpuMemCopyInfo->cmdType = acpu_dbcmd_memop_subtype_tiny_reg;
	AcpuMemCopyInfo->DstAddr = devPAddr;
	AcpuMemCopyInfo->SrcAddr = srcPADDR;
	AcpuMemCopyInfo->mask = mask;
	AcpuMemCopyInfo->value = value;
}

FORCE_INLINE
int Sharemem_ACPUTINYCOPY_D2D(PVRSRV_DEVICE_NODE * psDevNode, DMA_OP* psOp)
{
	int ret = 0;
	IMG_UINT32 ui32SendSize, ui32HostOffsetInPage, ui32HostPageIndex = 0;
	IMG_UINT32 ui32DevOffsetInPage, ui32DevPageIndex = 0,PageOffset = 0;
	IMG_UINT64 *pui64HostPhyTable = NULL;
	IMG_UINT64 *pui64DevPhyTable = NULL;
	IMG_UINT64 ui64HostPhyAddress,ui64DevPhyAddress,ui32DmaCmdSize;

	IMG_DEVMEM_SIZE_T uiLogicalSize;
	IMG_UINT32 ui32SizeInPages;

	ui32HostPageIndex = psOp->uiHOSTOffset >> PAGE_SHIFT;
	ui32HostOffsetInPage = psOp->uiHOSTOffset & (PAGE_SIZE - 1);

	ui32DevPageIndex = psOp->uiOffset >> PAGE_SHIFT;
	ui32DevOffsetInPage = psOp->uiOffset & (PAGE_SIZE - 1);

	PMR_LogicalSize(psOp->hHostPMR, &uiLogicalSize);
	ui32SizeInPages = uiLogicalSize >> PAGE_SHIFT;
	pui64HostPhyTable = OSAllocZMem(ui32SizeInPages * sizeof(IMG_UINT64));
	if (pui64HostPhyTable == NULL) {
		ret = -ENOMEM;
		goto err_no_mem1;
	}

	PMR_LogicalSize(psOp->hDevPMR, &uiLogicalSize);
	ui32SizeInPages = uiLogicalSize >> PAGE_SHIFT;
	pui64DevPhyTable = OSAllocZMem(ui32SizeInPages * sizeof(IMG_UINT64));
	if (pui64HostPhyTable == NULL) {
		ret = -ENOMEM;
		goto err_no_mem2;
	}

	ret = PMR_GetPages(psOp->hHostPMR, PAGE_SHIFT, &pui64HostPhyTable);
	if (ret) {
		PVR_DPF((PVR_DBG_ERROR, "Failed to get pages of host pmr \n"));
		goto err_phy_addr;
	}

	ret = PMR_GetPages(psOp->hDevPMR, PAGE_SHIFT, &pui64DevPhyTable);
	if (ret) {
		PVR_DPF((PVR_DBG_ERROR, "Failed to get pages of devide pmr \n"));
		goto err_phy_addr;
	}

	PVR_DPF((PVR_DBG_MESSAGE, "%s pui64HostPhyTable = 0x%llx\n",__func__,pui64HostPhyTable[0]));
	PVR_DPF((PVR_DBG_MESSAGE, "%s pui64DevPhyTable[0] = 0x%llx\n",__func__,pui64DevPhyTable[0]));

	ui32DmaCmdSize = 0;
	ui64DevPhyAddress = pui64DevPhyTable[ui32DevPageIndex + PageOffset] + ui32DevOffsetInPage;
	ui64HostPhyAddress = pui64HostPhyTable[ui32HostPageIndex + PageOffset] + ui32HostOffsetInPage;

	while (psOp->uiSize > 0) {
		ui32SendSize = MIN(PAGE_SIZE - ui32HostOffsetInPage, PAGE_SIZE - ui32DevOffsetInPage);

		if (psOp->uiSize > ui32SendSize) {
			psOp->uiSize -= ui32SendSize;
			ui32DmaCmdSize += ui32SendSize;

			if (ui32HostOffsetInPage > ui32DevOffsetInPage) {
				ui32HostOffsetInPage = 0;
				ui32DevOffsetInPage += ui32SendSize;
				if ((pui64HostPhyTable[ui32HostPageIndex + PageOffset] + PAGE_SIZE) ==  pui64HostPhyTable[ui32HostPageIndex + PageOffset+1]) {
					PVR_DPF((PVR_DBG_VERBOSE,"Host buffer continuous page %llx, %llx", pui64HostPhyTable[ui32HostPageIndex + PageOffset], pui64HostPhyTable[ui32HostPageIndex + PageOffset + 1]));
					ui32HostPageIndex++;
				} else {
					acpu_tiny_copy_package(psDevNode, ui64DevPhyAddress, ui64HostPhyAddress, ui32DmaCmdSize);
					PageOffset++;
					ui32DmaCmdSize = 0;
					ui64DevPhyAddress =  pui64DevPhyTable[ui32DevPageIndex + PageOffset -1] + ui32DevOffsetInPage;
					ui64HostPhyAddress = pui64HostPhyTable[ui32HostPageIndex + PageOffset] + ui32HostOffsetInPage;

				}
			} else if (ui32HostOffsetInPage == ui32DevOffsetInPage) {
				ui32HostOffsetInPage = 0;
				ui32DevOffsetInPage = 0;

				if ((pui64HostPhyTable[ui32HostPageIndex + PageOffset] + PAGE_SIZE) == pui64HostPhyTable[ui32HostPageIndex + PageOffset + 1]
					&& (pui64DevPhyTable[ui32DevPageIndex + PageOffset] + PAGE_SIZE) ==  pui64DevPhyTable[ui32DevPageIndex + PageOffset + 1]) {
					PVR_DPF((PVR_DBG_VERBOSE,"Host & Device buffer continuous page %llx, %llx", pui64HostPhyTable[ui32HostPageIndex + PageOffset],pui64HostPhyTable[ui32HostPageIndex + PageOffset +1]));
					ui32HostPageIndex++;
					ui32DevPageIndex++;
				} else {
					acpu_tiny_copy_package(psDevNode, ui64DevPhyAddress, ui64HostPhyAddress, ui32DmaCmdSize);
					PageOffset++;
					ui32DmaCmdSize = 0;
					ui64DevPhyAddress =  pui64DevPhyTable[ui32DevPageIndex + PageOffset] + ui32DevOffsetInPage;
					ui64HostPhyAddress = pui64HostPhyTable[ui32HostPageIndex + PageOffset] + ui32HostOffsetInPage;

				}
			} else {
				ui32DevOffsetInPage = 0;
				ui32HostOffsetInPage += ui32SendSize;
				if ((pui64DevPhyTable[ui32DevPageIndex + PageOffset] + PAGE_SIZE) ==  pui64DevPhyTable[ui32DevPageIndex + PageOffset + 1]) {
					PVR_DPF((PVR_DBG_VERBOSE,"Device buffer continuous page %llx, %llx", pui64DevPhyTable[ui32DevPageIndex + PageOffset], pui64DevPhyTable[ui32DevPageIndex + PageOffset + 1]));
					ui32DevPageIndex++;
				} else {
					acpu_tiny_copy_package(psDevNode, ui64DevPhyAddress, ui64HostPhyAddress, ui32DmaCmdSize);
					PageOffset++;
					ui32DmaCmdSize = 0;
					ui64DevPhyAddress =  pui64DevPhyTable[ui32DevPageIndex + PageOffset] + ui32DevOffsetInPage;
					ui64HostPhyAddress = pui64HostPhyTable[ui32HostPageIndex + PageOffset -1] + ui32HostOffsetInPage;
				}
			}
		} else {
			ui32DmaCmdSize += psOp->uiSize;
			acpu_tiny_copy_package(psDevNode, ui64DevPhyAddress, ui64HostPhyAddress, ui32DmaCmdSize);
			break;
		}
	}

	OSFreeMem(pui64DevPhyTable);
	OSFreeMem(pui64HostPhyTable);

	return 0;

err_phy_addr:
	OSFreeMem(pui64DevPhyTable);

err_no_mem2:
	OSFreeMem(pui64HostPhyTable);

err_no_mem1:
	return ret;
}
#endif


IMG_EXPORT PVRSRV_ERROR
DmaTransfer(CONNECTION_DATA *psConnection,
		    PVRSRV_DEVICE_NODE *psDevNode,
			IMG_UINT32 uiNumDMAs,
			IMG_UINT32 uiNumVDMAs,
			DMA_OP *pDma_Op,
			IMG_UINT32 uiFlags,
			PVRSRV_TIMELINE iUpdateFenceTimeline)
{
	PVRSRV_ERROR eError = PVRSRV_OK;
#ifdef SUPPORT_DMA_TRANSFER_B0
	DMA_INTERLEAV_T dma_interleav_info;
	IMG_UINT64 host_buf;
	IMG_UINT8 *tmp;
	IMG_UINT32 num = 0;
	DIR_B0_T Tdma_dir;
	IMG_UINT64		ui64HostAddr;
#endif
#if defined(NO_HARDWARE)
	/* On nohw the kernel call just advances the timeline to signal completion */

	SYNC_TIMELINE_OBJ sSwTimeline = {NULL, PVRSRV_NO_TIMELINE};

	if (iUpdateFenceTimeline != PVRSRV_NO_TIMELINE)
	{
		eError = SyncSWGetTimelineObj(iUpdateFenceTimeline, &sSwTimeline);
		PVR_LOG_RETURN_IF_ERROR(eError, "SyncSWGetTimelineObj");

		eError = SyncSWTimelineAdvanceKM(psDevNode, &sSwTimeline);
		PVR_LOG_RETURN_IF_ERROR(eError, "SyncSWTimelineAdvanceKM");

		eError = SyncSWTimelineReleaseKM(&sSwTimeline);
		PVR_LOG_RETURN_IF_ERROR(eError, "SyncSWTimelineReleaseKM");
	}
	return PVRSRV_OK;
#else
	int i, ret;
	//IMG_DEV_PHYADDR *psDevPhyAddr;
	SERVER_CLEANUP_DATA asServerData;
	IMG_DMA_ADDR sDmaBaseAddr, sDmaAddr;
	void*  pvOSData;
	IMG_UINT32 ui32SizeInPages;
	IMG_UINT32 uiOffsetInPage;
	IMG_UINT32 ui32Log2PageSize = OSGetPageShift();
	PVRSRV_RGXDEV_INFO *psDevInfo = psDevNode->pvDevice;
	IMG_PBYTE asAddressArray[DMA_MAX_TRANSLATION_STACK_ALLOC * (sizeof(IMG_DMA_ADDR) + sizeof(IMG_BOOL))];
	IMG_DMA_ADDR *psDmaAddr = NULL;
	IMG_BOOL *pbValid = NULL;
#if defined(SUPPORT_PANGU_ACPU) && defined(SUPPORT_DMA_TRANSFER)
	AcpuMemCopy_T *AcpuMemCopyInfo =  (AcpuMemCopy_T *)psDevInfo->puiTinyCopyVirAddr;
	IMG_UINT32 ui32TinyIndex, dbNum = INVALID_DB_INDEX, bitmap = 0;
#endif
#ifdef DEBUG
	IMG_UINT64 now_timer;
	IMG_UINT32 ui32TotalSize = 0;
	pr_debug("in DmaTransfer...uiNumVDMAs %d, uiNumDMAs %d\n", uiNumVDMAs, uiNumDMAs);
	now_timer = OSClockus64();
#endif
	sDmaBaseAddr.uiAddr = 0;
	if (!psDevNode->hDmaTxChan) {
		eError = PVRSRVInitialiseDMA(psDevNode);
		PVR_LOG_RETURN_IF_ERROR_VA(eError, "[%s]Dma device not ready.", __func__);
	}
	if(uiNumVDMAs)
	{
		//OSLockAcquire(psConnection->hDmaReqLock);
		if (!psConnection->bAcceptDmaRequests)
		{
			//OSLockRelease(psConnection->hDmaReqLock);
			return PVRSRV_OK;
		}
		//OSLockRelease(psConnection->hDmaReqLock);

		OSAtomicIncrement(&psConnection->ui32NumDmaTransfersInFlight);
#if defined(DMA_VERBOSE)
		PVR_DPF((PVR_DBG_ERROR, "Incremented to %d", OSAtomicRead(&psConnection->ui32NumDmaTransfersInFlight)));
#endif
		OSCachedMemSet(&asServerData, 0, sizeof(SERVER_CLEANUP_DATA));
		if (iUpdateFenceTimeline != PVRSRV_NO_TIMELINE)
		{
			eError = SyncSWGetTimelineObj(iUpdateFenceTimeline, &asServerData.sTimelineObject);
			PVR_LOG_GOTO_IF_ERROR(eError, "SyncSWGetTimelineObj", e0);
		}

		asServerData.psDevNode = psDevNode;
		asServerData.psConnection = psConnection;

		eError = OSDmaAllocData(psDevNode, uiNumVDMAs, &asServerData, Cleanup, &pvOSData);
		PVR_LOG_GOTO_IF_ERROR(eError, "OSDmaAllocData failed", e1);
	}

	for (i=0; i < uiNumDMAs; i++)
	{
		uiOffsetInPage = pDma_Op[i].uiOffset & (OSGetPageSize() - 1);
		ui32SizeInPages = (pDma_Op[i].uiSize + uiOffsetInPage + OSGetPageSize() - 1) >> OSGetPageShift();

		if(pDma_Op[i].eCmdType & DMA_CMD_PHY_FLAG)
		{
			sDmaBaseAddr.uiAddr = pDma_Op[i].ui64DevPAddr;
		}
		else if(pDma_Op[i].hDevPMR)
		{
			if (!PMR_IsSparse(pDma_Op[i].hDevPMR)&& (PMR_IsPhysContig(PMRGetFlavourData(pDma_Op[i].hDevPMR))))
			{
				IMG_BOOL bValid;
				eError = PMRLockSysPhysAddresses(pDma_Op[i].hDevPMR);
				PVR_LOG_GOTO_IF_ERROR(eError, "Lock Phys Addresses failed", e2);
				eError = PMR_DevPhysAddr(pDma_Op[i].hDevPMR,
										ui32Log2PageSize,
										1,
										0,
										(IMG_DEV_PHYADDR*)&sDmaBaseAddr,
										&bValid);
				PVR_LOG_GOTO_IF_ERROR(eError, "PMR_DevPhysAddr failed", e3);
				eError = PMRUnlockSysPhysAddresses(pDma_Op[i].hDevPMR);
				PVR_LOG_GOTO_IF_ERROR(eError, "Unlock Phys Addresses failed", e2);
			}
			else
			{
				//if size in page > 8, allocate new buffer, otherwize use stack array
				if(ui32SizeInPages > DMA_MAX_TRANSLATION_STACK_ALLOC)
				{
					psDmaAddr = OSAllocZMem(ui32SizeInPages * (sizeof(IMG_DMA_ADDR) + sizeof(IMG_BOOL)));
					PVR_LOG_GOTO_IF_NOMEM(psDmaAddr, eError, e2);
				}
				else
				{
					psDmaAddr = (IMG_DMA_ADDR *)&asAddressArray[0];
				}
				pbValid = (IMG_BOOL*)(psDmaAddr + ui32SizeInPages);

				eError = PMRLockSysPhysAddresses(pDma_Op[i].hDevPMR);
				PVR_LOG_GOTO_IF_ERROR(eError, "Lock Phys Addresses failed", e2);
				eError = PMR_DevPhysAddr(pDma_Op[i].hDevPMR,
										 OSGetPageShift(),
										 ui32SizeInPages,
										 pDma_Op[i].uiOffset,
										 (IMG_DEV_PHYADDR*)psDmaAddr,
										 pbValid);
				PVR_LOG_GOTO_IF_ERROR(eError, "PMR_DevPhysAddr failed", e3);
				eError = PMRUnlockSysPhysAddresses(pDma_Op[i].hDevPMR);
				PVR_LOG_GOTO_IF_ERROR(eError, "PMRUnlockSysPhysAddresses",e2);
			}

		}
		else
		{
			pr_err("i %d, uiNumVDMAs %d, uiNumDMAs %d type %d\n", i, uiNumVDMAs, uiNumDMAs, pDma_Op[i].eCmdType);
			PVR_ASSERT(i > 0);
		}

		sDmaAddr.uiAddr = sDmaBaseAddr.uiAddr + pDma_Op[i].uiOffset;
		switch(pDma_Op[i].eCmdType)
		{
			case DMA_CMD_VDMA_COPY:
				PVR_ASSERT(uiNumVDMAs);
				if (!PMR_IsSparse(pDma_Op[i].hDevPMR)&& (PMR_IsPhysContig(PMRGetFlavourData(pDma_Op[i].hDevPMR))))
				{
					eError = OSDmaPrepareTransfer(psDevNode,
													psDevNode->hDmaTxChan,
													&sDmaAddr, (IMG_UINT64*)pDma_Op[i].ui64HostAddr,
													pDma_Op[i].uiSize, (uiFlags & DMA_FLAG_MEM_TO_DEV), pvOSData, (uiFlags & DMA_FLAG_FROM_USER));
					PVR_LOG_GOTO_IF_ERROR(eError, "prepare transfer failed", e2);
					asServerData.uiCount++;
				}
				else
				{
					eError = OSDmaPrepareTransferSparse(psDevNode, psDevNode->hDmaTxChan,
																							psDmaAddr, pbValid,
																							(IMG_UINT64*)pDma_Op[i].ui64HostAddr, pDma_Op[i].uiSize,
																							uiOffsetInPage, ui32SizeInPages,
																							(uiFlags & DMA_FLAG_MEM_TO_DEV),
																							pvOSData);
					PVR_LOG_GOTO_IF_ERROR(eError, "prepare sparse transfer failed", e2);
					if(ui32SizeInPages > DMA_MAX_TRANSLATION_STACK_ALLOC)
					{
						OSFreeMem(psDmaAddr);
						//OSFreeMem(pbValid); //due to allocat address one time, so don't free pbValid
					}

					asServerData.uiCount++;
				}
#ifdef DEBUG
				ui32TotalSize += pDma_Op[i].uiSize;
#endif
			break;
			case DMA_CMD_VDMA_COPY_PHY:
				PVR_ASSERT(uiNumVDMAs);
				eError = OSDmaPrepareTransfer(psDevNode,
								psDevNode->hDmaTxChan,
								&sDmaAddr, (IMG_UINT64*)pDma_Op[i].ui64HostAddr,
								pDma_Op[i].uiSize, (uiFlags & DMA_FLAG_MEM_TO_DEV), pvOSData, (uiFlags & DMA_FLAG_FROM_USER));
				PVR_LOG_GOTO_IF_ERROR(eError, "prepare transfer failed", e2);
				asServerData.uiCount++;
#ifdef DEBUG
				ui32TotalSize += pDma_Op[i].uiSize;
#endif
				break;

#if defined(SUPPORT_PANGU_ACPU) && defined(SUPPORT_DMA_TRANSFER)
			case DMA_CMD_TINY_FILL:
				ui32TinyIndex = TinyBufferPoolGet(psDevInfo, IMG_FALSE);
				if(INVALID_DB_INDEX == ui32TinyIndex)
				{
					//TinyBuffer is not available, release have allocated
					Do_ACPUTINYCOPY(psDevInfo, bitmap, (uiFlags & DMA_FLAG_FROM_USER));
					bitmap = 0;

					//try it again
					ui32TinyIndex = TinyBufferPoolGet(psDevInfo, IMG_TRUE);
				}

				PVR_ASSERT(uiFlags & DMA_FLAG_MEM_TO_DEV);
				PVR_ASSERT(pDma_Op[i].uiSize <= sizeof(pDma_Op[i].as32FillData));
				memcpy(&AcpuMemCopyInfo[ui32TinyIndex].Data, &(pDma_Op[i].as32FillData[0]), pDma_Op[i].uiSize);
				Sharemem_ACPUTINYCOPY(&AcpuMemCopyInfo[ui32TinyIndex], sDmaAddr.uiAddr, pDma_Op[i].uiSize, 0);
				BIT_SET(bitmap, ui32TinyIndex);

				break;

			case DMA_CMD_TINY_COPY_OPT:
				Sharemem_ACPUTINYCOPY_D2D(psDevNode, &pDma_Op[i]);

				break;

			case DMA_CMD_TINY_COPY:
				ui32TinyIndex = TinyBufferPoolGet(psDevInfo, IMG_FALSE);
				if(INVALID_DB_INDEX == ui32TinyIndex)
				{
					//TinyBuffer is not available, release have allocated
					Do_ACPUTINYCOPY(psDevInfo, bitmap, (uiFlags & DMA_FLAG_FROM_USER));
					bitmap = 0;

					//try it again
					ui32TinyIndex = TinyBufferPoolGet(psDevInfo, IMG_TRUE);
				}

				if(uiFlags & DMA_FLAG_MEM_TO_DEV)
				{
					PVR_ASSERT(pDma_Op[i].uiSize <= MAX_TINYBUFFER_MAX_SIZE);
					if(uiFlags & DMA_FLAG_FROM_USER)
					{
						ret = copy_from_user(&(AcpuMemCopyInfo[ui32TinyIndex].Data[0]), (IMG_UINT64*)pDma_Op[i].ui64HostAddr, pDma_Op[i].uiSize);
					}
					else
					{
						//if use kmalloc/Vmalloc,should use memcpy
						//pr_debug("[%s] kmalloc use memcpy \n", __FUNCTION__);
						memcpy(&(AcpuMemCopyInfo[ui32TinyIndex].Data[0]), (IMG_UINT64*)pDma_Op[i].ui64HostAddr, pDma_Op[i].uiSize);
					}
					Sharemem_ACPUTINYCOPY(&AcpuMemCopyInfo[ui32TinyIndex], sDmaAddr.uiAddr, pDma_Op[i].uiSize, 0);
					BIT_SET(bitmap, ui32TinyIndex);
				}
				else if(uiFlags & DMA_FLAG_DEV_TO_DEV)
				{
					PVR_ASSERT(0);
				}
				else //DMA_FLAG_DEV_TO_MEM
				{
					//pr_debug("[%s] copy to user with D2M \n", __FUNCTION__);
					Sharemem_ACPUTINYCOPY(&AcpuMemCopyInfo[ui32TinyIndex], sDmaAddr.uiAddr, pDma_Op[i].uiSize, 1);
					psDevInfo->asArrayTinyCopyHostAddr[ui32TinyIndex] = pDma_Op[i].ui64HostAddr;
					BIT_SET(bitmap, ui32TinyIndex);
				}
				break;

			case DMA_CMD_BIT_COPY:
				ui32TinyIndex = TinyBufferPoolGet(psDevInfo, IMG_FALSE);
				if(INVALID_DB_INDEX == ui32TinyIndex)
				{
					//TinyBuffer is not available, release have allocated
					Do_ACPUTINYCOPY(psDevInfo, bitmap, (uiFlags & DMA_FLAG_FROM_USER));
					bitmap = 0;
					//try it again
					ui32TinyIndex = TinyBufferPoolGet(psDevInfo, IMG_TRUE);
				}

				Sharemem_ACPUREGCOPY(&AcpuMemCopyInfo[ui32TinyIndex], sDmaAddr.uiAddr,
								sDmaAddr.uiAddr, pDma_Op[i].mask, pDma_Op[i].value);
				BIT_SET(bitmap, ui32TinyIndex);
				break;

			case DMA_CMD_TINY_MEMSET:
				acpu_tiny_memset_package(psDevNode, sDmaAddr.uiAddr, pDma_Op[i].uiSize, &(pDma_Op[i].value));

				break;

			case DMA_CMD_VDMA_COPY_PMR:
				PVR_ASSERT(0);
				break;

			case DMA_CMD_RECT:
				#ifdef SUPPORT_DMA_TRANSFER_B0
				pr_err("%s line:%d\n",__func__,__LINE__);
				if(uiFlags & DMA_FLAG_DEV_TO_DEV)
				{
					if(pDma_Op[i].hHostPMR)
					{
						pr_err("hHostPMR = 0x%llx\n",(IMG_UINT64)pDma_Op[i].hHostPMR);
						eError = PMRLockSysPhysAddresses(pDma_Op[i].hHostPMR);
						pr_err("%s line:%d\n",__func__,__LINE__);
						PVR_ASSERT(PVRSRV_OK == eError);
						pr_err("%s line:%d\n",__func__,__LINE__);

						if (!PMR_IsSparse(pDma_Op[i].hHostPMR))
						{
							IMG_BOOL bValid;
							eError = PMR_DevPhysAddr(pDma_Op[i].hHostPMR,
													ui32Log2PageSize,
													1,
													0,
													(IMG_DEV_PHYADDR*)&sDmaBaseAddr,
													&bValid);
							PVR_ASSERT(PVRSRV_OK == eError);
						}
						else
						{
							PVR_ASSERT(0);
						}
					}
				//	pDma_Op[i].ui64HostAddr = sDmaBaseAddr.uiAddr + pDma_Op[i].hHostuiOffset;
					ui64HostAddr = sDmaBaseAddr.uiAddr + pDma_Op[i].HOSTOffset;
					pr_err("ui64HostAddr = 0x%llx sDmaAddr = 0x%llx Width = 0x%x Height = 0x%x MemStride = 0x%x DevStride = 0x%x\n",ui64HostAddr,sDmaAddr.uiAddr,pDma_Op[i].Width,pDma_Op[i].Height,pDma_Op[i].MemStride,pDma_Op[i].DevStride);
					Dma_2D_devtodev(psDevNode,
								pvOSData,
								psDevNode->hDmaTxChan,
								(IMG_UINT64 *)ui64HostAddr,
								sDmaAddr.uiAddr,
								pDma_Op[i].HOSTOffset,
								pDma_Op[i].DevOffset,
								pDma_Op[i].uiSize,
								pDma_Op[i].Width,
								pDma_Op[i].Height,
								pDma_Op[i].MemStride,
								pDma_Op[i].DevStride,
								(uiFlags & DMA_FLAG_DEV_TO_DEV));
				}
				else
				{
					Dma_2D_copy(psDevNode,
								pvOSData,
								psDevNode->hDmaTxChan,
								(IMG_UINT64 *)pDma_Op[i].ui64HostAddr,
								sDmaAddr.uiAddr,
								pDma_Op[i].HOSTOffset,
								pDma_Op[i].DevOffset,
								pDma_Op[i].uiSize,
								pDma_Op[i].Width,
								pDma_Op[i].Height,
								pDma_Op[i].MemStride,
								pDma_Op[i].DevStride,
								(uiFlags & DMA_FLAG_MEM_TO_DEV));
				}
				#else
				PVR_ASSERT(0);
				#endif
				break;
#ifdef SUPPORT_DMA_TRANSFER_B0
		case DMA_CMD_TDMA:
			host_buf = (IMG_UINT64)kmalloc(0x400000,GFP_KERNEL);
			memset((IMG_UINT8 *)host_buf,0x00,0x400000);

			dma_interleav_info.Dst_addr = sDmaAddr.uiAddr;
			dma_interleav_info.width = pDma_Op[i].T_Width;
			dma_interleav_info.height = pDma_Op[i].T_Height;
			dma_interleav_info.Srcstride = pDma_Op[i].T_XStride;
			dma_interleav_info.Dststride = pDma_Op[i].T_YStride;
			dma_interleav_info.depth = pDma_Op[i].T_Depth;
			dma_interleav_info.LenUnit = pDma_Op[i].T_LenUnit;
			dma_interleav_info.Start_addr = pDma_Op[i].ui64HostAddr;
			tmp = (IMG_UINT8 *)host_buf;

			pr_err("sDmaAddr.uiAddr = 0x%llx\n",sDmaAddr.uiAddr);
			if(uiFlags & DMA_FLAG_MEM_TO_DEV)
			{
			//	pr_err("LenUnit = %d width = %ld height = %ld depth = %ld\n",dma_interleav_info.LenUnit,dma_interleav_info.width,dma_interleav_info.height,dma_interleav_info.depth);
				if(dma_interleav_info.LenUnit == 0)
				{
					for(num=0;num < dma_interleav_info.width*dma_interleav_info.height*dma_interleav_info.depth*1;num++)
					{
					//	pr_err("num = %d\n",num);
						tmp[num] = num;
					}
				}
				else if(dma_interleav_info.LenUnit == 1)
				{
					for(num=0;num < dma_interleav_info.width*dma_interleav_info.height*dma_interleav_info.depth*2;num++)
					{
						tmp[num] = num;
					}
				}
				else if(dma_interleav_info.LenUnit == 2)
				{
					for(num=0;num < dma_interleav_info.width*dma_interleav_info.height*dma_interleav_info.depth*4;num++)
					{
					//	pr_err("num = %d\n",num);
						tmp[num] = num;
					}
				}
				else if(dma_interleav_info.LenUnit == 3)
				{
					for(num=0;num < dma_interleav_info.width*dma_interleav_info.height*dma_interleav_info.depth*6;num++)
					{
						tmp[num] = num;
					}
				}
				else if(dma_interleav_info.LenUnit == 4)
				{
					for(num=0;num < dma_interleav_info.width*dma_interleav_info.height*dma_interleav_info.depth*8;num++)
					{
						tmp[num] = num;
					}
				}
				else
				{
					pr_err("uninit host buf ...\n");
				}
				//dma_interleav_info.Start_addr = host_buf;
				Tdma_dir = MEM_TO_DEV;
			}
			else
			{
				if(pDma_Op[i].hHostPMR)
				{
					pr_err("hHostPMR = 0x%llx\n",(IMG_UINT64)pDma_Op[i].hHostPMR);
					eError = PMRLockSysPhysAddresses(pDma_Op[i].hHostPMR);
					pr_err("%s line:%d\n",__func__,__LINE__);
					PVR_ASSERT(PVRSRV_OK == eError);
					pr_err("%s line:%d\n",__func__,__LINE__);

					if (!PMR_IsSparse(pDma_Op[i].hHostPMR))
					{
						IMG_BOOL bValid;
						eError = PMR_DevPhysAddr(pDma_Op[i].hHostPMR,
												ui32Log2PageSize,
												1,
												0,
												(IMG_DEV_PHYADDR*)&sDmaBaseAddr,
												&bValid);
						PVR_ASSERT(PVRSRV_OK == eError);
					}
					else
					{
						PVR_ASSERT(0);
					}
				}
				else
				{
					PVR_ASSERT(0);
				}
				pr_err("%s line:%d sDmaBaseAddr.uiAddr = 0x%llx\n",__func__,__LINE__,sDmaBaseAddr.uiAddr);
				sDmaAddr.uiAddr = sDmaBaseAddr.uiAddr + pDma_Op[i].T_HOSTOffset;
				pr_err("%s line:%d sDmaAddr.uiAddr = 0x%llx\n",__func__,__LINE__,sDmaAddr.uiAddr);
				dma_interleav_info.Start_addr = sDmaAddr.uiAddr;
				Tdma_dir = DEV_TO_DEV;
			}

			Tdma_Transfer(psDevNode,psDevNode->hDmaTxChan, &dma_interleav_info, Tdma_dir);
			kfree((void *)host_buf);
			break;
#endif

#endif
			default:
				pr_err("not support transfer type %x", pDma_Op[i].eCmdType);
				PVR_ASSERT(0);
		}
	}
#if defined(SUPPORT_PANGU_ACPU) && defined(SUPPORT_DMA_TRANSFER)
	if(bitmap)
	{
		dbNum = Do_ACPUTINYCOPYNonBlock(psDevInfo, bitmap, (uiFlags & DMA_FLAG_FROM_USER));
	}
#endif
	if(uiNumVDMAs)
	{
		if (asServerData.uiCount == uiNumVDMAs)
		{
			OSDmaSubmitTransfer(pvOSData, psDevNode->hDmaTxChan);
		}
		else
		{
			/* One of the transfers could not be programmed, roll back */
			OSDmaForceCleanup(psDevNode->hDmaTxChan, pvOSData);
		}
		cleanup_thread(pvOSData);
	}
#if defined(SUPPORT_PANGU_ACPU) && defined(SUPPORT_DMA_TRANSFER)
	if(dbNum != INVALID_DB_INDEX)
	{
		Wait_ACPUTINYCOPY(psDevInfo, dbNum, bitmap);
		TinyBufferPoolPuts(psDevInfo, bitmap, (uiFlags & DMA_FLAG_FROM_USER));
	}
#endif
#ifdef DEBUG
	PVR_DPF((PVR_DBG_WARNING, "%s Dma transfer %d Bytes cost %lld us", __func__, ui32TotalSize, OSClockus64()-now_timer));
#endif

	return eError;
e3:
	PMRUnlockSysPhysAddresses(pDma_Op[i].hDevPMR);
e2:
	if(psDmaAddr != NULL && ui32SizeInPages > DMA_MAX_TRANSLATION_STACK_ALLOC)
	{
		OSFreeMem(psDmaAddr);
	}
	if(pvOSData != NULL)
	{
		OSDmaFreeData(pvOSData);
	}
e1:
	if (iUpdateFenceTimeline != PVRSRV_NO_TIMELINE)
	{
		SyncSWTimelineReleaseKM(&asServerData.sTimelineObject);
	}
e0:
	//OSLockRelease(psConnection->hDmaReqLock);
	return eError;
#endif
}

IMG_EXPORT PVRSRV_ERROR
DmaMemFill(CONNECTION_DATA *psConnection,
		    PVRSRV_DEVICE_NODE *psDevNode,
			PMR* psPMR,
			IMG_DEVMEM_OFFSET_T uiOffset,
			IMG_DEVMEM_SIZE_T uiSize,
			IMG_UINT32	ui32Pattern)
{
	PVRSRV_ERROR eError = PVRSRV_OK;
#if defined(NO_HARDWARE)
	return PVRSRV_OK;
#else
	IMG_DEV_PHYADDR sDevPhyAddr;
	//IMG_DMA_ADDR sDmaAddr;
	IMG_BOOL bValid;
#if defined(SUPPORT_PANGU_ACPU) && defined(SUPPORT_DMA_TRANSFER)
	eError = PMRLockSysPhysAddresses(psPMR);
	PVR_LOG_GOTO_IF_ERROR(eError, "PMRLockSysPhysAddresses", loop_e3);

	if (!PMR_IsSparse(psPMR) && (PMR_IsPhysContig(PMRGetFlavourData(psPMR))))
	{
		eError = PMR_DevPhysAddr(psPMR,
							 OSGetPageShift(),
							 1,
							 uiOffset,
							 &sDevPhyAddr,
							 &bValid);
		PVR_LOG_GOTO_IF_ERROR(eError, "PMR_DevPhysAddr", loop_e4);

		PVR_ASSERT(bValid == IMG_TRUE);

		ACPUTINYMEMSET(psDevNode, sDevPhyAddr.uiAddr, uiSize, &ui32Pattern);

	}
	else
	{
	//	IMG_UINT32 ui32SizeInPages;
	//	ui32SizeInPages = (uiSize + uiOffsetInPage + OSGetPageSize() - 1) >> OSGetPageShift();
		PVR_DPF((PVR_DBG_ERROR,"Dma memfill %lld ui32Pattern %d \n", uiSize, ui32Pattern));
		PVR_ASSERT(0);
	}

loop_e4:
	PMRUnlockSysPhysAddresses(psPMR);

#endif
loop_e3:
	return eError;
#endif
}

IMG_EXPORT PVRSRV_ERROR DmaSparseMemFill_lma(PVRSRV_DEVICE_NODE * psDevNode, IMG_UINT64 * pages, IMG_UINT32 pageNum, IMG_UINT32 pageSize, IMG_BYTE ui32Pattern)
{
	PVRSRV_ERROR eError;
	IMG_UINT32 page_idx = 0;
	IMG_UINT32 dma_idx = 0;
	IMG_UINT32 valid_idx = 0;
	IMG_UINT64 *psDmaAddr = NULL;
	IMG_UINT64 *psDmaSize = NULL;
	IMG_UINT32 valid_size = pageSize;
	IMG_UINT32 ui32MemSize;
	IMG_UINT32 pattern = ui32Pattern;

	if (pages != NULL) {
		psDmaAddr = OSAllocZMem(pageNum * (sizeof(IMG_UINT64)));
		PVR_GOTO_IF_ERROR(!psDmaAddr, err2);
		psDmaSize = OSAllocZMem(pageNum * (sizeof(IMG_UINT64)));
		PVR_GOTO_IF_ERROR(!psDmaSize, err1);

		psDmaAddr[0] = pages[0];
		psDmaSize[0] = pageSize;
		for (page_idx = 1; page_idx < pageNum; page_idx++) {
			if (pages[page_idx] == psDmaAddr[valid_idx] + valid_size) {
				valid_size += pageSize;
				if ((page_idx + 1) == pageNum)
					psDmaSize[valid_idx] = valid_size;
			} else {
				psDmaSize[valid_idx] = valid_size;
				valid_idx++;
				psDmaAddr[valid_idx] = pages[page_idx];
				valid_size = pageSize;
				if ((page_idx + 1) == pageNum) {
					psDmaSize[valid_idx] = pageSize;
				}
			}
		}

		for (dma_idx = 0; dma_idx <= valid_idx; dma_idx++) {
			while (psDmaSize[dma_idx] > 0) {
				ui32MemSize = MIN(MAX_ACPU_MEMSET_SIZE, psDmaSize[dma_idx]);
				eError = ACPUTINYMEMSET(psDevNode, psDmaAddr[dma_idx], ui32MemSize, &pattern);
				if (eError != PVRSRV_OK) {
					PVR_DPF((PVR_DBG_ERROR, "Failed to memset by ACPUTINYMEMSET"));
					return  eError;
				}
				psDmaAddr[dma_idx] = psDmaAddr[dma_idx] + ui32MemSize;
				psDmaSize[dma_idx] -= ui32MemSize;
			}
		}
	}

	OSFreeMem(psDmaAddr);
	OSFreeMem(psDmaSize);

	return PVRSRV_OK;

err1:
	OSFreeMem(psDmaAddr);

err2:
	PVR_DPF((PVR_DBG_ERROR, "Failed to allocate system memory"));

	return PVRSRV_ERROR_FAILED_TO_ALLOC_USER_MEM;
}

