/* XDXGPU Headers */
#include "xdx_device.h"
#include "pangu.h"

/* bridge headers */
#include "pvr_device.h"

/* PVR Headers */
#include "pvrsrv_device.h"
#include "syscommon.h"
#include "rgx_fwif_shared.h"
#include "rgxdevice.h"
#include "pvrsrv.h"

#include "xdx_irq.h"

static void _pvr_error_notify(IMG_HANDLE hSysData,
			      PVRSRV_ROBUSTNESS_NOTIFY_DATA *psErrorData)
{
	struct xdx_device *xdev = (struct xdx_device *)hSysData;
	uint32_t level;

	switch (psErrorData->eResetReason) {
	case RGX_CONTEXT_RESET_REASON_NONE:
	case RGX_CONTEXT_RESET_REASON_GUILTY_LOCKUP:
	case RGX_CONTEXT_RESET_REASON_INNOCENT_LOCKUP:
	case RGX_CONTEXT_RESET_REASON_GUILTY_OVERRUNING:
	case RGX_CONTEXT_RESET_REASON_INNOCENT_OVERRUNING:
	case RGX_CONTEXT_RESET_REASON_HARD_CONTEXT_SWITCH:
	case RGX_CONTEXT_RESET_REASON_GPU_ECC_OK:
	case RGX_CONTEXT_RESET_REASON_FW_ECC_OK:
		level = PVR_DBG_MESSAGE;
		break;
	case RGX_CONTEXT_RESET_REASON_GPU_ECC_HWR:
	case RGX_CONTEXT_RESET_REASON_FW_EXEC_ERR:
		level = PVR_DBG_WARNING;
		break;
	case RGX_CONTEXT_RESET_REASON_WGP_CHECKSUM:
	case RGX_CONTEXT_RESET_REASON_TRP_CHECKSUM:
	case RGX_CONTEXT_RESET_REASON_FW_ECC_ERR:
	case RGX_CONTEXT_RESET_REASON_FW_WATCHDOG:
	case RGX_CONTEXT_RESET_REASON_FW_PAGEFAULT:
	case RGX_CONTEXT_RESET_REASON_HOST_WDG_FW_ERR:
	default:
		level = PVR_DBG_ERROR;
		break;
	}

	if (psErrorData->pid > 0) {
		if (level == PVR_DBG_MESSAGE)
			dev_info(xdev->dev, "PID %d experienced error: %d\n",
				 psErrorData->pid, psErrorData->eResetReason);
		else if (level == PVR_DBG_WARNING)
			dev_warn(xdev->dev, "PID %d experienced error: %d\n",
				 psErrorData->pid, psErrorData->eResetReason);
		else if (level == PVR_DBG_ERROR)
			dev_err(xdev->dev, "PID %d experienced error: %d\n",
				psErrorData->pid, psErrorData->eResetReason);
	} else {
		if (level == PVR_DBG_MESSAGE)
			dev_info(xdev->dev, "Device experienced error: %d\n",
				 psErrorData->eResetReason);
		else if (level == PVR_DBG_WARNING)
			dev_warn(xdev->dev, "Device experienced error: %d\n",
				 psErrorData->eResetReason);
		else if (level == PVR_DBG_ERROR)
			dev_err(xdev->dev, "Device experienced error: %d\n",
				psErrorData->eResetReason);
	}

	if ((psErrorData->eResetReason ==
	     RGX_CONTEXT_RESET_REASON_WGP_CHECKSUM) ||
	    (psErrorData->eResetReason ==
	     RGX_CONTEXT_RESET_REASON_TRP_CHECKSUM)) {
		dev_err(xdev->dev, "ExtJobRef 0x%x, DM %d",
			psErrorData->uErrData.sChecksumErrData.ui32ExtJobRef,
			psErrorData->uErrData.sChecksumErrData.eDM);
	}
}

/* #define _DBG printk */
#define _DBG(...)

static PVRSRV_ERROR PrePower(IMG_HANDLE hSysData,
                             PVRSRV_SYS_POWER_STATE eNewPowerState,
                             PVRSRV_SYS_POWER_STATE eCurrentPowerState,
                             PVRSRV_POWER_FLAGS ePwrFlags)
{
   	struct xdx_device *psXdxDev = (struct xdx_device *)hSysData;
	IMG_DEV_PHYADDR sDevPAddr = {0};
	IMG_UINT64 uiHeapTotalSize, uiHeapUsedSize, uiHeapFreeSize;
	IMG_UINT64 uiSize = 0, uiOffset = 0;
	PVRSRV_ERROR eError;

	_DBG("(%s()) suspend snapshot state: current=%u, new=%u; flags: 0x%08x", __func__,
	     eCurrentPowerState,
	     eNewPowerState, ePwrFlags);

	/* The transition might be both from ON or OFF states to OFF state so check
	 * only for the *new* state. Also this is only valid for suspend requests. */
	if (eNewPowerState != PVRSRV_SYS_POWER_STATE_OFF ||
	    !BITMASK_HAS(ePwrFlags, PVRSRV_POWER_FLAGS_SUSPEND_REQ))
	{
		return PVRSRV_OK;
	}

	eError = LMA_HeapIteratorCreate(psXdxDev->gfx.pvrdev,
					PHYS_HEAP_USAGE_GPU_LOCAL,
	                                (PHYS_HEAP_ITERATOR**)&psXdxDev->gfx.pvr_heap_iter);
	PVR_LOG_GOTO_IF_ERROR(eError, "LMA_HeapIteratorCreate", return_error);

	eError = LMA_HeapIteratorGetHeapStats(psXdxDev->gfx.pvr_heap_iter, &uiHeapTotalSize,
	                                      &uiHeapUsedSize);
	PVR_LOG_GOTO_IF_ERROR(eError, "LMA_HeapIteratorGetHeapStats",
	                      return_error);
	uiHeapFreeSize = uiHeapTotalSize - uiHeapUsedSize;

	_DBG("(%s()) suspend snapshot heap stats: total=0x%" IMG_UINT64_FMTSPECx ", "
	     "used=0x%" IMG_UINT64_FMTSPECx ", free=0x%" IMG_UINT64_FMTSPECx,
	     __func__, uiHeapTotalSize, uiHeapUsedSize, uiHeapFreeSize);

    	psXdxDev->gfx.pvr_pwr_buffer = OSAllocMem(uiHeapUsedSize);
	PVR_LOG_GOTO_IF_NOMEM(psXdxDev->gfx.pvr_pwr_buffer, eError, destroy_iterator);

	while (LMA_HeapIteratorNext(psXdxDev->gfx.pvr_heap_iter, &sDevPAddr, &uiSize))
	{
		void *pvCpuVAddr;
		IMG_CPU_PHYADDR sCpuPAddr = {0};

		if (uiOffset + uiSize > uiHeapUsedSize)
		{
			PVR_DPF((PVR_DBG_ERROR, "uiOffset = %" IMG_UINT64_FMTSPECx ", "
			        "uiSize = %" IMG_UINT64_FMTSPECx, uiOffset, uiSize));

			PVR_LOG_GOTO_WITH_ERROR("LMA_HeapIteratorNext", eError,
			                        PVRSRV_ERROR_INVALID_OFFSET,
			                        free_buffer);
		}

        	PhysHeapDevPAddrToCpuPAddr(((PVRSRV_DEVICE_NODE*)psXdxDev->gfx.pvrdev)->apsPhysHeap[PVRSRV_PHYS_HEAP_GPU_LOCAL],
                           1,
                           &sCpuPAddr,
                           &sDevPAddr);

		pvCpuVAddr = OSMapPhysToLin(sCpuPAddr, uiSize,
		                            PVRSRV_MEMALLOCFLAG_CPU_UNCACHED_WC);

		_DBG("(%s()) suspend snapshot iterator: dev_paddr=%px, cpu_paddr=%px, cpu_vaddr=%px, "
		     "size=0x%05" IMG_UINT64_FMTSPECx, __func__,
		     (void *) sDevPAddr.uiAddr, (void *) sCpuPAddr.uiAddr,
		     pvCpuVAddr, uiSize);

		/* copy memory */
		memcpy((IMG_BYTE *) psXdxDev->gfx.pvr_pwr_buffer + uiOffset, pvCpuVAddr,
		       uiSize);
		/* and now poison it */
		memset(pvCpuVAddr, 0x9b, uiSize);

		uiOffset += uiSize;

		OSUnMapPhysToLin(pvCpuVAddr, uiSize);
	}

	return PVRSRV_OK;

free_buffer:
	OSFreeMem(psXdxDev->gfx.pvr_pwr_buffer);
	psXdxDev->gfx.pvr_pwr_buffer = NULL;
destroy_iterator:
	LMA_HeapIteratorDestroy(psXdxDev->gfx.pvr_heap_iter);
    psXdxDev->gfx.pvr_heap_iter = NULL;
return_error:
	return eError;
}

static PVRSRV_ERROR PostPower(IMG_HANDLE hSysData,
                              PVRSRV_SYS_POWER_STATE eNewPowerState,
                              PVRSRV_SYS_POWER_STATE eCurrentPowerState,
                              PVRSRV_POWER_FLAGS ePwrFlags)
{
   	struct xdx_device *psXdxDev = (struct xdx_device *)hSysData;
	IMG_DEV_PHYADDR sDevPAddr = {0};
	IMG_UINT64 uiSize = 0, uiOffset = 0;
	PVRSRV_ERROR eError;

	_DBG("(%s()) resume snapshot state: current=%u, new=%u; flags=0x%08x; buffer=%px", __func__,
	     eCurrentPowerState,
	     eNewPowerState, ePwrFlags,
	     psXdxDev->gfx.pvr_pwr_buffer);

	/* The transition might be both to ON or OFF states from OFF state so check
	 * only for the *current* state. Also this is only valid for resume requests. */
	if (eCurrentPowerState != PVRSRV_SYS_POWER_STATE_OFF ||
	    !BITMASK_HAS(ePwrFlags, PVRSRV_POWER_FLAGS_RESUME_REQ) ||
	    psXdxDev->gfx.pvr_pwr_buffer == NULL)
	{
		return PVRSRV_OK;
	}

	eError = LMA_HeapIteratorReset(psXdxDev->gfx.pvr_heap_iter);
	PVR_LOG_GOTO_IF_ERROR(eError, "LMA_HeapIteratorReset", free_buffer);

	while (LMA_HeapIteratorNext(psXdxDev->gfx.pvr_heap_iter, &sDevPAddr, &uiSize))
	{
		void *pvCpuVAddr;
		IMG_CPU_PHYADDR sCpuPAddr = {0};

        PhysHeapDevPAddrToCpuPAddr(((PVRSRV_DEVICE_NODE*)psXdxDev->gfx.pvrdev)->apsPhysHeap[PVRSRV_PHYS_HEAP_GPU_LOCAL],
                           1,
                           &sCpuPAddr,
                           &sDevPAddr);

		pvCpuVAddr = OSMapPhysToLin(sCpuPAddr, uiSize,
		                       PVRSRV_MEMALLOCFLAG_CPU_UNCACHED_WC);

		_DBG("(%s()) resume snapshot iterator: dev_paddr=%px, cpu_paddr=%px, cpu_vaddr=%px, "
		     "size=0x%05" IMG_UINT64_FMTSPECx, __func__,
		     (void *) sDevPAddr.uiAddr, (void *) sCpuPAddr.uiAddr,
		     pvCpuVAddr, uiSize);

		/* copy memory */
		memcpy(pvCpuVAddr, (IMG_BYTE *) psXdxDev->gfx.pvr_pwr_buffer + uiOffset,
		       uiSize);

		uiOffset += uiSize;

		OSUnMapPhysToLin(pvCpuVAddr, uiSize);
	}

	LMA_HeapIteratorDestroy(psXdxDev->gfx.pvr_heap_iter);
	psXdxDev->gfx.pvr_heap_iter = NULL;

	OSFreeMem(psXdxDev->gfx.pvr_pwr_buffer);
	psXdxDev->gfx.pvr_pwr_buffer = NULL;

	return PVRSRV_OK;

free_buffer:
	OSFreeMem(psXdxDev->gfx.pvr_pwr_buffer);
	psXdxDev->gfx.pvr_pwr_buffer = NULL;

	return eError;
}

PVRSRV_ERROR SysDevInit(void *pvOSDevice, PVRSRV_DEVICE_CONFIG **ppsDevConfig)
{
	struct device *dev = pvOSDevice;
	struct drm_device *ddev = dev_get_drvdata(dev);
	struct xdx_device *xdev = drm_to_xdev(ddev);
	PVRSRV_DEVICE_CONFIG *psDevConfig;
	RGX_DATA *psRGXData;
	RGX_TIMING_INFORMATION *psRGXTimingInfo;
	PHYS_HEAP_CONFIG *pasPhysHeaps;
	uint32_t alloc_size = sizeof(PVRSRV_DEVICE_CONFIG) +
			      sizeof(*psRGXData) + sizeof(*psRGXTimingInfo) +
			      sizeof(PHYS_HEAP_CONFIG) * HEAP_NUMS;

#ifdef CONFIG_FW_MAIN_ON_SYSMEM
	alloc_size += sizeof(PHYS_HEAP_CONFIG);
#endif

	psDevConfig = kmalloc(alloc_size, GFP_KERNEL);
	if (!psDevConfig)
		return PVRSRV_ERROR_OUT_OF_MEMORY;

	memset(psDevConfig, 0, alloc_size);

	psRGXData = (RGX_DATA *)((uint8_t *)psDevConfig + sizeof(*psDevConfig));
	psRGXTimingInfo = (RGX_TIMING_INFORMATION *)((uint8_t *)psRGXData +
						     sizeof(*psRGXData));
	pasPhysHeaps = (PHYS_HEAP_CONFIG *)((uint8_t *)psRGXTimingInfo +
					    sizeof(*psRGXTimingInfo));

	/* Setup RGX specific timing data */
	psRGXTimingInfo->ui32CoreClockSpeed = xdev->pm.current_gfxclk;
	psRGXTimingInfo->bEnableActivePM = IMG_FALSE;
	psRGXTimingInfo->bEnableRDPowIsland = IMG_FALSE;
	psRGXTimingInfo->ui32ActivePMLatencyms = 10;

	psRGXData->psRGXTimingInfo = psRGXTimingInfo;

	/* Setup DevConfig */
	psDevConfig->pvOSDevice = pvOSDevice;
	psDevConfig->hDevData = psRGXData;
	psDevConfig->pszName = "xdxgpu";
	psDevConfig->ui32IRQ = XDX_IRQ_CLIENTID_GPU;

	/* Setup heaps */
	pvr_init_phys_heaps(xdev, pasPhysHeaps,
			    &psDevConfig->ui32PhysHeapCount);
	psDevConfig->pasPhysHeaps = pasPhysHeaps;
	psDevConfig->eDefaultHeap = PVRSRV_PHYS_HEAP_GPU_LOCAL;

	psDevConfig->sRegsCpuPBase.uiAddr = xdev->rmmio_base + xdev->hw_info.gfx_regs.base;
	psDevConfig->ui32RegsSize = xdev->hw_info.gfx_regs.size;

#if defined(SUPPORT_DMA_TRANSFER)
	psDevConfig->pszDmaTxChanName = "xdx-dma";
	psDevConfig->pszDmaRxChanName = "xdx-dma";
	psDevConfig->bHasDma = true;
	/* Following two values are expressed in number of bytes */
	psDevConfig->ui32DmaTransferUnit = 8;
	psDevConfig->ui32DmaAlignment = 32;
#endif

	psDevConfig->eCacheSnoopingMode = PVRSRV_DEVICE_SNOOP_NONE;
	psDevConfig->bHasNonMappableLocalMemory = IMG_FALSE;
	psDevConfig->bHasFBCDCVersion31 = IMG_FALSE;

	/* Set psDevConfig->pfnSysDevErrorNotify callback */
	psDevConfig->pfnSysDevErrorNotify = _pvr_error_notify;

	/* power management on HW system */
	psDevConfig->pfnPrePowerState = PrePower;
	psDevConfig->pfnPostPowerState = PostPower;

	/* clock frequency */
	psDevConfig->pfnClockFreqGet = NULL;

	psDevConfig->hSysData = xdev;

	*ppsDevConfig = psDevConfig;

	return PVRSRV_OK;
}

void SysDevDeInit(PVRSRV_DEVICE_CONFIG *psDevConfig)
{

	kfree(psDevConfig);
}

struct pvr_irq {
	struct xdx_irq_src base;
	PFN_LISR pfnLISR;
	void *pvData;
	struct xdx_device *xdev;
};

static inline struct pvr_irq *xdx_to_pvr(struct xdx_irq_src *src)
{
	return container_of(src, struct pvr_irq, base);
}

static int pvr_irq_set(struct xdx_device *xdev, struct xdx_irq_src *src,
		       unsigned type, enum xdx_interrupt_state state)
{
	uint32_t offset = PANGU_REG_HOST_MISC_BASE + mmMSI_MISX_HW_ENABLE_PF0;
	uint32_t val = xdx_mmio_readl(xdev, offset);
	uint32_t all_gpu_core_mask;

	if (xdev->asic_type == CHIP_PANGU_A0) {
		all_gpu_core_mask = 
			BIT(xdx_get_msix_idx(xdev->asic_type, MSIX_GPU_CORE0)) |
			BIT(xdx_get_msix_idx(xdev->asic_type, MSIX_GPU_CORE1)) |
			BIT(xdx_get_msix_idx(xdev->asic_type, MSIX_GPU_CORE2)) |
			BIT(xdx_get_msix_idx(xdev->asic_type, MSIX_GPU_CORE3));
	} else {
		all_gpu_core_mask =
			BIT(xdx_get_msix_idx(xdev->asic_type, MSIX_GPU_CORE0)) |
			BIT(xdx_get_msix_idx(xdev->asic_type, MSIX_GPU_CORE1));
	}

	if(state == XDX_IRQ_STATE_ENABLE){
		val = val | all_gpu_core_mask;
	}else{
		val = val & (~all_gpu_core_mask);
	}
	xdx_mmio_writel(xdev, offset, val);
	return 0;
}

static int pvr_irq_process(struct xdx_device *xdev, struct xdx_irq_src *src,
			   struct xdx_irq_entry *entry)
{
	struct pvr_irq *irq = xdx_to_pvr(src);

	if (!xdev->irq.msi_enabled) {
		/* clear hardware irq status */
		xdx_mmio_writeb(xdev, PANGU_REG_DOORBELL_126, 0);
	}

	if (irq->pfnLISR(irq->pvData))
		return 0;

	return -EIO;
}

static struct xdx_irq_src_funcs pvr_src_funcs = {
	.set = pvr_irq_set,
	.process = pvr_irq_process,
};

PVRSRV_ERROR SysInstallDeviceLISR(IMG_HANDLE hSysData, IMG_UINT32 ui32IRQ,
				  const IMG_CHAR *pszName, PFN_LISR pfnLISR,
				  void *pvData, IMG_HANDLE *phLISRData)
{
	struct xdx_device *xdev = (struct xdx_device *)hSysData;
	struct pvr_irq *irq;

	irq = kmalloc(sizeof(struct pvr_irq), GFP_KERNEL);
	if (!irq)
		return PVRSRV_ERROR_OUT_OF_MEMORY;

	irq->base.id = XDX_IRQ_CLIENTID_GPU;
	irq->base.funcs = &pvr_src_funcs;
	irq->pfnLISR = pfnLISR;
	irq->pvData = pvData;
	irq->xdev = xdev;

	xdx_irq_register_src(xdev, &irq->base);
	xdev->gfx.irq_src = &irq->base;

	*phLISRData = (IMG_HANDLE)irq;
	return PVRSRV_OK;
}

PVRSRV_ERROR SysUninstallDeviceLISR(IMG_HANDLE hLISRData)
{
	struct pvr_irq *irq = (struct pvr_irq *)hLISRData;
	struct xdx_device *xdev = irq->xdev;
	int ret;

	ret = xdx_irq_unregister_src(xdev, &irq->base);
	if (ret) {
		dev_err(xdev->dev, "failed to unregister irq source\n");
		return PVRSRV_ERROR_NOT_READY;
	}

	kfree(irq);
	kfree(xdev->gfx.irq_src);
	xdev->gfx.irq_src = NULL;

	return PVRSRV_OK;
}

PVRSRV_ERROR SysDebugInfo(PVRSRV_DEVICE_CONFIG *psDevConfig,
			  DUMPDEBUG_PRINTF_FUNC *pfnDumpDebugPrintf,
			  void *pvDumpDebugFile)
{
	PVR_UNREFERENCED_PARAMETER(psDevConfig);
	PVR_UNREFERENCED_PARAMETER(pfnDumpDebugPrintf);
	PVR_UNREFERENCED_PARAMETER(pvDumpDebugFile);

	return PVRSRV_OK;
}
