#include "xdx_pm.h"
#include "xdx_device.h"
#include "acpu/xdxacpu.h"
#include "xdxthm_irq.h"
#include "../gfx/services/server/include/device.h"
#include "../gfx/services/server/devices/volcanic/rgxdevice.h"

static inline int pci_ps_set_check(struct xdx_device *xdev,int status)
{
	switch (status) {
	case 0:
		break;
	case -EIO:
		dev_warn(xdev->dev, "device doesn't support PCI PM\n");
		break;
	case -EINVAL:
		dev_warn(xdev->dev,
			 "can't enter requested pci power state\n");
		break;
	default:
		dev_warn(xdev->dev, "pci_set_power_state failed %d\n", status);
		break;
	}
	return status;
}

static int xdx_set_ps(struct xdx_device *xdev, pci_power_t state)
{
	int r;
	dev_dbg(xdev->dev,
		 "target D-state: %d\n",
		 state);
	r = pci_set_power_state(xdev->pdev, state);
	if (pci_ps_set_check(xdev, r))
		return r;
	r = pci_set_power_state(xdev_to_audio_pdev(xdev), state);
	dev_dbg(xdev->dev,
		 "gfx function D-state: %d, audio function D-state: %d\n",
		 xdev->pdev->current_state,
		 xdev_to_audio_pdev(xdev)->current_state);
	return pci_ps_set_check(xdev, r);
}

static int curr_ps_check(struct xdx_device *xdev,
			 enum xdx_power_state expected_ps)
{
	/* TODO: skip the check before related h02c doorbell cmd is ready */
	return 0;
}

static int xdx_ps0_to_ps1(struct xdx_device *xdev)
{
	int r;
	struct pci_dev *audio_pdev = xdev_to_audio_pdev(xdev);

	r = curr_ps_check(xdev, XDX_PS0);
	if (r)
		return r;

	r = pci_save_state(xdev->pdev);
	r = pci_save_state(audio_pdev);
	if (r) {
		dev_warn(xdev->dev, "pci_save_state failed\n");
		return r;
	}

	pci_disable_device(xdev->pdev);
	if (pci_is_enabled(audio_pdev))
		pci_disable_device(audio_pdev);

	r = xdx_set_ps(xdev, PCI_D1);
	if (r)
		return r;

	xdev->power_state = XDX_PS1;
	return 0;
	
}

static int xdx_ps0_to_ps2(struct xdx_device *xdev)
{
	int r;
	struct pci_dev *pdev = xdev->pdev;
	struct pci_dev *audio_pdev = xdev_to_audio_pdev(xdev);

	r = curr_ps_check(xdev, XDX_PS0);
	if (r)
		return r;

	r = pci_save_state(pdev);
	r = pci_save_state(xdev_to_audio_pdev(xdev));
	if (r) {
		dev_warn(xdev->dev, "pci_save_state failed\n");
		return r;
	}
	pci_disable_device(pdev);
	if (pci_is_enabled(audio_pdev))
		pci_disable_device(audio_pdev);
	r = xdx_set_ps(xdev, PCI_D3hot);

	if (r)
		return r;

	xdev->power_state = XDX_PS2;
	return 0;
}

static int xdx_set_ps3(struct xdx_device *xdev)
{
	int r = xdx_set_ps(xdev, PCI_D3cold);

	if (r)
		return r;

	xdev->power_state = XDX_PS3;
	return 0;
}

static int xdx_ps2_to_ps0(struct xdx_device *xdev)
{
	int r;
	struct pci_dev *pdev = NULL;
	r = curr_ps_check(xdev, XDX_PS2);
	if (r)
		return r;
	pdev = xdev->pdev;
	r = xdx_set_ps(xdev, PCI_D0);
	if (r)
		return r;
	pci_restore_state(pdev);
	pci_restore_state(xdev_to_audio_pdev(xdev));
	r = pci_enable_device(pdev);
	r = pci_enable_device(xdev_to_audio_pdev(xdev));
	pci_set_master(pdev);
	xdev->power_state = XDX_PS0;
	return 0;
}

static int xdx_ps1_to_ps0(struct xdx_device *xdev)
{
	int r;
	struct pci_dev *audio_pdev = xdev_to_audio_pdev(xdev);

	r = curr_ps_check(xdev, XDX_PS1);
	if (r)
		return r;

	r = xdx_set_ps(xdev, PCI_D0);
	if (r)
		return r;

	pci_restore_state(xdev->pdev);
	pci_restore_state(audio_pdev);
	r = pci_enable_device(xdev->pdev);
	r = pci_enable_device(audio_pdev);

	xdev->power_state = XDX_PS0;
	return 0;
}

int xdx_switch_power_state(struct xdx_device *xdev, enum xdx_power_state target)
{
	int r = 0;

	dev_dbg(xdev->dev, "switch xdxgpu power state from PS%d to PS%d\n",
		 xdev->power_state, target);

	if (unlikely(target == XDX_PS3))
		return xdx_set_ps3(xdev);

	switch (xdev->power_state) {
	case XDX_PS0:
		switch (target) {
		case XDX_PS1:
			r = xdx_ps0_to_ps1(xdev);
			break;
		case XDX_PS2:
			r = xdx_ps0_to_ps2(xdev);
			break;
		default:
			goto invalid_ps_switch;
		}
		break;
	case XDX_PS1:
		switch (target) {
		case XDX_PS0:
			r = xdx_ps1_to_ps0(xdev);
			break;
		case XDX_PS2:
			r = xdx_ps1_to_ps0(xdev);
			if (!r)
				r = xdx_ps0_to_ps2(xdev);
			break;
		default:
			goto invalid_ps_switch;
		}
		break;
	case XDX_PS2:
		switch (target) {
		case XDX_PS0:
			r = xdx_ps2_to_ps0(xdev);
			break;
		case XDX_PS1:
			r = xdx_ps2_to_ps0(xdev);
			if (!r)
				r = xdx_ps0_to_ps1(xdev);
			break;
		default:
			goto invalid_ps_switch;
		}
		break;
	/* TODO: cold reset to transfer from PS3 to PS0 again */
	default:
invalid_ps_switch:
		DRM_WARN("Doesn't support switch from PS%d to PS%d\n",
			 xdev->power_state, target);
		r = -EINVAL;
	}
	return curr_ps_check(xdev, target);
}

void xdx_dvfs_set_freq(struct xdx_device *xdev, u32 gfx_clk)
{
	IMG_HANDLE devData = ((PVRSRV_DEVICE_NODE *)xdev->gfx.pvrdev)->psDevConfig->hDevData;

	power_cmd_type power_cmd;
	power_cmd.dvfs.cmd = cCmcuHostDBCmdDDKPower;
	power_cmd.dvfs.subCmd = cDDKPowerDVFS;
	power_cmd.dvfs.type = DVFS_SET_FREQ;
	power_cmd.dvfs.channel = 1;
	power_cmd.dvfs.freq = gfx_clk;

	H02CDoorBellPowerSet(xdev, power_cmd.all);

	xdev->pm.current_gfxclk = gfx_clk;
	((RGX_DATA*) devData)->psRGXTimingInfo->ui32CoreClockSpeed = gfx_clk;
}

enum xdx_power_state xdx_power_state_get(struct xdx_device *xdev)
{
	/* TODO: send cmd to FW through h02c doorbell */
	return XDX_PS0;
}

long xdx_power_get(struct xdx_device *xdev, uint8_t subcmd,
		   uint8_t type, uint8_t channel)
{
	power_cmd_type power_cmd;

	power_cmd.thermal.cmd = cCmcuHostDBCmdDDKPower;
	power_cmd.thermal.subCmd = subcmd;
	power_cmd.thermal.type = type;
	power_cmd.thermal.channel = channel;

	return H02CDoorBellPowerGet(xdev, power_cmd.all);
}

void xdx_power_set(struct xdx_device *xdev, uint8_t subcmd,
		   uint8_t type, uint8_t channel, long val)
{
	power_cmd_type power_cmd;

	power_cmd.thermal.cmd = cCmcuHostDBCmdDDKPower;
	power_cmd.thermal.subCmd = subcmd;
	power_cmd.thermal.type = type;
	power_cmd.thermal.channel = channel;
	power_cmd.thermal.temp = val;

	H02CDoorBellPowerSet(xdev, power_cmd.all);
}

#define PANGU_REG_DOORBELL_SIZE 0x600
int xdx_pm_init(struct xdx_device *xdev) {
	int ret, err;

	dev_info(xdev->dev, "pm init start\n");
	xdev->thm.thm_rmmio_base =xdev->rmmio_base + PANGU_REG_DOORBELLS_BASE;
	xdev->thm.thm_rmmio_size = PANGU_REG_DOORBELL_SIZE;
	xdev->thm.thm_rmmio = ioremap(xdev->thm.thm_rmmio_base, xdev->thm.thm_rmmio_size); 
	if (xdev->thm.thm_rmmio == NULL) {
		dev_err(xdev->dev, "thm failed to map mmio registers\n");
		ret = -ENODEV;
		goto err1; 
	}

	err = xdxthm_isr_install(xdev);
	if(err != 0) {
    	dev_err(xdev->dev, "%s : failed to install xdx thm irq !!!", __func__);
		ret = -ENODEV;
        goto err2;
    }

	return 0;

err2:
	iounmap(xdev->thm.thm_rmmio);
err1:
	return ret;
}

int xdx_pm_fini(struct xdx_device *xdev) {
	int err;
	err = xdxthm_isr_uninstall(xdev);

	if(err != 0) {
    	dev_err(xdev->dev, "%s : failed to uninstall xdx thm isr !!!", __func__);
		return err;
    }

	return 0;
}

void xdx_runtime_suspend(struct xdx_device *xdev)
{
	int r;

	r =  xdx_switch_power_state(xdev, XDX_PS1);

	if (r) {
		dev_err(xdev->dev,
			"failed to switch xgpu power state to PS%u\n",
			XDX_PS1);
	}
}

void xdx_runtime_resume(struct xdx_device *xdev)
{
	int r;

	r =  xdx_switch_power_state(xdev, XDX_PS0);

	if (r) {
		dev_err(xdev->dev,
			"failed to switch xgpu power state to PS%u\n",
			XDX_PS0);
	}
}

/*
 * to support runtime low power mode with display on
 * one gddr phy (phy3) is still active during low power mode
 * on 2 phy GPU, it's 4K interleave on phy23
 * on 4 phy GPU, it's 4K interleave on phy0123
 * input is the dev_phy_addr of a gddr page
 * then return value is its offset to the next valid addr on phy3
 */
uint64_t rpm_valid_gddr_offset(struct xdx_device *xdev,
			       uint64_t dev_phy_addr)
{
	return PAGE_SIZE * (xdev->fw_st_info.gddr_count - 1) -
	       dev_phy_addr % (PAGE_SIZE * xdev->fw_st_info.gddr_count);
}
