#include <linux/module.h>
#include <linux/platform_device.h>

#include <drm/drm_atomic.h>
#include <drm/drm_atomic_helper.h>
#include <drm/drm_drv.h>
#include <drm/drm_fb_helper.h>
#include <drm/drm_file.h>
#include <drm/drm_gem.h>
#include <drm/drm_gem_framebuffer_helper.h>
#include <drm/drm_ioctl.h>
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 1, 0))
#include <drm/drm_probe_helper.h>
#else
#include <drm/drm_crtc_helper.h>
#endif
#include <drm/drm_vblank.h>

#include "xdx_vkms.h"
#include "xdx_gem.h"
#include "xdx_device.h"

bool enable_cursor;

static void vkms_atomic_commit_tail(struct drm_atomic_state *old_state)
{
	struct drm_device *dev = old_state->dev;
	struct drm_crtc *crtc;
	struct drm_crtc_state *old_crtc_state;
	int i;

	drm_atomic_helper_commit_modeset_disables(dev, old_state);

	drm_atomic_helper_commit_planes(dev, old_state, 0);

	drm_atomic_helper_commit_modeset_enables(dev, old_state);

	drm_atomic_helper_fake_vblank(old_state);

	drm_atomic_helper_commit_hw_done(old_state);

	drm_atomic_helper_wait_for_vblanks(dev, old_state);

	for_each_old_crtc_in_state (old_state, crtc, old_crtc_state, i) {
		struct vkms_crtc_state *vkms_state =
			to_vkms_crtc_state(old_crtc_state);

		flush_work(&vkms_state->composer_work);
	}

	drm_atomic_helper_cleanup_planes(dev, old_state);
}

static const struct drm_mode_config_funcs vkms_mode_funcs = {
	.fb_create = drm_gem_fb_create,
	.atomic_check = drm_atomic_helper_check,
	.atomic_commit = drm_atomic_helper_commit,
};

static const struct drm_mode_config_helper_funcs vkms_mode_config_helpers = {
	.atomic_commit_tail = vkms_atomic_commit_tail,
};

static int vkms_modeset_init(struct xdx_device *xdev)
{
	struct drm_device *ddev = xdev_to_drm(xdev);

	drm_mode_config_init(ddev);
	ddev->mode_config.funcs = &vkms_mode_funcs;
	ddev->mode_config.min_width = XRES_MIN;
	ddev->mode_config.min_height = YRES_MIN;
	ddev->mode_config.max_width = XRES_MAX;
	ddev->mode_config.max_height = YRES_MAX;
	ddev->mode_config.preferred_depth = 24;
	ddev->mode_config.helper_private = &vkms_mode_config_helpers;

	return xdx_vkms_output_init(xdev, 0);
}

static int vkms_early_init(void *handle)
{
	return 0;
}

static int vkms_late_init(void *handle)
{
	return 0;
}

static int vkms_sw_init(void *handle)
{
	struct xdx_device *xdev = (struct xdx_device *)handle;
	struct drm_device *ddev = xdev_to_drm(xdev);
	int ret;
#if (LINUX_VERSION_CODE < KERNEL_VERSION(5, 15, 0))
	ddev->irq_enabled = true;
#endif
	ret = drm_vblank_init(ddev, 1);
	if (ret) {
		dev_err(xdev->dev, "Failed to vblank\n");
		return ret;
	}

	ret = vkms_modeset_init(xdev);

	return ret;
}

static int vkms_sw_fini(void *handle)
{
	struct xdx_device *xdev = (struct xdx_device *)handle;
	struct vkms_device *vkmsdev = xdev_to_vkmsdev(xdev);
	struct drm_device *ddev = xdev_to_drm(xdev);

	drm_atomic_helper_shutdown(ddev);
	drm_mode_config_cleanup(ddev);
	destroy_workqueue(vkmsdev->output.composer_workq);

	return 0;
}

static int vkms_early_fini(void *handle)
{
	return 0;
}

static int vkms_hw_init(void *handle)
{
	return 0;
}

static int vkms_hw_fini(void *handle)
{
	return 0;
}
/*
static void vkms_late_fini(void *handle)
{

}
*/
static int vkms_suspend(void *handle)
{
	return 0;
}

static int vkms_resume(void *handle)
{
	return 0;
}

static bool vkms_is_idle(void *handle)
{
	return 0;
}

static int vkms_wait_for_idle(void *handle)
{
	return 0;
}

const struct xdx_ip_block_version vkms_ip_block_version =
{
	.name = "vkms",
	.type = XDX_IP_BLOCK_TYPE_DC,
	.major = 1,
	.minor = 0,
	.rev = 0,
	.funcs = {
		.early_init = vkms_early_init,
		.late_init = vkms_late_init,
		.sw_init = vkms_sw_init,
		.sw_fini = vkms_sw_fini,
		.early_fini = vkms_early_fini,
		.hw_init = vkms_hw_init,
		.hw_fini = vkms_hw_fini,
		.suspend = vkms_suspend,
		.resume = vkms_resume,
		.is_idle = vkms_is_idle,
		.wait_for_idle = vkms_wait_for_idle,		
	},
};
