#include "xdx_acpu.h"
#include "xdx_device.h"
#include "rgxdevice.h"
#include "xdxacpu_irq.h"
#include "rgxfwutils.h"

/* mmio register read/write functions */
uint64_t cmcu_doorbell_readq(struct xdx_acpu *acpu, uint32_t offset)
{
	u64 data;
	wmb();
	data = OSReadHWReg64(((void __iomem *)acpu->rmmio), offset);
	mb();
	return data;
}

void cmcu_doorbell_writeq(struct xdx_acpu *acpu, uint32_t offset, uint64_t v)
{
	wmb();
	OSWriteHWReg64(((void __iomem *)acpu->rmmio), offset, v);
	mb();
}

void
acpu_doorbell_submit(struct xdx_acpu *acpu, uint32_t dbNum,
		   uint64_t ui64Value)
{
	acpudb_pool_t *pool = acpu->acpu_db_pool;
	mutex_lock(&pool->pool_mutex);
	cmcu_doorbell_wreg64(acpu->rmmio_base, DOORBELL_OFFSET(dbNum),
			     ui64Value);
	WARN_ON(BIT_ISSET(pool->db_submit, dbNum));
	BIT_SET(pool->db_submit, dbNum);
	mutex_unlock(&pool->pool_mutex);
}

int xdx_acpu_fini(struct xdx_device *xdev)
{
	int err = xdxacpu_isr_uninstall(xdev);
	if (err != 0) {
		dev_err(xdev->dev, "%s : failed to uninstall xdx acpu isr !!!",
			__func__);
		return err;
	}
	acpu_doorbell_pool_destroy(xdev->acpu.acpu_db_pool);
	iounmap(xdev->acpu.rmmio);

	return 0;
}

static int pangu_pci_rebar_find_pos(struct pci_dev *pdev, int bar)
{
	unsigned int pos, nbars, i;
	u32 ctrl;

	pos = pci_find_ext_capability(pdev, PCI_EXT_CAP_ID_REBAR);
	if (!pos)
		return -ENOTSUPP;

	pci_read_config_dword(pdev, pos + PCI_REBAR_CTRL, &ctrl);
	nbars = (ctrl & PCI_REBAR_CTRL_NBAR_MASK) >> PCI_REBAR_CTRL_NBAR_SHIFT;

	for (i = 0; i < nbars; i++, pos += 8) {
		int bar_idx;

		pci_read_config_dword(pdev, pos + PCI_REBAR_CTRL, &ctrl);
		bar_idx = ctrl & PCI_REBAR_CTRL_BAR_IDX;
		if (bar_idx == bar)
			return pos;
	}

	return -ENOENT;
}

u32 pangu_pci_rebar_get_possible_sizes(struct pci_dev *pdev, int bar)
{
	int pos;
	u32 cap;

	pos = pangu_pci_rebar_find_pos(pdev, bar);
	if (pos < 0)
		return 0;

	pci_read_config_dword(pdev, pos + PCI_REBAR_CAP, &cap);
	cap &= PCI_REBAR_CAP_SIZES;
	return (fls(cap >> 4) - 1);
}

int pangu_device_resize_fb_bar(struct pci_dev *dev, int resno, u64 rbar_size)
{
	struct pci_bus *root;
	struct resource *res;
	unsigned i;
	u16 cmd;
	int result;

	/* Check if the root BUS has 64bit memory resources */
	root = dev->bus;
	while (root->parent)
		root = root->parent;

	pci_bus_for_each_resource(root, res, i) {
		if (res && res->flags & (IORESOURCE_MEM | IORESOURCE_MEM_64) &&
		    res->start > 0x100000000ull)
			break;
	}

	/* Trying to resize is pointless without a root hub window above 4GB */
	if (!res)
		return 0;

	/* Disable memory decoding while we change the BAR addresses and size */
	pci_read_config_word(dev, PCI_COMMAND, &cmd);
	pci_write_config_word(dev, PCI_COMMAND, cmd & ~PCI_COMMAND_MEMORY);

	pci_release_resource(dev, resno);
	pci_release_resource(dev, resno + 2);

	res = dev->resource + resno;

	result = pci_resize_resource(dev, resno, rbar_size);
	if (result == -ENOSPC)
		dev_info(&dev->dev,
			 "Not enough PCI address space for a large BAR.");
	else if (result == -ENOENT)
		dev_info(&dev->dev,
			 "A PCI device is busy and cannot reallocte resource,Manually modify the BIOS if necessary!!!!");
	else if (result && result != -ENOTSUPP)
		dev_info(&dev->dev, "Problem resizing BAR%d (%d).", resno,
			 result);

	pci_assign_unassigned_bus_resources(dev->bus);

	pci_write_config_word(dev, PCI_COMMAND, cmd);
	return 0;
}

static int xdx_acpu_suspend(void *handle)
{
#ifdef SUPPORT_DMA_TRANSFER
	struct xdx_device *xdev = (struct xdx_device *)handle;
	int r;

	r = acpu_wait_all_db_done(xdev->acpu.acpu_db_pool, 100);
	if (r)
		dev_err(xdev->dev,
			"wait acpu doorbell clear timeout (%d).\n", r);

	return r;
#else
        return 0;
#endif
}

static int xdx_acpu_resume(void *handle)
{
#ifdef SUPPORT_DMA_TRANSFER
	struct xdx_device *xdev = (struct xdx_device *)handle;
	PVRSRV_DEVICE_NODE *psDeviceNode = xdev->gfx.pvrdev;
	PVRSRV_RGXDEV_INFO *psDevInfo = psDeviceNode->pvDevice;
	PVRSRV_ERROR eError;
	IMG_DEV_PHYADDR acpuPhysAddr;
	IMG_BOOL bValid;

	eError = RGXGetPhyAddr(psDevInfo->psTinyCopyMemDesc->psImport->hPMR,
			       &acpuPhysAddr,
			       psDevInfo->psTinyCopyMemDesc->uiOffset,
			       OSGetPageShift(),
			       1,
			       &bValid);
	if (eError != PVRSRV_OK)
	{
		dev_err(xdev->dev,
			"failed to get tiny copy kccb addr (%d).\n", eError);
		return eError;
	}
	psDevInfo->psAcpuFWIfOsInit->psTinyCopyKCCB = acpuPhysAddr;
#endif
	return 0;
}

static int xdx_acpu_early_init(void *handle)
{
	return 0;
}

static int xdx_acpu_early_fini(void *handle)
{
	return 0;
}

static int xdx_acpu_sw_init(void *handle)
{
	struct xdx_device *xdev = (struct xdx_device *)handle;
	u32 cap, ctrl;
	ddkhandshake_cmd_type ddkhandshake_cmd;
	u64 fwcap;
	struct xdx_acpu *acpu = &xdev->acpu;
	int ret, err;

	dev_info(xdev->dev, "acpu init start\n");
	acpu->rmmio_base = xdev->rmmio_base + XGP_CMCU_DOORBELLS_BASE;
	acpu->rmmio_size = XGP_CMCU_DOORBELLS_SIZE;
	acpu->rmmio = ioremap(acpu->rmmio_base, acpu->rmmio_size);
	if (acpu->rmmio == NULL) {
		dev_err(xdev->dev, "acpu failed to map mmio registers\n");
		ret = -ENODEV;
		goto err1;
	}
	err = xdxacpu_isr_install(xdev);
	if (err != 0) {
		dev_err(xdev->dev, "%s : failed to install xdx acpu irq !!!",
			__func__);
		ret = -ENODEV;
		goto err2;
	}
	ret =
	    acpu_doorbell_pool_create(MAX_DOORBELL_POOL_SIZE, acpu->rmmio_base,
				   &(acpu->acpu_db_pool));
	if (PVRSRV_OK != ret) {
		PVR_DPF((PVR_DBG_ERROR, "%s: ACPUDoolBellPoolCreate failed",
			 __func__));
		goto err3;
	}
	cap = acpu_doorbell_get_timeout(acpu->acpu_db_pool, 1000000);
	ctrl = acpu_doorbell_get_timeout(acpu->acpu_db_pool, 1000000);
	if ((cap == INVALID_DB_INDEX) || (ctrl == INVALID_DB_INDEX)) {
		dev_err(xdev->dev, "%s get doorbell from pool timeout",
			__func__);
		ret = -ETIME;
		goto err4;
	}

	ddkhandshake_cmd.all = 0;
	ddkhandshake_cmd.cmd = acpu_dbcmd_ddkmisc;
	ddkhandshake_cmd.subcmd = acpu_dbcmd_ddkmisc_handshake_getcap;

	fwcap = cmcu_doorbell_rreg64(acpu->rmmio_base, DOORBELL_OFFSET(cap));
	dev_info(xdev->dev, "default runtime ACPU  Capability fwcap: 0x%llx.\n",
		 fwcap);
	acpu_doorbell_submit(acpu, cap, ddkhandshake_cmd.all);
	if (acpu_waitfor_value
	    ((volatile uint64_t
	      __iomem)((acpu->rmmio_base + DOORBELL_OFFSET(cap))), 0x1,
	     0x1) != PVRSRV_OK) {
		dev_err(xdev->dev, "%s DoorBell cap timeout", __func__);
		ret = -ETIME;
		goto err4;
	}

	fwcap = cmcu_doorbell_rreg64(acpu->rmmio_base, DOORBELL_OFFSET(cap));
	dev_info(xdev->dev, "runtime ACPU  Capability fwcap: 0x%llx.\n", fwcap);

	/*ACPU don't clear GPU int by default
	 *Use MSIX int type by default
	 */
	ddkhandshake_cmd.all = 0;
	ddkhandshake_cmd.cmd = acpu_dbcmd_ddkmisc;
	ddkhandshake_cmd.subcmd = acpu_dbcmd_ddkmisc_handshake_cntl;
	ddkhandshake_cmd.fwcntl.clr_gpu_int = 0;
	ddkhandshake_cmd.fwcntl.pcie_int_type = MSIX;
	acpu_doorbell_submit(acpu, ctrl, ddkhandshake_cmd.all);

	if (acpu_waitfor_value
	    ((uint64_t __iomem) (acpu->rmmio_base + DOORBELL_OFFSET(ctrl)),
	     0x1, 0x1) != PVRSRV_OK) {
		dev_err(xdev->dev, "%s DoorBell ctrl timeout", __func__);
		ret = -ETIME;
		goto err4;
	}
	fwcap = cmcu_doorbell_rreg64(acpu->rmmio_base, DOORBELL_OFFSET(ctrl));
	dev_info(xdev->dev, "runtime ACPU fwcntl: 0x%llx.\n", fwcap);
	dev_info(xdev->dev, "acpu init success\n");
	return PVRSRV_OK;

err4:
	acpu_doorbell_pool_destroy(acpu->acpu_db_pool);
err3:
	xdxacpu_isr_uninstall(xdev);
err2:
	iounmap(acpu->rmmio);
err1:
	return ret;
}

static int xdx_acpu_sw_fini(void *handle)
{
	struct xdx_device *xdev = (struct xdx_device *)handle;
	int err = xdxacpu_isr_uninstall(xdev);

	if (err != 0) {
		dev_err(xdev->dev, "%s : failed to uninstall xdx acpu isr !!!",
			__func__);
		return err;
	}
	acpu_doorbell_pool_destroy(xdev->acpu.acpu_db_pool);
	iounmap(xdev->acpu.rmmio);
	return 0;
}

static int xdx_acpu_hw_init(void *handle)
{
	return 0;
}

static int xdx_acpu_hw_fini(void *handle)
{
	return 0;
}

const struct xdx_ip_block_version acpu_ip_block_version = {
	.name = "acpu",
	.type = XDX_IP_BLOCK_TYPE_ACPU,
	.major = 1,
	.minor = 0,
	.rev = 0,
	.funcs = {
		.suspend = xdx_acpu_suspend,
		.resume = xdx_acpu_resume,
		.early_init = xdx_acpu_early_init,
		.early_fini = xdx_acpu_early_fini,
		.sw_init = xdx_acpu_sw_init,
		.sw_fini = xdx_acpu_sw_fini,
		.hw_init = xdx_acpu_hw_init,
		.hw_fini = xdx_acpu_hw_fini,
	},
};
