#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 1, 0)
#include <linux/idr.h>
#endif
#include <drm/drm_file.h>
#include <drm/drm_drv.h>
#include <drm/xdxgpu_drm.h>

#include "pangu.h"
#include "xdx_device.h"
#include "xdx_sysfs.h"
#include "xdx_drv.h"
#include "xdx_gfx.h"
#include "dm.h"
#include "xdx_vpu.h"
#include "xdx_audio.h"
#include "audio/xdxaudio_hal.h"
#include "xdx_vkms.h"
#include "xdx_acpu.h"
#include "pm/xdx_hwmon.h"
#include "pm/xdxthm_irq.h"
#include "pm/xdx_pm.h"
#include "dm_fb.h"

/* mmio register read/write functions */
uint32_t xdx_mmio_readl(struct xdx_device *xdev, uint32_t offset)
{
	if (offset < xdev->rmmio_size)
		return (readl(xdev->rmmio + offset));
	BUG();
	return 0;
}

void xdx_mmio_writel(struct xdx_device *xdev, uint32_t offset,
		     uint32_t v)
{
	if (offset < xdev->rmmio_size)
		writel(v, ((void __iomem *)xdev->rmmio) + offset);
	else
		BUG();
}

uint8_t xdx_mmio_readb(struct xdx_device *xdev, uint32_t offset)
{
	if (offset < xdev->rmmio_size)
		return (readb(xdev->rmmio + offset));
	BUG();
}

void xdx_mmio_writeb(struct xdx_device *xdev, uint32_t offset, uint8_t value)
{
	if (offset < xdev->rmmio_size)
		writeb(value, xdev->rmmio + offset);
	else
		BUG();
}

uint32_t xdx_device_rreg(struct xdx_device *xdev, uint32_t reg,
			 uint32_t acc_flags)
{
	if ((reg * 4) < xdev->rmmio_size)
		return readl(((void __iomem *)xdev->rmmio) + (reg * 4));
	BUG();
}

void xdx_device_wreg(struct xdx_device *xdev, uint32_t reg, uint32_t v)
{
	if ((reg * 4) < xdev->rmmio_size)
		writel(v, ((void __iomem *)xdev->rmmio) + (reg * 4));
	else
		BUG();
}

bool xdx_device_is_dc_supported(enum xdx_asic_type asic_type)
{
	switch (asic_type) {
	case CHIP_PANGU_A0:
	case CHIP_PANGU_B0:
		return true;
	default:
		DRM_WARN_ONCE("unsupported asic type: 0x%x\n", asic_type);
		return false;
	}
}

/**
 * @brief Get clock info from BIOS or firmware
 *
 * @param xdev
 * @return int
 */
static int xdx_device_get_clock_info(struct xdx_device *xdev)
{
	xdev->pm.default_gfxclk = xdev->fw_st_info.default_gpu_freq;
	xdev->pm.current_gfxclk = xdx_power_get(xdev, cDDKPowerDVFS, DVFS_GET_FREQ, 1) * 1000000;

	return 0;
}

static void xdx_device_enable_virtual_display(struct xdx_device *xdev)
{
#ifndef NO_HARDWARE
	xdev->flags.virtual_display = false;

	if (xdx_virtual_display) {
		const char *pci_address_name = pci_name(xdev->pdev);
		char *pciaddstr, *pciaddstr_tmp, *pciaddname_tmp, *pciaddname;

		pciaddstr = kstrdup(xdx_virtual_display, GFP_KERNEL);
		pciaddstr_tmp = pciaddstr;
		while ((pciaddname_tmp = strsep(&pciaddstr_tmp, ";"))) {
			pciaddname = strsep(&pciaddname_tmp, ",");
			if (!strcmp("all", pciaddname)
			    || !strcmp(pci_address_name, pciaddname)) {
				long num_crtc;
				int res = -1;

				xdev->flags.virtual_display = true;

				if (pciaddname_tmp)
					res = kstrtol(pciaddname_tmp, 10,
						      &num_crtc);

				if (!res) {
					if (num_crtc < 1)
						num_crtc = 1;
					if (num_crtc > 6)
						num_crtc = 6;
					xdev->vkmsdev.num_crtc = num_crtc;
				} else {
					xdev->vkmsdev.num_crtc = 1;
				}
				break;
			}
		}

		DRM_INFO("virtual display string:%s, %s:virtual_display:%d, num_crtc:%d\n",
			 xdx_virtual_display, pci_address_name,
			 xdev->flags.virtual_display, xdev->vkmsdev.num_crtc);

		kfree(pciaddstr);
	}
#else
	xdev->flags.virtual_display = true;
#endif /* NO_HARDWARE */
}

static void xdx_device_module_param_handle(struct xdx_device *xdev)
{
	if (xdx_virtual_vblank)
		xdev->flags.virtual_vblank = true;

	if (xdx_virtual_output)
		xdev->flags.virtual_output = true;

	if (xdx_fake_mode)
		xdev->flags.fake_mode = true;

	xdx_device_enable_virtual_display(xdev);
}

static int xdx_device_ip_early_init(struct xdx_device *xdev)
{
	int ret;
	uint32_t i;

	for (i = 0; i < xdev->num_ip_blocks; i++) {
		xdev->ip_blocks[i]->status.early_init = 1;

		if (!xdev->ip_blocks[i]->version.funcs.early_init)
			continue;

		ret = xdev->ip_blocks[i]->version.funcs.early_init(xdev);
		if (ret) {
			xdev->ip_blocks[i]->status.early_init = 0;
			dev_err(xdev->dev,
				"early_init of IP block <%s> failed %d\n",
				xdev->ip_blocks[i]->version.name, ret);
		}
	}

	return 0;
}

static int xdx_device_ip_early_fini(struct xdx_device *xdev)
{
	int ret;
	uint32_t i;

	for (i = 0; i < xdev->num_ip_blocks; i++) {
		if (!xdev->ip_blocks[i]->status.early_init)
			continue;

		xdev->ip_blocks[i]->status.early_init = 0;

		if (!xdev->ip_blocks[i]->version.funcs.early_fini)
			continue;

		ret = xdev->ip_blocks[i]->version.funcs.early_fini(xdev);
		if (ret) {
			dev_err(xdev->dev,
				"early_fini of IP block <%s> failed %d\n",
				xdev->ip_blocks[i]->version.name, ret);
		}
	}

	return 0;
}

static int xdx_device_ip_sw_fini(struct xdx_device *xdev)
{
	int i, ret;

	for (i = xdev->num_ip_blocks - 1; i >= 0; i--) {
		if (!xdev->ip_blocks[i]->status.sw_init)
			continue;

		if (!xdev->ip_blocks[i]->version.funcs.sw_fini) {
			dev_warn(xdev->dev, "sw_fini of IP block <%s> is NULL\n",
				  xdev->ip_blocks[i]->version.name);
			continue;
		}

		ret = xdev->ip_blocks[i]->version.funcs.sw_fini(xdev);
		if (ret) {
			dev_err(xdev->dev, "sw_fini of IP block <%s> failed %d\n",
				  xdev->ip_blocks[i]->version.name, ret);
			break;
		}

		xdev->ip_blocks[i]->status.sw_init = 0;
	}

	return ret;
}

static int xdx_device_ip_sw_init(struct xdx_device *xdev)
{
	int ret;
	uint32_t i;

	for (i = 0; i < xdev->num_ip_blocks; i++) {
		if (!xdev->ip_blocks[i]->status.early_init)
			continue;

		if (xdev->ip_blocks[i]->status.sw_init)
			continue;

		if (!xdev->ip_blocks[i]->version.funcs.sw_init) {
			dev_err(xdev->dev, "sw_init of IP block <%s> is NULL\n",
				  xdev->ip_blocks[i]->version.name);
			break;
		}

		ret = xdev->ip_blocks[i]->version.funcs.sw_init(xdev);
		if (ret) {
			xdev->ip_blocks[i]->status.sw_init = 0;
			dev_err(xdev->dev, "sw_init of IP block <%s> failed %d\n",
				  xdev->ip_blocks[i]->version.name, ret);
			break;
		}

		xdev->ip_blocks[i]->status.sw_init = 1;
	}

	if (ret)
		xdx_device_ip_sw_fini(xdev);

	return ret;
}

static int xdx_device_ip_hw_fini(struct xdx_device *xdev)
{
	int i, ret;

	for (i = xdev->num_ip_blocks - 1; i >= 0; i--) {
		if (!xdev->ip_blocks[i]->status.hw_init)
			continue;

		if (!xdev->ip_blocks[i]->version.funcs.hw_fini) {
			dev_warn(xdev->dev, "hw_fini of IP block <%s> is NULL\n",
				  xdev->ip_blocks[i]->version.name);
			continue;
		}

		ret = xdev->ip_blocks[i]->version.funcs.hw_fini(xdev);
		if (ret) {
			dev_err(xdev->dev, "hw_fini of IP block <%s> failed %d\n",
				  xdev->ip_blocks[i]->version.name, ret);
			break;
		}

		xdev->ip_blocks[i]->status.hw_init = 0;
	}

	return ret;
}

static int xdx_device_ip_hw_init(struct xdx_device *xdev)
{
	int ret;
	uint32_t i;

	for (i = 0; i < xdev->num_ip_blocks; i++) {
		if (!xdev->ip_blocks[i]->status.sw_init)
			continue;

		if (xdev->ip_blocks[i]->status.hw_init)
			continue;

		if (!xdev->ip_blocks[i]->version.funcs.hw_init) {
			dev_err(xdev->dev, "hw_init of IP block <%s> is NULL\n",
				  xdev->ip_blocks[i]->version.name);
			break;
		}

		ret = xdev->ip_blocks[i]->version.funcs.hw_init(xdev);
		if (ret) {
			xdev->ip_blocks[i]->status.hw_init = 0;
			dev_err(xdev->dev, "hw_init of IP block <%s> failed %d\n",
				  xdev->ip_blocks[i]->version.name, ret);
			break;
		}

		xdev->ip_blocks[i]->status.hw_init = 1;
	}

	if (ret)
		xdx_device_ip_hw_fini(xdev);

	return ret;
}

#if 1
static int xdx_device_ip_init(struct xdx_device *xdev)
{
	int ret;

	ret = xdx_device_ip_early_init(xdev);
	if (ret) {
		dev_err(xdev->dev, "failed to early init IPs\n");
		return ret;
	}

	ret = xdx_device_ip_sw_init(xdev);
	if (ret) {
		dev_err(xdev->dev, "failed to sw init IPs\n");
		goto failed_sw_init;
	}

	xdx_vmm_init(&xdev->vmm);

	xdev->soc_funcs.hw_init(xdev);

	ret = xdx_device_ip_hw_init(xdev);
	if (ret) {
		dev_err(xdev->dev, "failed to hw init IPs\n");
		goto failed_hw_init;
	}

#ifdef 	SUPPORT_VPU
	ret = xdx_vpu_init(xdev);
	if (ret) {
		dev_err(xdev->dev, "failed to init vpu\n");
		goto failed_hw_init;
	}
#endif

#if	defined(SUPPORT_DMA_TRANSFER) || defined(SUPPORT_DMA_TRANSFER_B0)
	ret = xdx_dma_dev_init(xdev);
	if (ret) {
		dev_err(xdev->dev, "failed to init dma\n");
		goto failed_hw_init;
	}
#endif

#ifdef 	SUPPORT_AUDIO
	ret = xdx_audio_init(xdev);
	if (ret) {
		dev_err(xdev->dev, "failed to init audio\n");
		if (ret != -ENXIO)
			goto failed_hw_init;
	}
#endif

#ifdef 	SUPPORT_PANGU_PM
	ret = xdx_pm_init(xdev);
	if( ret != 0 ) {
    	dev_err(xdev->dev, "%s : failed to install xdx thm irq !!!", __func__);
		goto failed_hw_init;
    }
#endif

return 0;

failed_hw_init:
	xdx_device_ip_sw_fini(xdev);
failed_sw_init:
	xdx_device_ip_early_fini(xdev);
	return ret;
}

static int xdx_device_ip_fini(struct xdx_device *xdev)
{

#ifdef SUPPORT_AUDIO
	xdx_audio_fini(xdev);
#endif

#if defined(SUPPORT_DMA_TRANSFER) ||  defined(SUPPORT_DMA_TRANSFER_B0)
	xdx_dma_dev_finit(xdev);
#endif

#ifdef SUPPORT_VPU
	xdx_vpu_fini(xdev);
#endif

#ifdef 	SUPPORT_PANGU_PM
	xdx_pm_fini(xdev);
#endif

	xdx_device_ip_hw_fini(xdev);

	xdx_vmm_fini(&xdev->vmm);

	xdx_device_ip_sw_fini(xdev);

	xdx_device_ip_early_fini(xdev);

	return 0;
}

#else
static int xdx_device_ip_init(struct xdx_device *xdev)
{
	struct drm_device *ddev = xdev_to_drm(xdev);
	int ret;

	xdx_device_module_param_handle(xdev);

#ifdef 	SUPPORT_PANGU_ACPU
	ret = xdx_acpu_init(xdev);
	if (ret)
		return ret;
#endif

	ret = xdx_gfx_init(xdev);
	if (ret)
		return ret;

	ret = xdx_dm_init(xdev);
	if (ret)
		return ret;

#ifdef 	SUPPORT_VPU
	ret = xdx_vpu_init(xdev);
	if (ret)
		return ret;
#endif

#ifdef 	SUPPORT_DMA_TRANSFER
	xdx_dma_init(xdev);
	if (ret)
		return ret;
#endif

#ifdef 	SUPPORT_AUDIO
	xdx_audio_init(xdev);
	if (ret)
		return ret;
#endif
	return ret;
}

static int xdx_device_ip_fini(struct xdx_device *xdev)
{
	int ret;

#ifdef 	SUPPORT_VPU
	xdx_vpu_fini(xdev);
#endif

#ifdef 	SUPPORT_PANGU_ACPU
	ret = xdx_acpu_fini(xdev);
	if (ret)
		return ret;
#endif

#if defined SUPPORT_PANGU_DC_HDMI
    ret = xdx_hdmi_deinit(xdev->ddev);
	if (ret)
		return ret;
#endif

#if defined SUPPORT_PANGU_DC_DP
	ret = xdx_dp_deinit(xdev->ddev);
	if (ret)
		return ret;
#endif

	ret = xdx_dm_fini(xdev);
	if (ret)
		return ret;
    //TODO
    //xdx_fbdev_fini((struct xdx_fbdev*)xdev->fbdev);

#ifdef 	SUPPORT_DMA_TRANSFER
	xdx_dma_fini(xdev);
#endif
#ifdef 	SUPPORT_AUDIO
	xdx_audio_fini(xdev);
	if (ret)
		return ret;
#endif

	ret = xdx_gfx_fini(xdev);

	xdx_irq_fini(xdev);

	return ret;
}

#endif

static void xdx_device_vram_init(struct xdx_device *xdev)
{
	uint64_t limit = (uint64_t)xdx_vram_limit << 20;
	uint64_t vis_vram_limit = (uint64_t)xdx_vis_vram_limit << 20;
	uint32_t regvalue = 0;
	uint8_t gddr_mask = 0;
	uint8_t count = 4, gddr_mask_tmp = 0;
	uint8_t channel_no = 0;
	uint8_t interleave = 0;
	uint8_t tmp = 0;
	uint64_t gddr_die_sizeGB = 1;
	uint64_t gddr_total_sizeGB = 0;

	xdev->vram_start = xdev->hw_info.gddr_dev_phyaddr_base;
	/**
	 * TODO:
	 * should be queried from firmware or read from a scratch
	 * register
	 * gddr_mask(bit0-3)--0:no gddr,1:with 1 gddr
	 * interleave(bit4-7)--0:no interleave 1:gddr01 or gddr23 2:gddr0123 3:dggr03
	 * gddr_die_sizeGB(bit8-11)--0:1GB 1:2GB 2:4GB 3:8GB
	 * gddr_total_sizeGB(bit12-19)
	 * resv(bit20-31)
	 */
	regvalue = xdx_mmio_readl(xdev, PANGU_REG_GDDR_INFO);
	gddr_mask = regvalue & 0xf;
	gddr_mask_tmp = gddr_mask;
	interleave = (regvalue & 0xf0) >> 4;
	tmp = (regvalue & 0xf00) >> 8;
	while (tmp != 0) {
		gddr_die_sizeGB = gddr_die_sizeGB*2;
		tmp--;
	}

	gddr_total_sizeGB = (regvalue & 0xff000) >> 12;
	while (count--) {
		if (gddr_mask_tmp & 1)
			channel_no++;
		gddr_mask_tmp = gddr_mask_tmp >> 1;
		dev_info(xdev->dev, "gddr_mask_tmp=%d, channel_no=%d\n", gddr_mask_tmp, channel_no);
	}

	if (interleave) {//interleave enabled
#if defined XDX_SUPPORT_VERFW
		if (gddr_total_sizeGB != gddr_die_sizeGB*channel_no) {
			dev_err(xdev->dev, "undefined GDDR total size, pls check fw version\n");
			return;
		}
#endif
		xdev->vram_size = gddr_total_sizeGB << 30;
	} else {
		xdev->vram_size = gddr_die_sizeGB << 30;
	}

	dev_info(xdev->dev, "gddr_mask=%d, interleave=%d, gddr_die_size=%llxGB\n",
			gddr_mask, interleave, gddr_die_sizeGB);
	dev_info(xdev->dev, "gddr_total_size=%llxGB, channel_no=%d, xdev->vram_size=%llx\n",
			gddr_total_sizeGB, channel_no, xdev->vram_size);

	xdev->real_vram_size = xdev->vram_size - xdev->hw_info.reserved_vram_size;
	xdev->vram_end = xdev->vram_start + xdev->real_vram_size - 1;
	if (limit && limit < xdev->real_vram_size)
		xdev->real_vram_size = limit;

	dev_info(xdev->dev, "VRAM: %lluM 0x%016llX - 0x%016llX (%lluM used)\n",
			xdev->vram_size >> 20,
			xdev->vram_start,
			xdev->vram_end,
			xdev->real_vram_size >> 20);

	/* Reduce size of CPU-visible VRAM if requested */
	if (vis_vram_limit &&
	    vis_vram_limit < xdev->visible_vram_size)
		xdev->visible_vram_size = vis_vram_limit;

	if (xdev->visible_vram_size == xdev->vram_size)
		xdev->visible_vram_size = xdev->real_vram_size;

	dev_info(xdev->dev, "Visable VRAM: %lluM 0x%016llX - 0x%016llX (%lluM used)\n",
			xdev->aper_size >> 20,
			xdev->aper_base,
			xdev->aper_base + xdev->aper_size - 1,
			xdev->visible_vram_size >> 20);
}

int xdx_device_init(struct xdx_device *xdev)
{
	struct drm_device *ddev = xdev_to_drm(xdev);
#ifndef NO_HARDWARE
	struct pci_dev *pdev = xdev->pdev;
#else
	struct platform_device *pdev = xdev->pdev;
#endif
	int ret;

	xdx_device_module_param_handle(xdev);

	switch(xdev->asic_type) {
	case CHIP_PANGU_A0:
	case CHIP_PANGU_B0:
		pangu_soc_setup(xdev);
		break;
	default:
		dev_err(&pdev->dev, "wrong asic type\n");
		return -ENODEV;
	}

	xdev->shutdown = false;
	xdev->gfx.gfx_timeout = msecs_to_jiffies(10000);
	xdev->vdma_timeout = xdev->gfx.gfx_timeout;
	xdev->gfx.compute_timeout = msecs_to_jiffies(60000);
	xdev->gfx.pvr_heap_iter = NULL;
	xdev->gfx.pvr_pwr_buffer = NULL;
	atomic_set(&xdev->in_s3, 0);
	atomic_set(&xdev->in_s4, 0);
	idr_init(&xdev->connection_idr);
	mutex_init(&xdev->connection_mutex);

#ifndef NO_HARDWARE
	dev_info(xdev->dev,
		 "initializing device (0x%04X:0x%04X 0x%04X:0x%04X 0x%02X).\n",
		 pdev->vendor, pdev->device, pdev->subsystem_vendor,
		 pdev->subsystem_device, pdev->revision);

	xdx_irq_init(xdev);

	/* Registers mapping */
	xdev->rmmio_base = pci_resource_start(xdev->pdev, 0);
	xdev->rmmio_size = pci_resource_len(xdev->pdev, 0);
	xdev->rmmio = ioremap(xdev->rmmio_base, xdev->rmmio_size);
    printk("Debug: xdev->rmmio address:0x%llx  value:0x%llx\n", (long long int)(&xdev->rmmio), (long long int)(xdev->rmmio));
	if (xdev->rmmio == NULL) {
		dev_err(xdev->dev, "failed to map mmio registers\n");
		return -ENOMEM;
	}
	dev_info(xdev->dev, "register mmio base: 0x%016llx\n",
		 (uint64_t)xdev->rmmio_base);
	dev_info(xdev->dev, "register mmio size: %uM\n",
		 (uint32_t)(xdev->rmmio_size >> 20));

	/* Get CPU visable VRAM */
	xdev->aper_base = pci_resource_start(xdev->pdev, 2);
	xdev->aper_size = pci_resource_len(xdev->pdev, 2);
	xdev->visible_vram_size = xdev->aper_size;
	/* reserve PAT memory space to WC for VRAM */
	arch_io_reserve_memtype_wc(xdev->aper_base, xdev->aper_size);
	/* Add an MTRR for the VRAM */
	xdev->vram_mtrr = arch_phys_wc_add(xdev->aper_base, xdev->aper_size);

	dev_info(xdev->dev, "visable vram base: 0x%016llx\n",
		 (uint64_t)xdev->aper_base);
	dev_info(xdev->dev, "visable vram size: %dM\n",
		 (uint32_t)(xdev->aper_size >> 20));

	/* must call query soc info firstly */
	xdev->soc_funcs.get_device_info(xdev);

	xdx_device_vram_init(xdev);

#ifdef SUPPORT_PANGU_ACPU
	xdev->shareram_base = pci_resource_start(xdev->pdev, 4);
	xdev->shareram_size = pci_resource_len(xdev->pdev, 4);
	xdev->share_ram = ioremap_wc(xdev->shareram_base, xdev->shareram_size);
	ret = cache_fw_static_info(xdev);
	if (ret) {
		dev_err(xdev->dev, "failed to read from fw share ram\n");
		goto err_clock_info;
	}
	xdxgpu_pm_sysfs_init(xdev);
#endif

	xdev->gtt_start = xdev->hw_info.host_dev_phyaddr_base;
	xdev->shareram_start = xdev->hw_info.shareram_dev_phyaddr_base;


	ret = xdx_device_get_clock_info(xdev);
	if (ret) {
		dev_err(xdev->dev, "failed to get clock info\n");
		goto err_clock_info;
	}
#else
	xdev->flags.no_hardware = 1;
	xdev->visible_vram_size = 256 * 1024 * 1024;
#endif

	/* early init functions */
	ret = xdx_device_ip_init(xdev);
	if (ret)
		goto err_clock_info;

	if (xdx_fbdev_init(ddev))
		goto faild_reg_fbdev;

	xdev->power_state = XDX_PS0;
	return 0;

faild_reg_fbdev:
	xdx_fbdev_fini(xdev->fbdev);

err_clock_info:
#ifndef NO_HARDWARE
	iounmap(xdev->rmmio);
	iounmap(xdev->share_ram);
#endif
	return ret;
}

int xdx_device_fini(struct xdx_device *xdev)
{
	struct xdx_fbdev* fbdev = xdev->fbdev;

	xdx_irq_fini(xdev);
	xdx_fbdev_fini(fbdev);
	xdx_device_ip_fini(xdev);
#ifndef NO_HARDWARE
	iounmap(xdev->rmmio);
	iounmap(xdev->share_ram);
	arch_phys_wc_del(xdev->vram_mtrr);
	arch_io_free_memtype_wc(xdev->aper_base, xdev->aper_size);
#endif
	idr_destroy(&xdev->connection_idr);
	mutex_destroy(&xdev->connection_mutex);
	return 0;
}

/**
 * xdx_device_ip_suspend - run suspend for hardware IPs
 *
 * @xdev: xdxgpu_device pointer
 *
 * Main suspend function for hardware IPs.  The list of all the hardware
 * IPs that make up the asic is walked, clockgating is disabled and the
 * suspend callbacks are run.  suspend puts the hardware and software state
 * in each IP into a state suitable for suspend.
 * Returns 0 on success, negative error code on failure.
 */
int xdx_device_ip_suspend(struct xdx_device *xdev)
{
	int r, i;

	for (i = xdev->num_ip_blocks -1; i >= 0; i--) {
		/* display is handled separately */
		if (xdev->ip_blocks[i]->version.type == XDX_IP_BLOCK_TYPE_DC)
			continue;
		DRM_DEBUG("suspending IP block <%s>\n",
			 xdev->ip_blocks[i]->version.name);

		r = xdev->ip_blocks[i]->version.funcs.suspend(xdev);

		if (r) {
			DRM_ERROR("suspend of IP block <%s> failed %d\n",
				  xdev->ip_blocks[i]->version.name, r);
			return r;
		}

		xdev->ip_blocks[i]->status.hw_init = false;
	}

	return 0;
}

int xdx_device_ip_resume(struct xdx_device *xdev)
{
	int r;
	uint32_t i;

	for (i = 0; i < xdev->num_ip_blocks; i++) {
		/* display is handled separately */
		if (xdev->ip_blocks[i]->version.type == XDX_IP_BLOCK_TYPE_DC)
			continue;
		DRM_DEBUG("resuming IP block <%s>\n",
			 xdev->ip_blocks[i]->version.name);

		r = xdev->ip_blocks[i]->version.funcs.resume(xdev);

		if (r) {
			DRM_ERROR("resume of IP block <%s> failed %d\n",
				  xdev->ip_blocks[i]->version.name, r);
			return r;
		}

		xdev->ip_blocks[i]->status.hw_init = true;
	}
	return 0;
}

static int xdx_device_early_suspend(struct xdx_device *xdev)
{
	struct drm_device *ddev = xdev_to_drm(xdev);
	int r;

	drm_kms_helper_poll_disable(ddev);
	r = xdev->dm.base.version.funcs.suspend(xdev);
	if (r) {
		dev_err(xdev->dev, "display suspend failed (%d).\n", r);
		return r;
	}
	xdev->dm.base.status.hw_init = false;
	return 0;
}

static int xdx_device_early_resume(struct xdx_device *xdev)
{
	struct drm_device *ddev = xdev_to_drm(xdev);
	struct drm_encoder *encoder;

	/*
	 * During S3/S4, PCIe would lose power supply and SOC FW
	 * would be reloaded. Which means display HW would be
	 * powered on here. So turnoff the display encoders to
	 * avoid abnormal display before gpu core and dc ip block
	 * have been resumed
	 */
	drm_modeset_lock_all(ddev);
	list_for_each_entry(encoder, &ddev->mode_config.encoder_list, head) {
		encoder->helper_private->prepare(encoder);
	}
	drm_modeset_unlock_all(ddev);

	return 0;
}

static int xdx_device_late_resume(struct xdx_device *xdev)
{
	struct drm_device *ddev = xdev_to_drm(xdev);
	int r;

	/* don't light-up the screen when pm thaw */
	if (atomic_read(&xdev->in_s4))
		return 0;

	r= xdev->dm.base.version.funcs.resume(xdev);
	if (r) {
		dev_err(xdev->dev, "display resume failed (%d).\n", r);
		return r;
	}
	xdev->dm.base.status.hw_init = true;
	drm_kms_helper_poll_enable(ddev);

	return 0;
}

/**
 * xdx_device_suspend - initiate device suspend
 *
 * @dev: drm dev pointer
 * @fbcon : notify the fbdev of suspend
 *
 * Puts the hw in the suspend state (all asics).
 * Returns 0 for success or an error on failure.
 * Called at driver suspend.
 */
int xdx_device_suspend(struct xdx_device *xdev, bool fbcon)
{
	int r;

	r = xdx_device_early_suspend(xdev);
	if (r)
		return r;

	r = xdx_device_ip_suspend(xdev);

	if (r) {
		dev_err(xdev->dev, "xdx_device_ip_suspend failed (%d).\n", r);
		return r;
	}
#ifdef SUPPORT_PANGU_ACPU
	r =  xdx_switch_power_state(xdev, XDX_PS2);

	if (r) {
		dev_err(xdev->dev,
			"failed to switch xgpu power state to PS%u\n",
			XDX_PS2);
		return r;
	}
#endif
	/* TODO: release host resources */
	/* TODO: evict memory */
	return 0;
}

/**
 * xdx_device_resume - initiate device resume
 *
 * @dev: drm dev pointer
 * @fbcon : notify the fbdev of resume
 *
 * Bring the hw back to operating state (all asics).
 * Returns 0 for success or an error on failure.
 * Called at driver resume.
 */
int xdx_device_resume(struct xdx_device *xdev, bool fbcon)
{
	int r;
	struct drm_device *ddev = xdev_to_drm(xdev);

#ifdef SUPPORT_PANGU_ACPU
	r =  xdx_switch_power_state(xdev, XDX_PS0);

	if (r) {
		dev_err(xdev->dev,
			"failed to switch xgpu power state to PS%u\n",
			XDX_PS0);
		return r;
	}
#endif

	r = xdx_device_early_resume(xdev);
	if (r)
		return r;

	/* TODO: reallocate host resources */
	r = xdx_device_ip_resume(xdev);
	if (r) {
		dev_err(xdev->dev, "xdx_device_ip_resume failed (%d).\n", r);
		return r;
	}

	r = xdx_device_late_resume(xdev);
	if (r)
		return r;
#ifdef SUPPORT_AUDIO
#if 0
       xdx_audio_resume(xdev);
#endif /* Enable audio resume after B0 audio driver is ready */
#endif

	/*
	 * dual-monitors link-status will changed(e.g. new cable in or one cable out) while in s3/s4;
	 * just send an event and let userspace tell us what to do.
	 *
	 * has no effect on normal use.
	 * */
	drm_helper_hpd_irq_event(ddev);

	return 0;
}

int xdx_device_reset(struct xdx_device *xdev)
{
	dev_warn(xdev->dev, "%s: no implemented\n", __func__);
	return -1;
}

int xdx_device_ip_block_add(struct xdx_device *xdev,
			    const struct xdx_ip_block *ip_block)
{
	if (!ip_block)
		return -EINVAL;

	dev_info(xdev->dev, "add ip block number %d <%s>\n", xdev->num_ip_blocks,
		ip_block->version.name);

	xdev->ip_blocks[xdev->num_ip_blocks++] = (struct xdx_ip_block *)ip_block;
	return 0;
}

int xdx_device_info_ioctl(struct drm_device *dev, void *data,
			  struct drm_file *filp)
{
	struct xdx_device *xdev = drm_to_xdev(dev);
	struct drm_xdxgpu_info *info = data;
	void __user *out = (void __user *)(uintptr_t)info->return_pointer;
	uint32_t size = info->return_size;

	if (!info->return_size || !info->return_pointer)
		return -EINVAL;

	switch (info->query) {
	case XDXGPU_INFO_DEVICE_INFO: {
		struct drm_xdxgpu_info_device dev_info = {};

		dev_info.asic_id = xdev->asic_type;
		return copy_to_user(out, &dev_info,
				    min((size_t)size, sizeof(dev_info))) ? -EFAULT : 0;
	}
	default:
		dev_err(xdev->dev, "Invalid request %d\n", info->query);
		return -EINVAL;
	}

	return 0;
}
