#include <linux/vmalloc.h>
#include <linux/spinlock.h>
#include <linux/shmem_fs.h>
#include <linux/dma-buf.h>
#include <linux/pfn_t.h>
#include <drm/drm_file.h>
#include <drm/drm_cache.h>

#include <drm/xdxgpu_drm.h>

#include "pangu.h"
#include "xdx_device.h"
#include "xdx_drv.h"
#include "xdx_gem.h"
#include "xdx_drm.h"

#include "pvr_gem.h"

#include "./gfx/include/img_types.h"
#include "./gfx/include/pvrsrv_error.h"
#include "./gfx/services/server/include/pmr.h"
#include "kernel_compatibility.h"
#include "hantro_priv.h"
#include "vpu_xmem.h"
#include "hantro_priv.h"
#include "pvr_dma_resv.h"
#include <linux/printk.h>

int xdx_bo_reserve(struct xdx_bo *bo, bool interruptible)
{
	struct drm_device *ddev = bo->base.dev;
	struct xdx_device *xdev = drm_to_xdev(ddev);
	int ret;
	//dump_stack();
	// printk("before dma_resv_lock!\n");
	
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 4, 0))
    //printk("bo->base.resv addr is %p\n",bo->base.resv);
	if (interruptible)
		ret = dma_resv_lock_interruptible(bo->base.resv, NULL);
	else
		ret = dma_resv_lock(bo->base.resv, NULL);
#else
	//printk("bo->resv addr ddd is %p\n",bo->resv);
	if (interruptible)
		ret = dma_resv_lock_interruptible(bo->resv, NULL);
	else
		ret = dma_resv_lock(bo->resv, NULL);
#endif
	// printk("after dma_resv_lock!\n");
	if (unlikely(ret != 0) && ret != -EINTR)
		dev_err(xdev->dev, "%p reserve failed\n", bo);

	return ret;
}

void xdx_bo_unreserve(struct xdx_bo *bo)
{
	#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 4, 0))
	dma_resv_unlock(bo->base.resv);
	#else
	dma_resv_unlock(bo->resv);
	#endif
}

static uint64_t *get_pages(struct xdx_bo *bo)
{
	int ret;
	uint32_t num_pages;

	if (!bo->pages) {
		ret = pvr_pmr_get_pages(
			bo->psPMR,
			get_order(bo->page_alignment << PAGE_SHIFT) + PAGE_SHIFT,
			&bo->pages, &num_pages);
		if (ret) {
			DRM_ERROR("Failed to get pages %p\n", bo);
			return NULL;
		}
		BUG_ON(bo->num_pages != num_pages);
	}

	return bo->pages;
}

uint64_t *xdx_bo_get_pages(struct xdx_bo *bo)
{
	return get_pages(bo);
}
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 1, 0))
static void put_pages(struct xdx_bo *bo)
{
	if (bo->pages) {
		pvr_pmr_put_pages(bo->psPMR, bo->pages);
		bo->pages = NULL;
	}
}

void xdx_bo_put_pages(struct xdx_bo *bo)
{
	put_pages(bo);
}
#endif
int xdx_bo_vmap(struct xdx_bo *bo)
{
	int ret;
	#if (LINUX_VERSION_CODE < KERNEL_VERSION(5, 4, 0))
    // printk("bo->resv addr c is %p\n",bo->resv);
	#endif
	if (bo->vaddr)
		return 0;

	ret = xdx_bo_reserve(bo, true);
	if (ret)
		goto out;

	ret = pvr_pmr_vmmap(bo->psPMR, &bo->vaddr, &bo->vmmap_priv);

	xdx_bo_unreserve(bo);
out:
	return ret;
}

void xdx_bo_vunmap(struct xdx_bo *bo)
{
	if (!bo->vaddr)
		return;

	pvr_pmr_vunmap(bo->psPMR, bo->vaddr, bo->vmmap_priv);
	bo->vaddr = NULL;
	bo->vmmap_priv = NULL;
}
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 1, 0))
static void xdx_gem_object_free(struct drm_gem_object *gobj)
{
	struct xdx_bo *bo = to_xdx_gem(gobj);
	struct xdx_device *xdev = drm_to_xdev(bo->base.dev);

	xdx_bo_vunmap(bo);

#ifndef NO_HARDWARE
	if (!xdev->flags.virtual_display) {
		struct xdx_bo_va *va, *tmp;

		list_for_each_entry_safe (va, tmp, &bo->va_list, node)
			xdx_vm_bo_remove(&xdev->vmm, va);
	}
#endif

#ifdef SUPPORT_VPU
	vpumem_remove(NULL, bo);
#endif

	xdx_bo_put_pages(bo);

	pvr_pmr_put(bo->psPMR);

	drm_gem_object_release(gobj);

	if (bo->cma_obj)
		kfree(bo->cma_obj);

	kvfree(bo);
}

/**
 * @brief Called from drm_gem_handle_create_tail
 * 
 * @param gobj 
 * @param file_priv 
 * @return int 
 */
static int xdx_gem_object_open(struct drm_gem_object *gobj,
			       struct drm_file *file_priv)
{
/*	struct xdx_bo *bo = to_xdx_gem(gobj);
	struct xdx_device *xdev = drm_to_xdev(bo->base.dev);
	struct xdx_drm_fpriv *fpriv = file_priv->driver_priv;
*/
	// TODO: do nothing now, for future use
	return 0;
}

static void xdx_gem_object_close(struct drm_gem_object *gobj,
				 struct drm_file *file_priv)
{
/*	struct xdx_bo *bo = to_xdx_gem(gobj);
	struct xdx_device *xdev = drm_to_xdev(bo->base.dev);
	struct xdx_drm_fpriv *fpriv = file_priv->driver_priv;
*/
	// TODO: do nothing now, for future use
}

struct dma_buf *xdx_gem_prime_export(struct drm_gem_object *gobj, int flags)
{
	struct xdx_bo *bo = to_xdx_gem(gobj);
	struct xdx_device *xdev = drm_to_xdev(bo->base.dev);
	struct dma_buf *buf;

	buf = pvr_export_dmabuf(xdev->gfx.pvrdev, bo->psPMR, gobj, flags);
	if (IS_ERR(buf))
		dev_err(xdev->dev, "failed to export obj %p to dmabuf\n", bo);

	return buf;
}
/*
static uint64_t devPAddr_to_cpuPAddr(struct xdx_device *xdev,
				     enum xdxgpu_gem_placement placement,
				     uint64_t addr)
{
	if (placement == XDXGPU_PL_VRAM)
		addr = addr - xdev->vram_start + xdev->aper_base;
	else if (placement == XDXGPU_PL_TT)
		addr = addr - xdev->gtt_start;
	else
		BUG_ON(1);

	return addr;
}
*/
#endif
int xdx_gem_object_mmap(struct drm_gem_object *gobj, struct vm_area_struct *vma)
{
	struct xdx_bo *bo = to_xdx_gem(gobj);
	int ret;
/*	struct xdx_bo *bo = to_xdx_gem(gobj);
	struct drm_device *ddev = bo->base.dev;
	struct xdx_device *xdev = drm_to_xdev(ddev);
	uint64_t *pages, page;
	pgprot_t page_prot;
	pfn_t pfn;
	uint32_t i;
	bool use_mixed_map = false;
*/
	BUG_ON(bo->pro.placement == XDXGPU_PL_PRIV);
	BUG_ON(bo->type == xdx_bo_type_kernel);

	ret = xdx_bo_reserve(bo, true);
	if (ret)
		goto out;
#if 0
	pages = get_pages(bo);
	if (!pages) {
		ret = -ENOMEM;
		goto no_pages;	
	}

	page_prot = vm_get_page_prot(vma->vm_flags);
	if (bo->flags & XDXGPU_GEM_CREATE_CPU_GTT_USWC)
		page_prot = pgprot_writecombine(page_prot);

	page_prot = pgprot_decrypted(page_prot);

	vma->vm_page_prot = page_prot;

	vma->vm_flags |= VM_IO;

	/* Don't include the mapping in core dumps */
	vma->vm_flags |= VM_DONTDUMP;

	/*
	 * Disable mremap because our nopage handler assumes all
	 * page requests have already been validated.
	 */
	vma->vm_flags |= VM_DONTEXPAND;

	/* Don't allow mapping to be inherited across a process fork */
	vma->vm_flags |= VM_DONTCOPY;

	for (i = 0; i < bo->num_pages; i++) {
		page = devPAddr_to_cpuPAddr(xdev, bo->pro.placement, pages[i]);
		pfn = phys_to_pfn_t(page, 0);

		if (!pfn_t_valid(pfn) || page_count(pfn_t_to_page(pfn)) == 0) {
			use_mixed_map = true;
			break;
		}
	}

	if (use_mixed_map)
		vma->vm_flags |= VM_MIXEDMAP;

	for (i = 0; i < bo->num_pages; i++) {
		page = devPAddr_to_cpuPAddr(xdev, bo->pro.placement, pages[i]);
		pfn = phys_to_pfn_t(page, 0);

		if (use_mixed_map) {
			vm_fault_t vmf;
	
			vmf = vmf_insert_mixed(vma, vma->vm_start + (i << PAGE_SHIFT),
					 pfn);

			if (vmf & VM_FAULT_ERROR)
				ret =  vm_fault_to_errno(vmf, 0);
		} else {
			ret = vm_insert_page(vma, vma->vm_start + (i << PAGE_SHIFT),
				       pfn_t_to_page(pfn));
		}

		if (ret)
			break;
	}

no_pages:
#else
	ret = pvr_pmr_mmap(bo->psPMR, vma);
#endif
	xdx_bo_unreserve(bo);
	drm_gem_object_put(gobj);
out:
	return ret;
}
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 1, 0)
#if ((LINUX_VERSION_CODE >= KERNEL_VERSION(5, 11, 0))&&(LINUX_VERSION_CODE < KERNEL_VERSION(5,18,0)))
static int xdx_gem_object_vmap(struct drm_gem_object *gobj,
			       struct dma_buf_map *map)
{
	struct xdx_bo *bo = to_xdx_gem(gobj);
	int ret;

	ret = xdx_bo_vmap(bo);
	if (ret)
		return ret;

	dma_buf_map_set_vaddr(map, bo->vaddr);
	return 0;
}

static void xdx_gem_object_vunmap(struct drm_gem_object *gobj,
				  struct dma_buf_map *map)
{
	struct xdx_bo *bo = to_xdx_gem(gobj);

	xdx_bo_vunmap(bo);
}
#elif (LINUX_VERSION_CODE >= KERNEL_VERSION(5,18,0))
static int xdx_gem_object_vmap(struct drm_gem_object *gobj,
			       struct iosys_map *map)
{
	struct xdx_bo *bo = to_xdx_gem(gobj);
	int ret;

	ret = xdx_bo_vmap(bo);
	if (ret)
		return ret;

	iosys_map_set_vaddr(map, bo->vaddr);
	return 0;
}

static void xdx_gem_object_vunmap(struct drm_gem_object *gobj,
				  struct iosys_map *map)
{
	struct xdx_bo *bo = to_xdx_gem(gobj);

	xdx_bo_vunmap(bo);
}
#else
static void *xdx_gem_object_vmap(struct drm_gem_object *gobj)
{
	struct xdx_bo *bo = to_xdx_gem(gobj);
	int ret;
    //printk("xdx_gem_object_vmap complete!\n");
	ret = xdx_bo_vmap(bo);
	if (ret)
		return NULL;

	return bo->vaddr;
}

static void xdx_gem_object_vunmap(struct drm_gem_object *gobj, void *vaddr)
{
	struct xdx_bo *bo = to_xdx_gem(gobj);
	//printk("xdx_gem_object_vunmap complete!\n");
	xdx_bo_vunmap(bo);
}
#endif

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 11, 0))
static vm_fault_t xdx_gem_object_vm_fault(struct vm_fault *vmf)
#else
static int xdx_gem_object_vm_fault(struct vm_area_struct *vma,
				   struct vm_fault *vmf)
#endif
{
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 11, 0))
	struct vm_area_struct *vma = vmf->vma;
#endif
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 10, 0))
	unsigned long vaddr = vmf->address;
#else
	unsigned long vaddr = (unsigned long)vmf->virtual_address;
#endif
	struct drm_gem_object *gobj = vma->vm_private_data;
	struct xdx_bo *bo = to_xdx_gem(gobj);
	struct xdx_device *xdev = drm_to_xdev(bo->base.dev);
	unsigned long page_offset, pfn;
	unsigned long addr;
	int ret = 0;

	if (bo->pro.placement != XDXGPU_PL_VRAM &&
	    bo->pro.placement != XDXGPU_PL_TT) {
		dev_err(xdev->dev, "no support heap %d\n", bo->pro.placement);
		return VM_FAULT_SIGSEGV;
	}

	if (!get_pages(bo)) {
		dev_err(xdev->dev, "no pages\n");
		return VM_FAULT_NOPAGE;
	}

	page_offset = ((vaddr - vma->vm_start) >>
		      PAGE_SHIFT) + vma->vm_pgoff - drm_vma_node_start(&bo->base.vma_node);

#if 0
	addr = devPAddr_to_cpuPAddr(xdev, bo->pro.placement,
				    bo->pages[page_offset]);
#else
	ret = pvr_pmr_get_cpu_paddr(bo->psPMR, (page_offset << PAGE_SHIFT), &addr);
	if (ret)
		return ret;
#endif

	pfn = addr >> PAGE_SHIFT;

#if defined(__loongarch64)
	vma->vm_page_prot = pgprot_noncached(vma->vm_page_prot);
#endif

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 20, 0))
	return vmf_insert_pfn(vma, vaddr, pfn);
#else
	{
		int err;

		err = vm_insert_pfn(vma, vaddr, pfn);
		switch (err) {
		case 0:
		case -EBUSY:
			return VM_FAULT_NOPAGE;
		case -ENOMEM:
			return VM_FAULT_OOM;
		default:
			return VM_FAULT_SIGBUS;
		}
	}
#endif
}

static const struct vm_operations_struct xdx_gem_vm_ops = {
	.fault = xdx_gem_object_vm_fault,
	.open = drm_gem_vm_open,
	.close = drm_gem_vm_close,
};
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 1, 0))
static const struct drm_gem_object_funcs xdx_gem_object_funcs = {
	.free = xdx_gem_object_free,
	.open = xdx_gem_object_open,
	.close = xdx_gem_object_close,
	.export = xdx_gem_prime_export,
	.vmap = xdx_gem_object_vmap,
	.vunmap = xdx_gem_object_vunmap,
  	.vm_ops = &xdx_gem_vm_ops,
};
#endif
#endif
struct drm_gem_object *xdx_gem_prime_import(struct drm_device *ddev,
					    struct dma_buf *dma_buf)
{
	struct xdx_device *xdev = drm_to_xdev(ddev);
	struct xdx_bo *bo;
	struct drm_gem_object *gobj = NULL;
	int ret;
	void *psPMR = NULL;
	struct xdx_bo_property pro;
	uint32_t size;

	ret = pvr_import_dmabuf(xdev->gfx.pvrdev, dma_buf, &psPMR, &gobj);
	if (ret) {
		dev_err(xdev->dev, "PVR driver import dmabuf failed\n");
		return ERR_PTR(ret);
	}

	if (gobj) {
		if (gobj->dev == ddev) {
			drm_gem_object_get(gobj);
			/* because xbo already has a refrence to PMR, shoud put the extra one */
			pvr_pmr_put(psPMR);
			return gobj;
		}
	}

	pvr_query_pmr_info(psPMR, &pro, &size);

	WARN_ON(size != dma_buf->size);

	ret = xdx_gem_object_create(xdev, NULL, (uint64_t)size, PAGE_SIZE, &pro,
				    xdx_bo_type_sg, &gobj);
	if (ret) {
		dev_err(xdev->dev, "%s: failed to import dmabuf: size %d\n",
			__func__, size);
		pvr_pmr_put(psPMR);
		return ERR_PTR(ret);
	}

	bo = to_xdx_gem(gobj);
	bo->psPMR = psPMR;

	return gobj;
}

static inline int xdx_bo_mmap_offset(struct xdx_bo *bo, u64 *offset_p)
{
	struct drm_device *ddev = bo->base.dev;
	struct xdx_device *xdev = drm_to_xdev(ddev);
	int ret;

	ret = drm_gem_create_mmap_offset(&bo->base);
	if (ret) {
		dev_err(xdev->dev, "failed to allocate mmap offset\n");
		return ret;
	}
#if defined(SUPPORT_VPU)
	*offset_p = drm_vma_node_offset_addr(&bo->base.vma_node) + VSI_MMAP_ADDRES_CEIL;
#else
	*offset_p = drm_vma_node_offset_addr(&bo->base.vma_node);
#endif
	return 0;
}

int xdx_gem_object_create(struct xdx_device *xdev, void *pvr_fpriv,
			  uint64_t size, uint32_t alignment,
			  struct xdx_bo_property *pro, enum xdx_bo_type type,
			  struct drm_gem_object **obj)
{
	struct xdx_bo *bo;
	struct xdx_bo_user *ubo;
	struct xdx_bo_create_params bp = { 0 };
	int ret;

#if defined(DUMP_DC_BUFFER) || defined(SUPPORT_XDXGPU_FBC)
	if (xdev->asic_type == CHIP_PANGU_A0 || xdev->asic_type == CHIP_PANGU_B0)
		pro->flags |= XDXGPU_GEM_ALLOC_CONTIGUOUS;
#endif

	bp.size = size;
	bp.alignment = alignment;
	bp.bo_ptr_size = sizeof(*ubo);
	bp.pro = *pro;
	bp.type = type;
	bp.pvr_fpriv = pvr_fpriv;

	ret = xdx_bo_create(xdev, &bp, (struct xdx_bo **)&ubo);
	if (ret)
		return ret;

	ubo->pid = get_task_pid(current->group_leader, PIDTYPE_PID);
	bo = &ubo->bo;
	#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 1, 0))
	bo->base.funcs = &xdx_gem_object_funcs;
	#endif
	*obj = &bo->base;

	return ret;
}

static bool xdx_bo_validate_size(struct xdx_device *xdev, uint64_t size,
				 enum xdxgpu_gem_placement placement,
				 uint32_t mapping)
{
	switch (placement) {
	case XDXGPU_PL_VRAM:
		return (size < xdev->visible_vram_size);
	case XDXGPU_PL_PRIV:
		return (size < (xdev->real_vram_size - xdev->aper_size));
	case XDXGPU_PL_TT:
	case XDXGPU_PL_SYSTEM:
	default:
		break;
	}

	return true;
}
#if (LINUX_VERSION_CODE < KERNEL_VERSION(5, 4, 0))
void xdx_gem_private_object_init(struct drm_device *dev,
				 struct drm_gem_object *obj, struct xdx_bo *bo, size_t size)
{
	BUG_ON((size & (PAGE_SIZE - 1)) != 0);

	obj->dev = dev;
	obj->filp = NULL;

	kref_init(&obj->refcount);
	obj->handle_count = 0;
	obj->size = size;
	dma_resv_init(&bo->_resv);
	if (!bo->resv)
		bo->resv = &bo->_resv;
	//printk("bo->base.resv addr is eee %p\n",bo->resv);
	drm_vma_node_reset(&obj->vma_node);
}
#endif

int xdx_bo_create(struct xdx_device *xdev, struct xdx_bo_create_params *bp,
		  struct xdx_bo **bo_ptr)
{
	struct drm_device *ddev = xdev_to_drm(xdev);
	struct xdx_bo *bo;
	int ret = 0;
	struct xdx_bo_create_params params = *bp;

	/* Memory should be aligned at least to a page size. */
	params.alignment = ALIGN(params.alignment, PAGE_SIZE);
	params.size = ALIGN(params.size, PAGE_SIZE);

	if (!xdx_bo_validate_size(xdev, params.size, params.pro.placement,
				  params.pro.mapping))
		return -ENOMEM;

	BUG_ON(params.bo_ptr_size < sizeof(struct xdx_bo));
	*bo_ptr = NULL;

	bo = kvzalloc(params.bo_ptr_size, GFP_KERNEL);
	if (NULL == bo)
		return -ENOMEM;
	#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 4, 0))
	drm_gem_private_object_init(ddev, &bo->base, params.size);
	#else
	xdx_gem_private_object_init(ddev, &bo->base, bo, params.size);	
	#endif

	if (!drm_arch_can_wc_memory() &&
	    params.pro.cpu_cache == XDXGPU_GEM_CPU_UNCACHED_WC) {
		dev_info_once(xdev->dev,
			      "system arch don't support wc, change to cache mode\n");
		params.pro.cpu_cache = XDXGPU_GEM_CPU_CACHE_COHERENT;
		params.pro.gpu_cache = XDXGPU_GEM_GPU_CACHE_COHERENT;
	}

	bo->pro = params.pro;
	bo->type = params.type;
	bo->page_alignment = params.alignment >> PAGE_SHIFT;
	bo->num_pages = bo->base.size / params.alignment;
	INIT_LIST_HEAD(&bo->va_list);

	if (bo->type != xdx_bo_type_sg && bo->type != xdx_bo_type_pvr) {
		ret = pvr_bo_create(xdev->gfx.pvrdev, bp->pvr_fpriv, &params,
				    &bo->psPMR);
		if (ret) {
			dev_err(xdev->dev,
				"PVR failed to create bo: size %lld, placement %d, cache (%d, %d), mapping (0x%x), flags (0x%x)\n",
				params.size, params.pro.placement,
				params.pro.cpu_cache, params.pro.gpu_cache,
				params.pro.mapping, params.pro.flags);
			goto err_pvr;
		}
		#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 4, 0))
		pvr_pmr_set_reservation(bo->psPMR, bo->base.resv);
		#else
		pvr_pmr_set_reservation(bo->psPMR, bo->resv);
		#endif
	}

#ifdef SUPPORT_VPU
	vpumem_create(NULL, bo, params.size);
#endif

	*bo_ptr = bo;
	return ret;

err_pvr:
	kfree(bo);
	return ret;
}

int xdx_drm_dumb_create(struct drm_file *filp, struct drm_device *ddev,
			struct drm_mode_create_dumb *args)
{
	struct xdx_device *xdev = drm_to_xdev(ddev);
	struct xdx_dm *dm = &xdev->dm;
	struct xdx_drm_fpriv *fpriv = filp->driver_priv;
	struct drm_gem_object *gobj;
	struct xdx_bo_property pro = { 0 };
	struct dc_hw_info dc_info = { 0 };
	uint32_t handle;
	int ret, i;

	for (i = 0; i < dm->dc_num; i++){
		if (!xdev->flags.virtual_display)
			xdev->dm.dc[i]->funcs->query_hw_info(xdev->dm.dc[i], &dc_info);
		//todo, need to add control for dc1
	}

	args->pitch = args->width * (ALIGN(args->bpp, 8) >> 3);
	if (!xdev->flags.virtual_display)
		args->pitch = ALIGN(args->pitch, dc_info.pitch_alignment);
	args->size = PAGE_ALIGN(args->pitch * args->height);

	pro.placement = XDXGPU_PL_VRAM;
	pro.gpu_cache = XDXGPU_GEM_GPU_UNCACHED_WC;
	pro.cpu_cache = XDXGPU_GEM_CPU_UNCACHED_WC;
	pro.mapping = (XDXGPU_GEM_MAPPING_GPU_RW | XDXGPU_GEM_MAPPING_CPU_RW);
#ifdef CONFIG_MMU_1K_16M
	ret = xdx_gem_object_create(xdev, fpriv->pvr_fpriv,
				    round_up(args->size, 1 << 24),
				    PAGE_SIZE << 12, &pro, xdx_bo_type_device,
				    &gobj);
#else
	ret = xdx_gem_object_create(xdev, fpriv->pvr_fpriv, args->size,
				    PAGE_SIZE, &pro, xdx_bo_type_device, &gobj);
#endif
	if (ret) {
		dev_err(xdev->dev,
			"failed to create dumb: bpp %d, width %d, height %d\n",
			args->bpp, args->width, args->height);
		return ret;
	}

	ret = drm_gem_handle_create(filp, gobj, &handle);
#if LINUX_VERSION_CODE<KERNEL_VERSION(5, 9, 0)
	drm_gem_object_put_unlocked(gobj);
#else
	drm_gem_object_put(gobj);
#endif
	if (ret)
		return ret;

	args->handle = handle;
	return ret;
}

int xdx_drm_dumb_mmap(struct drm_file *filp, struct drm_device *ddev,
		      uint32_t handle, uint64_t *offset_p)
{
	struct xdx_device *xdev = drm_to_xdev(ddev);
	struct drm_gem_object *gobj;
	struct xdx_bo *bo;
	int ret;

	gobj = drm_gem_object_lookup(filp, handle);
	if (gobj == NULL) {
		dev_err(xdev->dev, "%s: wrong handle 0x%x\n", __func__, handle);
		return -ENOENT;
	}

	bo = to_xdx_gem(gobj);
	if (!(bo->pro.mapping & XDXGPU_GEM_MAPPING_CPU_RW)) {
		#if LINUX_VERSION_CODE<KERNEL_VERSION(5, 9, 0)
		drm_gem_object_put_unlocked(gobj);
		#else
		drm_gem_object_put(gobj);
		#endif
		return -EPERM;
	}

	ret = xdx_bo_mmap_offset(bo, offset_p);
#if LINUX_VERSION_CODE<KERNEL_VERSION(5, 9, 0)
	drm_gem_object_put_unlocked(gobj);
#else
	drm_gem_object_put(gobj);
#endif

	return ret;
}

#if defined(SUPPORT_VPU)
extern int hantro_gem_dumb_create_combine(struct drm_file *file_priv,
					   struct drm_device *dev,
					   union drm_xdxgpu_gem_create *args);
#endif

int xdx_gem_create_ioctl(struct drm_device *dev, void *data,
			 struct drm_file *filp)
{
	struct xdx_device *xdev = drm_to_xdev(dev);
	struct xdx_drm_fpriv *fpriv = filp->driver_priv;
	union drm_xdxgpu_gem_create *args = data;
	struct drm_gem_object *gobj;
	uint32_t handle;
	int ret;
	struct xdx_bo_property pro = {0};

#if defined(SUPPORT_VPU)
	if (args->in.flags & XDXGPU_GEM_ALLOC_VPU) {
		return hantro_gem_dumb_create_combine(
			filp, dev, args);
	}
#endif

	pro.placement = args->in.placement;
	pro.gpu_cache = args->in.gpu_cache;
	pro.cpu_cache = args->in.cpu_cache;
	pro.mapping = args->in.mapping;
	pro.flags = args->in.flags;

	ret = xdx_gem_object_create(xdev, fpriv->pvr_fpriv, args->in.bo_size,
				    args->in.alignment, &pro,
				    xdx_bo_type_device, &gobj);
	if (ret) {
		if (ret != -ERESTARTSYS) {
			if (args->in.placement == XDXGPU_PL_VRAM) {
				dev_err(xdev->dev,
					"failed to alloc visable vram:  (%llu, %llu, %d)\n",
					args->in.bo_size, args->in.alignment,
					ret);
			}
		}
		return ret;
	}

	ret = drm_gem_handle_create(filp, gobj, &handle);
	#if LINUX_VERSION_CODE<KERNEL_VERSION(5, 9, 0)
	drm_gem_object_put_unlocked(gobj);
	#else
	drm_gem_object_put(gobj);
	#endif
	if (ret) {
		dev_err(xdev->dev, "%s: failed to create handle\n", __func__);
		return ret;
	}

	memset(args, 0, sizeof(*args));
	args->out.handle = handle;

	return ret;
}

int xdx_gem_info_query_ioctl(struct drm_device *ddev, void *data,
		       struct drm_file *filp)
{
	struct xdx_device *xdev = drm_to_xdev(ddev);
	union drm_xdxgpu_gem_query *args = data;
	uint32_t handle = args->in.handle;
	struct drm_gem_object *gobj;
	struct xdx_bo *bo;
	int ret = 0;

	gobj = drm_gem_object_lookup(filp, handle);
	if (gobj == NULL) {
		dev_err(xdev->dev, "%s: wrong handle 0x%x\n", __func__, handle);
		return -ENOENT;
	}

	bo = to_xdx_gem(gobj);

	if (!(bo->pro.mapping & XDXGPU_GEM_MAPPING_CPU_RW)) {
		dev_err(xdev->dev, "bo %p, handle %d couldn't mmap\n", bo,
			handle);
		args->out.addr_ptr = (uint64_t)-1;
		#if LINUX_VERSION_CODE<KERNEL_VERSION(5, 9, 0)
		drm_gem_object_put_unlocked(gobj);
		#else
		drm_gem_object_put(gobj);
		#endif
		return -EPERM;
	} else {
		ret = xdx_bo_mmap_offset(bo, &args->out.addr_ptr);
	}

	args->out.size = bo->base.size;
#if LINUX_VERSION_CODE<KERNEL_VERSION(5, 9, 0)
	drm_gem_object_put_unlocked(gobj);
#else
	drm_gem_object_put(gobj);
#endif

	return ret;
}

int xdx_gem_export_to_pvr_ioctl(struct drm_device *ddev, void *data,
				struct drm_file *filp)
{
	struct xdx_device *xdev = drm_to_xdev(ddev);
	union drm_xdxgpu_gem_export_to_pvr *args = data;
	uint32_t handle = args->in.handle;
	struct drm_gem_object *gobj;
	struct xdx_bo *bo;
	struct xdx_drm_fpriv *fpriv = filp->driver_priv;
	int ret;

	gobj = drm_gem_object_lookup(filp, handle);
	if (gobj == NULL) {
		dev_err(xdev->dev, "%s: wrong handle 0x%x\n", __func__, handle);
		return -ENOENT;
	}

	bo = to_xdx_gem(gobj);

	ret = pvr_pmr_export_handle(fpriv->pvr_fpriv, bo->psPMR,
				    &args->out.handle);
	if (ret) {
		dev_err(xdev->dev, "failed to export handle\n");
	}
#if LINUX_VERSION_CODE<KERNEL_VERSION(5, 9, 0)
	drm_gem_object_put_unlocked(gobj);
#else
	drm_gem_object_put(gobj);
#endif
	return ret;
}

int xdx_gem_import_from_pvr_ioctl(struct drm_device *ddev, void *data,
				  struct drm_file *filp)
{
	struct xdx_device *xdev = drm_to_xdev(ddev);
	union drm_xdxgpu_gem_import_from_pvr *args = data;
	struct xdx_drm_fpriv *fpriv = filp->driver_priv;
	int ret;
	void *psPMR;
	struct drm_gem_object *gobj;
	struct xdx_bo *bo;
	uint32_t handle;
	struct xdx_bo_property pro;
	uint32_t size;
	struct drm_file *file;
	bool bFound = false;

	ret = pvr_pmr_lookup(fpriv->pvr_fpriv, args->in.handle, &psPMR);
	if (ret) {
		dev_err(xdev->dev, "failed to lookup PMR: handle(0x%x)\n",
			args->in.handle);
		return ret;
	}

	mutex_lock(&ddev->filelist_mutex);
	list_for_each_entry (file, &ddev->filelist, lhead) {
		spin_lock(&file->table_lock);
		idr_for_each_entry (&file->object_idr, gobj, handle) {
			bo = to_xdx_gem(gobj);
			if (bo->psPMR == psPMR) {
				bFound = true;
				drm_gem_object_get(
					gobj); /* avoid released in other process */
				break;
			}
		}
		spin_unlock(&file->table_lock);

		if (bFound)
			break;
	}
	mutex_unlock(&ddev->filelist_mutex);

	if (bFound) {
		if (file != filp) {
			ret = drm_gem_handle_create(filp, gobj, &handle);
			if (ret) {
				dev_err(xdev->dev,
					"failed to create gem handle\n");
				pvr_pmr_put(psPMR);
				#if LINUX_VERSION_CODE<KERNEL_VERSION(5, 9, 0)
				drm_gem_object_put_unlocked(gobj);
				#else
				drm_gem_object_put(gobj);
				#endif
				return ret;
			}
		}

		args->out.handle = handle;
		pvr_pmr_put(psPMR);
		#if LINUX_VERSION_CODE<KERNEL_VERSION(5, 9, 0)
		drm_gem_object_put_unlocked(gobj);
		#else
		drm_gem_object_put(gobj);
		#endif
		return 0;
	}

	pvr_query_pmr_info(psPMR, &pro, &size);

	ret = xdx_gem_object_create(xdev, NULL, (uint64_t)size, PAGE_SIZE, &pro,
				    xdx_bo_type_pvr, &gobj);
	if (ret) {
		pvr_pmr_put(psPMR);
		dev_err(xdev->dev,
			"failed to import from pvr handle: size %d\n", size);
		return ret;
	}

	bo = to_xdx_gem(gobj);
	bo->psPMR = psPMR;

	ret = drm_gem_handle_create(filp, gobj, &handle);
	#if LINUX_VERSION_CODE<KERNEL_VERSION(5, 9, 0)
	drm_gem_object_put_unlocked(gobj);
	#else
	drm_gem_object_put(gobj);
	#endif
	if (ret)
		return ret;

	args->out.handle = handle;
	return ret;
}

uint64_t xdx_bo_gpu_offset(struct xdx_bo *bo)
{
	BUG_ON(bo->pages == NULL);
	//BUG_ON(!(bo->pro.flags & XDXGPU_GEM_ALLOC_CONTIGUOUS));
	return bo->pages[0];
}

int xdx_bo_create_reserved(struct xdx_device *xdev, unsigned long size,
			   int align, enum xdxgpu_gem_placement placement,
			   struct xdx_bo **bo_ptr, u64 *dev_addr,
			   void **cpu_addr)
{
	struct xdx_bo_create_params bp = { 0 };
	struct xdx_bo *bo;
	int ret;

	if (!bo_ptr)
		return -EINVAL;

	bp.size = size;
	bp.alignment = align;
	bp.bo_ptr_size = sizeof(*bo);
	bp.pro.placement = placement;
	bp.pro.cpu_cache = XDXGPU_GEM_CPU_UNCACHED_WC;
	bp.pro.gpu_cache = XDXGPU_GEM_GPU_UNCACHED_WC;
	bp.pro.flags |= XDXGPU_GEM_ALLOC_CONTIGUOUS;
	bp.pro.mapping = (XDXGPU_GEM_MAPPING_GPU_RW | XDXGPU_GEM_MAPPING_CPU_RW);
	/* means that the created bo only used in kernel */
	bp.type = xdx_bo_type_kernel;
	bp.pvr_fpriv = NULL;

	ret = xdx_bo_create(xdev, &bp, &bo);
	if (ret)
		return ret;
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 1, 0))
	bo->base.funcs = &xdx_gem_object_funcs;
#endif
	// printk("bo->base.resv addr ccc is %p\n",bo->resv);
    // printk("xdx_bo_create complete!\n");
	//dump_stack();
	ret = xdx_bo_reserve(bo, false);
	if (ret) {
		dev_err(xdev->dev, "(%d) failed to reserve kernel bo\n", ret);
		goto err_free;
	}

	if (!get_pages(bo)) {
		ret = -ENOMEM;
		goto err_alloc_pages;
	}

	if (dev_addr)
		*dev_addr = xdx_bo_gpu_offset(bo);

	if (cpu_addr) {
		ret = pvr_pmr_vmmap(bo->psPMR, &bo->vaddr, &bo->vmmap_priv);
		if (ret) {
			dev_err(xdev->dev, "%s: failed to vmap\n", __func__);
			goto err_vmap;
		}
		*cpu_addr = bo->vaddr;
	}

	xdx_bo_unreserve(bo);
	*bo_ptr = bo;
	return 0;

err_vmap:
err_alloc_pages:
	xdx_bo_unreserve(bo);
err_free:
#if LINUX_VERSION_CODE<KERNEL_VERSION(5, 9, 0)
	drm_gem_object_put_unlocked(&bo->base);
#else
	drm_gem_object_put(&bo->base);
#endif
	return ret;
}

int xdx_gem_mmap(struct file *filp, struct vm_area_struct *vma)
{
	struct drm_file *priv = filp->private_data;
	struct drm_device *dev = priv->minor->dev;
	struct drm_gem_object *obj = NULL;
	struct drm_vma_offset_node *node;
	int ret;

	drm_vma_offset_lock_lookup(dev->vma_offset_manager);
	node = drm_vma_offset_exact_lookup_locked(dev->vma_offset_manager,
						  vma->vm_pgoff,
						  vma_pages(vma));
	if (likely(node)) {
		obj = container_of(node, struct drm_gem_object, vma_node);
		/*
		 * When the object is being freed, after it hits 0-refcnt it
		 * proceeds to tear down the object. In the process it will
		 * attempt to remove the VMA offset and so acquire this
		 * mgr->vm_lock.  Therefore if we find an object with a 0-refcnt
		 * that matches our range, we know it is in the process of being
		 * destroyed and will be freed as soon as we release the lock -
		 * so we have to check for the 0-refcnted object and treat it as
		 * invalid.
		 */
		if (!kref_get_unless_zero(&obj->refcount))
			obj = NULL;
	}
	drm_vma_offset_unlock_lookup(dev->vma_offset_manager);

	if (!obj)
		return -EINVAL;

	if (!drm_vma_node_is_allowed(node, priv)) {
		#if LINUX_VERSION_CODE<KERNEL_VERSION(5, 9, 0)
		drm_gem_object_put_unlocked(obj);
		#else
		drm_gem_object_put(obj);
		#endif
		return -EACCES;
	}
#if LINUX_VERSION_CODE<KERNEL_VERSION(5, 14, 0)
	if (node->readonly)
#endif
 {
		if (vma->vm_flags & VM_WRITE) {
			#if LINUX_VERSION_CODE<KERNEL_VERSION(5, 9, 0)
			drm_gem_object_put_unlocked(obj);
			#else
			drm_gem_object_put(obj);
			#endif
			return -EINVAL;
		}

		vma->vm_flags &= ~VM_MAYWRITE;
	}

	ret = xdx_gem_object_mmap(obj, vma);
#if LINUX_VERSION_CODE<KERNEL_VERSION(5, 9, 0)
	drm_gem_object_put_unlocked(obj);
#else
	drm_gem_object_put(obj);
#endif

	return ret;
}

int xdx_bo_sparse_copy(struct xdx_bo *bo, void *vaddr)
{
	return sparse_pmr_copy(bo, vaddr);
}
