#include <linux/device.h>
#include <linux/stat.h>
#include <linux/sysfs.h>
#include <linux/pm_runtime.h>

#include "xdx_sysfs.h"
#include "xdx_device.h"
#include "physheap.h"
#include "pvrsrv.h"

static inline struct xdx_device *kdev_minor_to_xdev(struct device *kdev)
{
	struct drm_minor *minor = dev_get_drvdata(kdev);

	return drm_to_xdev(minor->dev);
}

static ssize_t xdx_dbdf_show(struct device *dev,
			     struct device_attribute *attr,
			     char *buf)
{
	uint32_t bus, did, fn, domain;
	struct xdx_device *xdev = kdev_minor_to_xdev(dev);

	bus = xdev->pdev->bus->number;
	domain = pci_domain_nr(xdev->pdev->bus);
	did = PCI_SLOT(xdev->pdev->devfn);
	fn = PCI_FUNC(xdev->pdev->devfn);
#if (LINUX_VERSION_CODE > KERNEL_VERSION(5, 10, 0))
	return sysfs_emit(buf, "%04x:%02x:%02x.%d\n", domain, bus, did, fn);
#else	
	return sprintf(buf, "%04x:%02x:%02x.%d\n", domain, bus, did, fn);
#endif
}

static ssize_t xdx_product_name_show(struct device *dev,
			     struct device_attribute *attr,
			     char *buf)
{
	char name[20];
	struct xdx_device *xdev = kdev_minor_to_xdev(dev);

	switch (xdev->asic_type) {
	case CHIP_PANGU_A0:
	case CHIP_PANGU_B0:
		strcpy(name, "PANGU");
		break;
	default:
		strcpy(name, "Unknown_Xdxgpu");
		break;
	}
#if (LINUX_VERSION_CODE > KERNEL_VERSION(5, 10, 0))
	return sysfs_emit(buf, "%s\n", name);
#else	
	return sprintf(buf, "%s\n", name);
#endif
}

static ssize_t xdx_enable_runpm_show(struct device *dev,
				     struct device_attribute *attr,
				     char *buf)
{
	char enable[2];
	struct xdx_device *xdev = kdev_minor_to_xdev(dev);

	if (xdev->enable_runpm)
		strcpy(enable, "1");
	else
		strcpy(enable, "0");
#if (LINUX_VERSION_CODE > KERNEL_VERSION(5, 10, 0))
	return sysfs_emit(buf, "%s\n", enable);
#else
	return sprintf(buf, "%s\n", enable);
#endif
}

static ssize_t xdx_enable_runpm_store(struct device *dev,
				      struct device_attribute *attr,
				      const char *buf,
				      size_t count)
{
	struct xdx_device *xdev = kdev_minor_to_xdev(dev);
	struct drm_device *ddev = xdev_to_drm(xdev);
	bool enable;
	int err;

	err = kstrtobool(buf, &enable);
	if (err)
		return err;
	
	if (enable == xdev->enable_runpm)
		return count;

	/* dynamically turn on/off runtime pm feature */
	xdev->enable_runpm = enable;
	if (enable) {
		pm_runtime_use_autosuspend(ddev->dev);
		pm_runtime_set_autosuspend_delay(ddev->dev, 5000);
		pm_runtime_allow(ddev->dev);
		pm_runtime_mark_last_busy(ddev->dev);
		pm_runtime_put_autosuspend(ddev->dev);
	} else {
		pm_runtime_get_sync(ddev->dev);
		pm_runtime_forbid(ddev->dev);
		pm_runtime_dont_use_autosuspend(ddev->dev);
	}

	return count;
}

static ssize_t xdx_gpu_info_show(struct device *dev, struct device_attribute *attr,
			     char *buf)
{
	PVRSRV_DEVICE_NODE *psDeviceNode;
	PHYS_HEAP_MEM_STATS *mem_stats;
	PVRSRV_PHYS_HEAP ids[1] = { PVRSRV_PHYS_HEAP_GPU_LOCAL };

	psDeviceNode = PVRSRVGetPVRSRVData()->psDeviceNodeList;
	mem_stats = kzalloc(sizeof(PHYS_HEAP_MEM_STATS), GFP_KERNEL);
	PhysHeapGetMemInfo(psDeviceNode, 1, ids, mem_stats);
#if (LINUX_VERSION_CODE > KERNEL_VERSION(5, 10, 0))
	return sysfs_emit(buf, "VRAM total size:0x%llx\n",
			  mem_stats[0].ui64TotalSize);
#else
	return sprintf(buf, "VRAM total size:0x%llx\n",
		       mem_stats[0].ui64TotalSize);
#endif
}

static DEVICE_ATTR_RO(xdx_dbdf);
static DEVICE_ATTR_RO(xdx_product_name);
static DEVICE_ATTR_RW(xdx_enable_runpm);

/* Due to a "-" sign in filename, generate the attr without macro. */
static struct device_attribute dev_attr_xdx_gpu_info = {
	.attr = { .name = "gpu-info", .mode = 0444 },
	.show = xdx_gpu_info_show,
};

static const struct attribute * const xdx_attrs[] = {
	&dev_attr_xdx_dbdf.attr,
	&dev_attr_xdx_product_name.attr,
	&dev_attr_xdx_enable_runpm.attr,
	NULL
};

static const struct attribute * const xdx_device_attrs[] = {
	&dev_attr_xdx_gpu_info.attr,
	NULL
};

void xdx_sysfs_init(struct xdx_device *xdev)
{
	struct device *kdev = xdev_to_drm(xdev)->primary->kdev;
	int r;

	/* sysfs under drm folder */
	r = sysfs_create_files(&kdev->kobj, (const struct attribute **)xdx_attrs);
	if (r)
		DRM_ERROR("xdxgpu sysfs init failed under drm folder\n");
	/* sysfs under device folder */
	r = sysfs_create_files(&xdev->pdev->dev.kobj, (const struct attribute **)xdx_device_attrs);
	if (r)
		DRM_ERROR("xdxgpu sysfs init failed under device folder\n");
}

void xdx_sysfs_fini(struct xdx_device *xdev)
{
	struct device *kdev = xdev_to_drm(xdev)->primary->kdev;

	sysfs_remove_files(&kdev->kobj, (const struct attribute **)xdx_attrs);
	sysfs_remove_files(&xdev->pdev->dev.kobj, (const struct attribute **)xdx_device_attrs);
}
